import json, time
from importlib import import_module
from django.views.generic import View
from django.http import HttpResponse, HttpResponseNotFound, HttpResponseForbidden
from django.conf import settings
from django.db import DatabaseError, IntegrityError, router, transaction
from django.views.decorators.csrf import csrf_exempt
from nicechat.utils.arbiter import arb_encrypt, arb_decrypt, mark_node_status, balancer_make_sign, balancer_check_sign
from nicechat.utils.arbiter_sync import get_user_relations, get_website_relations, scenario_read, scenario_clear

class Encrypt(dict):
    pass


class ArbiterS2SBase(View):
    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(ArbiterS2SBase, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        cmd = kwargs.get('cmd', 'none')
        if cmd != 'balancer':
            return HttpResponseNotFound()
        
        # if not balancer_check_sign(request.META.get('HTTP_X_BALANCER'), settings.BALANCER_SIGN):
        #     return HttpResponseForbidden()
        
        from apps.balancer.models import ArbiterHostMap, ArbiterHostDomains, ArbiterHostInfo
        from nicechat.utils.crypt import decrypt_dict_v

        hm = ArbiterHostMap.objects.last()
        if not hm:
            time.sleep(2)
            return HttpResponse("", status=412)

        domains = []
        for d in ArbiterHostDomains.objects.all():

            try:
                keys = json.loads(d.data)
            except json.JSONDecodeError:
                keys = {}

            domains.append(
                [
                    d.domain,
                    decrypt_dict_v(keys, settings.DATA_KEY_HS, settings.DATA_KEY_IV)
                ]
            )

        data = json.dumps({
            'nodes': hm.data,
            'domains': domains
        })
        client = getattr(request, 'scope', {}).get('client')

        if client:
            n, _ = ArbiterHostInfo.objects.update_or_create(
                address = client[0],
                defaults = {
                    'mode': "balancer",
                    'isok': True,
                    'timestamp': int(time.time()),
                    'stats': '{"status": "up"}'
                }
            )

        return HttpResponse(data, content_type='application/json')#, headers={
            #"X-Balancer": balancer_make_sign(settings.BALANCER_SIGN)})

    def post(self, request, *args, **kwargs):
        cmd = kwargs.get('cmd', 'none')
        is_valid, host, body = self.decode_request(request, cmd)
        if not is_valid:
            return HttpResponseForbidden()

        if hasattr(self, f'run_{cmd}') and callable(getattr(self, f'run_{cmd}')):
            if settings.NODE_MODE == settings.NODE_MODE_ARBITER:
                ret = getattr(self, f'run_{cmd}')(host, request, body)
                _ret_pub_key = host.host_pubkey if host else settings.NODE_PUBLIC_KEY
            else:
                ret = getattr(self, f'run_{cmd}')(request, body)
                _ret_pub_key = settings.NODE_ARBITER_PUB
        else:
            return HttpResponseNotFound()

        if ret is None:
            ret = 'null'

        if not isinstance(ret, str):
            ret = json.dumps(dict(ret))

        ret = arb_encrypt(ret, settings.NODE_PRIVATE_KEY, _ret_pub_key)

        resp = HttpResponse(ret, content_type='application/json')

        return resp
    
    @staticmethod
    def decode_request(request, cmd):
        x_key = request.META.get('HTTP_X_SIGN', None)
        content, is_correct = arb_decrypt(request.body, settings.NODE_PRIVATE_KEY, x_key)
        if not is_correct:
            return False, None, None
        anh = None
        if settings.NODE_MODE == settings.NODE_MODE_ARBITER and x_key:
            from apps.balancer.models import ArbiterNodeHost
            anh = ArbiterNodeHost.objects.filter(host_pubkey=x_key).first()

        try:
            content = json.loads(content)
        except json.JSONDecodeError:
            content = {}

        return True, anh, content


def get_node_stat():
    content = b''
    import subprocess
    content += b"Hostname\n"
    result = subprocess.run(['hostname'], stdout=subprocess.PIPE)
    content += result.stdout
    content += b"\n\nCPU\n"
    result = subprocess.run(['mpstat'], stdout=subprocess.PIPE)
    content += result.stdout
    content += b"\n\nMemory\n"
    result = subprocess.run(['free', '-m'], stdout=subprocess.PIPE)
    content += result.stdout
    content += b"\n\nStorage\n"
    result = subprocess.run(['df', '-h'], stdout=subprocess.PIPE)
    content += result.stdout
    content += b"\n\nProcess list\n"
    result = subprocess.run(['ps', '-a'], stdout=subprocess.PIPE)
    content += result.stdout
    content += b"\n\nBackground workers\n"
    result = subprocess.run(['celery', '-A', 'nicechat', 'inspect', 'stats'], stdout=subprocess.PIPE)
    content += result.stdout

    return {"data": content.decode()}


def send_test_email(email):
    if not email:
        return {'txt': 'Email is empty'}

    from django.core.mail import send_mail
    import smtplib

    try:
        send_mail(
            "Arbiter test email",
            "Nice.Chat Arbiter test email",
            settings.DEFAULT_FROM_EMAIL,
            [email],
            fail_silently=False,
        )
    except smtplib.SMTPResponseException as err:
        return {'txt': f'{err.smtp_error}'}
    except smtplib.SMTPException as e:
        return {'txt': f'{e}'}
    except Exception as e:
        return {'txt': f'{e}'}

    return {'txt': 'ok'}


class ArbiterS2S(ArbiterS2SBase):

    def options(self, request, *args, **kwargs):
        resp = HttpResponse(status=204)
        setattr(resp, 'is_rendered', True)
        return resp
    
    def run_monitor(self, host, request, data):
        mark_node_status(data, host, request)
        return {}

    def run_getstat(self, host, request, data):
        return get_node_stat()

    def run_testemail(self, host, request, data):
        return send_test_email(data.get('email'))

    def run_upsync(self, host, request, data):
        an = host.node
        from apps.balancer.models import ArbiterNode
        from apps.users.models import User
        from apps.website.models import Website

        page = data.get('page', None)

        if page is None:
            return {
                'error': 'pagination',
                'next': False
            }
        PAGE_STEP = 10
        websites = []
        users = []

        _users = list(User.objects.filter(
            website__node2website__node=an
        ))
        _users_sa = list(User.objects.filter(
            is_superuser=True
        ))
        _websites = list(Website.objects.filter(
            node2website__node=an
        ))
        _req = _websites + _users + _users_sa

        for rec in _req[page*PAGE_STEP:(page+1)*PAGE_STEP]:
            if isinstance(rec, Website):
                websites.append(get_website_relations(rec))
            if isinstance(rec, User):
                users.append(get_user_relations(rec))
        
        have_next = (page+1)*PAGE_STEP <= len(_req)

        return {
            'websites': websites,
            'users': users,
            'next': have_next
        }


    def run_init(self, host, request, content):

        if not host or not host.node:
            return Encrypt({'err': 'host not authorized'})

        creds = host.node.get_credentials()

        return Encrypt({
            'NODE': host.node.name,

            'ALL_DOMAINS': settings.ENV_ALL_DOMAINS,

            'MAIN_DOMAIN': settings.ENV_MAIN_DOMAIN,
            'MAIN_NAME': settings.ENV_MAIN_NAME,
            'GOOGLE_MAP_KEY': settings.ENV_GOOGLE_MAP_KEY,

            'SEND_ERRORS_EMAILS': settings.ENV_SEND_ERRORS_EMAILS,
            'INNER_NOTIFICATION_EMAILS': settings.ENV_INNER_NOTIFICATION_EMAILS,

            'SMTP_HOST': settings.ENV_SMTP_HOST,
            'SMTP_PORT': settings.ENV_SMTP_PORT,
            'SMTP_USER': settings.ENV_SMTP_USER,
            'SMTP_MODE': settings.ENV_SMTP_MODE,
            'SMTP_PASSWORD': settings.ENV_SMTP_PASSWORD,
            'EMAIL_FROM': settings.ENV_EMAIL_FROM,

            'SECRET_KEY': settings.ENV_SECRET_KEY,
            'SENTRY_DSN': settings.ENV_SENTRY_DSN,
            'PUBLIC_ROOT': settings.ENV_PUBLIC_ROOT,

            'POSTGRES_NAME': creds.get('postgres_dbname'),
            'POSTGRES_USER': creds.get('postgres_user'),
            'POSTGRES_PASSWORD': creds.get('postgres_password'),
            'POSTGRES_HOST': creds.get('postgres_host'),
            'POSTGRES_PORT': creds.get('postgres_port', 5432) or 5432,

            'REDIS_HOST': creds.get('redis_host'),
            'REDIS_SSL': creds.get('redis_ssl', False),
            'REDIS_PORT': creds.get('redis_port', 6379) or 6379,
            'REDIS_DB': creds.get('redis_dbnumber'),
            'REDIS_USER': creds.get('redis_user'),
            'REDIS_PASS': creds.get('redis_password'),

        })

    def run_getchatid(self, host, request, data):
        from apps.balancer.models import UniqueChatId
        obj = UniqueChatId.objects.create(node_name=host.node.name)
        return {'chat_id': obj.pk}


class ArbiterS2SWorker(ArbiterS2SBase):

    def run_user(self, request, data):
        from nicechat.utils.arbiter_sync import set_user_with_relations
        set_user_with_relations(data)
        return {}

    def run_getstat(self, request, data):
        return get_node_stat()

    def run_testemail(self, request, data):
        return send_test_email(data.get('email'))

    def run_todjangoadmin(self, request, data):
        from django.contrib.auth import SESSION_KEY, HASH_SESSION_KEY, BACKEND_SESSION_KEY, get_user_model

        user_id = data.get('user_id')
        session_id = data.get('session_id')
        backend = data.get('backend')
        user = get_user_model().objects.filter(pk=user_id).first()
        if not user:
            return {'ok': False}
        user_hash = user.get_session_auth_hash()
        # print(user_id, user_hash, session_id)

        session_store = import_module(settings.SESSION_ENGINE).SessionStore

        scope_session = session_store(session_id)

        data = {
            SESSION_KEY: user_id,
            HASH_SESSION_KEY: user_hash,
            BACKEND_SESSION_KEY: backend
        }
        obj = scope_session.model(
            session_key=session_id,
            session_data=scope_session.encode(data),
            expire_date=scope_session.get_expiry_date(),
        )

        using = router.db_for_write(scope_session.model, instance=obj)
        try:
            with transaction.atomic(using=using):
                obj.save(
                    using=using
                )
        except (IntegrityError, DatabaseError):
            return {'ok': False}

        return {'ok': True}

    def run_checkexists(self, request, data):
        website_id = data.get('website_id')
        if not website_id:
            return {"state": "err"}

        from apps.website.models import Website
        import random
        wss = Website.objects.filter(pk=website_id)

        if wss.exists():
            ws = wss.first()
            code = ''.join(random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789') for _ in range(32))
            ws.props['clr-code'] = code
            ws.save()
            return {"state": "clr", "code": code}
        else:
            return {"state": "ok"}
        
    def run_check_seq(self, request, data):
        from django.core import management
        management.call_command("fix_db_seq")
        return {"state": "ok"}
    
    def run_exec_clear(self, request, data):
        website_id = data.get('website_id')
        code = data.get('code')

        if not website_id:
            return {"state": "err"}
        from apps.website.models import Website
        import time
        ws = Website.objects.filter(pk=website_id).first()
        if not ws:
            return {"state": "err"}
        
        if ws.props.get('clr-code') != code:
            return {"state": "err"}

        _t = time.time()
        print('Clearing ', ws)
        try:
            print(ws.archivedchat_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}
            print(ws.chat_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}
            print(ws.callbacks_set.all().delete())
            print(ws.product_set.all().delete())
            print(ws.productscategory_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}
            print(ws.formresult_set.all().delete())
            print(ws.form_set.all().delete())
            print(ws.quickreply_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}

            print(ws.usersiterole_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}
            print(ws.visitor_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}
            print(ws.user_set.all().delete())
            if time.time() - _t > 20:
                return {"state": "processing"}
            print(ws.delete())
        except Exception as e:
            import traceback
            traceback.print_exception(e)
            return {"state": "processing"}
        
        print('Clearing ', ws , 'done')

        return {"state": "ok"}

    
    def run_doread(self, request, data):
        website_id = data.get('website_id')
        action = data.get('action')
        if not website_id or not action:
            return None
        
        from apps.website.models import Website
        ws = Website.objects.filter(pk=website_id).first()

        if not ws:
            return None
    
        from apps.balancer.transfer import processing
        if callable(getattr(processing, f'do_read_{action}', None)):
            ret = getattr(processing, f'do_read_{action}')(ws)
            if ret != False and ret != None:
                return {'dumps': ret}
            else:
                return None
        else:
            return None

    def run_poweroff(self, request, data):
        website_id = data.get('website_id')
        new_node = data.get('new_node')

        from channels.layers import get_channel_layer
        from asgiref.sync import async_to_sync
        channel_layer = get_channel_layer()
        import time

        mod_msg = json.dumps({
            "type": "ws_mod_prefix",
            "id": website_id,
            "node_prefix": new_node,
            "path_prefix": f"/{new_node}/"
        })
        async_to_sync(channel_layer.group_send)('website_global-%s' % website_id, {
            "type": "websocket.send", 'text': mod_msg })
        time.sleep(2)

        async_to_sync(channel_layer.group_send)('website_global-%s' % website_id, {
            "type": "websocket.send", 'text': mod_msg })
        async_to_sync(channel_layer.group_send)('website_global-%s' % website_id, {
            "type": "websocket.send", 'text': '{"type": "ws_offline"}'})
        time.sleep(2)

        async_to_sync(channel_layer.group_send)('website_global-%s' % website_id, {
            "type": "websocket.send", 'text': 'setoffline'})
        
        return {'sent': True}


    def run_dowrite(self, request, data):
        website_id = data.get('website_id')
        action = data.get('action')
        t_pk = data.get('t_pk')
        data_path = data.get('data_path')
        if not website_id or not action or not data_path or not t_pk:
            return None
        
        from apps.website.models import Website
        ws = Website.objects.filter(pk=website_id).first()

        if not ws:
            return None
        
        _s, _d = scenario_read(data_path)

        from apps.balancer.transfer import processing
        if callable(getattr(processing, f'do_write_{action}', None)):
            ret = getattr(processing, f'do_write_{action}')(ws, _d, _s, t_pk)
            if ret != False and ret != None:
                #scenario_clear(data_path)
                return ret
            else:
                return None
        else:
            return None


    def run_website(self, request, data):
        from nicechat.utils.arbiter_sync import set_website_with_relations
        set_website_with_relations(data)
        return {}

    def run_userdel(self, requset, data):
        from apps.users.models import User
        user = User.objects.filter(pk=data['pk']).first()
        if user:
            user.delete()
            return {'status': 'ok'}
        return {'status': 'notfound'}
    
    def run_websitedel(self, requset, data):
        from apps.webiste.models import Website
        webiste = Website.objects.filter(pk=data['pk']).first()
        if webiste:
            webiste.delete()
            return {'status': 'ok'}
        return {'status': 'notfound'}

    def run_w2userpostclear(self, request, data):
        from apps.users.models import User
        from apps.website.models import Website
        ws = Website.objects.filter(pk=data.get('site', None)).first()
        if not ws:
            return {'state': 'err'}
        ws.users.clear()
        return {'state': 'ok'}


class UIPropsView(View):
    def get(self, request, *args, **kwargs):

        from apps.balancer.models import ArbiterSetting

        s = { d['key']: d['val'] for d in ArbiterSetting.objects.filter(
           key__in=['global_title', 'global_domain', 'global_base_icon']
        ).values('key', 'val')}

        domain = s.get('global_domain', 'chatusnow.com')
        title = s.get('global_title', 'Chat Us Now')
        icon = s.get('global_base_icon', f"https://widget.{domain}/img/noicon.png")

        data = json.dumps({
                "MAIN_DOMAIN": domain,
                "CHAT_NAME": title,
                "WIDGET_DOMAIN": f"widget.{domain}",
                "SUPPORT_DOMAIN": f"support.{domain}",
                "DEFAULT_WIDGET_ICON": icon
            })

        data = f"window.CHAT_UI_PROPS = {data};"

        return HttpResponse(data, content_type='application/javascript')
