import re
from django import forms
from apps.balancer.models import ArbiterNode, ArbiterTOTP
from apps.users.models import User
from django.conf import settings
from django.contrib.auth.forms import AuthenticationForm, PasswordChangeForm as PasswordChangeFormBase, \
    SetPasswordForm as SetPasswordFormBase


class NiceViewMixin():
    def __init__(self, *args, **kwargs):
        super(NiceViewMixin, self).__init__(*args, **kwargs)
        placeholders = getattr(self, 'placeholders', {})
        for visible in self.visible_fields():
            visible.field.widget.attrs['class'] = 'form-control'
            if placeholders and placeholders.get(visible.name):
                visible.field.widget.attrs['placeholder'] = placeholders.get(visible.name)
            if isinstance(visible.field.widget, forms.widgets.CheckboxInput):
                visible.field.widget.attrs['class'] = "checkbox-field"


class SetPasswordForm(NiceViewMixin, SetPasswordFormBase):
    pass


class PasswordChangeForm(NiceViewMixin, PasswordChangeFormBase):
    pass


class CreateManagerForm(NiceViewMixin, forms.Form):
    first_name = forms.CharField(max_length=100, required=False)
    last_name = forms.CharField(max_length=100, required=False)
    email = forms.CharField(max_length=100)
    password = forms.CharField(max_length=100, required=False, widget=forms.PasswordInput())

    notes = 'Leave "Password" empty for auto-generation'

    def clean_email(self):
        email = self.cleaned_data.get('email', '').lower()
        if User.objects.filter(email=email).exists():
            raise forms.ValidationError("This email is already used")
        return email


class AssignHostForm(NiceViewMixin, forms.Form):
    notes = 'Please select node, to which need assign this host'

    def __init__(self, *args, **kwargs):
        nodes_list = [(None, ' - Please choice Node - ')]
        [nodes_list.append((a.pk, a.name,)) for a in ArbiterNode.objects.all()]

        super(AssignHostForm, self).__init__(*args, **kwargs)
        self.fields["assign_to"] = forms.ChoiceField(
            choices=nodes_list,
            widget=forms.Select(choices=nodes_list, attrs={'class':'form-control'})
        )


class CreateNodeForm(NiceViewMixin, forms.Form):
    name = forms.CharField(max_length=100)
    description = forms.CharField(required=False, widget=forms.Textarea(attrs={"rows":"5"}))

    def clean_description(self):
        return self.cleaned_data.get('description', '-')

    def clean_name(self):
        name = self.cleaned_data.get('name', '').lower()
        if not name.isalnum() or name[0].isdigit():
            raise forms.ValidationError("The name must be alphanumeric")
        if ArbiterNode.objects.filter(name=name).exists():
            raise forms.ValidationError("The name is already used")
        return name


class CredentailsNodeForm(NiceViewMixin, forms.Form):
    postgres_host = forms.CharField(max_length=100)
    postgres_port = forms.IntegerField(required=False)
    postgres_dbname = forms.CharField(max_length=100)
    postgres_user = forms.CharField(max_length=100)
    postgres_password = forms.CharField(max_length=100, widget=forms.PasswordInput(render_value=True))

    redis_host = forms.CharField(max_length=100)
    redis_ssl = forms.BooleanField(required=False)
    redis_port = forms.IntegerField(required=False)
    redis_dbnumber = forms.IntegerField()
    redis_user = forms.CharField(max_length=100, required=False)
    redis_password = forms.CharField(max_length=100, widget=forms.PasswordInput(render_value=True), required=False)


class DomainEditForm(NiceViewMixin, forms.Form):
    DOMAIN_REGEXP = re.compile('^((?!-))(xn--)?[a-z0-9][a-z0-9-_]{0,61}[a-z0-9]{0,}\\.?((xn--)?([a-z0-9\\-.]{1,61}|[a-z0-9-]{1,30})\\.?[a-z]{2,})$')

    domain = forms.CharField(max_length=150)
    certificate = forms.CharField(required=False, widget=forms.Textarea(attrs={"rows":"15"}))
    privatekey = forms.CharField(required=False, widget=forms.Textarea(attrs={"rows":"10"}))

    def clean(self):
        domain = self.cleaned_data.get("domain", "").lower()

        if not self.DOMAIN_REGEXP.match(domain):
            raise forms.ValidationError("Domain is not correct")

        self.cleaned_data['domain'] = domain

        certificate = self.cleaned_data.get("certificate", "")
        privatekey = self.cleaned_data.get("privatekey", "")

        if (privatekey and not certificate) or (not privatekey and certificate):
            raise forms.ValidationError("PrivateKey and Certificate muse be filled both")
        return self.cleaned_data


class CredentailsArbiterForm(NiceViewMixin, forms.Form):
    # placeholders = {'email_from': ''}

    redis_host = forms.CharField(max_length=100)
    redis_ssl = forms.BooleanField(required=False)
    redis_port = forms.IntegerField(required=False)
    redis_dbnumber = forms.IntegerField()
    redis_user = forms.CharField(max_length=100, required=False)
    redis_password = forms.CharField(max_length=100, widget=forms.PasswordInput(render_value=True), required=False)

    google_map_key = forms.CharField(max_length=100, required=False)

    smtp_host = forms.CharField(max_length=100, required=False)
    smtp_port = forms.IntegerField(initial=587)
    smtp_user = forms.CharField(max_length=100, required=False)
    smtp_password = forms.CharField(max_length=100, widget=forms.PasswordInput(render_value=True), required=False)
    smtp_mode = forms.CharField(max_length=100, initial='tls',  widget=forms.Select(choices=(('tls', 'tls',), ('ssl', 'ssl',), ('none', 'none',))))
    email_from = forms.CharField(max_length=400, required=False, label='Email from: Name <email> or email')

    debug = forms.CharField(max_length=100, initial='no', widget=forms.Select(choices=(('yes', 'yes',), ('no', 'no',))))

    sentry_dsn = forms.CharField(max_length=300, required=False)

    send_errors_emails = forms.CharField(widget=forms.Textarea, required=False)
    inner_notification_emails = forms.CharField(widget=forms.Textarea, required=False)

    def clean(self):
        redis_host = self.cleaned_data.get("redis_host")
        redis_port = self.cleaned_data.get("redis_port")
        redis_ssl = self.cleaned_data.get("redis_ssl")
        redis_dbnumber = self.cleaned_data.get("redis_dbnumber")
        redis_user = self.cleaned_data.get("redis_user")
        redis_password = self.cleaned_data.get("redis_password")
        redis_auth = f'{redis_user}:{redis_password}@' if redis_user else ''

        redis_url = f'rediss://{redis_auth}{redis_host}:{redis_port}/{redis_dbnumber}?ssl_cert_reqs=none' if \
            redis_ssl else f'redis://{redis_auth}{redis_host}:{redis_port}/{redis_dbnumber}'
        
        from nicechat.settingvalidation import validate_settings
        code, _msg = validate_settings(settings.DATABASES, redis_url, settings.PUBLIC_ROOT, settings.MEDIA_ROOT, settings.STATIC_ROOT)
        if code != 0:
            raise forms.ValidationError(f"{_msg}")

        return self.cleaned_data


class ForNewNodeForm(NiceViewMixin, forms.Form):
    pass


class TotpSkipForm(NiceViewMixin, forms.Form):
    otpbase = forms.CharField(widget=forms.HiddenInput())

    def __init__(self, user, *args, **kwargs):
        super(TotpSkipForm, self).__init__(*args, **kwargs)


class TotpSkip2Form(NiceViewMixin, forms.Form):

    def __init__(self, user, *args, **kwargs):
        super(TotpSkipForm, self).__init__(*args, **kwargs)


class TotpAskForm(NiceViewMixin, forms.Form):
    otpbase = forms.CharField(widget=forms.HiddenInput(), required=False)
    otp = forms.CharField(max_length=100, widget=forms.PasswordInput, required=True)

    def __init__(self, user, *args, **kwargs):
        super(TotpAskForm, self).__init__(*args, **kwargs)
        self._user = user


    def clean(self):
        otp = self.cleaned_data.get("otp")
        otpbase = self.cleaned_data.get("otpbase")
        if otpbase:
            if not ArbiterTOTP.validate_otp_static(otp, otpbase):
                raise forms.ValidationError("Wrong OTP")
        elif self._user:
            if not self._user.atotp.first() or not self._user.atotp.first().validate_otp(otp):
                raise forms.ValidationError("Wrong OTP")
        else:
            raise forms.ValidationError("Wrong OTP")

        return self.cleaned_data


class WebsitesStartClearForm(NiceViewMixin, forms.Form):
    notes = "You cannot directly transfer this Website to the selected Node, because this Node contains outdated version of this Website (the Node previously served this Website). To transfer the Website to this Node, you need to clear outdated data on the Node."


class WebsitesStartTransferForm(NiceViewMixin, forms.Form):
    # transfer_to = forms.ChoiceField(choices=[])
    
    def __init__(self, params, *args, **kwargs):
        self.request_params = params
        transfer_to_vals = params.get('transfer_to_vals')
        super(WebsitesStartTransferForm, self).__init__(*args, **kwargs)
        self.fields["transfer_to"] = forms.ChoiceField(
            choices=transfer_to_vals,
            widget=forms.Select(choices=transfer_to_vals, attrs={'class':'form-control'})
        )

    def clean_transfer_to(self):
        transfer_to = self.cleaned_data.get('transfer_to', '')
        if not transfer_to:
           raise forms.ValidationError("Target node is required")
        for _q, _ in self.request_params.get('transfer_to_vals', []):
            if f"{transfer_to}" == f"{_q}":
                return transfer_to
        raise forms.ValidationError("Specified target cannot be used")

class DeleteNodeForm(NiceViewMixin, forms.Form):
    name = forms.CharField(max_length=100)

    notes = 'Fill "name" of current node for deleting'

    def __init__(self, params, *args, **kwargs):
        self.request_params = params
        super(DeleteNodeForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        name = self.cleaned_data.get('name', '').lower()
        node = ArbiterNode.objects.get(pk=self.request_params.get('pk'))
        if node.name != name:
            raise forms.ValidationError("Wrond node 'name' value")
        return name


class CompanyEditForm(NiceViewMixin, forms.Form):
    label = forms.CharField(max_length=100, required=True)
    allow_branding = forms.BooleanField(label="Allow Customer to change widget branding (link and copyright)")


class EmailHostForm(NiceViewMixin, forms.Form):
    email_address = forms.EmailField(max_length=150)


class DeleteArbiterDomainForm(NiceViewMixin, forms.Form):
    domain = forms.CharField(max_length=150)

    notes = 'Fill "domain" of current domain for deleting'

    def __init__(self, params, *args, **kwargs):
        self.request_params = params
        super(DeleteArbiterDomainForm, self).__init__(*args, **kwargs)

    def clean_domain(self):
        domain = self.cleaned_data.get('domain', '').lower()
        if not domain == self.request_params.get('domain'):
            raise forms.ValidationError("Wrond 'domain' value")
        return domain
    

class DeleteHostForm(NiceViewMixin, forms.Form):
    host_address = forms.CharField(max_length=100)

    notes = 'Fill "host_address" of current host for deleting'

    def __init__(self, params, *args, **kwargs):
        self.request_params = params
        super(DeleteHostForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        host_address = self.cleaned_data.get('host_address', '').lower()
        host = ArbiterNodeHost.objects.get(pk=self.request_params.get('pk'))
        if host.host_address != host_address:
            raise forms.ValidationError("Wrond host 'host_address' value")
        return host_address


class DeleteUserForm(NiceViewMixin, forms.Form):
    email = forms.CharField(max_length=100)

    notes = 'Fill "email" of user for deleting'

    def __init__(self, manager, *args, **kwargs):
        self.request_manager = manager
        super(DeleteUserForm, self).__init__(*args, **kwargs)

    def clean_name(self):
        email = self.cleaned_data.get('email', '').lower()
        if self.request_manager.email != email:
            raise forms.ValidationError("Wrond 'email' value")
        return email


class ArbiterAuthenticationForm(NiceViewMixin, AuthenticationForm):
    totp = forms.CharField(required=False, widget=forms.PasswordInput(), label="OTP")

    def confirm_login_allowed(self, user):
        if not user.is_superuser:
            raise forms.ValidationError("This account cannot be used here.")
        super(ArbiterAuthenticationForm, self).confirm_login_allowed(user)

        if user.atotp.exists():
            code = self.cleaned_data.get('totp', None)
            if not code or len(code) != 6 or not code.isdigit():
                raise forms.ValidationError("TOTP code is requied (6 digit)")
            for totp in user.atotp.all():
                if totp.validate_otp(code):
                    return
            raise forms.ValidationError("Wrong TOTP code")



class AppearanceForm(NiceViewMixin, forms.Form):

    use_files = True
    opts = [(d, d,) for d in settings.ENV_ALL_DOMAINS]

    title = forms.CharField(max_length=150, required=False, label='Title')
    domain = forms.ChoiceField(choices=opts, label='Main domain')
    c_text = forms.CharField(max_length=150, required=False, label='Copyright text')
    c_url = forms.URLField(max_length=150, required=False, label='Copyright link')
    base_icon = forms.FileField(required=False)

    labels = {
        'title': 'Application',
        'c_text': 'Widget',
    }
    preview = None
    preview_txt = None


    def __init__(self, *args, **kwargs):
        # self.request_params = params
        # kwargs['initial'] = params
        if preview := kwargs.get('initial', {}).get('base_icon'):
            setattr(self, 'preview', preview)
            setattr(self, 'preview_txt', 'Current global icon')
        super(AppearanceForm, self).__init__(*args, **kwargs)
