import os
import time
import json
import base64
import datetime
from typing import Collection

from django.conf import settings
from django.db import models
from cryptography.hazmat.primitives.twofactor.totp import TOTP
from cryptography.hazmat.primitives.hashes import SHA1
from nicechat.utils.crypt import ncencrypt, ncdecrypt

class ArbiterSetting(models.Model):
    key = models.CharField(max_length=50, primary_key=True)
    val = models.CharField(max_length=200, blank=True, null=True)


class ArbiterNode(models.Model):
    NODE_MODE_CHOICES = (
        (settings.NODE_MODE_ARBITER, 'Arbiter',),
        (settings.NODE_MODE_WORKER, 'Worker',),
    )
    name=models.CharField(max_length=64, unique=True)
    mode=models.CharField(max_length=16, db_index=True, choices=NODE_MODE_CHOICES)
    for_new=models.BooleanField(default=False, db_index=True)
    credentials = models.TextField(null=True, blank=True, default="")
    credentials_hash = models.CharField(max_length=64, default="")
    description=models.TextField(default="-")

    def __str__(self):
        return f'Node {self.name} [{self.mode}]'
    

    def set_credentials(self, data):
        self.credentials = ncencrypt(json.dumps(data).encode(), settings.DATA_KEY_HS, settings.DATA_KEY_IV)

    def get_credentials(self):
        if not self.credentials:
            return {}
        udata = ncdecrypt(self.credentials, settings.DATA_KEY_HS, settings.DATA_KEY_IV)
        return json.loads(udata)


class ArbiterNodeHost(models.Model):
    node = models.ForeignKey(ArbiterNode, related_name='hosts', on_delete=models.CASCADE)
    host_pubkey = models.CharField(max_length=128)
    host_address = models.CharField(max_length=128)

    def __str__(self):
        return f"Node-host: {self.host_address} #{self.node_id}"


class ArbiterHostInfo(models.Model):
    address = models.CharField(max_length=128, primary_key=True)
    mode = models.CharField(max_length=20)
    timestamp = models.BigIntegerField(default=0, db_index=True)
    isok = models.BooleanField(default=True, db_index=True)
    hostname=models.CharField(max_length=196)
    host = models.ForeignKey(ArbiterNodeHost, related_name='hostinfo', blank=True, null=True, on_delete=models.CASCADE)
    stats = models.TextField(default='{}')

    @property
    def recency(self):
        return "%i secods ago" % (int(time.time()) - self.timestamp)

    def get_stat(self):
        try:
            return json.loads(self.stats)
        except Exception:
            return {}

    @property
    def error(self):
        return self.get_stat().get('error', '')

    def __str__(self):
        return f"Host: {self.address} {self.hostname} - {self.mode} {'on' if self.isok else 'off'}"


class ArbiterNodeWorkersStat(models.Model):
    record_key=models.CharField(max_length=256, unique=True)
    node = models.ForeignKey(ArbiterNode, related_name='workers', on_delete=models.CASCADE)
    hostname=models.CharField(max_length=196)
    host = models.ForeignKey(ArbiterNodeHost, related_name='workers', blank=True, null=True, on_delete=models.CASCADE)
    data_timestamp=models.BigIntegerField(default=0, db_index=True)
    stat_http_req=models.IntegerField(default=0)
    stat_http_s2s=models.IntegerField(default=0)
    stat_ws_req=models.IntegerField(default=0)
    stat_active_ws=models.IntegerField(default=0)
    stats = models.TextField(default='{}')

    @property
    def recency(self):
        return "%i secods ago" % (int(time.time()) - self.data_timestamp)

class ArbiterNodeWorkersStatHead(models.Model):
    record_key=models.CharField(max_length=256, unique=True)
    worker_hostname=models.CharField(max_length=196)
    worker_mode=models.CharField(max_length=64)
    worker_process=models.CharField(max_length=64)
    worker_node=models.CharField(max_length=64, db_index=True)
    worker_host=models.CharField(max_length=128, db_index=True)
    

class ArbiterNodeWorkersStatData(models.Model):
    head=models.ForeignKey(ArbiterNodeWorkersStatHead, related_name='data', on_delete=models.PROTECT)
    data_timestamp=models.BigIntegerField(default=0, db_index=True)
    stat_http_req=models.IntegerField(default=0)
    stat_http_s2s=models.IntegerField(default=0)
    stat_ws_req=models.IntegerField(default=0)
    stat_active_ws=models.IntegerField(default=0)


class ArbiterHostMap(models.Model):
    timestamp=models.BigIntegerField(db_index=True)
    data=models.JSONField(default=dict)

    @property
    def normalizetime(self):
        return datetime.datetime.fromtimestamp(int(self.timestamp)).strftime('%m/%d/%Y %H:%M:%S')


class ArbiterNodeWebsite(models.Model):
    STATUS_ACTIVE = 0
    STATUS_PRETARGET = 1
    STATUS_POSTSOURCE = 2
    STATUS_CHOICES = (
        (STATUS_ACTIVE, 'Active'),
        (STATUS_PRETARGET, 'Inactive (target-candidate)'),
        (STATUS_POSTSOURCE, 'Inactive (postmigration)'),
    )
    node = models.ForeignKey(ArbiterNode, related_name='node2website', on_delete=models.CASCADE)
    website = models.ForeignKey('website.Website', related_name='node2website', on_delete=models.CASCADE)
    status = models.IntegerField(default=STATUS_ACTIVE, choices=STATUS_CHOICES, db_index=True)

    @property
    def text_status(self):
        for val, txt in self.STATUS_CHOICES:
            if val == self.status:
                return txt
        return '-'
    class Meta:
        unique_together = ['node', 'website']


class UniqueChatId(models.Model):
    date_used = models.DateTimeField(auto_now_add=True)
    node_name=models.CharField(max_length=64)



class ArbiterTOTP(models.Model):
    user = models.ForeignKey('users.User', related_name="atotp", on_delete=models.CASCADE)
    secret = models.CharField(max_length=64)
    created = models.DateTimeField(auto_now_add=True)
    INTERVAL = 30
    DIGITS = 6
    ALG = SHA1

    @staticmethod
    def create_new(user, secret=None):
        if not secret:
            secret = ArbiterTOTP.get_token()
        ArbiterTOTP.objects.create(
            user=user,
            secret=secret
        )

    def get_topt(self):
        return TOTP(base64.b32decode(self.secret), ArbiterTOTP.DIGITS, ArbiterTOTP.ALG(), ArbiterTOTP.INTERVAL)

    @staticmethod
    def get_token():
        return base64.b32encode(os.urandom(20)).decode()

    @staticmethod
    def get_url(secret, email):
        return TOTP(base64.b32decode(secret), ArbiterTOTP.DIGITS, ArbiterTOTP.ALG(), ArbiterTOTP.INTERVAL).get_provisioning_uri(email, 'NC-Arbiter')

    def get_user_url(self, secret=None):
        return self.get_topt().get_provisioning_uri(self.user.email, 'NC-Arbiter')

    @staticmethod
    def validate_otp_static(code, secret):
        return ArbiterTOTP._validate_otp(code, TOTP(base64.b32decode(secret), ArbiterTOTP.DIGITS, ArbiterTOTP.ALG(), ArbiterTOTP.INTERVAL))

    def validate_otp(self, code):
        return ArbiterTOTP._validate_otp(code, self.get_topt())

    @staticmethod
    def _validate_otp(code, totp):
        ts = time.time()
        for r in [0, -30, -60, 30]:
            if code.encode() == totp.generate(ts+r):
                return True
        return False


class ArbiterWebsiteTransfer(models.Model):
    node_from = models.ForeignKey(ArbiterNode, related_name='transfer_to', on_delete=models.CASCADE)
    node_to = models.ForeignKey(ArbiterNode, related_name='transfer_from', on_delete=models.CASCADE)
    website = models.ForeignKey('website.Website', related_name='transfers', on_delete=models.CASCADE)
    step = models.IntegerField(db_index=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    fished = models.DateTimeField(blank=True, null=True, db_index=True)

    def run(self):
        from apps.balancer.transfer import transfer_execute
        transfer_execute(self)

    @property
    def progress(self):
        if self.step == 100:
            return 'completed'
        else:
            return int(self.step/70.0*100)

    def log(self, text):
        ArbiterWebsiteTransferLog.objects.create(
            info=text, 
            transfer=self,
            step=self.step
        )


class ArbiterWebsiteTransferPK(models.Model):
    transfer = models.IntegerField(db_index=True)
    class_name = models.CharField(max_length=256, db_index=True)
    old_id = models.BigIntegerField(db_index=True)    
    new_id = models.BigIntegerField()    
    class Meta:
        unique_together = ['transfer', 'class_name', 'old_id']


class ArbiterWebsiteTransferLog(models.Model):
    transfer = models.ForeignKey(ArbiterWebsiteTransfer, related_name='logs', on_delete=models.CASCADE)
    step = models.IntegerField()
    timestamp = models.DateTimeField(auto_now_add=True)
    info = models.TextField(default="")


class ArbiterProps(models.Model):
    key = models.CharField(max_length=45, primary_key=True)
    data = models.TextField(default="")
    class Meta:
        managed=False
        db_table='arbiter_settings'


class ArbiterHostDomains(models.Model):
    domain = models.CharField(max_length=150, primary_key=True)
    data = models.TextField(default="{}")
    class Meta:
        managed=False
        db_table='arbiter_hostdomains'
