import string
import random
import time
import traceback
import json
import re
import base64
import socket
import hashlib
from datetime import datetime
from django.views.generic import View, TemplateView, FormView
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.db.models import Count, Sum, Min, F, Value, Avg, Max
from django.views.decorators.csrf import csrf_exempt
from django.urls import reverse
from django.conf import settings
from django.contrib.auth.views import LoginView, LogoutView
from nicechat.utils.arbiter import make_new_node_key, get_fullnode_key, update_data_workernode, host_exec
from apps.balancer.models import ArbiterNode, ArbiterNodeHost, ArbiterHostMap, ArbiterNodeWorkersStat, \
    ArbiterWebsiteTransfer, ArbiterNodeWorkersStatData, ArbiterTOTP, ArbiterNodeWebsite, ArbiterSetting, \
    ArbiterProps, ArbiterHostDomains, ArbiterHostInfo
from apps.balancer.forms import CreateNodeForm, ArbiterAuthenticationForm, CreateManagerForm, AssignHostForm, \
    DeleteNodeForm, CredentailsNodeForm, ForNewNodeForm, WebsitesStartTransferForm, PasswordChangeForm, \
    SetPasswordForm, DeleteUserForm, TotpSkipForm, TotpSkip2Form, TotpAskForm, WebsitesStartClearForm, \
    AppearanceForm, DeleteHostForm, EmailHostForm, CompanyEditForm, CredentailsArbiterForm, DomainEditForm, \
    DeleteArbiterDomainForm
from apps.website.models import Website
from apps.company.models import Company
from apps.users.models import User
from apps.balancer.utils import save_site_icon
from nicechat.utils.crypt import ncdecrypt, ncencrypt

wsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
wsock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

class SuperLoginRequired:

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated or not request.user.is_superuser:
            return HttpResponseNotFound()
        return super().dispatch(request, *args, **kwargs)

class NodesInviteView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = CreateNodeForm

    def get_success_url(self):
        return reverse('arbiter:main')

    def get_context_data(self, **kwargs):
        context = super(NodesInviteView, self).get_context_data(**kwargs)
        context['name'] = 'Add new node'
        return context

    def form_valid(self, form):
        for_new = False if ArbiterNode.objects.filter(mode=settings.NODE_MODE_WORKER).exists() else True
        node = ArbiterNode.objects.create(
            name=form.cleaned_data['name'],
            mode=settings.NODE_MODE_WORKER,
            description=form.cleaned_data['description'],
            for_new=for_new
        )
        return HttpResponseRedirect(reverse('arbiter:node_creds', kwargs={'pk': node.pk}))
    

class NodesCredentailsView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = CredentailsNodeForm
    PWD_PLACEHOLDER = '********'

    def get_success_url(self):
        return reverse('arbiter:main')

    def get_context_data(self, **kwargs):
        context = super(NodesCredentailsView, self).get_context_data(**kwargs)
        node = ArbiterNode.objects.get(pk=self.kwargs['pk'])
        context['name'] = f'Credentials for node {node.name}'
        context['node'] = node
        return context

    def get_form_kwargs(self):
        form_kwargs = super(NodesCredentailsView, self).get_form_kwargs()
        node = ArbiterNode.objects.get(pk=self.kwargs['pk'])
        creds = node.get_credentials()
        creds['postgres_password'] = self.PWD_PLACEHOLDER
        creds['redis_password'] = self.PWD_PLACEHOLDER if creds.get('redis_password') else ''
        form_kwargs['initial'] = creds
        return form_kwargs

    def form_valid(self, form):
        data = form.cleaned_data
        node = ArbiterNode.objects.get(pk=self.kwargs['pk'])
        creds = node.get_credentials()
        if data.get('postgres_password') == self.PWD_PLACEHOLDER:
            data['postgres_password'] = creds.get('postgres_password')
        if data.get('redis_password') == self.PWD_PLACEHOLDER:
            data['redis_password'] = creds.get('redis_password')
        node.set_credentials(data)
        node.save()
        wsock.sendto(f'confupdate:{node.name}'.encode(), ('127.0.0.1', 5006))
        return HttpResponseRedirect(reverse('arbiter:node_det', kwargs={'pk': node.pk}))
    

class ArbiterDomainsView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/domains.html"

    def get_context_data(self, **kwargs):
        context = super(ArbiterDomainsView, self).get_context_data(**kwargs)
        domains = []
        from nicechat.utils.crypt import decrypt_dict_v
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.serialization import load_pem_private_key

        for d in ArbiterHostDomains.objects.all():
            try:
                data = decrypt_dict_v(json.loads(d.data), settings.DATA_KEY_HS, settings.DATA_KEY_IV)
                peers = x509.load_pem_x509_certificates(data.get('crt', '').encode())
                private_key = load_pem_private_key(data.get('key', '').encode(), None, default_backend())
            except Exception as e:
                domains.append(
                    {
                        'domain': d.domain,
                        'crtinfo': [{
                            'subject': 'Error',
                            'serial_number': 'n/a',
                            'public_key': str(e)
                        }]
                    }
                )
                continue

            crtinfo = []
            priv_key = self.get_pub_hash(private_key.public_key())
            match_pk = False
            for peer in peers:
                pub = self.get_pub_hash(peer.public_key())
                # subject = []
                subject = peer.subject.rfc4514_string() if isinstance(peer.subject, x509.Name) else f"{peer.subject}"
                         

                alt_names = []
                for e in peer.extensions:
                    if isinstance(e.value, x509.SubjectAlternativeName):
                        alt_names = e.value.get_values_for_type(x509.DNSName)
                ci = {
                    "subject": f"{subject}",
                    "serial_number": f"{peer.serial_number}",
                    "public_key": f"{pub}",
                }
                if alt_names:
                    ci['alt_names'] = ", ".join(alt_names)
                if pub == priv_key:
                    ci['match_pk'] = True
                    match_pk = True
                crtinfo.append(ci)

            domains.append(
                {
                    'domain': d.domain,
                    'crtinfo': crtinfo
                }
            )

        context['domains'] = domains
        context['match_pk'] = match_pk
        return context

    def get_pub_hash(self, public_key):
        from cryptography.hazmat.primitives import serialization
        k = public_key.public_bytes(serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo)
        k = re.sub(b"-----[^\-]+-----", b"", k).replace(b"\n", b"")
        # print(k)
        return hashlib.sha256(base64.b64decode(k)).hexdigest()

class ArbiterDomainEditView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = DomainEditForm

    def get_form_kwargs(self):
        form_kwargs = super(ArbiterDomainEditView, self).get_form_kwargs()
        domain = ArbiterHostDomains.objects.filter(domain=self.kwargs.get('domain')).first()
        if domain:
            form_kwargs['initial'] = {'domain': domain.domain}
        return form_kwargs

    def get_context_data(self, **kwargs):
        context = super(ArbiterDomainEditView, self).get_context_data(**kwargs)
        domain = ArbiterHostDomains.objects.filter(domain=self.kwargs.get('domain')).first()
        if domain:
            context['name'] = f'Edit host domain {domain.domain}'
            context['buttons'] = [
                {'url': reverse('arbiter:domains_delete', kwargs={'domain': domain.domain}), 'label': 'Delete'}
            ]
        else:
            context['name'] = 'Add new host domain'
        return context

    def form_valid(self, form):

        domain = ArbiterHostDomains.objects.filter(domain=self.kwargs.get('domain')).first()
        refresh_all = False
        if not domain:
            refresh_all = True
            domain = ArbiterHostDomains(domain=form.cleaned_data['domain'])
            from nicechat.settingvalidation import generate_selfsigned_cert
            certificate, privatekey = generate_selfsigned_cert(domain.domain, add_sub=True)
            certificate = certificate.decode()
            privatekey = privatekey.decode()
            from nicechat.utils.crypt import encrypt_dict_v
        
            domain.data = json.dumps(encrypt_dict_v(
                {
                    'crt': certificate, 
                    'key': privatekey
                },
                settings.DATA_KEY_HS, settings.DATA_KEY_IV))
        elif domain.domain != form.cleaned_data['domain']:
            domain.domain = form.cleaned_data['domain']
            refresh_all = True

        # print(form.cleaned_data)

        certificate = form.cleaned_data.get('certificate', "")
        privatekey = form.cleaned_data.get('privatekey', "")

        if privatekey and certificate:
            from nicechat.utils.crypt import encrypt_dict_v
        
            domain.data = json.dumps(encrypt_dict_v(
                {
                    'crt': certificate, 
                    'key': privatekey
                },
                settings.DATA_KEY_HS, settings.DATA_KEY_IV))
            
        domain.save()
        
        if (privatekey and certificate) or refresh_all:
            hm = ArbiterHostMap.objects.last()
            ArbiterHostMap.objects.create(
                timestamp=int(time.time()),
                data=hm.data
            )
        if refresh_all:
            wsock.sendto(b'confupdate:all', ('127.0.0.1', 5006))

        return HttpResponseRedirect(reverse('arbiter:domains_list'))


class ArbiterDomainDeleteView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = DeleteArbiterDomainForm

    def get_success_url(self):
        return reverse('arbiter:domains_list')

    def get_context_data(self, **kwargs):
        context = super(ArbiterDomainDeleteView, self).get_context_data(**kwargs)
        domain = ArbiterHostDomains.objects.filter(domain=self.kwargs['domain']).first()
        if ArbiterHostDomains.objects.all().count() == 1:
            context['message'] = 'You cannot delete once domain'
        else:
            context['name'] = f'Do you realy want to delete domain "{domain.domain}"?'
            context['domain'] = domain
        return context
    
    def get_form_kwargs(self):
        form_kwargs = super(ArbiterDomainDeleteView, self).get_form_kwargs()
        form_kwargs['params'] = self.kwargs
        return form_kwargs

    def form_valid(self, form):
        domain = form.cleaned_data.get('domain', '').lower()
        domain_rec = ArbiterHostDomains.objects.get(domain=self.kwargs.get('domain'))
        if domain_rec.domain == domain:
            domain_rec.delete()
            wsock.sendto(b'confupdate:all', ('127.0.0.1', 5006))
        return HttpResponseRedirect(reverse('arbiter:domains_list'))


class ArbiterCredentailsView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = CredentailsArbiterForm

    PWD_PLACEHOLDER = '*******'

    def get_success_url(self):
        return reverse('arbiter:main')

    def get_context_data(self, **kwargs):
        context = super(ArbiterCredentailsView, self).get_context_data(**kwargs)
        context['name'] = f'System props'
        return context

    def get_form_kwargs(self):
        form_kwargs = super(ArbiterCredentailsView, self).get_form_kwargs()
        creds = {
            m.key: ncdecrypt(m.data, settings.DATA_KEY_HS, settings.DATA_KEY_IV, tostr=True) for m in ArbiterProps.objects.all()
        }
        creds['redis_password'] = self.PWD_PLACEHOLDER if creds.get('redis_password') else ''
        creds['smtp_password'] = self.PWD_PLACEHOLDER if creds.get('smtp_password') else ''
        creds['redis_ssl'] = True if creds.get('redis_ssl') in ['1', 'true', 'yes'] else False
        form_kwargs['initial'] = creds
        return form_kwargs

    def form_valid(self, form):
        creds = [
            m.key for m in ArbiterProps.objects.all()
        ]
        for key, val in form.cleaned_data.items():
            if key in ['redis_password', 'smtp_password'] and val == self.PWD_PLACEHOLDER:
                if key in creds:
                    creds.remove(key)
                continue
            if isinstance(val, bool):
                val = '1' if val else '0'
            if not isinstance(val, str):
                val = f"{val}"
            ArbiterProps.objects.update_or_create(
                key = key,
                defaults={
                    'data': ncencrypt(val.encode(), settings.DATA_KEY_HS, settings.DATA_KEY_IV)
                }
            )
            if key in creds:
                creds.remove(key)
        ArbiterProps.objects.filter(key__in=creds).delete()
        wsock.sendto(b'confupdate:all', ('127.0.0.1', 5006))
        return HttpResponseRedirect(reverse('arbiter:arbiter_settings'))


class AssignHostView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = AssignHostForm

    def get_context_data(self, **kwargs):
        context = super(AssignHostView, self).get_context_data(**kwargs)
        host = ArbiterHostInfo.objects.get(pk=self.kwargs['address'])
        context['name'] = f'Assign host {host.address} to node'
        context['host'] = host
        return context

    def form_valid(self, form):
        host = ArbiterHostInfo.objects.get(pk=self.kwargs['address'])
        token = ''.join(random.choice(string.digits + string.ascii_uppercase + string.ascii_lowercase) for i in range(32))

        assign_to = form.cleaned_data.get('assign_to')
        node = ArbiterNode.objects.get(pk=assign_to)
        stat = {
            'status': 'initassign',
            'ts': int(time.time()),
        }
        host.status = json.dumps(stat)
        host.save()
        print(stat)
        import requests
        error = False
        try:
            req = requests.post(
                f'https://{host.address}:8000/initassign/',
                data={},
                verify=False,
                timeout=10)
            print(req)
            if req.status_code != 202:
                error = True
            else:
                node_pub_key_b = req.content
                from cryptography.hazmat.primitives.asymmetric import x448
                from cryptography.hazmat.primitives import serialization
                node_pub_key_b_obj = x448.X448PublicKey.from_public_bytes(node_pub_key_b)
                data = {}
                if node.mode == settings.NODE_MODE_WORKER:
                    data = {
                        'NODE': node.name,
                        'APUB': settings.NODE_ARBITER_PUB,
                    }
                    acpath = 'applyassign/worker'
                elif node.mode == settings.NODE_MODE_ARBITER:
                    acpath = 'applyassign/arbiter'
                    data = {
                        'ARBITER_CA': settings.ENV_ARBITER_CA,
                        'BALANCER_SIGN': settings.ENV_BALANCER_SIGN,
                        'SECRET_KEY': settings.ENV_SECRET_KEY,
                        'DATA_KEY': settings.ENV_DATA_KEY,
                        'POSTGRES_NAME': settings.ENV_POSTGRES_NAME, 
                        'POSTGRES_USER': settings.ENV_POSTGRES_USER, 
                        'POSTGRES_PASSWORD': settings.ENV_POSTGRES_PASSWORD, 
                        'POSTGRES_HOST': settings.ENV_POSTGRES_HOST, 
                    }
                else:
                    error = True
                    data = None
                    acpath = ''

                if data:
                    print(data)
                    from nicechat.utils.arbiter import arb_encrypt
                    _data = arb_encrypt(json.dumps(data), settings.NODE_PRIVATE_KEY, node_pub_key_b_obj)
                    print("!!!", acpath, 'aa')
                    req2 = requests.post(
                        f'https://{host.address}:8000/{acpath}/',
                        data=_data,
                        headers={"x-arbiter": settings.NODE_ARBITER_PUB},
                        verify=False,
                        timeout=10)
                    print("aa>", req2)
                    if req2.status_code != 201:
                        error = True
                    else:
                        if node.mode == settings.NODE_MODE_WORKER:
                            nodepubkey = node_pub_key_b_obj
                            pubkey = node_pub_key_b_obj.public_bytes(
                                encoding=serialization.Encoding.Raw,
                                format=serialization.PublicFormat.Raw
                            )
                            pubkey = base64.b32encode(b'}\x02\xb4\x08' + pubkey).decode()
                            host_x = ArbiterNodeHost.objects.create(
                                node=node,
                                host_pubkey=pubkey,
                                host_address=host.address
                            )
                        # req3 = requests.post(
                        #     f'https://{host.address}:8000/applyreboot/',
                        #     data={},
                        #     verify=False,
                        #     timeout=10)
                        # print(req.content)
                        return HttpResponseRedirect(reverse('arbiter:hosts_list'))

            
        except Exception as e:
            traceback.print_exception(e)
            error = True

        if error:
            return self.get(self.request, *self.args, **self.kwargs)

        return HttpResponseRedirect(reverse('arbiter:host_info', kwargs={'address': host.address}))


class HostInfoView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/hostinfo.html"

    def get_context_data(self, **kwargs):
        context = super(HostInfoView, self).get_context_data(**kwargs)
        context['host'] = ArbiterHostInfo.objects.get(pk=self.kwargs['address'])
        return context


class HostDetailView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/host.html"

    def get_context_data(self, **kwargs):
        context = super(HostDetailView, self).get_context_data(**kwargs)
        host = ArbiterNodeHost.objects.get(pk=self.kwargs.get('pk'))
        context['host'] = host
        context['node'] = host.node
        ret = host_exec(host, 'getstat', {})
        
        if ret and isinstance(ret, str):
            try:
                _ret = json.loads(ret)
                _ret = _ret['data']
                ret = _ret
            except Exception:
                pass
        context['state'] = ret
        return context


class HostDeleteView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = DeleteHostForm

    def get_success_url(self):
        return reverse('arbiter:managers_list')

    def get_context_data(self, **kwargs):
        context = super(HostDeleteView, self).get_context_data(**kwargs)
        host = ArbiterNodeHost.objects.get(pk=self.kwargs.get('pk'))
        context['name'] = f'Do you realy want to delete Host "{host.host_address}"?'
        context['host'] = host
        return context
    
    def get_form_kwargs(self):
        form_kwargs = super(HostDeleteView, self).get_form_kwargs()
        form_kwargs['params'] = self.kwargs
        return form_kwargs

    def form_valid(self, form):
        host_address = form.cleaned_data.get('host_address', '').lower()
        host = ArbiterNodeHost.objects.get(pk=self.kwargs.get('pk'))
        if host.host_address == host_address:
            host.delete()
        return HttpResponseRedirect(reverse('arbiter:node_det', kwargs={'pk': host.node_id}))


class HostEmailView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = EmailHostForm

    def get_success_url(self):
        return reverse('arbiter:managers_list')

    def get_context_data(self, **kwargs):
        context = super(HostEmailView, self).get_context_data(**kwargs)
        context['name'] = f'Send test email'
        return context

    def form_valid(self, form):
        email_address = form.cleaned_data.get('email_address', '').lower()
        host = ArbiterNodeHost.objects.get(pk=self.kwargs.get('pk'))
        ret = host_exec(host, 'testemail', {'email': email_address})
        
        if ret and isinstance(ret, str):
            try:
                _ret = json.loads(ret)
                _ret = _ret['txt']
                ret = _ret
            except Exception:
                pass

        context = self.get_context_data()
        context['message'] = ret

        return self.render_to_response(context)


class WebsitesDetailView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/website.html"

    def get_context_data(self, **kwargs):
        context = super(WebsitesDetailView, self).get_context_data(**kwargs)
        website = Website.objects.get(pk=self.kwargs.get('pk'))
        context['website'] = website
        return context


class NodeDeleteView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = DeleteNodeForm

    def get_success_url(self):
        return reverse('arbiter:managers_list')

    def get_context_data(self, **kwargs):
        context = super(NodeDeleteView, self).get_context_data(**kwargs)
        node = ArbiterNode.objects.get(pk=self.kwargs.get('pk'))
        context['name'] = f'Do you realy want to delete Node "{node.name}"?'
        context['node'] = node
        return context
    
    def get_form_kwargs(self):
        form_kwargs = super(NodeDeleteView, self).get_form_kwargs()
        form_kwargs['params'] = self.kwargs
        return form_kwargs

    def form_valid(self, form):
        name = form.cleaned_data.get('name', '').lower()
        node = ArbiterNode.objects.get(pk=self.kwargs.get('pk'))
        if node.name == name:
            node.delete()
        return HttpResponseRedirect(reverse('arbiter:nodes_list'))


class AppearancePermissionsView(SuperLoginRequired, TemplateView):
    template_name = 'base/form.html'


class AppearanceView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = AppearanceForm

    PROPS = ['title', 'domain', 'c_text', 'c_url', 'base_icon']

    def get_success_url(self):
        return reverse('arbiter:system_appearance')
    
    def get_form_kwargs(self):
        form_kwargs = super(AppearanceView, self).get_form_kwargs()
        _g_props = [f'global_{f}' for f in self.PROPS]
        form_kwargs['initial'] = {a.key[7:]: a.val for a in
        ArbiterSetting.objects.filter(
            key__in=_g_props
        )}
        return form_kwargs

    def get_context_data(self, **kwargs):
        context = super(AppearanceView, self).get_context_data(**kwargs)
        context['name'] = 'Edit appearance'
        context['buttons'] = [
            {'url': reverse('arbiter:system_appearance_perm'), 'label': 'Edit permissions'}
        ]
        return context

    def form_valid(self, form):
        _g_props = [f'global_{f}' for f in self.PROPS]
        data = {a.key[7:]: a.val for a in
        ArbiterSetting.objects.filter(
            key__in=_g_props
        )}
        for k, v in form.cleaned_data.items():
            if k != 'base_icon':
                data[k] = v
            else:
                if not isinstance(v, str):
                    content = v.read()
                    _s, _v = save_site_icon(content)
                    if _s:
                        data[k] = _v 
        if not data.get('title'):
            data['title'] = 'Batam Inc.'
        if not data.get('c_text'):
            data['c_text'] = data['title']
        if not data.get('c_url'):
            data['c_url'] = f"https://{data['domain']}/"

        for f in self.PROPS:
            ArbiterSetting.objects.update_or_create(
                key=f'global_{f}',
                defaults={
                    'val': data[f]
                }
            )
        
        return HttpResponseRedirect(reverse('arbiter:system_appearance'))


class NodeDetailView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/node.html"

    def get_context_data(self, **kwargs):
        context = super(NodeDetailView, self).get_context_data(**kwargs)
        node =  ArbiterNode.objects.get(pk=self.kwargs.get('pk'))
        context['node'] = node
        context['websites'] = node.node2website.all()
        context['hosts'] = node.hosts.all().annotate(
            workers_count=Count('workers')
        )
        return context


class AgentDetailView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/agent.html"

    def get_context_data(self, **kwargs):
        context = super(AgentDetailView, self).get_context_data(**kwargs)
        agent =  User.objects.get(is_superuser=False, pk=self.kwargs.get('pk'))
        context['agent'] = agent
        context['websites'] = agent.website_set.all()
        return context


class NodeForNewSwitchView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = ForNewNodeForm

    def get_success_url(self):
        return reverse('arbiter:managers_list')

    def get_context_data(self, **kwargs):
        context = super(NodeForNewSwitchView, self).get_context_data(**kwargs)
        node = ArbiterNode.objects.get(pk=self.kwargs.get('pk'))

        context['node'] = node
        if node.for_new:
            if not ArbiterNode.objects.filter(for_new=True).exclude(pk=node.pk).exists():
                context['message'] = 'At least one node must to accept new customers!'
                context['continue'] = reverse('arbiter:node_det', kwargs={'pk': node.pk})
                return context
            context['name'] = f'Do you realy want to stop accepting new customers at "{node.name}"?'
        else:
            context['name'] = f'Do you realy want accepting new customers at "{node.name}"?'
        return context

    def form_valid(self, form):
        node = ArbiterNode.objects.get(pk=self.kwargs.get('pk'))
        node.for_new = not node.for_new
        node.save()
        return HttpResponseRedirect(reverse('arbiter:node_det', kwargs={'pk': node.pk}))


class AgentSwitchView(SuperLoginRequired, View):
    def post(self, request, *args, **kwargs):
        agent =  User.objects.get(is_superuser=False, pk=self.kwargs.get('pk'))
        agent.is_active = self.kwargs.get('action', 'on') == 'on'
        agent.save()
        ret = request.META.get('HTTP_REFERER', None)
        mgrs = reverse('arbiter:agents_list')
        if not ret or not mgrs in ret:
            ret = mgrs
        return HttpResponseRedirect(ret)


class AgentPwdView(SuperLoginRequired, FormView):
    template_name = "base/form.html"
    form_class = SetPasswordForm

    def get_context_data(self, **kwargs):
        context = super(AgentPwdView, self).get_context_data(**kwargs)
        agent =  User.objects.get(is_superuser=False, pk=self.kwargs.get('pk'))
        context['name'] = f'Change password for {agent}'
        return context
    
    def get_form_kwargs(self):
        form_kwargs = super(AgentPwdView, self).get_form_kwargs()
        agent =  User.objects.get(is_superuser=False, pk=self.kwargs.get('pk'))
        form_kwargs['user'] = agent
        return form_kwargs

    def form_valid(self, form):
        agent =  User.objects.get(is_superuser=False, pk=self.kwargs.get('pk'))
        agent.set_password(form.cleaned_data['new_password1'])
        agent.save()
        return HttpResponseRedirect(reverse('arbiter:agent_det', kwargs={'pk': agent.pk}))


class ManagerDetailView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/manager.html"

    def get_context_data(self, **kwargs):
        context = super(ManagerDetailView, self).get_context_data(**kwargs)
        manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        context['manager'] = manager
        return context


class ManagerPwdView(SuperLoginRequired, FormView):
    template_name = "base/form.html"

    def get_form_class(self):
        if self.kwargs.get('pk') != self.request.user.pk: 
            return SetPasswordForm
        else:
            return PasswordChangeForm
    
    def get_form_kwargs(self):
        form_kwargs = super(ManagerPwdView, self).get_form_kwargs()
        manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        form_kwargs['user'] = manager
        return form_kwargs

    def get_context_data(self, **kwargs):
        context = super(ManagerPwdView, self).get_context_data(**kwargs)
        manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        context['manager'] = manager
        context['name'] = f'Change password for {manager}'
        return context

    def form_valid(self, form):
        manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        manager.set_password(form.cleaned_data['new_password1'])
        manager.save()
        if manager.pk == self.request.user.pk:
            return HttpResponseRedirect(reverse('arbiter:main'))
        return HttpResponseRedirect(reverse('arbiter:manager_det', kwargs={'pk': manager.pk}))


class ManagerOtpView(SuperLoginRequired, FormView):
    template_name = "base/form.html"

    def get_form_class(self):
        if self.kwargs.get('pk') != self.request.user.pk: 
            manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
            if manager.atotp.exists():
                return TotpSkip2Form
            else:
                return TotpSkipForm
        else:
            return TotpAskForm

    def get_form_kwargs(self):
        form_kwargs = super(ManagerOtpView, self).get_form_kwargs()
        manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        form_kwargs['user'] = manager
        return form_kwargs

    def get_context_data(self, **kwargs):
        manager = User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        form_class = self.get_form_class()

        if manager.atotp.exists():
            form_name = f'Disable OTP for {manager}'
            initial={}
            qr_token = None
        else:
            form_name = f'Enable OTP for {manager}'
            qr_token = ArbiterTOTP.get_token()            
            initial = {'otpbase': qr_token}

        form_kwargs = self.get_form_kwargs()
        form_kwargs['initial'] = initial
        form = form_class(**form_kwargs)
        context = super(ManagerOtpView, self).get_context_data(form=form)

        context['manager'] = manager
        context['name'] = form_name
        if qr_token:
            context['qr'] = ArbiterTOTP.get_url(qr_token, str(manager))
        return context


    def form_valid(self, form):
        manager = User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        otpbase = form.cleaned_data.get('otpbase', None)
        if otpbase:
            ArbiterTOTP.objects.create(
                secret=otpbase,
                user=manager
            )
        else:
            manager.atotp.first().delete()
        return HttpResponseRedirect(reverse('arbiter:manager_det', kwargs={'pk': manager.pk}))



class ManagerSwitchView(SuperLoginRequired, View):

    def post(self, request, *args, **kwargs):
        manager =  User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        manager.is_active = self.kwargs.get('action', 'on') == 'on'
        manager.save()
        ret = request.META.get('HTTP_REFERER', None)
        mgrs = reverse('arbiter:managers_list')
        if not ret or not mgrs in ret:
            ret = mgrs
        return HttpResponseRedirect(ret)


class ManagerDelView(SuperLoginRequired, FormView):
    template_name = "base/form.html"
    form_class = DeleteUserForm


    def get_context_data(self, **kwargs):
        context = super(ManagerDelView, self).get_context_data(**kwargs)
        manager = User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        context['name'] = f'Do you realy want to delete "{manager}"?'
        context['manager'] = manager
        return context
    
    def get_form_kwargs(self):
        form_kwargs = super(ManagerDelView, self).get_form_kwargs()
        manager = User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        form_kwargs['manager'] = manager
        return form_kwargs

    def form_valid(self, form):
        email = form.cleaned_data.get('email', '').lower()
        manager = User.objects.get(is_superuser=True, pk=self.kwargs.get('pk'))
        if manager.email == email:
            manager.delete()
        return HttpResponseRedirect(reverse('arbiter:managers_list'))


class ManagerAddView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class = CreateManagerForm

    def get_success_url(self):
        return reverse('arbiter:managers_list')

    def get_context_data(self, **kwargs):
        context = super(ManagerAddView, self).get_context_data(**kwargs)
        context['name'] = 'Add manager'
        return context

    def form_valid(self, form):
        user = User.objects.create(
            first_name=form.cleaned_data.get('first_name', '-'),
            last_name=form.cleaned_data.get('last_name', '-'),
            email=form.cleaned_data['email'],
            is_superuser=True,
        )

        pwd = form.cleaned_data.get('password', None)

        autopwd = True if not pwd else False

        if autopwd:
            pwd = ''.join(random.choice(string.digits + string.ascii_uppercase + string.ascii_lowercase) for i in range(8))

        user.set_password(pwd)
        user.save()

        if autopwd:
            return self.render_to_response({
                "continue": reverse('arbiter:manager_det', kwargs={'pk': user.pk}),
                "message": f"New password {pwd}"
            })

        return HttpResponseRedirect(reverse('arbiter:manager_det', kwargs={'pk': user.pk}))


class DashboardView(TemplateView):
    template_name = "arbiter/index.html"
    def get_context_data(self, **kwargs):
        if not self.request.user.is_authenticated or not self.request.user.is_superuser:
            return {}
        context = super(DashboardView, self).get_context_data(**kwargs)
        context['stats'] = ArbiterNodeWorkersStat.objects.annotate(
            dummy_group_by = Value(1)).values('dummy_group_by').annotate(
            host_name=F('hostname'),
            node_name=F('node__name'),
            s_http_req=Sum('stat_http_req') * 2,
            s_http_s2s=Sum('stat_http_s2s') * 2,
            s_ws_req=Sum('stat_ws_req') * 2,
            s_active_ws=Max('stat_active_ws'),
            workers=Count('record_key'),
            min_data=int(time.time())-Min('data_timestamp'),
        ).values(
            'host_name', 'node_name', 's_http_req', 's_http_s2s',
            's_ws_req', 's_active_ws', 'workers', 'min_data', 
        ).order_by('node_name', 'host_name')

        dts = {}
        _nodes = set()
        _ts = set()
        for rec in ArbiterNodeWorkersStatData.objects.annotate(
            dummy_group_by = Value(1)).values('dummy_group_by').annotate(
            worker_node=F('head__worker_node'),
            ts=F('data_timestamp') / 10800,
            stat_http_req_sum=Sum('stat_http_req'),
            stat_ws_req_sum=Sum('stat_ws_req'),
            stat_active_ws_avg=Avg('stat_active_ws'),
        ).filter(
            data_timestamp__gte = int(time.time() - 432000)
        ).values('worker_node', 'ts', 'stat_http_req_sum', 'stat_ws_req_sum', 'stat_active_ws_avg'):
            dts[(rec['worker_node'], rec['ts'], 'http_req')] = rec['stat_http_req_sum']
            dts[(rec['worker_node'], rec['ts'], 'ws_req')] = rec['stat_ws_req_sum']
            dts[(rec['worker_node'], rec['ts'], 'active_ws')] = rec['stat_active_ws_avg']
            _nodes.add(rec['worker_node'])
            _ts.add(rec['ts'])

        _nodes = list(sorted(list(_nodes)))
        _ts = list(sorted(list(_ts)))
        context['ts'] = [
            datetime.fromtimestamp(_t*10800).strftime('%b %d, %H:00') for _t in _ts
        ]
        nodes = {'http_req': {},'ws_req': {},'active_ws': {},}
        for n in _nodes:
            _http_req = []
            _active_ws = []
            _ws_req = []
            for _t in _ts:
                _http_req.append(dts.get((n, _t, 'http_req'), 0))
                _ws_req.append(dts.get((n, _t, 'ws_req'), 0))
                _active_ws.append(dts.get((n, _t, 'active_ws'), 0))
                
            nodes['http_req'][n] = {
                'name': n,
                'data': _http_req
            }
            nodes['active_ws'][n] = {
                'name': n,
                'data': _active_ws
            }
            nodes['ws_req'][n] = {
                'name': n,
                'data': _ws_req
            }
        context['charts'] = nodes
        context['nodes'] = ArbiterNode.objects.all()

        ws_stats = {}

        websites = {str(w.pk):w.title for w in Website.objects.all()}

        for rec in ArbiterNodeWorkersStat.objects.all():
            if rec.stats:
                dt = json.loads(rec.stats)
                for _k, _v  in dt.items():
                    if _v and _k and ':' in _k:
                        _, _ws = _k.split(':', 1)
                        _ws = websites.get(_ws, f'#{_ws}')                      
                        ws_stats[rec.node.name] = ws_stats.get(rec.node.name, {})
                        ws_stats[rec.node.name][_ws] = ws_stats[rec.node.name].get(_ws, 0) + _v

        context['site_stats'] = ws_stats

        return context


class HostsListView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/hosts.html"

    def get_context_data(self, **kwargs):
        context = super(HostsListView, self).get_context_data(**kwargs)
        objs = list(ArbiterHostInfo.objects.filter(timestamp__gte=int(time.time() - 240)))

        assigned = []
        hosts = []
        print(objs)
        for host in ArbiterNodeHost.objects.all():
            have_node = False
            for obj in objs:
                if obj.host == host:
                    have_node = True
                    hosts.append({
                        'pk': host.pk,
                        'node_id': host.node_id,
                        'name': host.node.name,
                        'host_address': host.host_address,
                        'obj_address': obj.address,
                        'obj_hostname': obj.hostname,
                        'ready': obj.isok,
                        'recency': obj.recency,
                        'mode': obj.mode,
                        'error': obj.error,
                        'attr': 'host',
                    })
                    assigned.append(obj)
            if not have_node:
                hosts.append({
                    'pk': host.pk,
                    'node_id': host.node_id,
                    'name': host.node.name,
                    'host_address': host.host_address,
                    'obj_address': '',
                    'obj_hostname': '',
                    'ready': False,
                    'recency': '',
                    'mode': host.node.mode,
                    'error': '',
                    'attr': 'host',
                })
        for obj in objs:
            if not obj in assigned:
                hosts.append({
                    'pk': None,
                    'node_id': None,
                    'name': obj.hostname,
                    'host_address': obj.hostname,
                    'obj_address': obj.address,
                    'obj_hostname': obj.hostname,
                    'ready': obj.isok,
                    'recency': obj.recency,
                    'mode': obj.mode,
                    'error': obj.error,
                    'attr': 'obj',
                })

        context['hosts'] = hosts
        
        return context


class ManagersListView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/managers.html"

    def get_context_data(self, **kwargs):
        context = super(ManagersListView, self).get_context_data(**kwargs)
        context['managers'] = User.objects.filter(is_superuser=True)
        return context


class AgentsListView(TemplateView):
    template_name = "arbiter/agents.html"

    def get_context_data(self, **kwargs):
        context = super(AgentsListView, self).get_context_data(**kwargs)
        context['agents'] = User.objects.exclude(is_superuser=True).annotate(
            websites_count=Count('website__id'),
            #nodes_count=Count('website__node2website__node_id'),
        )

        return context


class WebsitesStartClearView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class =  WebsitesStartClearForm

    def get_context_data(self, **kwargs):
        context = super(WebsitesStartClearView, self).get_context_data(**kwargs)
        website = Website.objects.get(pk=self.kwargs.get('pk'))
        node = ArbiterNode.objects.get(pk=self.kwargs.get('node'))
        context['website'] = website
        context['name'] = f'Do you realy want to clear outdated version of "{website}" at node "{node.name}"?'
        return context
    
    def form_valid(self, form):
        return HttpResponseRedirect(reverse('arbiter:website_clear_exec', kwargs=self.kwargs) + "?code=" + self.request.GET.get('code'))


class WebsitesProcessClearView(SuperLoginRequired, TemplateView):
    template_name = 'base/msg.html'

    def get_context_data(self, **kwargs):
        website = Website.objects.get(pk=self.kwargs.get('pk'))
        node = ArbiterNode.objects.get(pk=self.kwargs.get('node'))
        code = self.request.GET.get('code')
        context = super(WebsitesProcessClearView, self).get_context_data(**kwargs)
        readed = update_data_workernode([node], 'exec_clear', {
            'website_id': website.pk,
            'code': code
        })
        if not readed:
            context['message'] = 'Cannot apply clearing...'
        else:
            state = json.loads(readed).get('state')
            if state == 'err':
                context['message'] = 'Cannot apply clearing...'
            elif state == 'ok':
                nodes = website.node2website.all()
                awt = ArbiterWebsiteTransfer.objects.create(
                    node_from = nodes.first().node,
                    node_to = node,
                    website=website,
                    step=0
                )
                from apps.balancer.tasks import website_transfer_executor
                website_transfer_executor.delay()
                context['message'] = 'Clearing completed. Starting transfet'
                context['continue'] = reverse('arbiter:website_moves_det', kwargs={'pk': awt.pk, 'site_id': website.pk})
            else:
                context['doref'] = True
                context['message'] = 'Clearing...'
        return context


class WebsitesStartTransferView(SuperLoginRequired, FormView):
    template_name = 'base/form.html'
    form_class =  WebsitesStartTransferForm

    def get_success_url(self):
        return reverse('arbiter:managers_list')

    def get_context_data(self, **kwargs):
        context = super(WebsitesStartTransferView, self).get_context_data(**kwargs)
        website = Website.objects.get(pk=self.kwargs.get('pk'))
        nodes = website.node2website.all()
        context['website'] = website
        if nodes.count() > 1:
            context['message'] = 'Website already on transfering.'
            context['continue'] = reverse('arbiter:website_det', kwargs={'pk': website.pk})
            return context
        context['name'] = f'Do you realy want to move "{website}"?'
        return context
    
    def get_form_kwargs(self):
        form_kwargs = super(WebsitesStartTransferView, self).get_form_kwargs()
        transfer_to_vals = [(None, ' - Please choice Node - '),]
        website = Website.objects.get(pk=self.kwargs.get('pk'))
        _node = website.node2website.first()
        transfer_to_vals.extend([
            (_n.pk, _n.name) for _n in ArbiterNode.objects.filter(
                mode=settings.NODE_MODE_WORKER).exclude(pk=_node.node.pk)
        ])
        form_kwargs['params'] = {
            'pk': self.kwargs.get('pk'),
            'transfer_to_vals': transfer_to_vals
        }
        return form_kwargs

    def form_valid(self, form):
        website = Website.objects.get(pk=self.kwargs.get('pk'))
        nodes = website.node2website.all()
        transfer_to = form.cleaned_data.get('transfer_to', None)
        if nodes.count() == 1 and nodes.filter(node__pk=transfer_to).count() == 0:
            node_to = ArbiterNode.objects.filter(pk=transfer_to).first()
            if node_to:
                    readed = update_data_workernode([node_to], 'checkexists', {
                        'website_id': website.pk,
                    })
                    if not readed:
                        return HttpResponseRedirect(reverse('arbiter:website_move', kwargs={'pk': website.pk}))
                    readed = json.loads(readed)
                    state = readed.get('state', 'err')
                    if state == 'ok':
                        awt = ArbiterWebsiteTransfer.objects.create(
                            node_from = nodes.first().node,
                            node_to = node_to,
                            website=website,
                            step=0
                        )
                        from apps.balancer.tasks import website_transfer_executor
                        website_transfer_executor.delay()
                        return HttpResponseRedirect(reverse('arbiter:website_moves_det', kwargs={'pk': awt.pk, 'site_id': website.pk}))
                    if state == 'clr':
                        return HttpResponseRedirect(reverse('arbiter:website_clear', kwargs={'pk': website.pk, "node": node_to.pk}) + "?code=" + readed.get('code'))
        return HttpResponseRedirect(reverse('arbiter:website_move', kwargs={'pk': website.pk}))


class ArbiterWebsiteTransfersListView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/transfers.html"

    def get_context_data(self, **kwargs):
        context = super(ArbiterWebsiteTransfersListView, self).get_context_data(**kwargs)
        context['website'] = Website.objects.get(pk=self.kwargs.get('pk'))
        context['transfers'] = ArbiterWebsiteTransfer.objects.filter(
            website=context['website']).order_by('step', 'created')
        return context


class ArbiterWebsiteTransferDetailView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/transfer.html"

    def get_context_data(self, **kwargs):
        context = super(ArbiterWebsiteTransferDetailView, self).get_context_data(**kwargs)
        context['transfer'] = ArbiterWebsiteTransfer.objects.get(pk=self.kwargs.get('pk'))
        context['logs'] = context['transfer'].logs.order_by('-timestamp')
        return context


class WebsitesListView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/websites.html"

    def get_context_data(self, **kwargs):
        context = super(WebsitesListView, self).get_context_data(**kwargs)
        context['websites'] = Website.objects.all()
        return context


class NodesListView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/nodes.html"

    def get_context_data(self, **kwargs):
        context = super(NodesListView, self).get_context_data(**kwargs)
        context['nodes'] = ArbiterNode.objects.all().annotate(
            hosts_count=Count('hosts'),
        )
        return context


class CustomerListView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/customers.html"

    def get_context_data(self, **kwargs):
        context = super(CustomerListView, self).get_context_data(**kwargs)
        context['customers'] = Company.objects.all()
        return context


class CustomerDetailView(SuperLoginRequired, FormView):
    template_name = "base/form.html"
    form_class = CompanyEditForm

    def get(self, request, *args, **kwargs):
        self.object = Company.objects.filter(pk=kwargs.get('pk')).first()
        if not self.object:
            return HttpResponseNotFound()
        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = Company.objects.filter(pk=kwargs.get('pk')).first()
        if not self.object:
            return HttpResponseNotFound()
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self):
        form_kwargs = super(CustomerDetailView, self).get_form_kwargs()
        form_kwargs['initial'] = {
            'label': self.object.label,
            'allow_branding': self.object.allow_branding
        }
        return form_kwargs

    def form_valid(self, form):
        label = form.cleaned_data.get('label', '')
        allow_branding = form.cleaned_data.get('allow_branding', False)
        obj = self.object
        obj.allow_branding = allow_branding
        obj.label=label
        obj.save()
        return HttpResponseRedirect(reverse('arbiter:customers_list'))


class RouteMapView(SuperLoginRequired, TemplateView):
    template_name = "arbiter/route_map.html"

    def get_context_data(self, **kwargs):
        context = super(RouteMapView, self).get_context_data(**kwargs)
        context['hostmap'] = ArbiterHostMap.objects.last()
        return context


class ArbiterLoginView(LoginView):
    form_class = ArbiterAuthenticationForm
    template_name = "base/login.html"

    def get_success_url(self):
        return reverse('arbiter:main')


class ArbiterLogoutView(LogoutView):
    def get_success_url(self):
        return reverse('arbiter:login')


class FirstRunView(View):
    def get(self, request, *args, **kwargs):
        step = kwargs.get('step')

        if User.objects.filter(is_superuser=True).exists():
            return HttpResponseNotFound()

        if not step in ['static', 'migrate', 'populate', 'new_user']:
            return HttpResponseRedirect(reverse('arbiter:first_run_step', kwargs={'step': 'static'}))

        from django.core.management import call_command
        import os, io
        if step == 'static':
            os.makedirs(settings.MEDIA_ROOT, exist_ok=True)
            os.makedirs(settings.STATIC_ROOT, exist_ok=True)
            buf = io.StringIO()
            call_command('collectstatic', interactive=False, stdout=buf)
            buf.seek(0)
            ret = buf.read()
            return self.make_resp(reverse('arbiter:first_run_step', kwargs={'step': 'migrate'}), ret)

        if step == 'migrate':
            os.makedirs(settings.MEDIA_ROOT, exist_ok=True)
            os.makedirs(settings.STATIC_ROOT, exist_ok=True)
            buf = io.StringIO()
            call_command('migrate', stdout=buf)
            buf.seek(0)
            ret = buf.read()
            return self.make_resp(reverse('arbiter:first_run_step', kwargs={'step': 'populate'}), ret)

        if step == 'populate':
            os.makedirs(settings.MEDIA_ROOT, exist_ok=True)
            os.makedirs(settings.STATIC_ROOT, exist_ok=True)
            buf = io.StringIO()
            call_command('first_run', nosuperuser=True, stdout=buf)
            buf.seek(0)
            ret = buf.read()
            return self.make_resp(reverse('arbiter:first_run_step', kwargs={'step': 'new_user'}), ret)

        if step == 'new_user':
            return self.new_user_form()

            # call_command('migrate')
            # call_command('first_run')

    def post(self, request, *args, **kwargs):
        step = kwargs.get('step')

        if User.objects.filter(is_superuser=True).exists():
            return HttpResponseNotFound()
        if step == 'new_user':
            return self.new_user_form()
        else:
            return HttpResponce("", status=405)

    def new_user_form(self):
        url = reverse('arbiter:first_run_step', kwargs={'step': 'new_user'})
        
        login = '' if not self.request.method == 'POST' else self.request.POST.get('login', '')
        password = '' if not self.request.method == 'POST' else self.request.POST.get('password', '')
        confirm = '' if not self.request.method == 'POST' else self.request.POST.get('confirm', '')

        from django.core.exceptions import ValidationError
        from django.core.validators import validate_email

        error = []
        if self.request.method == 'POST':
            if login:
                try:
                    validate_email(login)
                except ValidationError as e:
                    error.append('Bad email at "Login"')
            else:
                error.append('"Login" is required')
            
            if not password:
                error.append('"Password" is required')
            else:
                if len(password) < 8:
                    error.append('"Password" must have at least 8 chars')
                if password != confirm:
                    error.append('"Confirmation" not match "Password"')
        
            if not error:
                from django.contrib.auth import get_user_model
                User = get_user_model()
                usr = User(
                    email=login,
                    is_staff=True,
                    is_superuser=True
                )
                usr.set_password(password)
                usr.save()
                return HttpResponseRedirect(reverse('arbiter:login'))

        error = '<br/>'.join(error)

        return HttpResponse(f'''<html><head><title>...</title></head><body>
                <form action="{url}"
                method="POST"
                enctype="application/x-www-form-urlencoded">
                <table border=0>
                <tr><td colspan=2>{error}</td></tr>
                <tr><td colspan=2>Create your superuser (allow arbiter dashboad access)</td></tr>
                <tr><td>Login</td><td><input type="text" name="login" required value="{login}" /> </td></tr>
                <tr><td>Password</td><td><input type="password" name="password" required value="{password}" /> </td></tr>
                <tr><td>Confirmation</td><td><input type="password" name="confirm" required value="{confirm}" /> </td></tr>
                <tr><td colspan=2>
                <input type="submit" value="Submit" /> </td></tr>
                </table>
            </form>
            </body></html>''', content_type='text/html')

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def make_resp(self, addr_to, text):
        return HttpResponse(f'<html><head><title>...</title><meta http-equiv="refresh" content="10; url={addr_to}"></head><body><pre>{text}</pre></body></html>', content_type='text/html')


class GoToNodeAdminView(SuperLoginRequired, View):
    def get(self, request, *args, **kwargs):
        from django.contrib.auth import SESSION_KEY, HASH_SESSION_KEY, BACKEND_SESSION_KEY
        node =  ArbiterNode.objects.get(pk=self.kwargs.get('pk'))

        from apps.balancer.tasks import sync_user
        sync_user(self.request.user, target=node)

        ret = update_data_workernode([node], 'todjangoadmin', {
            'user_id': self.request.session.get(SESSION_KEY),
            'backend': self.request.session.get(BACKEND_SESSION_KEY),
            'session_id': self.request.session.session_key,
        })

        if isinstance(ret, str):
            ret = json.loads(ret)
        if ret and ret.get('ok') == True:
            return HttpResponseRedirect(f"/{node.name}/admin/")
        else:
            return HttpResponseRedirect(f"/arbiter/admin/")
