
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from rest_framework import mixins
from rest_framework.generics import RetrieveUpdateAPIView
from rest_framework import viewsets, permissions
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from nicechat.utils.permissions import IsCompanyOwner
from nicechat.utils.serializers import raise_serializer_error
from nicechat.utils.billing import execute_billing, get_billing_provider
from .models import Transaction, Plan, PaymentMethod, PromoCode, Invoice
from .serializers import TransactionSerializer, PlanSerializer, CompanyPlanSerializer, PaymentMethodSerializer, \
    ChangePlanSerializer, InvoiceSerializer


class FilterByOwnCompanyMixin(object):
    """
    Limits the queryset of viewset class to only those objects that belong 
     to current user's company where he is an owner
    
    """
    def get_queryset(self):
        self.queryset = self.queryset.filter(company=self.request.user.own_company)

        return super(FilterByOwnCompanyMixin, self).get_queryset()


class FilterByCompanyMixin(object):
    """
    Limits the queryset of viewset class to only those objects that belong 
     to current user's company

    """

    def get_queryset(self):
        self.queryset = self.queryset.filter(company=self.request.user.company)

        return super(FilterByCompanyMixin, self).get_queryset()


class PlanViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    queryset = Plan.objects.all()
    serializer_class = PlanSerializer


class TransactionViewSet(FilterByOwnCompanyMixin, viewsets.ModelViewSet):
    queryset = Transaction.objects.all().order_by('-id')
    serializer_class = TransactionSerializer
    permission_classes = [IsAuthenticated, IsCompanyOwner]


class InvoiceViewSet(FilterByOwnCompanyMixin, viewsets.ModelViewSet):
    queryset = Invoice.objects.all().order_by('-id')
    serializer_class = InvoiceSerializer
    permission_classes = [IsAuthenticated, IsCompanyOwner]


class CompanyPlanView(FilterByCompanyMixin, RetrieveUpdateAPIView):
    serializer_class = CompanyPlanSerializer
    permission_classes = [IsAuthenticated]

    def get_object(self):
        return self.request.user.company.billing_plan

    def get_serializer_class(self):
        if self.request.method.lower() == 'put':
            return ChangePlanSerializer

        return super(CompanyPlanView, self).get_serializer_class()

    def get_permissions(self):
        if self.request.method.lower() == 'put':
            return [IsAuthenticated(), IsCompanyOwner()]
        return super(CompanyPlanView, self).get_permissions()


class PaymentMethodViewSet(viewsets.ModelViewSet):
    queryset = PaymentMethod.objects.all()
    serializer_class = PaymentMethodSerializer
    permission_classes = [IsAuthenticated, IsCompanyOwner]

    def get_queryset(self):
        return self.request.user.company.payment_methods.all()


class BillingClinetToken(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    http_method_names = ['get', 'options']

    def get(self, request, format=None):
        return Response(
            {
                'token': execute_billing('get_client_token', args=[request.user]),
                'provider': get_billing_provider()
            })


class ConfigView(APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, format=None):
        return Response({
            'DEBUG': settings.DEBUG,
            'billingEnabled': settings.BILLING_ENABLED
        })


class ValidatePromoCodeView(APIView):
    permission_classes = (permissions.IsAuthenticated, IsCompanyOwner,)
    http_method_names = ['post', 'options']

    def post(self, request):
        if 'code' not in request.data:
            raise_serializer_error(_('Providing "code" parameter is required'))

        result = {'valid': False}

        try:
            code_obj = PromoCode.objects.get(code=request.data['code'])
            plan_obj = None
            payment_period = None

            if code_obj.plans.count():
                if request.data.get('plan_id') is None:
                    raise_serializer_error({
                        'plan_id': _('Providing "plan_id" parameter is required for this promo code')
                    })

                plan_obj = Plan.objects.get(pk=request.data.get('plan_id'))

            if code_obj.payment_period:
                payment_period = request.data.get('payment_period')
                if payment_period is None:
                    raise_serializer_error({
                        'payment_period': _('Providing "payment_period" parameter is required for this promo code')
                    })

            if code_obj.is_valid_for_company(request.user.company, plan_obj, payment_period):
                result = {'valid': True,
                          'discount_percent': code_obj.discount_percent}
        except (PromoCode.DoesNotExist, Plan.DoesNotExist):
            pass

        return Response(result)
