from django.conf import settings
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.core.validators import MaxValueValidator, MinValueValidator, MinLengthValidator
from nicechat.utils.billing import execute_billing


class BillingCycles:
    MONTH = 'month'
    YEAR = 'year'

    CHOICES = (
        (MONTH, 'Monthly',),
        (YEAR, 'Annually',),
    )


class BillingProps(models.Model):
    k = models.CharField(max_length=25, primary_key=True)
    v = models.CharField(max_length=256)


class PlanFeature(models.Model):
    name = models.CharField(max_length=25, primary_key=True)
    title = models.CharField(max_length=128)

    def __str__(self):
        return self.title


class Plan(models.Model):
    name = models.CharField(max_length=25)
    month_price = models.DecimalField(max_digits=10, decimal_places=2,
                                      help_text=_("Price per agent seat per month if client chooses monthly billing"))
    annual_price = models.DecimalField(max_digits=10, decimal_places=2, blank=True, null=True,
                                       help_text=_("Price per agent seat per month if client chooses yearly billing"))
    description = models.TextField(blank=True, null=True)
    features = models.ManyToManyField(PlanFeature, blank=True)
    order = models.PositiveSmallIntegerField(blank=True, null=True)
    # a set of fields for holding plan restrictions:
    sites_num = models.PositiveSmallIntegerField(blank=True, null=True,
                                                 help_text=_("The max number of sites available"))
    agents_num = models.PositiveSmallIntegerField(blank=True, null=True,
                                                  help_text=_("The max number of agent seats "
                                                              "available for each company"))
    products_num = models.PositiveSmallIntegerField(blank=True, null=True,
                                                    help_text=_("The max number of products that "
                                                                "can be purchased from chat"))
    chat_stats_period = models.PositiveSmallIntegerField(blank=True, null=True,
                                                         help_text=_("The max period for chat statistics "
                                                                     "to be available (in days)"))
    sales_stats_period = models.PositiveSmallIntegerField(blank=True, null=True,
                                                          help_text=_("The max period for sales statistics "
                                                                      "to be available (in days)"))
    visitors_stats_period = models.PositiveSmallIntegerField(blank=True, null=True,
                                                             help_text=_("The max period for sales statistics "
                                                                         "to be available (in days)"))
    chat_history_period = models.PositiveSmallIntegerField(blank=True, null=True,
                                                           help_text=_("The max period for chats history "
                                                                       "to be available (in days)"))
    visitor_data_period = models.PositiveSmallIntegerField(blank=True, null=True,
                                                           help_text=_("The max period for visitors data "
                                                                       "to be available (in days)"))
    billing_prefix = models.CharField(max_length=25, blank=True, null=True)
    promotion = models.TextField(default='',
                                   help_text='Each row is a separate record at plans list')

    def __str__(self):
        return ': '.join([self.name, str(self.month_price), str(self.annual_price)])

    @cached_property
    def is_trial(self):
        return self.settings.filter(prop=PlanSettings.TRIAL).exists()

    @cached_property
    def is_free(self):
        return self.settings.filter(prop=PlanSettings.FREE).exists()

    def get_billing_id(self, billing_cycle):
        if not self.billing_prefix:
            return None

        assert billing_cycle in [BillingCycles.MONTH, BillingCycles.YEAR]

        return self.billing_prefix + '_' + billing_cycle

    @classmethod
    def get_by_billing_id(cls, billing_id):
        """
        Get the company plan data based on plan id.

        :return: a tuple with two items. First is a corresponding plan object, second one is a payment period constant
        """
        try:
            assert isinstance(billing_id, str) and len(billing_id.split('_')) == 2
            INTERVALS = {
                BillingCycles.MONTH: 1,
                BillingCycles.YEAR: 12
            }
            interval = INTERVALS.get(billing_id.split('_')[1], None)
            if not interval:
                raise AssertionError('Not interval')
            return cls.objects.get(billing_prefix=billing_id.split('_')[0]), interval
        except (AssertionError, cls.DoesNotExist):
            raise ValueError("Invalid plan Briantree id provided")


class PlanSettings(models.Model):
    FREE = 'free'
    TRIAL = 'trial'
    PROPS = (
        (FREE, 'Free',),
        (TRIAL, 'Trial',),
    )
    plan = models.ForeignKey(Plan, related_name='settings', on_delete=models.CASCADE)
    prop = models.CharField(max_length=8, choices=PROPS, unique=True)


class Transaction(models.Model):
    SUCCESS = 'success'
    PENDING = 'pending'
    REQACT = 'reqact'
    FAILED = 'failed'
    STATUSES = (
        (SUCCESS, 'Success',),
        (PENDING, 'Pending',),
        (REQACT, 'Require action',),
        (FAILED, 'Failed',),
    )

    company = models.ForeignKey('company.Company', related_name='transactions', on_delete=models.CASCADE)
    plan = models.ForeignKey(Plan, on_delete=models.CASCADE)
    plan_period = models.CharField(max_length=8, choices=BillingCycles.CHOICES)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    payment_method = models.ForeignKey('PaymentMethod', on_delete=models.CASCADE)
    timestamp = models.DateTimeField()
    result = models.TextField()
    amount = models.DecimalField(max_digits=10, decimal_places=2)
    sysinfo = models.TextField(null=True)
    status = models.CharField(max_length=7, choices=STATUSES, default=PENDING)
    billing_id = models.CharField(max_length=32, db_index=True)
    billing_provider = models.CharField(max_length=32, default='nicechat_mocks')

    def __str__(self):
        return ': '.join([self.company.label, self.plan.name, str(self.amount)])

    @property
    def result_short(self):
        return self.status

    class Meta:
        unique_together = ['billing_id', 'billing_provider']


class Invoice(models.Model):
    SUCCESS = 'success'
    CANCELED = 'cancel'
    PENDING = 'pending'
    DRAFT = 'draft'
    REQACT = 'reqact'
    FAILED = 'failed'
    STATUSES = (
        (SUCCESS, 'Success',),
        (CANCELED, 'Canceled',),
        (PENDING, 'Pending',),
        (DRAFT, 'Draft',),
        (REQACT, 'Require action',),
        (FAILED, 'Failed',),
    )

    company = models.ForeignKey('company.Company', related_name='invoices', on_delete=models.CASCADE)
    plan = models.ForeignKey(Plan, on_delete=models.CASCADE)
    plan_period = models.CharField(max_length=8, choices=BillingCycles.CHOICES)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    number = models.CharField(max_length=32)
    link = models.CharField(max_length=256, default='')
    timestamp = models.DateTimeField()
    result = models.TextField()
    amount = models.DecimalField(max_digits=10, decimal_places=2)
    sysinfo = models.TextField(null=True)
    status = models.CharField(max_length=7, choices=STATUSES, default=PENDING)
    billing_id = models.CharField(max_length=32, db_index=True)
    billing_provider = models.CharField(max_length=32, default='nicechat_mocks')

    def __str__(self):
        return ': '.join([self.company.label, self.plan.name, str(self.amount)])

    @property
    def result_short(self):
        return self.status

    class Meta:
        unique_together = ['billing_id', 'billing_provider']


class CompanyPlan(models.Model):
    company = models.OneToOneField('company.Company', related_name='billing_plan', on_delete=models.CASCADE)
    plan = models.ForeignKey(Plan, on_delete=models.CASCADE)
    agents_count = models.PositiveSmallIntegerField()
    begin = models.DateTimeField()
    end = models.DateTimeField()
    payment_period = models.CharField(max_length=8, choices=BillingCycles.CHOICES, default=BillingCycles.MONTH)

    billing_subscription_id = models.CharField(max_length=32, blank=True, null=True)
    billing_provider = models.CharField(max_length=32, default='nicechat_mocks')
    notes = models.TextField(default='{}')

    def __str__(self):
        return f'{self.company}:{self.plan.name}:{self.agents_count}'

    @cached_property
    def month_total_payment(self):
        return self.agents_count * self.plan.month_price

    @cached_property
    def annual_total_payment(self):
        return self.agents_count * self.plan.annual_price * 12

    @cached_property
    def annual_saving(self):
        return self.month_total_payment * 12 - self.annual_total_payment

    def cancel_subscription(self):
        if self.billing_subscription_id is not None:
            execute_billing('cancel_subscription', [self.billing_subscription_id, self])


class PaymentMethod(models.Model):
    CREDIT_CARD = 'credit_card'
    PAYPAL = 'paypal'

    TYPE_CHOICES = (
        (CREDIT_CARD, _('Credit card')),
        (PAYPAL, 'PayPal')
    )

    company = models.ForeignKey('company.Company', related_name='payment_methods', on_delete=models.CASCADE)
    card_number = models.CharField(max_length=64, blank=True, null=True)
    is_primary = models.BooleanField(default=False)
    type = models.CharField(max_length=16, choices=TYPE_CHOICES, default=CREDIT_CARD)
    billing_token = models.CharField(max_length=32, blank=True, null=True)
    billing_provider = models.CharField(max_length=32, default='nicechat_mocks')

    class Meta:
        ordering = ['-is_primary']


class PromoCodeUsage(models.Model):
    company = models.ForeignKey('company.Company', on_delete=models.CASCADE)
    code = models.ForeignKey('PromoCode', on_delete=models.CASCADE)
    date_used = models.DateTimeField(auto_now_add=True)


class PromoCode(models.Model):
    code = models.CharField(max_length=128, primary_key=True,
                            help_text=_('Textual representation of a code'))
    discount_percent = models.PositiveIntegerField(default=1, validators=[
        MaxValueValidator(100),
        MinValueValidator(1)
    ], help_text=_("The discount in percents"))
    # todo discount_fixed_amount
    is_active = models.BooleanField(default=True)
    billing_id = models.CharField(max_length=128, validators=[MinLengthValidator(1)])
    billing_provider = models.CharField(max_length=32, default='nicechat_mocks')
    usages = models.ManyToManyField('company.Company', through=PromoCodeUsage, related_name='used_promo_codes')

    companies = models.ManyToManyField('company.Company', related_name='allowed_promo_codes', blank=True)
    plans = models.ManyToManyField('Plan', related_name='allowed_promo_codes', blank=True)
    payment_period = models.CharField(max_length=8, choices=BillingCycles.CHOICES, blank=True, null=True)

    def __str__(self):
        result = self.code

        if self.discount_percent:
            result += u' [{}%]'.format(self.discount_percent)

        # if self.max_usages:
        #     result += u' [max usages: {}]'.format(self.max_usages)

        if self.companies.count():
            result += u' [companies: {}]'.format(', '.join([str(c) for c in self.companies.all()]))

        return result

    def is_valid_for_company(self, company, plan=None, payment_period=None):
        if not self.is_active:
            return False

        if self.companies.count() and company not in self.companies.all():
            return False

        if self.plans.count():
            assert plan is not None, _('Promo code has plan restrictions set up, '
                                       'so providing plan parameter is required')
            if plan not in self.plans.all():
                return False

        if self.payment_period:
            assert payment_period in [BillingCycles.MONTH, BillingCycles.YEAR]

            if payment_period != self.payment_period:
                return False

        # if self.max_usages and self.usages.count() >= self.max_usages:
        #     return False

        return True
