from datetime import timedelta
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.template.loader import render_to_string
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from nicechat.utils.serializers import raise_serializer_error
from nicechat.utils.billing import execute_billing, BillingError
from nicechat.utils.mail import notify_emails
from .models import Transaction, Plan, CompanyPlan, PaymentMethod, BillingCycles, PromoCode, PromoCodeUsage, Invoice


class InvoiceSerializer(serializers.ModelSerializer):
    class Meta:
        model = Invoice
        fields = ('id', 'number', 'link', 'timestamp', 'amount', 'billing_provider', 'plan_period', 'plan', 'status')



class TransactionSerializer(serializers.ModelSerializer):
    payment_method = serializers.SerializerMethodField()

    def get_payment_method(self, instance):
        return instance.payment_method.card_number

    class Meta:
        model = Transaction
        fields = ('id', 'timestamp', 'result_short', 'amount', 'payment_method', 'billing_provider', 'status')


class PlanSerializer(serializers.ModelSerializer):
    annual_price = serializers.IntegerField()
    month_price = serializers.IntegerField()
    is_trial = serializers.BooleanField()

    class Meta:
        model = Plan
        fields = '__all__'
        read_only_fields = ('features',)


class CompanyPlanSerializer(serializers.ModelSerializer):
    month_total_payment = serializers.IntegerField()
    annual_total_payment = serializers.IntegerField()
    annual_saving = serializers.IntegerField()
    plan = PlanSerializer(read_only=True)

    class Meta:
        model = CompanyPlan
        fields = '__all__'


class PaymentMethodSerializer(serializers.ModelSerializer):
    nonce = serializers.CharField(max_length=512, required=True, write_only=True)

    class Meta:
        model = PaymentMethod
        exclude = ('company', 'billing_token')
        read_only_fields = ('card_number', 'type', 'is_primary', 'billing_provider',)

    def create(self, validated_data):
        request_user = self.context['request'].user
        validated_data['company'] = request_user.own_company

        if not settings.BILLING_ENABLED:
            raise ValidationError({
                'Billing error': 'Payment methods processing is currently disabled'
            })
        try:
            execute_billing('create_payment_method', args=[request_user, validated_data])
        except BillingError as e:
            raise raise_serializer_error({'payment_method': e.message})

        return super(PaymentMethodSerializer, self).create(validated_data)


class ChangePlanSerializer(serializers.ModelSerializer):
    plan = serializers.PrimaryKeyRelatedField(queryset=Plan.objects.all())
    agents_count = serializers.IntegerField(min_value=1)
    payment_period = serializers.ChoiceField(choices=BillingCycles.CHOICES, required=False)
    payment_method = serializers.IntegerField(required=False)
    promo_code = serializers.CharField(required=False)

    class Meta:
        model = CompanyPlan
        fields = ('plan', 'agents_count', 'payment_period', 'payment_method', 'promo_code')

    def validate_plan(self, value):
        """
        No one can change billing plan to "Trial"
        """
        if value.is_trial:
            raise serializers.ValidationError(_("Wrong value"))

        return value

    def validate_payment_method(self, value):
        # value is a payment method id
        try:
            PaymentMethod.objects.get(pk=value, company=self.context['request'].user.company)
        except PaymentMethod.DoesNotExist:
            raise serializers.ValidationError(_("Not found"))

        return value

    def validate(self, data):
        """
        Make 'payment_period' and 'payment_method' fields optional for the "Free" plan only
        """
        # todo can't change to the same plan as current if agents count is not changed
        if not data['plan'].is_free:
            if not data.get('payment_period'):
                raise_serializer_error({'payment_period': _("Payment period is required for non-free plans")})
            if not data.get('payment_method'):
                raise_serializer_error({'payment_method': _("Payment method is required for non-free plans")})

        return data

    def _cancel_instance_subscription(self, instance):
        try:
            instance.cancel_subscription()
            instance.billing_subscription_id = None
        except BillingError as e:
            raise_serializer_error(e.message)


    def update(self, instance, validated_data):
        validated_plan = validated_data.get('plan')
        validated_payment_period = validated_data.get('payment_period')

        if validated_plan.name == 'Free':  # switching to "Free" plan
            self._cancel_instance_subscription(instance)

            validated_data['agents_count'] = validated_plan.agents_num
        else:  # only create subscription for non-free plans
            try:
                assert instance.company.primary_payment_method is not None
            except AssertionError:
                raise_serializer_error(_("The plan can't be changed if no payment method is set up"))

            self._cancel_instance_subscription(instance)

            price_per_agent = validated_plan.annual_price * 12 \
                if validated_payment_period == BillingCycles.YEAR \
                else validated_plan.month_price

            company_plan_price = price_per_agent * validated_data['agents_count']
            payment_method_obj = PaymentMethod.objects.get(pk=validated_data['payment_method'],
                                                           company=self.context['request'].user.company)

            promo_code_obj = None
            if 'promo_code' in validated_data:
                try:
                    _promo_code_obj = PromoCode.objects.get(code=validated_data['promo_code'])

                    if _promo_code_obj.is_valid_for_company(instance.company, validated_plan, validated_payment_period):
                        promo_code_obj = _promo_code_obj
                    else:
                        raise PromoCode.DoesNotExist  # to avoid error processing duplication
                except PromoCode.DoesNotExist:
                    raise_serializer_error({'promo_code': _("Couldn't find the corresponding promo code object")})

            print(validated_plan, validated_payment_period)
            billing_id = validated_plan.get_billing_id(validated_payment_period)

            try:
                status, details = execute_billing('apply_payment',
                                                  args=[
                                                      billing_id,
                                                      validated_plan.pk,
                                                      company_plan_price,
                                                      payment_method_obj.billing_token,
                                                      promo_code_obj,
                                                      validated_data['agents_count'],
                                                      self.context['request'].user
                                                  ])

                if status:
                    instance.billing_subscription_id = details.get('subscription_id')
                    instance.billing_provider = details.get('billing_provider')
                    instance.begin = timezone.now()
                    delta = timedelta(days=30 if validated_payment_period == BillingCycles.YEAR else 365)
                    instance.end = timezone.now() + delta
                    if details.get('promo_code_obj'):
                        PromoCodeUsage.objects.create(code=details.get('promo_code_obj'), company=instance.company)

            except BillingError as e:
                print(e.__dict__)
                raise_serializer_error(e.message)

        ret = super(ChangePlanSerializer, self).update(instance, validated_data)

        template = 'users/email/plan_changed'
        context = {
            'plan': instance.plan,
            'user': instance.company.owner,
            'company': instance.company.label
        }
        try:
            subject = render_to_string('%s.subject' % template, context)
            print(instance.company.owner.email, subject)
            notify_emails(
                [instance.company.owner.email],
                subject,
                '%s.txt' % template,
                '%s.html' % template,
                context
            )
        except Exception as e:
            print(e)

        return ret
