import json
from nicechat.admins import ncadmin
from django.contrib import admin
from django.conf import settings
from django.db.models import Q
from django.http.response import HttpResponseNotFound
from ..chat.models import Visitor, Chat, Message, Attachment, VisitedPage, ArchivedChat, Idempotency, BatamStatistics, Connector
from ..chat.views import render_attach_response

def chat_mode(cls, obj):
    for key, val in Idempotency.MODES_CHOICE:
        if key == obj.mode:
            return '%s #%s' % (val, obj.id)
    return ' - #%s' % obj.id
chat_mode.short_description = 'Room'

def get_client_name(cls, obj):
    if obj.mode == Idempotency.MODE_P2P:
        return '-'
    else:
        return obj.client_name
get_client_name.short_description = 'Name'


class ArchivedChatAdmin(admin.ModelAdmin):
    change_form_template = 'admin/ext_chat_change_form.html'
    list_display = ('chat_mode', 'ts_open', 'ts_close', 'get_client_name', 'companies', 'operators_list', 'operators_count' )

    search_fields = ('id',)

    get_client_name = get_client_name
    chat_mode = chat_mode

    def get_search_results(self, request, queryset, search_term):
        queryset = ArchivedChat.objects.all().order_by('-ts_close')
        if search_term:
            queryset = queryset.filter(
                Q(name__icontains=search_term) |
                Q(client_name__icontains=search_term) |
                Q(messages__icontains=search_term) |
                Q(operators__first_name__icontains=search_term) |
                Q(operators__last_name__icontains=search_term) |
                Q(operators__company__name__icontains=search_term)
            ).distinct()

        return queryset, True

    def get_urls(self):
         from django.urls import re_path
         urls = super(ArchivedChatAdmin, self).get_urls()
         urls += [
             re_path(r'^(?P<object_id>.+)/change/download/(?P<msg_id>\d+)/(?P<file_name>.*)$', self.admin_site.admin_view(self.download))
         ]
         return urls

    def download(self, request, *args, **kwargs):
        object_id = kwargs.get('object_id')
        achat = ArchivedChat.objects.get(pk=object_id)
        messages = {msg.get('id'): msg for msg in json.loads(achat.messages)}
        attachments = {att.get('message_id'): att for att in json.loads(achat.attachments)}
        msg = messages.get(int(kwargs.get('msg_id')))
        attach = attachments.get(int(kwargs.get('msg_id')))
        if not msg:
            return HttpResponseNotFound()
        if not attach:
            return HttpResponseNotFound()
        size = json.loads(msg.get('data', '{}')).get('size', 0)
        return render_attach_response(msg.get('text'), attach.get('url'), size)

    def change_view(self, request, object_id, form_url='', extra_context=None):
        extra_context = extra_context or {}
        achat = ArchivedChat.objects.get(pk=object_id)
        extra_context['chat_messages'] = json.loads(achat.messages)
        return super(ArchivedChatAdmin, self).change_view(
            request, object_id, form_url, extra_context=extra_context,
        )

class ChatAdmin(admin.ModelAdmin):
    change_form_template = 'admin/ext_chat_change_form.html'
    list_display = ('chat_mode', 'ts_open', 'get_client_name', 'is_open', 'companies', 'operators' )

    search_fields = ('id',)

    get_client_name = get_client_name
    chat_mode = chat_mode

    def get_urls(self):
         from django.urls import re_path
         urls = super(ChatAdmin, self).get_urls()
         urls += [
             re_path(r'^(.+)/change/download/(?P<msg_id>\d+)/(?P<file_name>.*)$', self.admin_site.admin_view(self.download))
         ]
         return urls

    def download(self, request, *args, **kwargs):
        msg = Message.objects.get(id=kwargs.get('msg_id'))
        attach = Attachment.objects.filter(message_id=msg.pk).first()
        return render_attach_response(msg.text, attach.url, attach.size)

    def get_search_results(self, request, queryset, search_term):
        queryset = Chat.objects.all().order_by('-id')
        if search_term:
            queryset = queryset.filter(
                Q(message__text__icontains=search_term) |
                Q(message__sender_name__icontains=search_term) |
                Q(operators_current__first_name__icontains=search_term) |
                Q(operators_current__last_name__icontains=search_term) |
                Q(operators_current__company__name__icontains=search_term)
            ).distinct()

        return queryset, True

    def change_view(self, request, object_id, form_url='', extra_context=None):
            extra_context = extra_context or {}
            chat = Chat.objects.get(pk=object_id)
            extra_context['chat_messages'] = chat.message_set.order_by('timestamp')
            return super(ChatAdmin, self).change_view(
                request, object_id, form_url, extra_context=extra_context,
            )

class MessageAdmin(admin.ModelAdmin):
    list_display = ('type', 'sender_id', 'sender_name', 'text', 'data_type', 'state', 'chat' )

class AttachmentAdmin(admin.ModelAdmin):
    list_display = ('url', 'file_type', 'size', 'upload_ts', 'chat', 'message',)

class BatamStatisticsAdmin(admin.ModelAdmin):
    fieldsets = [
        (None, {'fields':()}), 
        ]
    actions = None

    def __init__(self, *args, **kwargs):
        super(BatamStatisticsAdmin, self).__init__(*args, **kwargs)
        self.list_display_links = None

    def has_change_permission(self, *args, **kwargs):
        return False

    view_on_site = False

    list_display = (
        'client', 'website', 'name', 'email', 'total_duration', 'total_visits', 'total_chats', 
        'total_chats_agentinit', 'latest_chat_id', 'latest_page_url', 'latest_page_title', 'sys_browser', 
        'sys_os', 'sys_device', 'geo_country', 'geo_city', 'page_ref'
    )

ncadmin.register(Visitor)
ncadmin.register(VisitedPage)
ncadmin.register(Chat, ChatAdmin)
# ncadmin.register(Message, MessageAdmin)
ncadmin.register(Attachment, AttachmentAdmin)
ncadmin.register(ArchivedChat, ArchivedChatAdmin)
ncadmin.register(BatamStatistics, BatamStatisticsAdmin)

if settings.CONNECTOR_ENABLED:
    ncadmin.register(Connector)

