import datetime
import json

from django.conf import settings
from django.http import Http404
from django.db import models
from django.db.models import Q
from django.core.cache import caches, InvalidCacheBackendError
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy as _
from django.utils.timezone import make_aware

from rest_framework import permissions
from rest_framework import status
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.generics import get_object_or_404
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()

from nicechat.analytics import event_track, VISITOR_CHAT, JOIN_CHAT, BACKEND_VISITOR_INVITE, LEAVE_CHAT, \
    VISITOR_PURCHASE_STEP_1, VISITOR_PURCHASE_STEP_2
from ..website.models import Website, Product, Form
from nicechat.utils import constants
from nicechat.utils.attach import save_attach_file
from nicechat.utils.constants import WebsocketEventTypes

from ..website.models import Website
from ..users.models import User
from ..visitors.models import Client, FormResult
from .models import Message, Chat, ArchivedChat, Visitor, Attachment
from .serializers import MessageSerializer, ChatSerializer, JoinSerializer, UserHistorySerializer, \
    ArchivedChatShortSerializer, ArchivedChatSerializer
from .utils import join_watch_list, leave_watch_list, set_sizes_by_user, move_chat_to_archive, \
    fill_order_by_product, filter_tag, filter_tags


class ChatPermission(permissions.BasePermission):
    def has_object_permission(self, request, view, obj):
        try:
            assert request.user.company_id == obj.website.company_id \
                   and obj.website in request.user.website_set.all()
        except AssertionError:
            return False
        else:
            return True


class MessageViewSet(viewsets.ModelViewSet):
    queryset = Message.objects.all()
    serializer_class = MessageSerializer
    pagination_class = None

    def perform_create(self, serializer):
        try:
            chat_id = int(self.kwargs.get('chat_id', '0'))
        except ValueError:
            chat_id = None
        if not chat_id and settings.ALLOW_CHAT_AUTOCREATE:

            # if user is employee the chatroom must be specified (we can not autocreate chatroom, because we don't
            #       know client-id, we must create Chat for specified client and send Message to this new Chat)
            if self.request.user.is_authenticated:
                raise PermissionDenied(_("Autocreation is possible only for visitors"))

            # TODO Unique website identification instead of using 'site_id'?
            current_site = Website.objects.filter(
                pk=self.request.query_params.get('site_id') or self.kwargs.get('site_id'))
            chat_obj = Chat.create(client_id=self.request.data.get('sender_id'),
                                   ts_open=datetime.datetime.now(),
                                   website=current_site[0])
            chat_id = chat_obj.pk
        else:
            chat_obj = get_object_or_404(Chat, id=chat_id)
            # if chat is present, but not open - raise 404 too
            #     (we can not add message to closed chat by political)) reasons)
            if not chat_obj.is_open and chat_obj.is_serialized:
                raise Http404

        save_kwargs = {'chat_id': chat_id}

        if not self.request.user.is_authenticated:
            save_kwargs['type'] = Message.TYPE_CLIENT
            if self.request.data.get('as_event', False):
                save_kwargs['type'] = Message.TYPE_EVENT_CLI
            if chat_obj.client_id != self.request.data.get('sender_id'):
                raise PermissionDenied("Wrong sender_id")
            if self.request.data.get('data_type') == Message.DATA_FORM_FILLED:
                raise PermissionDenied(_("User can not write form data without validation"))
        else:
            save_kwargs['type'] = Message.TYPE_OPERATOR
            if self.request.data.get('as_event', False):
                save_kwargs['type'] = Message.TYPE_EVENT_OPER
            save_kwargs['sender_id'] = self.request.user.user_tag
            if not self.request.data.get('sender_name'):
                save_kwargs['sender_name'] = self.request.user.get_full_name or str(self.request.user)

        # ws.send_chat_msg(chat_id, ws.MsgTypes.NEW_MESSAGE, self.request.data.get('text'))

        msg = serializer.save(**save_kwargs)
        if msg and msg.pk:
            async_to_sync(channel_layer.group_send)("cli_speak-%s" % msg.chat_id, {"type": "websocket.send",  "text":  json.dumps({
                "type": WebsocketEventTypes.FRONTEND_CHAT_ID_UPDATE,
                "new_id": msg.pk,
                "x_id": self.request.data.get('x_tag'),
                "chat": msg.chat_id
                })
            })


    @action(detail=True, methods=['POST'])
    def set_data_type(self, request, **kwargs):
        msg = self.get_object()
        if 'data_type' in self.request.data:
            data_type = self.request.data.get('data_type')
            if (msg.data_type == Message.DATA_SUGGEST and data_type in (Message.DATA_SUGGEST_CANCEL,)) or \
                    (msg.data_type == Message.DATA_FORM and data_type in (Message.DATA_FORM_CANCEL,)):
                msg.data_type = data_type
                msg.save(update_fields=('data_type',))
                return Response(status=status.HTTP_200_OK)

        return Response(status=status.HTTP_406_NOT_ACCEPTABLE)

    def get_queryset(self):
        # Nobody can read ALL messages (only by chat-room-id with security checks)
        try:
            qs = {'pk': self.kwargs.get('chat_id')}
            if self.request.user.is_authenticated:
                qs['website__users'] = self.request.user
            else:
                qs['client_id'] = self.request.nice_chat_client_id
            chat = Chat.objects.get(**qs)
        except Chat.DoesNotExist:
            raise PermissionDenied()

        return chat.message_set.all().order_by('timestamp')


class ChatViewSet(viewsets.ModelViewSet):
    queryset = Chat.objects.all()
    serializer_class = ChatSerializer

    def perform_create(self, serializer):
        super(ChatViewSet, self).perform_create(serializer)
        if getattr(serializer, 'instance', None) and \
                self.request.user and self.request.user.is_authenticated:
            chat = serializer.instance

            if self.request.data.get('notify_send', False):
                event_track(BACKEND_VISITOR_INVITE, self.request.user.pk, {
                    'siteId': chat.website_id,
                    'clientID': chat.client_id,
                    'chatId': chat.pk,
                    'mp_country_code': chat.visitor.geo_country if chat.visitor else '',
                    '$city': chat.visitor.geo_city if chat.visitor else '',
                    'os': chat.visitor.sys_os if chat.visitor else '',
                    '$browser': chat.visitor.sys_browser if chat.visitor else '',
                })

            website = self.request.current_site
            if website:
                sender_name = self.request.user.get_full_name or str(self.request.user)

                self.send_welcome(chat, self.request.user.user_tag, sender_name, website)

    def get_serializer_class(self):
        if self.action == 'join':
            return UserHistorySerializer

        elif self.action == 'invite':
            return JoinSerializer

        return super(ChatViewSet, self).get_serializer_class()

    @action(detail=True, methods=['POST'], permission_classes=[IsAuthenticated])
    def upload(self, request, **kwargs):
        chat = self.get_object()

        if not chat.operators_current.filter(pk=request.user.pk).exists():
            return Response(status=status.HTTP_403_FORBIDDEN)

        r_file, created = save_attach_file(request, 'file', chat.pk, True)

        if not created:
            text = u"~!ATTBLOCK%s:%s" % (r_file['error'], r_file['name'])

            data = {"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_SERVER_NOTIFY,
                                        "data": {
                                            "text": text,
                                            "level": 'error',
                                        }}
                   )}
            at_key = "adm-website_mon-u-%i-%i" % (chat.website_id, request.user.pk)
            async_to_sync(channel_layer.group_send)(at_key, data)

            return Response("{}", status=status.HTTP_200_OK)

        # only after file wrote - create Message and Attachment
        message = Message.objects.create(
            chat=chat,
            sender_id=self.request.user.user_tag,
            sender_name=self.request.user.get_full_name or str(self.request.user),
            text=r_file['name'],
            data=json.dumps({
                'file_type': r_file['content_type'],
                'preview': r_file.get('preview', None),
                'size': r_file['size'],
            }),
            data_type=Message.DATA_ATTACH,
            type=Message.TYPE_OPERATOR
        )

        Attachment.objects.create(
            website=chat.website,
            chat=chat,
            message=message,
            file_type=r_file['content_type'],
            size=r_file['size'],
            url=r_file['path']
        )

        return Response(status=status.HTTP_201_CREATED)

    @action(detail=True, methods=['POST'])
    def set_client(self, request, **kwargs):
        chat = self.get_object()
        if 'client_name' in self.request.data:
            client_name = self.request.data.get('client_name')
            if self.do_set_client(client_name, chat):
                return Response(status=status.HTTP_204_NO_CONTENT)

        return Response(status=status.HTTP_406_NOT_ACCEPTABLE)

    @action(detail=True, methods=['POST'])
    def add_tag(self, request, **kwargs):
        chat = self.get_object()
        if 'tag' in self.request.data:
            tag = filter_tag(self.request.data.get('tag'))
            tags = chat.tags.split(',')
            tags.append(tag)
            tags = filter_tags(tags)
            chat.tags = ','.join(tags)
            chat.save(update_fields=['tags'])
            return Response(status=status.HTTP_204_NO_CONTENT)
        return Response(status=status.HTTP_406_NOT_ACCEPTABLE)

    @action(detail=True, methods=['POST'])
    def rem_tag(self, request, **kwargs):
        chat = self.get_object()
        if 'tag' in self.request.data:
            tag = filter_tag(self.request.data.get('tag'))
            tags = chat.tags.split(',')
            if tag in tags:
                tags.remove(tag)
            chat.tags = ','.join(tags)
            chat.save(update_fields=['tags'])
            return Response(status=status.HTTP_204_NO_CONTENT)
        return Response(status=status.HTTP_406_NOT_ACCEPTABLE)

    @staticmethod
    def do_set_client(client_name, chat):
        token = "visitor-client-name-%s-%s" % (chat.website_id, chat.client_id)
        client = Client.objects.filter(website_id=chat.website_id, client_id=chat.client_id).first()
        try:
            cache = caches['default']
            cache.set(token, client_name, 86400)
        except InvalidCacheBackendError:
            pass

        if client:
            client.client_name = client_name
            client.save(update_fields=('client_name',))

            models.signals.post_save.send(sender=Chat, instance=chat, created=False, update_fields=[])
            return True
        return False

    @action(detail=False, methods=['get'], permission_classes=[IsAuthenticated])
    def search(self, request, **kwargs):
        search_txt = request.GET.get('q', '')
        if not search_txt:
            return self.list(request, **kwargs)
        else:
            queryset = self.filter_queryset(self.get_queryset())
            queryset = queryset.filter(
                Q(message__text__icontains=search_txt) |
                Q(message__sender_name__icontains=search_txt) |
                Q(operators_current__first_name__icontains=search_txt) |
                Q(operators_current__last_name__icontains=search_txt)
            ).distinct()

            page = self.paginate_queryset(queryset)
            if page is not None:
                serializer = self.get_serializer(page, many=True)
                return self.get_paginated_response(serializer.data)

            serializer = self.get_serializer(queryset, many=True)
            return Response(serializer.data)

    @staticmethod
    def parse_message(msg):
        try:
            if msg and msg.data:
                setattr(msg, 'data', json.loads(msg.data))
        except ValueError:
            pass
        return msg

    def get_queryset(self):
        website = self.request.current_site
        if not website or not self.request.user.is_authenticated:
            if self.request.nice_chat_client_id and self.action in ('close', 'rate', 'cliopen'):
                return Chat.objects.filter(client_id=self.request.nice_chat_client_id)
            else:
                return Chat.objects.none()

        return Chat.objects.filter(website=website, website__users=self.request.user)

    @action(detail=True, methods=['POST'], permission_classes=[IsAuthenticated])
    def invite(self, request, pk, **kwargs):
        chat = self.get_object()
        operator_id = self.request.data.get('operator_id')
        if not operator_id:
            return Response(status=status.HTTP_404_NOT_FOUND)

        operator_in_chat = Chat.objects.filter(operators_current=operator_id, pk=pk, website_id=chat.website.company.id)

        if operator_in_chat.exists():
            return Response(status=status.HTTP_302_FOUND)
        else:
            added_user = User.objects.filter(id=operator_id, company=self.request.user.company)
            # todo do we need a check here for if user is an agent?

            if added_user.exists():
                user = added_user.first()
                chat_obj = Chat.objects.get(pk=pk)
                chat_obj.operators_current.add(user.pk)
                txt = "~!JOINED:%s" % user.get_full_name or str(user)
                icon = user.avatar_thumbs().get('large', None)
                icon_url = icon.url if icon and icon.exists() else None

                chat_obj.put_tech_message_wdata(txt, {
                    'name': user.get_full_name or str(user),
                    'avatar': icon_url
                })

                models.signals.post_save.send(sender=Chat, instance=chat_obj, created=False, update_fields=[])

                return Response(status=status.HTTP_204_NO_CONTENT)
            else:
                return Response(status=status.HTTP_403_FORBIDDEN)

    @action(detail=True, methods=['POST'], permission_classes=[IsAuthenticated, ChatPermission])
    def join(self, request, pk, **kwargs):
        chat = self.get_object()
        self.check_object_permissions(request, chat)
        self.process_chat_join(chat, request.user)
        return Response(status=status.HTTP_204_NO_CONTENT)
    
    @staticmethod
    def process_chat_join(chat, request_user):
        chat.operators_current.add(request_user)
        chat.operators_history.add(request_user)
        txt = "~!JOINED:%s" % request_user.get_full_name
        event_track(JOIN_CHAT, request_user.id, {
            'chatId': chat.id,
            'clientID': chat.visitor.client_id if chat.visitor else '',
            'mp_country_code': chat.visitor.geo_country if chat.visitor else '',
            '$city': chat.visitor.geo_city if chat.visitor else '',
            'os': chat.visitor.sys_os if chat.visitor else '',
            '$browser': chat.visitor.sys_browser if chat.visitor else '',
        })

        icon = request_user.avatar_thumbs().get('large', None)
        icon_url = icon.url if icon and icon.exists() else None

        chat.put_tech_message_wdata(txt, {
            'name': request_user.get_full_name,
            'avatar': icon_url
        })
        join_watch_list(chat, request_user)
        visitor = Visitor.objects.filter(client_id=chat.client_id, website_id=chat.website_id).first()
        if visitor and visitor.state == Visitor.STATE_CALLING:
            visitor.state = Visitor.STATE_CHATING
            visitor.save(update_fields=('state',))
        c_key = "cli_wait-%s-%s" % (chat.website_id, chat.client_id)
        async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
            "type": WebsocketEventTypes.FRONTEND_CHAT_GA_EVENT, "event": "agjoin", "pk": chat.pk
        })})
        models.signals.post_save.send(sender=Chat, instance=chat, created=False, update_fields=[])

    @action(detail=True, methods=['GET'], permission_classes=[IsAuthenticated])
    def leave(self, request, pk, **kwargs):
        user = get_object_or_404(User, id=self.request.user.id, company=self.request.user.company)
        return self.provess_chat_leave(user, pk)

    @staticmethod
    def provess_chat_leave(request_user, pk):
        chat_obj = Chat.objects.get(pk=pk)
        if chat_obj.operators_current.filter(pk=request_user.pk).exists():
            chat_obj.operators_current.remove(request_user)
            txt = "~!LEAVE:%s" % request_user.get_full_name or str(request_user)
            chat_obj.put_tech_message(txt)
            leave_watch_list(chat_obj, request_user)
            event_track(LEAVE_CHAT, request_user.id, {'chatId': chat_obj.id})
            visitor = Visitor.objects.filter(client_id=chat_obj.client_id, website_id=chat_obj.website_id).first()
            if visitor and visitor.state == Visitor.STATE_CHATING:
                if chat_obj.operators_current.count() == 0:
                    visitor.state = Visitor.STATE_CALLING
                    visitor.save(update_fields=('state',))

            models.signals.post_save.send(sender=Chat, instance=chat_obj, created=False, update_fields=[])

        return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=True, methods=['POST'], permission_classes=[IsAuthenticated])
    def set_read(self, request, pk, **kwargs):
        chat = self.get_object()
        self.check_object_permissions(request, chat)
        try:
            msg_id = int(request.data.get('msg_id', '0'))
        except ValueError:
            msg_id = 0
        if msg_id:
            if set_sizes_by_user(chat, request.user, msg_id):
                return Response(status=status.HTTP_202_ACCEPTED)
            else:
                return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            return Response(status=status.HTTP_304_NOT_MODIFIED)

    @staticmethod
    def send_welcome(chat, sender_id, sender_name, website):

        welcome = website.props.get(website.WELCOME_MESSAGE, settings.GLOBAL_WELCOME_MESSAGE)
        if isinstance(welcome, list) or isinstance(welcome, tuple):
            for msg in welcome:
                Message.objects.create(chat=chat, sender_id=sender_id, sender_name=sender_name,
                                       text=msg, type=Message.TYPE_OPERATOR)
        elif isinstance(welcome, str):
            Message.objects.create(chat=chat, sender_id=sender_id, sender_name=sender_name,
                                   text=welcome, type=Message.TYPE_OPERATOR)

    @action(detail=True, methods=['POST'], permission_classes=[IsAuthenticated, ChatPermission])
    def finish(self, request, pk, **kwargs):
        chat = self.get_object()
        self.check_object_permissions(request, chat)

        chat.close()
        return Response(status=status.HTTP_204_NO_CONTENT)


class ArchivedChatViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = ArchivedChat.objects.all()
    SORTING = {
        '-open': '-ts_open',
        '+open': 'ts_open',
        'open': 'ts_open',
        '-close': '-ts_close',
        '+close': 'ts_close',
        'close': 'ts_close',
    }

    @staticmethod
    def conv_date(dt):
        if isinstance(dt, str):
            if dt.isdigit():
                return make_aware(datetime.datetime.fromtimestamp(int(dt)))
            try:
                return make_aware(datetime.datetime.strptime(dt, '%Y-%m-%d'))
            except ValueError:
                    pass
            try:
                return make_aware(datetime.datetime.strptime(dt, '%m/%d/%Y'))
            except ValueError:
                pass
        return None


    def get_queryset(self):
        website = self.request.current_site
        if not website or not self.request.user.is_authenticated:
            return ArchivedChat.objects.none()
        queryset = ArchivedChat.objects.filter(website=website)

        search_txt = self.request.GET.get('q', None)
        agents = self.request.GET.get('agents', None)
        date_from = self.conv_date(self.request.GET.get('date_from', None))
        date_to = self.conv_date(self.request.GET.get('date_to', None))
        sort_by = self.request.GET.get('sort_by', None)

        date_to = None if not date_to else date_to.replace(hour=23, minute=59, second=59)

        if search_txt:
            if search_txt.isdigit():
                queryset = queryset.filter(
                    Q(operators__id=int(search_txt)) | Q(id=int(search_txt))
                )
            else:
                queryset = queryset.filter(messages__icontains=search_txt)

        if agents and isinstance(agents, str):
            if ',' in agents:
                agents = [ag.strip() for ag in agents.split(',') if ag and ag.strip() and ag.strip().isdigit()]
            elif agents.strip().isdigit():
                agents = [agents.strip()]
            else:
                agents = None
            if agents:
                queryset = queryset.filter(operators__id__in=agents)
        
        if date_from and date_to:
            queryset = queryset.filter(
                (Q(ts_open__lte=date_from) & Q(ts_close__gte=date_to)) |
                (Q(ts_open__gte=date_from) & Q(ts_open__lte=date_to)) |
                (Q(ts_close__gte=date_from) & Q(ts_close__lte=date_to))
            )
        elif date_from and not date_to:
            queryset = queryset.filter(
                Q(ts_open__gte=date_from) | Q(ts_close__gte=date_from) 
            )
        elif date_to and not date_from:
            queryset = queryset.filter(
                Q(ts_open__lte=date_to) | Q(ts_close__lte=date_to) 
            )
        
        queryset = queryset.order_by(self.SORTING.get(sort_by, '-ts_close'))
        return queryset


    def get_serializer_class(self):
        if self.action == 'retrieve':
            return ArchivedChatSerializer
        else:
            return ArchivedChatShortSerializer

    def get_permissions(self):
        if self.action in ['list', 'retrieve']:
            # todo needs testing
            self.permission_classes = [IsAuthenticated]
        return super(ArchivedChatViewSet, self).get_permissions()

    @action(detail=False, methods=['get'], permission_classes=[IsAuthenticated])
    def search(self, request, **kwargs):
        return self.list(request, **kwargs)



