import json
import random
from asgiref.sync import async_to_sync

from django.db import models
from django.conf import settings
from django.utils import timezone
from django.utils.functional import cached_property
from django.core.cache import caches, InvalidCacheBackendError

from channels.layers import get_channel_layer

from ..website.models import Website
from nicechat.utils.arbiter import arbiter_request
from nicechat.utils.constants import WebsocketEventTypes


class Connector(models.Model):
    CONNECTOR_DEFAULT = 'nicechat'
    name = models.CharField(max_length=10, default=CONNECTOR_DEFAULT, unique=True)
    connection_key = models.CharField(max_length=100, default=None, blank=True, null=True)

    def __str__(self):
        return ' '.join(['Connector', "-", self.name, "(with key)" if self.connection_key else '' ])


class Idempotency(models.Model):
    MODE_CLASSIC = 1
    MODE_P2P = 2
    MODE_GROUP = 3
    MODES_CHOICE = (
        (MODE_CLASSIC, 'Guest-based chat',),
        (MODE_P2P, 'Private P2P',),
        (MODE_GROUP, 'Group room',),
    )
    value = models.CharField(max_length=100, primary_key=True)
    mode = models.IntegerField(choices=MODES_CHOICE, default=MODE_CLASSIC, db_index=True)
    connector = models.ForeignKey(Connector, default=0, on_delete=models.PROTECT)

    @property
    def client_id(self):
        peer = self.peers.first()
        return peer.peer if peer else None

    @cached_property
    def clients_list(self):
        return list(self.peers.values_list('peer', flat=True))

    @staticmethod
    def create_by_sid(session_id):
        Idempotency.objects.get_or_create(value=session_id)


class IdempotencyPeer(models.Model):
    node = models.ForeignKey(Idempotency, related_name='peers', on_delete=models.CASCADE)
    peer = models.CharField(max_length=100, db_index=True)
    name = models.CharField(max_length=100, default='Guest')
    info = models.CharField(max_length=1024, default='{}')

    class Meta:
        unique_together = ('node', 'peer')


class ChatVisitorBase(models.Model):
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    client = models.ForeignKey(Idempotency, on_delete=models.CASCADE)

    ts_open = models.DateTimeField(auto_now_add=True)
    ts_close = models.DateTimeField(blank=True, null=True, db_index=True)

    sys_ip = models.CharField(max_length=30, blank=True, null=True)
    sys_lang = models.CharField(max_length=30, blank=True, null=True)
    sys_lang_full = models.CharField(max_length=100, blank=True, null=True)
    sys_browser = models.CharField(max_length=30, blank=True, null=True)
    sys_browser_ver = models.CharField(max_length=60, blank=True, null=True)
    sys_os = models.CharField(max_length=30, blank=True, null=True)
    sys_os_ver = models.CharField(max_length=60, blank=True, null=True)
    sys_device = models.CharField(max_length=120, blank=True, null=True)
    sys_time = models.IntegerField(default=0)
    sys_info = models.TextField(blank=True, null=True)
    page_ref = models.CharField(max_length=200, default='')
    page_utm = models.JSONField(default=dict)
    geo_country = models.CharField(max_length=30, blank=True, null=True)
    geo_city = models.CharField(max_length=100, blank=True, null=True)
    geo_lat = models.FloatField(default=-90)
    geo_lon = models.FloatField(default=-180)
    obj_props = models.JSONField(default=dict)
    tags = models.CharField(max_length=980, db_index=True, default='') # max 50 tags at 16 chars

    fields_for_copy = ('sys_ip', 'sys_lang', 'sys_lang_full', 'sys_browser', 'sys_os', 'sys_time', 'sys_info',
                       'geo_country', 'geo_city', 'geo_lat', 'geo_lon', 'sys_browser_ver', 'sys_os_ver', 'sys_device')

    class Meta:
        abstract = True

    def get_client_pk(self):
        from ..visitors.models import Client
        client, is_created = Client.objects.get_or_create(
            website_id=self.website_id,
            client_id=self.client_id,
            defaults={
                'client_name': settings.DEFAULT_CLIENT_NAME
            }
        )
        return client.pk

    def get_client(self):
        from ..visitors.models import Client
        client, is_created = Client.objects.get_or_create(
            website_id=self.website_id,
            client_id=self.client_id,
            defaults={
                'client_name': settings.DEFAULT_CLIENT_NAME
            }
        )
        return client

    def get_client_name(self):

        token = "visitor-client-name-%s-%s" % (self.website_id, self.client_id)
        try:
            cache = caches['default']
            ret = cache.get(token)
        except InvalidCacheBackendError:
            ret = None

        if not ret:
            from ..visitors.models import Client
            client, is_created = Client.objects.get_or_create(
                website_id=self.website_id,
                client_id=self.client_id,
                defaults={
                    'client_name': settings.DEFAULT_CLIENT_NAME
                }
            )
            ret = client.client_name
            name_user = client.props.get("NAME", None)
            if name_user:
                ret = "%s (%s)" % (ret, name_user)
            try:
                cache = caches['default']
                cache.set(token, ret, 86400)
            except InvalidCacheBackendError:
                pass

        return ret


class Visitor(ChatVisitorBase):
    STATE_ONLINE='online'
    STATE_OFFLINE='offline'
    STATE_CALLING='calling'
    STATE_CHATING='chating'
    STATE_ENUM = (
        (STATE_ONLINE, 'ON-line',),
        (STATE_OFFLINE, 'OFF-line',),
        (STATE_CALLING, 'CALLing',),
        (STATE_CHATING, 'CHATing',),
    )
    invite_sent = models.DateTimeField(blank=True, null=True)
    invite_discard = models.DateTimeField(blank=True, null=True)
    state = models.CharField(choices=STATE_ENUM, default=STATE_OFFLINE, max_length=10)
    page_url = models.CharField(max_length=200, default='')
    page_title = models.CharField(max_length=200, default=' - ')
    session = models.CharField(max_length=100, blank=True, null=True)

    @property
    def sub_state(self):
        token = "visitor-state-%s-%s" % (self.website_id, self.client_id)
        try:
            cache = caches['default']
            ret = cache.get(token)
            return ret if ret else 'away'
        except InvalidCacheBackendError:
            pass
        return None

    @property
    def active_chat_id(self):
        chat = self.active_chat.filter(is_open=True).first()
        return chat.pk if chat else None

    @property
    def cart(self):
        cl = self.get_client()
        return cl.cart if cl else []

    @property
    def client_typing(self):
        try:
            cache = caches['default']
            return True if cache.get('chat-typing-s-%s-client' % self.client_id, None) else False
        except InvalidCacheBackendError:
            return False

    @cached_property
    def client_name(self):
        return self.get_client_name()

    @property
    def active_chat_obj(self):
        return self.active_chat.filter(is_open=True).first()

    def __str__(self):
        return ' - '.join(['Visitor', str(self.pk), self.website.domain, self.client_id, self.state])


class VisitorStatWide(ChatVisitorBase):
    website = models.IntegerField(db_index=True)
    client = models.CharField(max_length=100, db_index=True)
    duration = models.IntegerField(default=0)
    processed = models.BooleanField(default=False, db_index=True)
    visited_pages = models.JSONField(default=list)
    ts_open = models.BigIntegerField(default=0, db_index=True)
    ts_close = models.BigIntegerField(default=0, db_index=True)
    batam_processed = models.BooleanField(default=False, db_index=True)

    obj_props = None
    geo_lat = None
    geo_lon = None
    sys_info = None
    tags = None
    sys_lang_full = None


class BatamStatistics(models.Model):
    client = models.IntegerField(primary_key=True, verbose_name='CustomerID')
    website = models.IntegerField(db_index=True)
    name = models.CharField(max_length=100, db_index=True, verbose_name='Name')
    email = models.CharField(max_length=100, db_index=True, verbose_name='Email')
    total_duration = models.IntegerField(default=0, verbose_name='Secs on all pages')
    total_visits = models.IntegerField(default=0, verbose_name='No of visits')
    total_chats = models.IntegerField(default=0, verbose_name='No of chat')
    total_chats_agentinit = models.IntegerField(default=0, verbose_name='Invited by agent')
    first_visit_ts = models.BigIntegerField(default=0)               
    last_visit_ts = models.BigIntegerField(default=0, db_index=True)
    first_chat_ts = models.BigIntegerField(default=0)
    last_chat_ts = models.BigIntegerField(default=0, db_index=True)
    latest_chat_id = models.IntegerField(default=-1, db_index=True, verbose_name='Last chat ID')
    latest_page_url = models.CharField(max_length=512, verbose_name='Last page')
    latest_page_title = models.CharField(max_length=512, verbose_name='Last page')
    sys_browser = models.CharField(max_length=100, default='', verbose_name='Browser')
    sys_os = models.CharField(max_length=100, default='', verbose_name='OS')
    sys_device = models.CharField(max_length=100, default='', verbose_name='Device')
    geo_country = models.CharField(max_length=30, default='', verbose_name='Country')
    geo_city = models.CharField(max_length=100, default='', verbose_name='City')
    page_ref = models.CharField(max_length=200, default='', verbose_name='Came from')
    page_utm = models.JSONField(default=dict)
        
    # Sent greetings
    # Accept greetings
    # Ignored greetings
    # Last greeting
    # Group


class VisitedPage(models.Model):
    visitor = models.ForeignKey(Visitor, related_name='visited_pages', on_delete=models.CASCADE)
    page_url = models.CharField(max_length=200, default='')
    page_title = models.CharField(max_length=200, default=' - ')
    opened_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):
        return ': '.join([self.page_title, self.page_url])


class Chat(ChatVisitorBase):
    is_open = models.BooleanField(default=True)
    is_serialized = models.BooleanField(default=False, db_index=True)
    result_score = models.PositiveSmallIntegerField(blank=True, null=True)
    result_description = models.TextField(blank=True, null=True)
    of_client_score = models.PositiveSmallIntegerField(blank=True, null=True)
    of_client_description = models.TextField(blank=True, null=True)
    result_purchase = models.CharField(max_length=100, blank=True, null=True)
    is_connector = models.BooleanField(default=False, db_index=True)
    connector_name = models.CharField(max_length=10, default='')
    operators_current = models.ManyToManyField(settings.AUTH_USER_MODEL, related_name='operators_current', blank=True)
    operators_history = models.ManyToManyField(settings.AUTH_USER_MODEL, related_name='operators_history', blank=True)
    visitor = models.ForeignKey(Visitor, related_name='active_chat', blank=True, null=True, on_delete=models.SET_NULL)
    invited_by_agent = models.PositiveIntegerField(blank=True, null=True)

    def __str__(self):
        return ' - '.join(['Chat', str(self.pk), self.website.domain, self.get_client_name(), self.client_id])

    def put_tech_message(self, txt, is_synthetic=False):
        Message(text=str(txt), sender_id="0", sender_name="System", is_synthetic=is_synthetic,  chat=self, type=Message.TYPE_TECH).save()

    def put_tech_message_wdata(self, txt, data, is_synthetic=False):
        Message(text=str(txt), sender_id="0", data=json.dumps(data), is_synthetic=is_synthetic, sender_name="System", chat=self, type=Message.TYPE_TECH).save()

    def clear_call_messages(self):
        for msg in self.message_set.filter(sender_id="0", type=Message.TYPE_TECH, text='~!CALL_AGT_BEGIN:1'):
            msg.text='~!CALL_AGT_BEGIN_DONE:1'
            msg.save(update_fields=['text'])
        for msg in self.message_set.filter(sender_id="0", type=Message.TYPE_TECH, text='~!CALL_CLI_BEGIN:1'):
            msg.text='~!CALL_CLI_BEGIN_DONE:1'
            msg.save(update_fields=['text'])

    @staticmethod
    def create(force_empty=False, **kwargs): # OLD method
        client_id = kwargs.get('client_id', None)
        if client_id and not force_empty:
            cl = Idempotency.objects.filter(pk=client_id, mode=Idempotency.MODE_CLASSIC).first()
            if not cl:
                cl = Idempotency.objects.create(
                    pk=client_id,
                    mode=Idempotency.MODE_CLASSIC
                )
                IdempotencyPeer.objects.get_or_create(peer=client_id, node=cl)
            kwargs['client'] = cl
            del kwargs['client_id']
            if settings.NODE_MODE != settings.NODE_MODE_STANDALONE:
                arb_data = arbiter_request(settings.NODE_ARBITER_HOST, 'getchatid')
                if arb_data and arb_data.get('chat_id'):
                    kwargs['pk'] = arb_data.get('chat_id')
        return Chat.objects.create(**kwargs)

    @cached_property
    def unread_count(self):
        return self.pk

    @property
    def client_typing(self):
        try:
            cache = caches['default']
            return True if cache.get('chat-typing-%s-client' % self.pk, None) else False
        except InvalidCacheBackendError:
            return False

    @cached_property
    def mode(self):
        return self.client.mode

    @cached_property
    def client_name(self):
        return self.get_client_name()

    @property
    def cart(self):
        cl = self.get_client()
        return cl.cart if cl else []

    @cached_property
    def typing(self):
        return self

    @cached_property
    def website_b(self):
        return self.obj_props.get('website_b', None)

    @cached_property
    def operators_info(self):
        if not settings.OPERATORS_INFO:
            return {}
        data = {}
        for op in self.operators_current.select_related('company').all():
            data[op.pk] = {
                'name': op.get_full_name,
                'user_id': op.pk,
                'avatar': op.avatar_thumbs(),
                'company': op.company.name if op.company else '',
                'company_id': op.company.pk if op.company else ''
            }
        return data

    @cached_property
    def operators(self):
        return u', '.join([op['name'] for op in self.operators_info.values() if op.get('name')])

    @cached_property
    def companies(self):
        cmps = set()
        for op in self.operators_info.values():
            if op.get('company'):
                cmps.add(op['company'])
        return u', '.join(list(cmps))

    def close(self):
        self.is_open = False
        self.ts_close = timezone.now()
        self.save(update_fields=['is_open', 'ts_close'])

        # remove visitor
        if self.visitor:
            self.visitor.state = Visitor.STATE_ONLINE
            self.visitor.ts_close = None
            self.visitor.save(update_fields=('state', 'ts_close',))

        self.visitor = None
        self.save(update_fields=['visitor'])

        # notify about chat closed
        self.put_tech_message('~!CLOSEDBYAGENT:')
        channel_layer = get_channel_layer()
        c_key = "cli_speak-%s" % self.pk
        async_to_sync(channel_layer.group_send)(
            c_key,
            {
                "type": "websocket.send",
                "text": json.dumps({
                    "type": WebsocketEventTypes.FRONTEND_CHAT_CLOSE,
                    "pk": self.pk
                })
            }
        )

        # schedule chat archive
        from apps.chat.utils import move_chat_to_archive
        move_chat_to_archive(self, True)


class Message(models.Model):
    STATE_NEW = 0
    STATE_SENT = 1
    STATE_DELIVERED = 2
    STATE_WAS_READ = 3
    STATE_ENUM = (
        (STATE_NEW, 'New',),
        (STATE_SENT, 'Sent',),
        (STATE_DELIVERED, 'Delivered',),
        (STATE_WAS_READ, 'Was read',),
    )
    TYPE_CLIENT = 'cli'
    TYPE_OPERATOR = 'empl'
    TYPE_TECH = 'tech'
    TYPE_EVENT_CLI = 'evtc'
    TYPE_EVENT_OPER = 'evta'
    TYPE_ENUM = (
        (TYPE_CLIENT, 'Client',),
        (TYPE_OPERATOR, 'Operator',),
        (TYPE_TECH, 'Tech-info',),
        (TYPE_EVENT_CLI, 'Event (for client)',),
        (TYPE_EVENT_OPER, 'Event (for operator)',),
    )
    DATA_SUGGEST = 'suggest'
    DATA_SUGGEST_DONE = 'd_suggst'
    DATA_SUGGEST_CANCEL = 'c_suggst'
    DATA_ATTACH = 'attach'
    DATA_ANSWERS = 'ans'
    DATA_ANSWERS_REPLY = 'ansrpl'
    DATA_FORM = 'form'
    DATA_FORM_DONE = 'f_done'
    DATA_FORM_CANCEL = 'f_cancel'
    DATA_FORM_FILLED = 'f_filled'
    DATA_PRODUCT_VIEWED = 'p_viewed'
    DATA_PRODUCT_BUY = 'p_buy'
    DATA_ = 'f_sent'
    DATA_ENUM = (
        (DATA_SUGGEST, 'Suggest sent',),
        (DATA_SUGGEST_DONE, 'Suggest complete (order created)',),
        (DATA_SUGGEST_CANCEL, 'Suggest canceled',),
        (DATA_ATTACH, 'Attach sent',),
        (DATA_ANSWERS, "Message with answers",),
        (DATA_FORM, 'Form sent',),
        (DATA_FORM_DONE, 'Form filled',),
        (DATA_FORM_CANCEL, 'Form canceled',),
        (DATA_PRODUCT_VIEWED, 'Product viewed',),
        (DATA_PRODUCT_BUY, 'Product want to buy',),
        (DATA_FORM_FILLED, 'Filled form data',),
    )
    chat = models.ForeignKey(Chat, on_delete=models.CASCADE)
    sender_id = models.CharField(max_length=60, blank=True, null=True)
    sender_name = models.CharField(max_length=60, blank=True, null=True)
    timestamp = models.DateTimeField(auto_now=True)
    text = models.TextField()
    data = models.TextField(blank=True, null=True)
    hidden = models.TextField(blank=True, null=True)
    is_synthetic=models.BooleanField(default=False)
    is_processed=models.BooleanField(default=False)
    data_type = models.CharField(choices=DATA_ENUM, default=None, max_length=8, blank=True, null=True)
    state = models.SmallIntegerField(choices=STATE_ENUM, default=STATE_SENT)
    type = models.CharField(choices=TYPE_ENUM, default=TYPE_CLIENT, max_length=5, db_index=True)

    def __str__(self):
        return ': '.join([str(self.chat), self.text[:100]])


class Attachment(models.Model):
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    chat = models.ForeignKey(Chat, on_delete=models.CASCADE)
    message = models.ForeignKey(Message, on_delete=models.CASCADE)
    file_type = models.CharField(max_length=100, blank=True, null=True)
    upload_ts = models.DateTimeField(auto_now=True)
    size = models.IntegerField(blank=True, null=True)
    url = models.URLField()


class ChatReadStack(models.Model):
    chat = models.ForeignKey(Chat, related_name='read_stack', on_delete=models.CASCADE)
    operator = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='read_stack', 
                                 blank=True, null=True, on_delete=models.CASCADE)
    client_id = models.CharField(max_length=100, db_index=True, blank=True, null=True)
    chat_top_msg_id = models.IntegerField(blank=True, null=True)
    user_top_msg_id = models.IntegerField(blank=True, null=True)
    chat_msg_count = models.IntegerField(default=0)
    user_msg_read = models.IntegerField(default=0)
    unread_count = models.IntegerField(default=0, db_index=True)
    user_msg_notify_id = models.IntegerField(default=0)
    user_msg_notify_cnt = models.IntegerField(default=0)
    user_msg_notify_ts = models.BigIntegerField(default=0)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        assert self.client_id is not None or self.operator is not None
        super(ChatReadStack, self).save(force_insert, force_update, using, update_fields)


class ChatOuterReplyKeys(models.Model):
    chat = models.ForeignKey(Chat, related_name='reply_keys', on_delete=models.CASCADE)
    operator = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='reply_keys', on_delete=models.CASCADE)
    key = models.CharField(max_length=64, primary_key=True)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if not self.key:
            self.key = ''.join(random.choice('ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789') for _ in range(32))
        super(ChatOuterReplyKeys, self).save(force_insert, force_update, using, update_fields)

    class Meta:
        unique_together = ('chat', 'operator')


class ArchivedChat(ChatVisitorBase):
    client_name = models.CharField(max_length=100, blank=True, null=True)
    name = models.CharField(max_length=60)
    result_score = models.PositiveSmallIntegerField(blank=True, null=True)
    result_description = models.TextField(blank=True, null=True)
    of_client_score = models.PositiveSmallIntegerField(blank=True, null=True)
    of_client_description = models.TextField(blank=True, null=True)
    result_purchase = models.CharField(max_length=100, blank=True, null=True)
    operators = models.ManyToManyField(settings.AUTH_USER_MODEL, related_name='archived_operators_history', blank=True)
    operators_count = models.PositiveIntegerField(default=0, db_index=True)
    batam_processed = models.BooleanField(default=False, db_index=True)

    messages = models.TextField(default="{}")
    attachments = models.TextField(default="[]")

    ts_clean_attachments = models.DateTimeField(blank=True, null=True, db_index=True)

    invited_by_agent = models.PositiveIntegerField(blank=True, null=True)

    def __str__(self):
        return ': '.join([self.website.domain, self.client_name, str(self.of_client_score)])

    @cached_property
    def website_b(self):
        return self.obj_props.get('website_b', None)

    @cached_property
    def mode(self):
        return self.client.mode

    @cached_property
    def operators_info(self):
        if not settings.OPERATORS_INFO:
            return {}
        data = {}
        for op in self.operators.select_related('company').all():
            data[op.pk] = {
                'name': op.get_full_name,
                'avatar': op.avatar_thumbs(),
                'company': op.company.name if op.company else ''
            }
        return data

    @cached_property
    def operators_list(self):
        return u', '.join([op['name'] for op in self.operators_info.values() if op.get('name')])

    @cached_property
    def companies(self):
        cmps = set()
        for op in self.operators_info.values():
            if op.get('company'):
                cmps.add(op['company'])
        return u', '.join(list(cmps))

    class Meta:
        ordering = ('-id',)

