import email
import re
import os
import json
import base64
import email.header
from time import time, sleep
from datetime import timedelta, datetime
from poplib import POP3_SSL
from django.db.models import F, Max
from django.utils import timezone
from django.conf import settings
from django.template.loader import get_template
from django.core.mail import EmailMultiAlternatives

from nicechat.utils.email_reply_parser import EmailReplyParser
from nicechat.utils.attach import simple_save
from .models import Chat, ChatOuterReplyKeys, ChatReadStack, Message, Idempotency, Attachment


email_re = re.compile(".*\+([A-Za-z0-9]+)\@.*", flags=re.MULTILINE)
subject_re = re.compile(".*\[([A-Za-z0-9]+)\].*", flags=re.MULTILINE)


def process_notificator_sender():
    if not settings.NOTIFICATOR_ENABLE:
        print('Notificator is disabled')
    if (time() // 300) % 10 == 0:
        cleanup_unread()
    get_unread()


def cleanup_unread():
    c = Chat.objects.filter(
        is_open=True,
    ).annotate(
        message_id=Max('message__id'),
        message_ts=Max('message__timestamp')
    ).values('pk', 'message_id', 'message_ts')

    ts = timezone.now() - timedelta(minutes=settings.NOTIFICATOR_IGNORE_MIN)
    
    for ch in c:
        if ch.get('message_ts') < ts and ch.get('message_id'):
            ChatReadStack.objects.filter(chat__pk=ch.get('pk')).update(
                user_msg_notify_id=ch.get('message_id'),
                user_msg_notify_ts = time())


def get_unread():
    un = ChatReadStack.objects.filter(
        unread_count__gt=0,
        chat__is_open=True,
        chat_top_msg_id__gt=F('user_msg_notify_id'),
        user_msg_notify_ts__lt=time() - settings.NOTIFICATOR_STEP_MIN * 60
    ).exclude(chat__client__mode = Idempotency.MODE_CLASSIC)
    for unread in un:
        try:
            _process_notification(unread)
        except Exception as e:
            print(e)

def _process_notification(unread):
    msgx = unread.chat.message_set.filter(pk__gt=unread.user_msg_notify_id).exclude(type=Message.TYPE_TECH).order_by('id').values('text', 'sender_name', 'timestamp')
    if not msgx:
        max_pk = unread.chat.message_set.aggregate(max_pk = Max('pk'))['max_pk']
        unread.user_msg_notify_id = max_pk
        unread.user_msg_notify_ts = time()
        unread.save(update_fields=('user_msg_notify_id', 'user_msg_notify_ts'))
        print('EMAIL-SHIFT', unread.chat.pk, max_pk)
        print(msgx)
        return False

    msg = []
    dt = timezone.now() - timedelta(minutes=settings.NOTIFICATOR_STEP_MIN)
    _pd = None
    _pu = None
    _pp = None
    for m in msgx:
        if m.get('timestamp') and m.get('timestamp') < dt:
            _d = m.get('timestamp').strftime('%d %b, %Y')
            _u = m.get('timestamp').strftime('%H:%M')
            m['t'] = ''
            m['d'] = ''
            _p = m.get('sender_name')
            if _pp == _p:
                m['sender_name'] = ''
            else:
                m['t'] = _u
            if _pu != _u:
                m['t'] = _u
            if _pd != _d:
                m['d'] = _d
            msg.append(m)
            _pd = _d
            _pu = _u
            _pp = _p
        else:
            continue
    if not msg:
        return False

    cr, _ = ChatOuterReplyKeys.objects.get_or_create(chat=unread.chat, operator=unread.operator)
    cr_key = cr.key

    if unread.chat.client.mode == Idempotency.MODE_P2P:
        user = unread.chat.operators_current.exclude(pk=unread.operator.pk).first()
        chat_name = user.get_full_name or str(user)
        chat_mode = 'private chat with'
    elif unread.chat.client.mode == Idempotency.MODE_GROUP:
        chat_name = unread.chat.client_name
        chat_mode = 'group chat'
    else:
        chat_name = 'chat #%s' % unread.chat.pk
    
    context = {
        'chat_name': chat_name,
        'chat_mode': chat_mode,
        'key': cr_key,
        'messages': msg
    }
    email = settings.NOTIFICATOR_TARGET(unread.operator)

    template = get_template('chat/email/notification.html')
    text = get_template('chat/email/notification.txt')
    subject = get_template('chat/email/notification.subject')
    text_content = text.render(context)
    html_content = template.render(context)
    subject_content = subject.render(context)
    
    reply_to = settings.NOTIFICATOR_REPLY if settings.NOTIFICATOR_REPLY else settings.DEFAULT_FROM_EMAIL
    if not settings.NOTIFICATOR_PLUS_REPLY:
        subject_content += ' [' + cr_key + ']'
    try:
        subject_content = subject_content.replace('&amp;', '&').replace('&quot;', '"').replace('&#39;', "'").replace("\n", " ").replace("\\\\n", " ")
        if isinstance(subject_content, str):
            cc = subject_content.encode('utf-8')
        subject_content = '=?UTF-8?B?' + base64.b64encode(cc).decode('ascii') + '?='
    except Exception:
        pass

    headers = {
        'Reply-To': reply_to.replace('@', '+%s@' % cr_key) if settings.NOTIFICATOR_PLUS_REPLY else reply_to, 
        'X-Chat-ID': cr_key
    }

    msg = EmailMultiAlternatives(
        subject_content, text_content,
        settings.DEFAULT_FROM_EMAIL,
        [email], headers=headers)
    msg.attach_alternative(html_content, "text/html")

    try:
        msg.send()
        print('EMAIL-OK', unread.chat.pk, subject_content, email)
    except Exception as e:
        print('EMAIL-ERRR', unread.chat.pk, str(e))
        return False
    max_pk = unread.chat.message_set.aggregate(max_pk = Max('pk'))['max_pk']
    unread.user_msg_notify_id = max_pk
    unread.user_msg_notify_ts = time()
    unread.save(update_fields=('user_msg_notify_id', 'user_msg_notify_ts'))
    sleep(0.5)
    return True


def email_message_parser():
    if not settings.NOTIFICATOR_ENABLE:
        print('Notificator is disabled')
    if not settings.POP3_SERVER or not settings.POP3_USER or not settings.POP3_PASS:
        print('POP3 configuration incompleted')
    _read_messages()


def _parse_reply_r(get_payload, level):
    if level > 10:
        return None

    for p in get_payload:
        # print(level, '.>>>>>>>', p)
        # print(p.get_content_type())
        # print(p.get('Content-Disposition'))
        if p.get_content_type() == 'text/plain' and 'attachment' not in str(p.get('Content-Disposition')):
            pl = p.get_payload(decode=True)
            return EmailReplyParser.parse_reply(pl)
        elif p.get_content_type() in ['multipart/alternative', 'multipart/related']:
            return _parse_reply_r(p.get_payload(), level + 1)


def _parse_reply(lines, msg_uid):
    msg = email.message_from_string(u'\r\n'.join(lines))
    # print(msg)
    reply = None
    attachs = []
    reply = _parse_reply_r(msg.get_payload(), 0)

    if settings.NOTIFICATOR_ALLOW_ATTACH:
        for p in msg.walk():
            if p.get_content_maintype() == 'multipart' or p.get('Content-Disposition') is None:
                continue
            fileName = p.get_filename()
            mime = p.get_content_type()
            content = p.get_payload(decode=True)
            if fileName:
                attachs.append((fileName, mime, content)) 

    if not attachs and settings.NOTIFICATOR_ALLOW_ATTACH:
        for p in msg.get_payload():
            if 'attachment' in str(p.get('Content-Disposition')):
                fileName = p.get_filename()
                mime = p.get_content_type()
                content = p.get_payload(decode=True)
                if fileName:
                    attachs.append((fileName, mime, content)) 

    # print('>==>==>', reply)
    if not reply and not attachs:
        print(msg_uid, 'No reply found')
        return False

    addr_to = msg['to']
    addr_to = email_re.search(addr_to)
    replyKey = None
    if addr_to:
        replyKey = addr_to.group(1)
    if not replyKey:
        subj = msg['subject']
        subj = subject_re.search(subj)
        if subj:
            replyKey = subj.group(1)
    if not replyKey:
        print (msg_uid, 'replyKey not found')
        return False
    rk = ChatOuterReplyKeys.objects.filter(pk=replyKey).first()
    if not rk:
        print (msg_uid, 'chat not found by replyKey')
        return False
    UID = "msgUID%s" % msg_uid
    for mj in rk.chat.message_set.filter(type=Message.TYPE_OPERATOR).exclude(hidden__isnull=True):
        if mj.hidden and UID in mj.hidden:
            print (msg_uid, 'reply duplicate')
            return False

    sender_name = rk.operator.get_full_name or str(rk.operator)
    if reply:
        Message(
            text=reply,
            sender_id=rk.operator.pk,
            hidden=json.dumps({"UID": "msgUID%s" % msg_uid}),
            sender_name=sender_name,
            chat=rk.chat,
            type=Message.TYPE_OPERATOR).save()
    
    for (_a_name, a_mime, a_cnt) in attachs:
        a_name = _parse_u_encode(_a_name)
        r_file, state = simple_save(a_cnt, a_mime, a_name, datetime.now().strftime('%Y%m%d'))
        if state:
            message = Message.objects.create(
                chat=rk.chat,
                sender_id=rk.operator.pk,
                sender_name=sender_name,
                text=r_file['name'],
                data=json.dumps({
                    'file_type': r_file['content_type'],
                    'size': r_file['size'],
                    'preview': r_file.get('preview', None),
                }),
                hidden=json.dumps({"UID": "msgUID%s" % msg_uid}),
                data_type=Message.DATA_ATTACH,
                type=Message.TYPE_OPERATOR
            )

            Attachment.objects.create(
                website=rk.chat.website,
                chat=rk.chat,
                message=message,
                file_type=r_file['content_type'],
                size=r_file['size'],
                url=r_file['path']
            )

    print (msg_uid, 'ok! processed')

    return True

def _parse_u_encode(s):
    if isinstance(s, str) and s and '=?' in s:
        try:
            return u''.join(
                word.decode(encoding or 'utf8') if isinstance(word, bytes) else word
                for word, encoding in email.header.decode_header(s))
        except Exception:
            pass
    return s

def _read_messages():
    box = POP3_SSL(settings.POP3_SERVER)
    box.user(settings.POP3_USER)
    box.pass_(settings.POP3_PASS)

    response, lst, octets = box.uidl()

    for msgrec in lst:
        try:
            msg_num, msg_uid = msgrec.split()
            resp, lines, octets = box.retr(msg_num)
            _parse_reply(lines, msg_uid)
            pth = os.path.join(settings.BASE_DIR, 'pop3', '%s.eml' % msg_uid)
            if not os.path.exists(pth):
                with open(pth,  'w') as f:
                    f.write(u'\r\n'.join(lines))
            box.dele(msg_num)
        except Exception:
            pass

    box.quit()
