from django.conf import settings
from rest_framework import serializers, fields
import json
import hashlib
from django.core.cache import caches, InvalidCacheBackendError
from ..chat.models import Message, Chat, ArchivedChat, ChatReadStack
from .utils import filter_tags
from nicechat.utils.attach import make_seed

class DataFieldSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        if obj:
            try:
                return json.loads(obj)
            except ValueError:
                pass
        return None

    def to_internal_value(self, data):
        if isinstance(data, dict) or isinstance(data, list):
            try:
                data = json.dumps(data)
            except ValueError:
                data = None
        if isinstance(data, str):
            return data
        return None


class TagsSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        if obj:
            return obj.split(',')
        return []

    def to_internal_value(self, data):
        if isinstance(data, list):
            return ','.join(filter_tags(data))
        return ""


class MessageSerializer(serializers.ModelSerializer):
    sender_id = serializers.CharField(default=getattr(serializers.CurrentUserDefault(), 'pk', 'err-None'))
    data = DataFieldSerializer(many=False, required=False)
    seed = serializers.SerializerMethodField()

    def get_seed(self, msg):
        if msg and msg.data_type == Message.DATA_ATTACH:
            return make_seed(msg.pk)
        return None

    class Meta:
        model = Message
        fields = ('id', 'text', 'data', 'data_type', 'sender_id', 'is_synthetic', 'sender_name', 'chat_id', 'state', 'type', 'timestamp', 'seed')
        read_only_fields = ('id', 'chat_id', 'chat', 'timestamp', 'seed')


class UnreadByChatSerializer(serializers.BaseSerializer):

    def to_representation(self, obj):
        if hasattr(self, 'context') and isinstance(self.context, dict):
            request = self.context.get('request', None)
            if request and hasattr(request, 'user') and request.user.is_authenticated:
                unread = ChatReadStack.objects.filter(operator=request.user,
                                                      chat_id=obj).first()

                return unread.unread_count if unread else 0

        return -1


class ChatTypingSerializer(serializers.CurrentUserDefault, serializers.BaseSerializer):

    def to_representation(self, obj):
        try:
            cache = caches['default']
        except InvalidCacheBackendError:
            return []

        is_auth = False
        if hasattr(self, 'context') and isinstance(self.context, dict):
            request = self.context.get('request', None)
            if request and hasattr(request, 'user') and request.user.is_authenticated:
                is_auth = True

        result = []

        for user in obj.operators_current.all():
            text = cache.get('chat-typing-%s-%s' % (obj.pk, user.pk), None)
            if text:
                result.append({
                    'id': user.pk,
                    'name': user.get_full_name or str(user),
                    'text': text
                })
        if is_auth:
            text = cache.get('chat-typing-%s-client' % obj.pk, None)
            if text:
                result.append({
                    'id': 0,
                    'name': 'client',
                    'text': text
                })

        return result

class MiniUserSerializer(serializers.Serializer):
    avatar = serializers.DictField(read_only=True, child=fields.ImageField())
    name = serializers.ReadOnlyField()
    company = serializers.ReadOnlyField()


class ChatSerializer(serializers.ModelSerializer):
    # visitor = VisitorSerializer(many=False, read_only=True, required=False)
    unread_count = UnreadByChatSerializer(many=False, read_only=True, required=False)
    typing = ChatTypingSerializer(many=False, read_only=True, required=False)
    cart = serializers.ReadOnlyField()
    tags = TagsSerializer(required=False)
    client_name = serializers.ReadOnlyField()
    client_typing = serializers.ReadOnlyField()
    mode = serializers.ReadOnlyField()
    operators_info = serializers.DictField(read_only=True, child=MiniUserSerializer())

    class Meta:
        model = Chat
        fields = '__all__'


class ChatBotSerializer(ChatSerializer):
    client = serializers.SerializerMethodField()

    def get_client(self, obj):
        return hashlib.sha256(f"{obj.client}".encode()).hexdigest()


class UserHistorySerializer(serializers.ModelSerializer):
    class Meta:
        model = Chat
        fields = ('operators_current', 'operators_history')


class JoinSerializer(serializers.Serializer):
    operator_id = serializers.IntegerField()


class ArchivedChatShortSerializer(serializers.ModelSerializer):
    class Meta:
        model = ArchivedChat
        fields = ('id', 'name', 'client_name', 'ts_open', 'ts_close', 'of_client_score')


class MessagesFieldSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        try:
            return json.loads(obj)
        except ValueError:
            return []


class ArchivedChatSerializer(serializers.ModelSerializer):
    messages = MessagesFieldSerializer(many=False, read_only=True, required=False)
    client_name = serializers.ReadOnlyField()

    class Meta:
        model = ArchivedChat
        fields = '__all__'
