import json
from django.utils import timezone
from django.db.models.signals import post_save, post_delete, pre_delete, pre_save, m2m_changed
from django.dispatch import receiver
from django.conf import settings
from .models import Visitor, Chat, Message, ChatReadStack, VisitorStatWide
# from ..visitors.serializers import VisitorSerializer
from .serializers import ChatBotSerializer
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()
from nicechat.utils.constants import WebsocketEventTypes
from nicechat.utils.callback import make_callback, EVENT_NEW_CHAT
from .utils import set_sizes_from_chat
from .tasks import chat_activity_wdt
from django.utils.module_loading import import_string
from apps.connector.dispatch import connector_prepare_message_to
from apps.connector.consumers import prepare_send_data

ChatSerializer = import_string(settings.CHAT_SERIALIZER)
MessageSerializer = import_string(settings.MESSAGE_SERIALIZER)
VisitorSerializer = import_string(settings.VISITOR_SERIALIZER)


@receiver(post_save, sender=Visitor)
def visitor_post_save(sender, instance=None, **kwargs):
    if instance:
        a_key = "adm-website_mon-%i" % instance.website_id
        async_to_sync(channel_layer.group_send)(a_key, {"type": "websocket.send",  "text":  json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_VISITOR_UPDATE,
                                               "data": dict(VisitorSerializer(instance=instance)
                                                            .to_representation(instance=instance)),
                                               "pk": instance.pk,
                                               "created": kwargs.get('created', None)
                                               })})
        if instance.active_chat_obj:
            chat_post_save_small(instance.active_chat_obj)


@receiver(pre_delete, sender=Visitor)
def visitor_pre_delete(sender, instance=None, **kwargs):
    if instance:
        visited_pages = []
        for p in instance.visited_pages.all():
            visited_pages.append({'page_url': p.page_url, 'page_title': p.page_title, 'opened_at': str(p.opened_at)})
        try:
            duration = int((timezone.now() - instance.ts_open).total_seconds())
        except (ValueError, TypeError):
            duration = 0

        VisitorStatWide(
            website = instance.website_id,
            client = instance.client_id,
            ts_open = instance.ts_open.timestamp() if instance.ts_open else timezone.now().timestamp(),
            ts_close = timezone.now().timestamp(),
            sys_lang = instance.sys_lang,
            sys_time = instance.sys_time,
            sys_os = instance.sys_os,
            sys_browser = instance.sys_browser,
            sys_browser_ver = instance.sys_browser_ver,
            sys_os_ver = instance.sys_os_ver,
            sys_device = instance.sys_device,
            geo_country = instance.geo_country,
            geo_city = instance.geo_city,
            visited_pages = visited_pages,
            page_utm = instance.page_utm,
            page_ref = instance.page_ref,
            duration = duration
        ).save()
    

@receiver(post_delete, sender=Visitor)
def visitor_post_delete(sender, instance=None, **kwargs):
    if instance:
        a_key = "adm-website_mon-%i" % instance.website_id
        async_to_sync(channel_layer.group_send)(a_key, {"type": "websocket.send",  "text":  json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_VISITOR_REMOVE,
                                               "data": {"id": instance.pk},
                                               "pk": instance.pk
                                               })
                           })
    if instance.active_chat_obj:
        chat_post_save_small(instance.active_chat_obj)


def check_adm_available(chat):
    if chat.is_open == False:
        return False
    if chat.operators_current.count() == 0:
        return False
    return True


@receiver(post_save, sender=Chat)
def chat_post_save(sender, instance=None, **kwargs):
    if instance:
        created = kwargs.get('created', False)
        a_key = "adm-website_mon-%i" % instance.website_id
        a_chat_key = "chat-%s" % instance.pk
        msg = {"type": "websocket.send", "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_CHAT_CREATE if created else
        WebsocketEventTypes.BACKEND_SITE_CHAT_UPDATE,
                                   "data": dict(ChatSerializer(instance=instance)
                                                .to_representation(instance=instance)),
                                   "pk": instance.pk
                                   })}
        async_to_sync(channel_layer.group_send)(a_key, msg)
        async_to_sync(channel_layer.group_send)(a_chat_key, msg)
        if instance.website_b:
            ax_key = "adm-website_mon-%i" % instance.website_b
            async_to_sync(channel_layer.group_send)(ax_key, msg)
        if created:
            c_key = "cli_wait-%s-%s" % (instance.website_id, instance.client_id)
            async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
                "type": WebsocketEventTypes.FRONTEND_CHAT_CREATED, "pk": instance.pk
            })})
            async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
                "type": WebsocketEventTypes.FRONTEND_CHAT_GA_EVENT, "event": "open", "pk": instance.pk
            })})
            if instance.obj_props.get('website_b'):
                c_key = "cli_wait-%s-%s" % (instance.obj_props.get('website_b'), instance.client_id)
                async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
                    "type": WebsocketEventTypes.FRONTEND_CHAT_CREATED, "pk": instance.pk
                })})
                async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
                    "type": WebsocketEventTypes.FRONTEND_CHAT_GA_EVENT, "event": "open", "pk": instance.pk
                })})
        if instance and not instance.is_open:
            c_key = "cli_speak-%s" % instance.pk
            async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
                "type": WebsocketEventTypes.FRONTEND_CHAT_CLOSED, "pk": instance.pk
            })})
        if created:
            chat_activity_wdt(instance.pk, instance.website.props)
            make_callback.apply_async(args=(EVENT_NEW_CHAT, instance.website_id, "Chat", instance.pk,), countdown=2)

        bot_type = None
        if not instance.is_open:
            bot_type = WebsocketEventTypes.BOT_ON_ENDROOM
        elif not created:
            bot_type = WebsocketEventTypes.BOT_ON_EDTROOM

        if bot_type:
            databot = {
                "type": "websocket.send",
                "text": json.dumps({
                    "type": bot_type,
                    "data": ChatBotSerializer(instance=instance).to_representation(instance=instance),
                    "chat_id": instance.pk
                })
            }
            async_to_sync(channel_layer.group_send)("bot_site-%s" % instance.website_id, databot)


@receiver(pre_save, sender=Chat)
def chat_pre_save(sender, instance=None, **kwargs):
    if instance and not instance.is_open:
        old = Chat.objects.filter(pk=instance.pk).first()
        if old and old.is_open != instance.is_open:
            c_key = "cli_wait-%s-%s" % (instance.website_id, instance.client_id)
            async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
                "type": WebsocketEventTypes.FRONTEND_CHAT_GA_EVENT, "event": "close", "pk": instance.pk
            })})


def chat_post_save_small(instance):
    if not instance:
        return
    a_key = "adm-website_mon-%i" % instance.website_id
    a_chat_key = "chat-%s" % instance.pk
    msg = {"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_CHAT_UPDATE,
                               "data": dict(ChatSerializer(instance=instance)
                                            .to_representation(instance=instance)),
                               "pk": instance.pk
                               })}
    async_to_sync(channel_layer.group_send)(a_key, msg)
    async_to_sync(channel_layer.group_send)(a_chat_key, msg)
    if instance.obj_props.get('website_b'):
        a_key =  "adm-website_mon-%i" % instance.obj_props.get('website_b')
        async_to_sync(channel_layer.group_send)(a_key, msg)



@receiver(post_delete, sender=Chat)
def chat_post_delete(sender, instance=None, **kwargs):
    if instance:
        a_key = "adm-website_mon-%i" % instance.website_id
        a_chat_key = "chat-%s" % instance.pk
        msg = {"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_CHAT_REMOVE,
                                   "data": {"id": instance.pk},
                                   "pk": instance.pk
                                   })
               }
        async_to_sync(channel_layer.group_send)(a_key, msg)
        async_to_sync(channel_layer.group_send)(a_chat_key, msg)
        if instance.obj_props.get('website_b'):
            a_key =  "adm-website_mon-%i" % instance.obj_props.get('website_b')
            async_to_sync(channel_layer.group_send)(a_key, msg)

        c_key = "cli_wait-%s-%s" % (instance.website_id, instance.client_id)
        async_to_sync(channel_layer.group_send)(c_key, {"type": "websocket.send",  "text":  json.dumps({
            "type": WebsocketEventTypes.FRONTEND_CHAT_ARCHIVED, "pk": instance.pk
        })})


@receiver(post_save, sender=Message)
def message_post_save(sender, instance=None, **kwargs):
    if instance:
        a_key = "chat-%s" % instance.chat_id  # employee chatroom channel
        c_key = "cli_speak-%s" % instance.chat_id  # cleint chatroom channel

        data = {"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_CHAT_ROOM_NEW_MESSAGE
        if kwargs.get('created', False) else
        WebsocketEventTypes.BACKEND_CHAT_ROOM_UPD_MESSAGE,
                                    "data": dict(MessageSerializer(instance=instance)
                                                 .to_representation(instance=instance)),
                                    "pk": instance.pk,
                                    "chat": instance.chat_id
                                    })}
        async_to_sync(channel_layer.group_send)(a_key, data)
        async_to_sync(channel_layer.group_send)(c_key, data)

        databot = dict(MessageSerializer(instance=instance).to_representation(instance=instance))
        if instance.type == Message.TYPE_CLIENT:
            databot['sender_id'] = 'client'
        databotmsg = {
            "type": "websocket.send",  
            "text": json.dumps({
                "type": WebsocketEventTypes.BOT_ON_MESSAGE,
                "data": databot,
                "msg_id": instance.pk,
                "chat_id": instance.chat_id
            })}
        async_to_sync(channel_layer.group_send)("bot_site-%s" % instance.chat.website_id, databotmsg)

        if not instance.is_processed and instance.type != Message.TYPE_CLIENT and instance.chat.is_connector:
            async_to_sync(channel_layer.group_send)(
                "connector-%s-%s" % (instance.chat.website_id, instance.chat.connector_name), {
                "type": "websocket.send",
                "data": databot,
                "msg_id": instance.pk,
                "client_id": instance.chat.client_id,
                "is_tech": instance.type == Message.TYPE_TECH,
                "is_agent": instance.type == Message.TYPE_OPERATOR,
            })
            for rec in prepare_send_data({
                "data": databot,
                "msg_id": instance.pk,
                "client_id": instance.chat.client_id,
                "is_tech": instance.type == Message.TYPE_TECH,
                "is_agent": instance.type == Message.TYPE_OPERATOR,
            }, instance.chat.connector_name):
                connector_prepare_message_to.send_robust(
                    sender=instance.chat,
                    connector_name=instance.chat.connector_name,
                    website_id=instance.chat.website_id,
                    data=rec
                )

        for operator in instance.chat.operators_current.all():
            if instance.sender_id != operator.user_tag:
                set_sizes_from_chat(instance.chat, operator)

        if kwargs.get('created', False):
            data = {"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_MSG_NOTIFY,
                                        "data": {
                                            "text": instance.text if instance.text else "-",
                                            "data_type": instance.data_type,
                                            "sender": instance.sender_name,
                                            "sender_id": instance.sender_id,
                                            "pk": instance.pk,
                                            "chat": instance.chat_id
                                        }}
                                       )}
            website_b = instance.chat.website_b
            for operator in instance.chat.operators_current.all():
                at_key = "adm-website_mon-u-%i-%i" % (instance.chat.website_id, operator.pk)
                async_to_sync(channel_layer.group_send)(at_key, data)
                if website_b:
                    at_key = "adm-website_mon-u-%i-%i" % (website_b, operator.pk)
                    async_to_sync(channel_layer.group_send)(at_key, data)


@receiver(post_save, sender=ChatReadStack)
def read_stack_post_save(sender, instance=None, **kwargs):
    if instance:
        a_key = "adm-website_mon-u-%i-%i" % (instance.chat.website_id, instance.operator_id)
        msg ={"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_CHAT_UNREAD,
                                               "data": {
                                                   "chat": instance.chat_id,
                                                   "msg_total": instance.chat_msg_count,
                                                   "msg_was_read": instance.user_msg_read,
                                                   "unread": instance.unread_count
                                               }})
                           }
        async_to_sync(channel_layer.group_send)(a_key, msg)
        # async_to_sync(channel_layer.group_send)("adm-user_mon-%i" % instance.operator_id, msg)
        if instance.chat.website_b:
            a_key = "adm-website_mon-u-%i-%i" % (instance.chat.website_b, instance.operator_id)
            async_to_sync(channel_layer.group_send)(a_key, msg)


@receiver(post_delete, sender=ChatReadStack)
def read_stack_post_delete(sender, instance=None, **kwargs):
    if instance:
        a_key = "adm-website_mon-u-%i-%i" % (instance.chat.website_id, instance.operator_id)
        msg = {"type": "websocket.send",  "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_CHAT_UNREAD,
                                               "data": {
                                                   "chat": instance.chat_id,
                                                   "msg_total": 0,
                                                   "msg_was_read": 0,
                                                   "unread": 0
                                               }})
                           }
        async_to_sync(channel_layer.group_send)(a_key, msg)
        if instance.chat.website_b:
            a_key = "adm-website_mon-u-%i-%i" % (instance.chat.website_b, instance.operator_id)
            async_to_sync(channel_layer.group_send)(a_key, msg)


# @receiver(m2m_changed, sender=Chat.operators_current.through)
# def chat_m2m_operators_current_change(instance, action, **kwargs):
#     from apps.ncbot.constants import ChatMode

#     if action == 'post_add':
#         chat_state = instance.get_state()
#         chat_state.mode = ChatMode.AGENT
#         chat_state.save(update_fields=('mode',))

#         # make bot leave
#         databot = {
#             "type": "websocket.send",
#             "text": json.dumps({
#                 "type": WebsocketEventTypes.BOT_ON_ENDROOM,
#                 "data": ChatBotSerializer(instance=instance).to_representation(instance=instance),
#                 "chat_id": instance.pk
#             })
#         }
#         async_to_sync(channel_layer.group_send)("bot_site-%s" % instance.website_id, databot)

#     # make bot join
#     if action in ['post_remove',  'post_clear']:
#         if not instance.operators_current.all().exists():
#             databot = {
#                 "type": "websocket.send",
#                 "text": json.dumps({
#                     "type": WebsocketEventTypes.BOT_ON_NEWROOM,
#                     "data": ChatBotSerializer(instance=instance).to_representation(instance=instance),
#                     "chat_id": instance.pk
#                 })
#             }
#             async_to_sync(channel_layer.group_send)("bot_site-%s" % instance.website_id, databot)

#     if action in ['post_add', 'post_remove', 'post_clear']: