import json
import re
import time
from datetime import datetime
from .models import ChatReadStack, Chat, ArchivedChat, Message, Attachment
from ..visitors.models import Client, Order
from ..website.models import Product, AgentStatus, Form
from ..website.serializers import FormForWidgetSerializer
from ..video.models import Call
from nicechat.utils.callback import make_callback, EVENT_UNANSWERED
from django.utils import timezone
from django.conf import settings

TAG_TPL = re.compile('[^A-Za-z0-9_]')


def filter_tags(tags):
    tg = set()
    for tag in tags:
        t = filter_tag(tag)
        if t:
            tg.add(t)
    return list(tg)


def filter_tag(tag):
    if not isinstance(tag, str):
        return None
    tag = TAG_TPL.sub('', tag).lower()
    if not tag:
        return None
    return tag[:16]


def is_list_has_value(search_list, value):
    for d in search_list:
        if d in value:
            return d
    return None


def join_watch_list(chat, user):
    set_sizes_from_chat(chat, user)


def leave_watch_list(chat, user):
    ChatReadStack.objects.filter(chat=chat, operator=user).delete()


def set_sizes_from_chat(chat, user):
    item = ChatReadStack.objects.filter(chat=chat, operator=user).first()

    if not item:
        item = ChatReadStack(chat=chat, operator=user)

    item.chat_msg_count = item.chat.message_set.count()
    msg = item.chat.message_set.last()
    if msg:
        item.chat_top_msg_id = msg.pk

    item.unread_count = item.chat_msg_count - item.user_msg_read
    if item.unread_count < 0:
        item.unread_count = 0
    item.save()


def set_sizes_by_user(chat, user, msg_id):
    item = ChatReadStack.objects.filter(chat=chat, operator=user).first()

    if item:
        l = list(chat.message_set.all().order_by('pk').values_list('pk', flat=True))
        if msg_id in l:
            item.user_top_msg_id = msg_id
            item.user_msg_read = l.index(msg_id) + 1

        item.unread_count = item.chat_msg_count - item.user_msg_read
        if item.unread_count < 0:
            item.unread_count = 0
        item.save()
        return True
    else:
        return False

TS_FORMATS = [
    "%Y-%m-%d %H:%M:%S.%f%z",
    "%Y-%m-%d %H:%M:%S%z"
    "%Y-%m-%dT%H:%M:%S.%f%z",
    "%Y-%m-%dT%H:%M:%S%z"
]

def parse_timestamp(datetime_str):
    for f in TS_FORMATS:
        try:
            return datetime.strptime(datetime_str, f)
        except ValueError:
            continue
    return timezone.now()
        

def move_chat_from_archive(ac):
    chat_to = Chat.objects.filter(pk=ac.pk).first()

    if chat_to:
        chat_to.is_serialized = False
        chat_to.is_open = True
        chat_to.ts_close = None
        chat_to.ts_open = timezone.now()
    else:
        client, created = Client.objects.get_or_create(client_id=ac.client_id, website=ac.website)
        chat_to = Chat(client_id=ac.client_id, website=ac.website)
        for f in Chat._meta.get_fields():
            name = str(f.name)
            if hasattr(chat_to, name) and hasattr(ac, name):
                setattr(chat_to, name, getattr(ac, name))
        setattr(chat_to, 'client_name', getattr(ac, 'client_name'))
        chat_to.is_serialized = False
        chat_to.is_open = True
        chat_to.ts_close = None
        chat_to.ts_open = timezone.now()
        chat_to.save()

        for attach in json.loads(ac.attachments):
            Attachment.objects.get_or_create(
                id=attach['id'],
                defaults={
                    'chat': chat_to,
                    'file_type': attach['file_type'],
                    'url': attach['url'],
                    'upload_ts': parse_timestamp(attach['upload_ts']),
                    'message_id': attach['message_id']
                }
            )
        for msg in json.loads(ac.messages):
            Message.objects.get_or_create(
                id=msg['id'],
                defaults={
                    'chat': chat_to,
                    'sender_name': msg['sender_name'],
                    'timestamp': parse_timestamp(msg['timestamp']),
                    'text': msg['text'],
                    'data': msg['data'],
                    'data_type': msg['data_type'],
                    'type': msg['type']
                }
            )
            
        for opr in ac.operators.all():
            chat_to.operators_history.add(opr)
           
    chat_to.save()
    chat_to.put_tech_message('~!REOPENED')
    ac.delete()
    return chat_to



def move_chat_to_archive(chat, force_del=False):

    if chat and not chat.is_serialized and chat.ts_close is not None:

        chat_took_place = False
        for msg in chat.message_set.order_by('id').all():
            if msg.sender_id != '0' or msg.type != Message.TYPE_OPERATOR:
                chat_took_place = True
                break

        if settings.ARCHIVE_ALL:
            chat_took_place = True

        client, created = Client.objects.get_or_create(client_id=chat.client_id, website=chat.website)

        if not chat_took_place:
            chat.is_serialized = True
            chat.save(update_fields=['is_serialized'])
            client.chat_count += 1
            client.void_chat_count += 1
            client.last_activity_ts = timezone.now()
            client.save(update_fields=('chat_count', 'void_chat_count', 'last_activity_ts',))
            return

        arc = ArchivedChat(client_id=chat.client_id, website=chat.website)

        for f in Chat._meta.get_fields():
            name = str(f.name)
            if hasattr(arc, name) and hasattr(chat, name):
                setattr(arc, name, getattr(chat, name))
        setattr(arc, 'client_name', getattr(chat, 'client_name'))

        messages = []
        p_view = 0
        p_buy = 0
        for msg in chat.message_set.order_by('id').all():
            messages.append({
                'id': msg.id,
                'sender_name': msg.sender_name,
                'timestamp': str(msg.timestamp),
                'text': msg.text,
                'data': msg.data,
                'data_type': msg.data_type,
                'type': msg.type
            })
            if msg.data_type:
                if msg.data_type == Message.DATA_PRODUCT_VIEWED:
                    p_view += 1
                if msg.data_type == Message.DATA_PRODUCT_BUY:
                    p_buy += 1
        arc.website = chat.website
        attachments = []
        for attach in chat.attachment_set.all():
            attachments.append({
                'id': attach.id,
                'file_type': attach.file_type,
                'upload_ts': str(attach.upload_ts),
                'url': attach.url,
                'message_id': attach.message_id
            })
        arc.attachments = json.dumps(attachments)
        arc.messages = json.dumps(messages)
        arc.save()
        for opr in chat.operators_history.all():
            arc.operators.add(opr)
            arc.operators_count += 1
        arc.ts_open = chat.ts_open
        arc.save()
        chat.is_serialized = True
        chat.save(update_fields=['is_serialized'])

        client.chat_count += 1
        if p_view:
            client.product_view_count += p_view
        if p_buy:
            client.purchase_count += p_buy
        client.last_activity_ts = timezone.now()
        client.save(update_fields=('chat_count', 'void_chat_count', 'last_activity_ts',))
        if force_del:
            chat.joinedbot.all().delete()
            chat.delete()


def fill_info_by_tags(tags, chat):
    need_to_write = False
    for nm in tags.keys():
        if nm.startswith("PRIMARY_"):
            need_to_write = True
            break
    if need_to_write:
        client, created = Client.objects.get_or_create(client_id=chat.client_id, website=chat.website)
        for nm, val in tags.items():
            if nm.startswith("PRIMARY_"):
                client.props[nm.replace("PRIMARY_", "")] = val
        client.save()


def fill_order_by_product(data, chat, products):

    client, created = Client.objects.get_or_create(client_id=chat.client_id, website=chat.website)

    order = Order.objects.create(client=client, website=chat.website, chat_id=chat.pk)
    order.product_detail = {}
    for opr in chat.operators_history.all():
        order.agents.add(opr)
    order.qty = 1

    try:
        products = json.loads(products)
        product_ids = [prod.get('pk') for prod in products if prod.get('pk')]
    except ValueError:
        product_ids = []

    for prod in Product.objects.filter(pk__in=product_ids):
        order.products.add(prod)

    for obj in data:
        label = obj.get('label', '')
        if label in ['Billing Address', 'Shipping Address']:
            rez = {}
            for itm in obj.get('data', {}):
                rez[itm.get('label', '-')] = itm.get('value')
            if label == 'Billing Address':
                order.billing_address = rez
            else:
                order.shipping_address = rez

    order.product_detail = {
        'sysIp': chat.sys_ip,
        'sysBrowser': chat.sys_browser,
        'sysOs': chat.sys_os,
        'sysDevice': chat.sys_device,
        'sysLangFull': chat.sys_lang_full,
        'sysTime': chat.sys_time,
        'geoCountry': chat.geo_country,
        'geoCity': chat.geo_city
    }

    order.save()

    return order


def do_chat_activity_wdt(chat_pk, is_final):

    chat = Chat.objects.filter(pk=chat_pk).first()

    if chat and not chat.operators_current.count() and chat.is_open:

        is_active = AgentStatus.objects.filter(website_id=chat.website_id)\
            .exclude(state=AgentStatus.STATUS_AWAY).exists()

        if not is_active or is_final:

            make_callback.apply_async(args=(EVENT_UNANSWERED, chat.website_id, "Chat", chat.pk,), countdown=5)

            is_message = True if chat.message_set.filter(data_type__in=[Message.DATA_FORM, Message.DATA_FORM_DONE,
                                                                        Message.DATA_FORM_FILLED,
                                                                        Message.DATA_FORM_CANCEL]).exists() else False
            if not is_message:
                is_message = True if chat.message_set.filter(type=Message.TYPE_TECH, text__startswith='~!JOINED').exists() else False

            if not is_message:
                form = Form.objects.filter(website=chat.website, mode=Form.MODE_SITE_CATCH_FORM).first()

                if form:
                    data = dict(FormForWidgetSerializer(form).to_representation(form))

                    Message.objects.create(
                        chat=chat,
                        sender_id=0,
                        sender_name=chat.website.title,
                        text=form.label,
                        data=json.dumps(data),
                        data_type=Message.DATA_FORM,
                        hidden=json.dumps({'form_id': form.pk}),
                        type=Message.TYPE_OPERATOR
                    )


def do_celery_chat_call_wdt(chat_pk):
    cl = Call.objects.filter(chat_id=chat_pk, is_active=True, is_started=False).first()
    cl.is_active=False
    cl.save()
    cl.chat.put_tech_message_wdata("~!CALL_NOANS:1", {})
    cl.chat.clear_call_messages()


def batam_assign_agents_to_chat(chat_pk, no_list=[], force=False, small_wait=False):
    if small_wait:
        time.sleep(3)
    chat = Chat.objects.filter(pk=chat_pk).first()

    if not chat:
        return
    
    website = chat.website
    if not website.auto_assign_enable:
        return

    if chat.operators_current.count() > 0 and not force:
        return
    
    ag_w = None
    ag_c = None

    for ag in AgentStatus.objects.filter(
            website=website,
            is_active=True,
            is_current=True,
            state=AgentStatus.STATUS_ONLINE
        ):
        if isinstance(no_list, list) and ag.operator in no_list:
            continue
        if ag.operator.operators_current.count() < website.auto_assign_limit:
            if ag_w is None or ag_c > ag.operator.operators_current.count():
                ag_w = ag.operator
                ag_c = ag.operator.operators_current.count()
    if ag_w:
        from apps.chat.api import ChatViewSet
        print(ag_w)
        ChatViewSet.process_chat_join(chat, ag_w)
        return True
