import json
import sys
import time
import asyncio

from websockets.sync.client import connect as sync_connect
from websockets import connect as async_connect
from websockets.exceptions import ConnectionClosed
from apps.connector.utils import make_message_body

CH_NEW_ROOM = 'new_room'
CH_RCVD_MSG = 'rcvd_msg'
CH_CLOSE_ROOM = 'close_room'
CH_COMMIT = 'msg_commit'
CH_ON_SEND_MESSAGE = 'on_msg'
CH_ON_SEND_AGENT_JOIN= 'on_msg_join'
CH_ON_SEND_AGENT_LEAVE = 'on_msg_leave'
CH_ON_SEND_SUGGESTION = 'on_msg_product'


class ConnectorLibBase(object):

    def __init__(self, host, site_id, connector, node_id, key=None):
        _host = host.rstrip('/')
        k = f"/{key}" if key else ""
        self.ws = None
        npref = f'/{node_id}' if node_id else ''
        self.ws_url = f"{_host}{npref}/api/v1/connector/{connector}/{site_id}{k}/"
        self.website = {}

    def new_chat_room(self, room_id, client_name='Guest'):
        self.ws.send(json.dumps({
            "type": CH_NEW_ROOM, 
            "client_id": room_id, 
            "name": client_name, 
        }))        

    def receive_message(self, room_id, text='', client_name='Guest', reply_cmd=None, reply_event=None):
        self.ws.send(json.dumps({
            "type": CH_RCVD_MSG, 
            "client_id": room_id,
            "opts": make_message_body(text=text, client_name=client_name, reply_cmd=reply_cmd, reply_event=reply_event)
        }))        

    def close_chat_room(self, room_id):
        self.ws.send(json.dumps({
            "type": CH_CLOSE_ROOM, 
            "client_id": room_id, 
        }))        

    def commit(self, event_id):
        self.ws.send(json.dumps({
            "type": CH_COMMIT, 
            "event_id": event_id, 
        }))        

    def on_connect(self):
        pass

    def on_send_message(self, room_id, event_id, text, sender_name, buttons=None):
        pass

    def on_agent_join(self, room_id, event_id, agent_name, agent_img):
        pass

    def on_agent_leave(self, room_id, event_id, agent_name):
        pass

    def on_suggestion(self, room_id, event_id, product, sender_name):
        pass

    def run(self):
        while True:
            try:
                self.run_inner()
            except (ConnectionClosed, ConnectionError):
                print('reconnecting...')
                time.sleep(0.5)
            except Exception as e:
                sys.exit()

    def run_inner(self):
        self.ws = sync_connect(self.ws_url)
        self.on_connect()
        print('connected!')
        while True:
            result = json.loads(self.ws.recv())
            _type = result.get('type')
            if _type == CH_ON_SEND_MESSAGE:
                self.on_send_message(
                    room_id = result.get('room_id'),
                    text = result.get('text'),
                    buttons = result.get('buttons', None),
                    event_id = result.get('event_id'),
                    sender_name = result.get('sender_name'),
                )
            elif _type == CH_ON_SEND_AGENT_LEAVE:
                self.on_agent_leave(
                    room_id = result.get('room_id'),
                    event_id = result.get('event_id'),
                    agent_name = result.get('agent_name'),
                )
            elif _type == CH_ON_SEND_AGENT_JOIN:
                self.on_agent_join(
                    room_id = result.get('room_id'),
                    event_id = result.get('event_id'),
                    agent_name = result.get('agent_name'),
                    agent_img = result.get('agent_img'),
                )
            elif _type == CH_ON_SEND_SUGGESTION:
                self.on_suggestion(
                    room_id = result.get('room_id'),
                    event_id = result.get('event_id'),
                    sender_name = result.get('sender_name'),
                    product = result.get('product'),
                )


class ConnectorLibAsyncBase(object):

    def __init__(self, host, site_id, connector, node_id, key=None):
        _host = host.rstrip('/')
        k = f"/{key}" if key else ""
        npref = f'/{node_id}' if node_id else ''
        self.ws_url = f"{_host}{npref}/api/v1/connector/{connector}/{site_id}{k}/"
        self.ws = None
        self.website = {}

    async def new_chat_room(self, room_id, client_name='Guest'):
        await self.ws.send(json.dumps({
            "type": CH_NEW_ROOM, 
            "client_id": room_id, 
            "name": client_name, 
        }))        

    async def receive_message(self, room_id, text='', client_name='Guest', reply_cmd=None, reply_event=None):
        await self.ws.send(json.dumps({
            "type": CH_RCVD_MSG, 
            "client_id": room_id,
            "opts": make_message_body(text=text, client_name=client_name, reply_cmd=reply_cmd, reply_event=reply_event)
        }))        

    async def close_chat_room(self, room_id):
        await self.ws.send(json.dumps({
            "type": CH_CLOSE_ROOM, 
            "client_id": room_id, 
        }))        

    async def commit(self, event_id):
        await self.ws.send(json.dumps({
            "type": CH_COMMIT, 
            "event_id": event_id, 
        }))        

    async def on_connect(self):
        pass

    async def on_send_message(self, room_id, event_id, text, sender_name, buttons=None):
        pass

    async def on_agent_join(self, room_id, event_id, agent_name, agent_img):
        pass

    async def on_agent_leave(self, room_id, event_id, agent_name):
        pass

    async def on_suggestion(self, room_id, event_id, product, sender_name):
        # product = {
        #       'pk': int,
        #       'title': str,
        #       'price': str (Ex: '11.00'), 
        #       'currency': str, iso lower ('usd'),
        #       'original_id': str (Product ID/SKU),
        #       'url': str|none,
        #       'attr': {},
        #       'imageUrl': str, fullurl 'https://widget.nice.chat/media/...'
        # }
        pass

    async def run(self):
        while True:
            try:
                await self.run_inner()
            except (ConnectionClosed, ConnectionError):
                print('reconnecting...')
                await asyncio.sleep(0.5)

    async def run_inner(self):
        await self.on_connect()
        self.ws = await async_connect(self.ws_url)
        while True:
            result = json.loads(await self.ws.recv())
            _type = result.get('type')
            if _type == CH_ON_SEND_MESSAGE:
                await self.on_send_message(
                    room_id = result.get('room_id'),
                    text = result.get('text'),
                    buttons = result.get('buttons', None),
                    event_id = result.get('event_id'),
                    sender_name = result.get('sender_name'),
                )
            elif _type == CH_ON_SEND_AGENT_LEAVE:
                await self.on_agent_leave(
                    room_id = result.get('room_id'),
                    event_id = result.get('event_id'),
                    agent_name = result.get('agent_name'),
                )
            elif _type == CH_ON_SEND_AGENT_JOIN:
                await self.on_agent_join(
                    room_id = result.get('room_id'),
                    event_id = result.get('event_id'),
                    agent_name = result.get('agent_name'),
                    agent_img = result.get('agent_img'),
                )
            elif _type == CH_ON_SEND_SUGGESTION:
                await self.on_suggestion(
                    room_id = result.get('room_id'),
                    event_id = result.get('event_id'),
                    sender_name = result.get('sender_name'),
                    product = result.get('product'),
                )
