import json
import os
import time
import traceback
from channels.generic.websocket import WebsocketConsumer
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
from threading import Thread
from queue import SimpleQueue
from django.conf import settings
from nicechat.utils.constants import WebsocketEventTypes
from nicechat import watcher
channel_layer = get_channel_layer()
from apps.connector.monitor import cnn_monitor_packet
from apps.chat.models import Message, Connector
from django.utils.module_loading import import_string
from .dispatch import connector_prepare_message_to
from nicechat.utils.arbiter_sync import WatcherMiddleware


ws_cnn_q = SimpleQueue()

def process_ws_cnn():
    print("Start connector processing", os.getpid())
    while True:
        try:
            text_data, socket = ws_cnn_q.get()
            if text_data == "stop_all" and not watcher.can_i_run():
                break
            try:
                if isinstance(text_data, str):
                    dt = json.loads(text_data)
                elif isinstance(text_data, dict):
                    dt = text_data
                else:
                    continue
            except ValueError:
                continue
            if isinstance(socket, dict):
                cnn_monitor_packet(
                    dt,
                    {'connector': socket.get('connector_name'), 'site_id': socket.get('site_id')},
                    socket.get('connector_id'),
                    socket.get('connector_name'),
                    socket.get('send')
                )
            if isinstance(socket, WebsocketConsumer):
                cnn_monitor_packet(
                    dt,
                    socket.scope['url_route'].get('kwargs', {}),
                    socket.connector_id,
                    socket.connector_name,
                    socket.send
                )
        except Exception as e:
            traceback.print_exception(e)
    print("Stop connector processing", os.getpid())

ws_cnn_t = Thread(target=process_ws_cnn)
ws_cnn_t.isDaemon = True


def put_ws_cnn(data, context):
    ws_cnn_q.put((data, context))
    if not getattr(ws_cnn_t, 'ident'):
        ws_cnn_t.start()


def prepare_send_data(msg, connector_name):
    msgs = []
    event_id = msg['msg_id']
    client_id = msg['client_id']
    is_tech = msg['msg_id']
    is_agent = msg['is_agent']

    client_id = client_id[len(connector_name)+1:]
    msg = msg['data']
    # print(msg)
    if is_agent and not msg['data_type'] == Message.DATA_SUGGEST:
        buttons = None
        attach = None
        if msg['data_type'] == Message.DATA_ATTACH:
            preview = msg['data'].get('preview', None)
            if preview:
                attach = {'url': preview}
            # from apps.chat.models import Attachment
            # att = Attachment.objects.filter(
            #     message_id = event_id
            # ).first()
            # if att:
            #     print(att.__dict__)
        if msg['data_type'] == Message.DATA_ANSWERS:
            _buttons = msg['data'].get('buttons', None)
            if isinstance(_buttons, dict):
                buttons = [
                    {
                        'title': _tl,
                        'payload': _pl
                    }
                    for _pl, _tl in _buttons.items()
                ]
        msgs.append({
            'type': WebsocketEventTypes.CH_ON_SEND_MESSAGE,
            'event_id': event_id,
            'room_id': client_id,
            'text': msg.get('text', ''),
            'buttons': buttons,
            'attach': attach,
            'sender_name': msg.get('sender_name', 'Agent')
        })
    elif is_tech and '~!JOINED:' in msg.get('text', ''):
        dt = msg.get('data', None)
        msgs.append({
            'type': WebsocketEventTypes.CH_ON_SEND_AGENT_JOIN,
            'event_id': event_id,
            'room_id': client_id,
            'agent_name': dt.get('name') if dt else msg.get('text', '')[9:],
            'agent_img': dt.get('avatar') if dt else None
        })
    elif is_tech and '~!LEAVE:' in msg.get('text', ''):
        dt = msg.get('data', None)
        msgs.append({
            'type': WebsocketEventTypes.CH_ON_SEND_AGENT_LEAVE,
            'event_id': event_id,
            'room_id': client_id,
            'agent_name': dt.get('name') if dt else msg.get('text', '')[8:],
            'agent_img': dt.get('avatar') if dt else None
        })
    elif msg['data_type'] == Message.DATA_SUGGEST:
        msgs.append({
            'type': WebsocketEventTypes.CH_ON_SEND_SUGGESTION,
            'event_id': event_id,
            'room_id': client_id,
            'sender_name': msg['sender_name'],
            'product': msg['data'],
        })
    return msgs


class ConnectorConsumer(WebsocketConsumer):
    method_mapping = {
        "websocket.connect": "connect",
        "websocket.receive": "receive",
        "websocket.send": "websocket_send",
        "websocket.disconnect": "disconnect",
    }

    def __init__(self, *args, **kwargs):
        super(ConnectorConsumer, self).__init__(*args, **kwargs)
        self.connector_name = None
        self.connector_id = None
        self.site_id = None
        self.run_watch = False
        self.watcher_th = None

    def __del__(self):
        self.run_watch = False

    def watcher(self):
        c = 0
        while self.run_watch:
            time.sleep(1)
            c+= -1
            if c < 0:
                c = settings.CONNECTOR_REPEAT_TIMEOUT
                try:
                    self.process_untracked()
                except Exception as e:
                    traceback.print_exception(e)

        print('Watcher done', self.connector_name)

    def process_untracked(self):
        MessageSerializer = import_string(settings.MESSAGE_SERIALIZER)

        msgs = Message.objects.filter(
            is_processed=False,
            type__in=[Message.TYPE_OPERATOR, Message.TYPE_TECH],
            chat__is_connector=True,
            chat__connector_name=self.connector_name,
            chat__is_open=True,
            chat__website_id=self.site_id,
        ).order_by('timestamp')
        for msg in msgs:
            self.websocket_send({
                "data": dict(MessageSerializer(instance=msg).to_representation(instance=msg)),
                "msg_id": msg.pk,
                "client_id": msg.chat.client_id,
                "is_tech": msg.type == Message.TYPE_TECH,
                "is_agent": msg.type == Message.TYPE_OPERATOR,
            })

    def connect(self, *args, **kwargs):
        if not getattr(ws_cnn_t, 'ident'):
            ws_cnn_t.start()
        scope_kwargs = self.scope['url_route'].get('kwargs', {})
        # TODO Auth
        connector = Connector.objects.filter(name=scope_kwargs.get('connector')).first()
        if not connector:
            if settings.CONNECTOR_ALLOW_AUTOCREATE:
                connector = Connector.objects.create(name=scope_kwargs.get('connector'))
            else:
                self.close()
                return
        if connector.connection_key:
            key = scope_kwargs.get('key')
            if not key or key != connector.connection_key:
                self.close()
                return
        self.connector_name = connector.name
        self.connector_id = connector.pk
        self.site_id = scope_kwargs.get('site_id')
        self.accept()
        async_to_sync(self.channel_layer.group_add)("connector-%s-%s" % (self.site_id, scope_kwargs.get('connector')), self.channel_name)
        self.run_watch = True
        self.watcher_th = Thread(target=self.watcher)
        self.watcher_th.isDaemon = True
        self.watcher_th.start()

    def receive(self, text_data="", bytes_data=None):
        ws_cnn_q.put((text_data, self))
        site_id = self.scope['url_route'].get('kwargs', {}).get('site_id')
        WatcherMiddleware.stats[f"ws:{site_id}"] = 1 + WatcherMiddleware.stats.get(f"ws:{site_id}", 0)

    def websocket_send(self, msg):
        for rec in prepare_send_data(msg, self.connector_name):
            self.send(text_data=json.dumps(rec))

    def disconnect(self, *args, **kwargs):
        self.run_watch = False
        async_to_sync(self.channel_layer.group_discard)("connector-%s-%s" % (self.site_id, self.connector_name), self.channel_name)
