import asyncio
from telegram.ext import (
    Application,
    CommandHandler,
    MessageHandler,
    filters,
    ContextTypes,
    CallbackContext,
    CallbackQueryHandler,
    ConversationHandler,
)
from telegram import InlineKeyboardButton, InlineKeyboardMarkup, Update, constants
from .consts import CHAT, CMD_START, CMD_STOP, CMD_TXT, CMD_JOIN, CMD_LEAVE, CMD_COMMIT, CMD_PRODUCT

class Bot():
    def __init__(self, tg_key, cmd_stream, recv_stream, connector):
        self.tg_key = tg_key
        self.cmd_stream = cmd_stream
        self.recv_stream = recv_stream
        self.connector = connector

    def run(self):
        application = Application.builder().token(self.tg_key).post_init(self.on_bot_start_up).build()

        conv_handler = ConversationHandler(
            entry_points=[CommandHandler('start', self.start), MessageHandler(filters.TEXT & ~filters.COMMAND, self.do_chat)],
            states={
                CHAT: [MessageHandler(filters.TEXT & ~filters.COMMAND, self.do_chat)],
            },
            fallbacks=[MessageHandler(filters.TEXT & ~filters.COMMAND, self.do_chat), CommandHandler('start', self.start), CommandHandler('stop', self.stop), ]
        )
        application.add_handler(conv_handler)
        application.add_handler(CallbackQueryHandler(self.buttons))
        application.run_polling(allowed_updates=Update.ALL_TYPES)

    async def on_bot_start_up(self, application: Application):
        asyncio.create_task(self.queue_processor(application))
        asyncio.create_task(self.connector.run())
    
    async def queue_processor(self, application):
        while True:
            rec = await self.recv_stream.get()
            if rec['type'] == CMD_TXT:
                reply_markup = None
                if rec.get('buttons'):
                    reply_markup = InlineKeyboardMarkup([
                        [InlineKeyboardButton(r['title'], callback_data=f"{rec['event_id']}%{r['payload']}")]
                            for r in rec.get('buttons', [])
                    ])
                await application.bot.send_message(rec['room_id'], rec['text'], reply_markup=reply_markup)
            elif rec['type'] == CMD_PRODUCT:
                product = rec.get('product')
                if product:
                    url = product.get('url')
                    currency = product.get('currency')
                    img = product.get('imageUrl')
                    price = product.get('price')
                    title = product.get('title')
                    txt = [f'[{title}]({url})' if url else title]
                    if price:
                        txt.append(f'Price: {price} {currency}')
                    txt = "\n".join(txt)
                    if img:
                        await application.bot.sendPhoto(rec['room_id'], img)
                    await application.bot.send_message(rec['room_id'], txt, parse_mode=constants.ParseMode.MARKDOWN)
            elif rec['type'] == CMD_JOIN:
                await application.bot.send_message(rec['room_id'], f"*ℹ️ {rec['agent_name']} joined to chat!*", parse_mode=constants.ParseMode.MARKDOWN)
            elif rec['type'] == CMD_LEAVE:
                await application.bot.send_message(rec['room_id'], f"*ℹ️ {rec['agent_name']} left the chat!*", parse_mode=constants.ParseMode.MARKDOWN)
            if rec.get('event_id'):
                await self.cmd_stream.put({"type": CMD_COMMIT, "event_id": rec['event_id']})

    async def start(self, update: Update, context: CallbackContext) -> int:
        await self.cmd_stream.put(
            {
                'type': CMD_START,
                'id': update.message.chat.id,
                'name': f"{update.message.chat.first_name} {update.message.chat.last_name}",
            })
        return CHAT

    async def stop(self, update: Update, context: CallbackContext) -> int:
        await self.cmd_stream.put(
            {
                'type': CMD_STOP,
                'id': update.message.chat.id,
            })
        return await self.start(update, context)
    
    async def buttons(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        if update.callback_query:
            query = update.callback_query
            await query.answer()
            payload = query.data
            payload_txt = query.data
            for row in query.message.reply_markup.inline_keyboard:
                if row:
                    var = row[0]
                    if var.callback_data ==payload:
                        payload_txt = var.text
            await query.edit_message_text(text=f"{query.message.text}\n\nReply: {payload_txt}")
            await self.cmd_stream.put(
            {
                'type': CMD_TXT,
                'text': payload_txt,
                'payload': payload,
                'name': f"{query.from_user.first_name} {query.from_user.last_name}",
                'id': query.from_user.id,
            })

    async def do_chat(self, update: Update, context: CallbackContext) -> int:
        await self.cmd_stream.put(
            {
                'type': CMD_TXT,
                'text': update.message.text,
                'name': f"{update.message.chat.first_name} {update.message.chat.last_name}",
                'id': update.message.chat.id,
            })
        return CHAT