import urllib3
import urllib
import time
import json
import datetime
import pytz
import hashlib
import traceback
from uuid import uuid4
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django.db import IntegrityError
from django.http import Http404, HttpResponse
from django.core.exceptions import PermissionDenied
from nicechat.utils.images import image_post_create, image_post_processing
from django.core.cache import caches, InvalidCacheBackendError
from ..website.tasks import process_import
from ..website.models import Product, ProductsCategory, AccessToken
from ..website.utils.service import filter_products_props
from ..visitors.models import Order
from rest_framework.response import Response
from rest_framework import viewsets, status, exceptions
from .serializers import MarketplaceOrderSerializer, MarketplaceProductSerializer, MarketplaceCategorySerializer
from drf_yasg.utils import swagger_auto_schema


@swagger_auto_schema(auto_schema=None)
class MarketplaceApiAuthMixin(object):

    def dispatch(self, request, *args, **kwargs):
        token = self.check_signature(request)
        if not token:
            return self.make_response(401)

        self.current_website = token.website
        try:
            result = super(MarketplaceApiAuthMixin, self).dispatch(request, *args, **kwargs)
        except IntegrityError:
            return self.make_response(412)
        except PermissionDenied:
            return self.make_response(403)
        except Http404:
            return self.make_response(404)
        except Exception as ex:
            print("!!!! ERROR !!!!", str(ex))
            traceback.print_exc()
            return self.make_response(500)

        result = result.render()
        return self.make_response(result.status_code, result.content)

    @staticmethod
    def make_response(code, content=None):

        data = {}
        try:
            if content:
                data['result'] = json.loads(content)
        except ValueError:
            pass

        if code < 300:
            data['status'] = 'ok'
        else:
            data['status'] = 'error'
        data['code'] = code

        if code == 200 or code == 204:
            data['description'] = 'Accepted'
            code = 200      # update 204->200 (else user not see response content)
        elif code == 201:
            data['description'] = 'Created'
        elif code == 202:
            data['description'] = 'Modified'
        elif code == 401:
            data['description'] = 'Access token or signature is non-valid'
        elif code == 403:
            data['description'] = 'You have not permission to access this object'
        elif code == 404:
            data['description'] = 'Object not found'
        elif code == 412:
            data['description'] = 'Object with this ID already exists'
        elif code == 500:
            data['description'] = 'Server-side error'

        return HttpResponse(
            json.dumps(data),
            status=code,
            content_type='application/json'
        )

    def get_queryset(self):
        queryset = super(MarketplaceApiAuthMixin, self).get_queryset().filter(website=self.current_website)
        return queryset

    @staticmethod
    def check_signature(request):

        data = request.GET if request.method == 'GET' else request.POST

        if request.META.get('CONTENT_TYPE') == 'application/json':
            key = request.META.get('HTTP_X_API_KEY', None)
            sign = request.META.get('HTTP_X_API_SIGN', None)
        else:
            key = request.META.get('HTTP_X_API_KEY', data.get('_key', None))
            sign = request.META.get('HTTP_X_API_SIGN', data.get('_sign', None))

        at = AccessToken.objects.filter(token=key).first()
        if not at:
            return False

        if request.META.get('CONTENT_TYPE') == 'application/json':
            hstr1 = request.body.decode('utf-8')
            hstr2 = None
        else:
            hstr1 = u'&'.join(['%s=%s' % (j, data.get(j)) for j in sorted(data.keys()) if j not in ['_key', '_sign']])
            hstr2 = urllib.urlencode([(j, data.get(j, '').encode('utf-8')) for j in sorted(data.keys()) if j not in ['_key', '_sign']])

        hstr1 = u'%s%s%s%s' % (at.secret, request.path, hstr1, request.method)
        hstr2 = (u'%s%s%s%s' % (at.secret, request.path, hstr2, request.method)) if hstr2 else None

        try:
            hsh1 = hashlib.sha256(hstr1).hexdigest()
        except UnicodeEncodeError:
            hsh1 = hashlib.sha256(str(hstr1).encode('utf-8')).hexdigest()

        if hstr2:
            try:
                hsh2 = hashlib.sha256(hstr2).hexdigest()
            except UnicodeEncodeError:
                hsh2 = hashlib.sha256(str(hstr2).encode('utf-8')).hexdigest()
        else:
            hsh2 = None

        if hsh1 == sign or hsh2 == sign:
            return at
        else:
            return False


class MarketplaceImportViewSet(MarketplaceApiAuthMixin, viewsets.ViewSet):

    FORMAT_YML = 'yml'
    FORMAT_GOOGLE_CSV = 'google-csv'
    FORMAT_FB_CSV = 'facebook-csv'
    FORMAT_CSV = 'csv'
    FORMAT_CHOICE = (
        (FORMAT_GOOGLE_CSV, _("Google CSV")),
        #(FORMAT_FB_CSV, _("Facebook CSV")),
        (FORMAT_CSV, _("CSV (custom)" )),
    )

    def list(self, request, *args, **kwargs):
        return Response(data={"Field 'format' allowed values": self.FORMAT_CHOICE}, status=status.HTTP_200_OK)

    def post(self, request, *args, **kwargs):
        mode = request.data.get('format')
        url = request.data.get('url')
        callback = request.data.get('callback')

        token = None

        if url:
            if mode == self.FORMAT_GOOGLE_CSV:
                if callback:
                    process_import.delay(self.current_website.pk, None, None, request.data, google_merchant_csv=url,
                                         callback=callback)
                else:
                    token = self.build_token()
                    process_import.delay(self.current_website.pk, None, None, request.data, google_merchant_csv=url,
                                         token=token)

        if token:
            return Response(data={'token': token}, status=status.HTTP_202_ACCEPTED)
        elif callback:
            return Response(data={'state': 'enqueued'}, status=status.HTTP_202_ACCEPTED)
        else:
            return Response(status=status.HTTP_400_BAD_REQUEST)

    @staticmethod
    def build_token():
        for t in range(1, 10):
            token = uuid4()
            try:
                cache = caches['default']
                ret = cache.get(token)
                if not ret:
                    cache.set(token, {'updated': time.time(), 'pushed': time.time(), 'state': 'enqueued'}, 86400)
                    return token
            except InvalidCacheBackendError:
                pass
        return None


class MarketplaceImportResultViewSet(MarketplaceApiAuthMixin, viewsets.ViewSet):

    def retrieve(self, request, *args, **kwargs):
        try:
            cache = caches['default']
            ret = cache.get(kwargs.get('pk'))
            if ret:
                ret['pushed'] = datetime.datetime.fromtimestamp(ret['pushed'], pytz.UTC).isoformat()
                ret['updated'] = datetime.datetime.fromtimestamp(ret['updated'], pytz.UTC).isoformat()
                return Response(data=ret)
            else:
                return Response(status=status.HTTP_404_NOT_FOUND)
        except InvalidCacheBackendError:
            pass
        return Response(status=status.HTTP_501_NOT_IMPLEMENTED)


class MarketplaceOrderViewSet(MarketplaceApiAuthMixin, viewsets.ReadOnlyModelViewSet):
    queryset = Order.objects.all()
    serializer_class = MarketplaceOrderSerializer


class MarketplaceProductViewSet(MarketplaceApiAuthMixin, viewsets.ModelViewSet):
    queryset = Product.objects.all()
    serializer_class = MarketplaceProductSerializer

    def get_object(self):
        obj = self.get_queryset().filter(original_id=self.kwargs.get('pk')).first()
        if not obj:
            raise Http404
        else:
            return obj

    def perform_update(self, serializer):
        kwargs = {"website": self.current_website}
        category = self.request.data.get('category', None)
        if category is not None:
            category = ProductsCategory.objects.filter(original_id=category, website=self.current_website).first()
            if category:
                kwargs['category'] = category
        props = self.request.data.get('props', None)
        if isinstance(props, str):
            try:
                props = json.loads(props)
            except ValueError:
                props = {}
        if isinstance(props, dict):
            props = filter_products_props(props)
        else:
            props = {}
        kwargs['props'] = props
        serializer.save(**kwargs)
        self.postprocess_image(serializer.instance)

    def perform_create(self, serializer):
        obj_id = self.request.data.get('id', None)
        if not obj_id:
            raise exceptions.ValidationError("Field 'id' is required")
        kwargs = {"website": self.current_website, "original_id": obj_id}
        category = self.request.data.get('category', None)
        if category is not None:
            category = ProductsCategory.objects.filter(original_id=category, website=self.current_website).first()
            if category:
                kwargs['category'] = category
        props = self.request.data.get('props', None)
        if isinstance(props, str):
            try:
                props = json.loads(props)
            except ValueError:
                props = {}
        if isinstance(props, dict):
            props = filter_products_props(props)
        else:
            props = {}
        kwargs['props'] = props
        serializer.save(**kwargs)
        self.postprocess_image(serializer.instance)

    def postprocess_image(self, instance):
        if not instance:
            return None

        if instance.image_url and instance.image_url.startswith('http') \
                and not instance.image_url.startswith(settings.STATIC_URL):

            file = urllib3.urlopen(instance.image_url).read()
            image, image_url = image_post_create(self.current_website, file, instance.image_url)
            if image:
                instance.image_url = image_url
                image_post_processing(image)
                instance.save(update_fields=('image_url',))


class MarketplaceCategoryViewSet(MarketplaceApiAuthMixin, viewsets.ModelViewSet):
    queryset = ProductsCategory.objects.all()
    serializer_class = MarketplaceCategorySerializer

    def get_object(self):
        obj = self.get_queryset().filter(original_id=self.kwargs.get('pk')).first()
        if not obj:
            raise Http404
        else:
            return obj

    def perform_update(self, serializer):
        parent_id = self.request.data.get('parent', None)
        parent = ProductsCategory.objects.filter(original_id=parent_id, website=self.current_website).first() \
            if parent_id else None
        serializer.save(website=self.current_website, parent=parent)

    def perform_create(self, serializer):
        obj_id = self.request.data.get('id', None)
        if not obj_id:
            raise exceptions.ValidationError("Field 'id' is required")

        parent_id = self.request.data.get('parent', None)
        parent = ProductsCategory.objects.filter(original_id=parent_id, website=self.current_website).first() \
            if parent_id else None
        serializer.save(website=self.current_website, original_id=obj_id, parent=parent)
