import hashlib
import json
from rest_framework import status

from nicechat.utils.tests import TestCaseWithSites
from ..website.models import AccessToken


class MarketplaceAuthTest(TestCaseWithSites):
    num_of_companies = 1
    users_per_company = 1
    site_per_company = 1

    @staticmethod
    def make_sign(payload, secret, action, request_method):
        if isinstance(payload, dict):
            hstr = u'&'.join(['%s=%s' % (j, payload.get(j)) for j in sorted(payload.keys())])
        elif isinstance(payload, str):
            hstr = payload
        else:
            return None

        hstr = u'%s%s%s%s' % (secret, action, hstr, request_method)

        try:
            return hashlib.sha256(hstr).hexdigest()
        except UnicodeEncodeError:
            return hashlib.sha256(str(hstr).encode('utf-8')).hexdigest()

    def setUp(self):
        super(MarketplaceAuthTest, self).setUp()

        self.key = 'FEDH0I9UO9DF08VH'
        self.secret = 'JVFzW6ywhw42KnyAZpSasTxaDUukIThD'

        self.token = AccessToken.objects.create(
            token=self.key,
            secret=self.secret,
            website=self.website_0_0
        )

        self.marketplace_url = '/marketplaces/v1/'
        self.products_url = self.marketplace_url + 'products/'
        self.categories_url = self.marketplace_url + 'categories/'

    def test_product(self):

        payload_ctg_1 = {
            'id': 'C-ID-1',
            'name': 'Example-1',
        }

        # create category. Send signature as POST vars
        sign = self.make_sign(payload_ctg_1, self.secret, self.categories_url, 'POST')
        payload_ctg_1['_key'] = self.key
        payload_ctg_1['_sign'] = sign

        response = self.client.post(self.categories_url, payload_ctg_1)
        self.assertEquals(response.status_code, status.HTTP_201_CREATED)

        payload_prod_1 = {
            'id': 'ID-1',
            'category': 'C-ID-1',
            'title': 'Example-1',
            'price': 100,
            'currency': 'usd',
            'in_stock': True
        }

        # send POST with in-headers signature

        response = self.client.post(self.products_url, payload_prod_1, HTTP_X_API_KEY=self.key,
            HTTP_X_API_SIGN=self.make_sign(payload_prod_1, self.secret, self.products_url, 'POST'))

        self.assertEquals(response.status_code, status.HTTP_201_CREATED)

        payload_prod_2 = json.dumps({
            'category': 'C-ID-1',
            'title': 'Example-1',
            'price': 200,
            'currency': 'usd',
            'in_stock': False
        })

        # send PUT with json data (property 'json=' not used, because signed not dictionary, but full string)

        response = self.client.put(self.products_url + 'ID-1/', payload_prod_2,
                                   content_type='application/json', HTTP_X_API_KEY=self.key,
                                   HTTP_X_API_SIGN=self.make_sign(payload_prod_2, self.secret,
                                                                  self.products_url + 'ID-1/', 'PUT'))

        self.assertEquals(response.status_code, status.HTTP_200_OK)

        payload_prod_2 = json.dumps({
            'title': 'Example-2',
            'price': 200,
        })

        response = self.client.patch(self.products_url + 'ID-1/', payload_prod_2,
                                     content_type='application/json', HTTP_X_API_KEY=self.key,
                                     HTTP_X_API_SIGN=self.make_sign(payload_prod_2, self.secret,
                                                                  self.products_url + 'ID-1/', 'PATCH'))

        self.assertEquals(response.status_code, status.HTTP_200_OK)

        # receive GET

        response = self.client.get(self.products_url,HTTP_X_API_KEY=self.key,
                                   HTTP_X_API_SIGN=self.make_sign({}, self.secret, self.products_url, 'GET'))

        self.assertEquals(response.status_code, status.HTTP_200_OK)