# -*- coding: utf-8 -*-

import string
import random
from rest_framework.pagination import PageNumberPagination
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import Count
from rest_framework import viewsets, status
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated, AllowAny
from rest_framework.response import Response

from nicechat.utils.permissions import IsCompanyOwnerOrAdmin
from .models import AIAgent, AIAgentLog, AIPreset, AIPresetQA, AIPresetFile
from .serializers import AIAgentSerializer, AIAgentLogSerializer, AIPresetSerializer, AIPresetQASerializer
from .tasks import create_or_update_agent

class FilterByWebsiteOrCompanyMixin(object):
    def get_queryset(self):
        queryset = super(FilterByWebsiteOrCompanyMixin, self).get_queryset()

        if self.request.user and self.request.user.is_authenticated and \
            (self.request.user.is_admin_or_owner(self.request.current_site) or self.request.user.is_superuser):
            return self.queryset.filter(website=self.request.current_site)
        elif self.request.user and self.request.user.is_authenticated:
            return queryset.none()
        else:
            raise PermissionDenied()

    def perform_create(self, serializer):
        serializer.save(website=self.request.current_site)


class AIAgentViewSet(FilterByWebsiteOrCompanyMixin, viewsets.ModelViewSet):
    serializer_class = AIAgentSerializer
    queryset = AIAgent.objects.all()
    permission_classes = [IsAuthenticated, IsCompanyOwnerOrAdmin]

    def perform_create(self, serializer):
        save_kwagrs = {
            'website': self.request.current_site,
            'assistant_id': 'new_' + ''.join(random.choice(string.digits + string.ascii_uppercase + string.ascii_lowercase) for _ in range(28))

        }
        if not serializer.validated_data.get('instructions'):
            save_kwagrs['instructions'] = "You are a customer support. Use your knowledge base to best respond to customer queries. Answer in a business style. Be briefly."

        a = serializer.save(**save_kwagrs)
        AIAgentLog.objects.create(
            agent=a,
            log='Created'
        )
        create_or_update_agent.apply_async(args=(a.pk,), countdown=15)


    def perform_update(self, serializer):
        a = serializer.save()
        AIAgentLog.objects.create(
            agent=a,
            log='Edit'
        )
        create_or_update_agent.apply_async(args=(a.pk,), countdown=15)


class AIAgentLogViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = AIAgentLogSerializer
    queryset = AIAgentLog.objects.all()
    permission_classes = [IsAuthenticated, IsCompanyOwnerOrAdmin]

    def get_queryset(self):
        queryset = super(AIAgentLogViewSet, self).get_queryset()
        queryset = queryset.filter(agent=self.kwargs['agent_id'])

        if self.request.user and self.request.user.is_authenticated and \
                (self.request.user.own_company or self.request.user.is_superuser):
            return queryset.filter(agent__website=self.request.current_site)
        elif self.request.user and self.request.user.is_authenticated:
            return queryset.none()
        else:
            raise PermissionDenied()


class AIPresetViewSet(FilterByWebsiteOrCompanyMixin, viewsets.ModelViewSet):
    serializer_class = AIPresetSerializer
    queryset = AIPreset.objects.all()
    permission_classes = [IsAuthenticated, IsCompanyOwnerOrAdmin]

    @action(detail=True, methods=['POST'], permission_classes=[IsAuthenticated])
    def apply(self, request, **kwargs):
        instance = self.get_object()
        if instance.state in ['new', 'draft']:
            instance.state = 'applying'
            instance.save()
            return Response(status=status.HTTP_204_NO_CONTENT) 
        return Response(data={"error": f"Wrong current state '{instance.state}' ('draft' required)"}, status=status.HTTP_400_BAD_REQUEST)

    def perform_create(self, serializer):
        file = serializer.validated_data.get('file', None)
        obj = serializer.save(
            website=self.request.current_site,
            company=self.request.user.company,
            file_id='new',
            qa_mode = False if file else True,
        )
        if file:
            AIPresetFile.objects.create(
                preset=obj,
                data=file.read(),
                filename=file.name
            )



class LargeResultsSetPagination(PageNumberPagination):
    page_size = 500
    page_size_query_param = 'page_size'
    max_page_size = 10000

class AIPresetQAViewSet(viewsets.ModelViewSet):
    serializer_class = AIPresetQASerializer
    queryset = AIPresetQA.objects.all()
    permission_classes = [IsAuthenticated, IsCompanyOwnerOrAdmin]
    pagination_class = LargeResultsSetPagination

    def get_queryset(self):
        queryset = super(AIPresetQAViewSet, self).get_queryset()
        queryset = queryset.filter(preset_id=self.kwargs['preset_id'])

        if self.request.user and self.request.user.is_authenticated and \
                (self.request.user.own_company or self.request.user.is_superuser):
            return queryset.filter(preset__website=self.request.current_site)
        elif self.request.user and self.request.user.is_authenticated:
            return queryset.none()
        else:
            raise PermissionDenied()

    def perform_create(self, serializer):
        obj = serializer.save(preset_id=self.kwargs['preset_id'])
        preset = obj.preset
        preset.state = 'draft' if preset.state != 'new' else 'new'
        preset.save()

    def perform_update(self, serializer):
        obj = serializer.save()
        preset = obj.preset
        preset.state = 'draft' if preset.state != 'new' else 'new'
        preset.save()

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        preset = instance.preset
        self.perform_destroy(instance)
        preset.state = 'draft' if preset.state != 'new' else 'new'
        preset.save()
        return Response(status=status.HTTP_204_NO_CONTENT)
