import json
import sqlite3
from websockets.sync.client import connect


BOT_LIST_ROOMS = 'bot_rooms'
BOT_GET_WEBSITE = 'bot_website'
BOT_SEND_TEXT = 'bot_send_text'
BOT_SEND_JOINED = 'bot_send_joined'
BOT_SEND_LEAVE = 'bot_send_leave'
BOT_GET_ROOM = 'bot_get_room'
BOT_ON_MESSAGE = 'bot_message'
BOT_ON_NEWROOM = 'bot_new_room'
BOT_ON_EDTROOM = 'bot_edit_room'
BOT_ON_ENDROOM = 'bot_end_room'
BOT_SEND_TECH = 'bot_send_tech'
BOT_SEND_FORM = 'bot_send_form'
BOT_LIST_FORMS = 'bot_list_forms'


class BotLibBase(object):

    def __init__(self, host, site_id, persistent_db_enable=False, persistent_db_path="/tmp/persistent.sqlite3"):
        _host = host.rstrip('/')
        if persistent_db_enable:
            self.persistent = sqlite3.connect(persistent_db_path)
            cur = self.persistent.cursor()
            try:
                cur.execute('CREATE TABLE "persistent_vals" ( "chat_id" INTEGER NOT NULL, "value" TEXT, PRIMARY KEY ("chat_id"))')
            except Exception as e:
                pass
            cur.close()
        else:
            self.persistent = None
        self.ws = connect(f"{_host}/bot-api/{site_id}/")
        self.website = {}

    def on_receive(self, chat_id, message):
        pass

    def on_receive_tech(self, chat_id, message):
        pass

    def on_new_room(self, chat_id, room):
        pass

    def on_exists_room(self, chat_id, room):
        pass

    def on_get_room(self, chat_id, room):
        pass

    def on_end_room(self, chat_id, room):
        pass

    def get_session(self, chat_id):
        if not self.persistent:
            return {}
        cur = self.persistent.cursor()
        try:
            rec = cur.execute('select "value" from "persistent_vals" where "chat_id" = ?;', [chat_id]).fetchone()
            if rec:
                data = json.loads(rec[0])
            else:
                data = {}
        except Exception as e:
            data = {}
        cur.close()
        return data  
        
    def set_session_val(self, chat_id, key, val):
        if not self.persistent:
            return False
        cur = self.persistent.cursor()
        try:
            rec = cur.execute('select "value" from "persistent_vals" where "chat_id" = ?;', [chat_id]).fetchone()
            if rec:
                data = json.loads(rec[0])
            else:
                data = {}
        except Exception as e:
            data = {}
        data[key] = val
        data = json.dumps(data)
        try:
            rec = cur.execute('insert or replace into "persistent_vals" ("chat_id", "value") values (?,  ?);', [chat_id, data])
            ret = True
        except Exception as e:
            ret = False
        cur.close()
        return ret
            
    def _clean_session(self, chat_id):
        if not self.persistent:
            return
        cur = self.persistent.cursor()
        try:
            cur.execute('delete from "persistent_vals" where "chat_id" = ?', chat_id)
        except Exception:
            pass
        cur.close()

    def get_room(self, chat_id):
        self.ws.send(json.dumps({"type": BOT_GET_ROOM, "room_id": chat_id}))

    def send(self, chat_id, text, data=None, bot_props=None, name='Bot'):
        self.ws.send(json.dumps({
            "type": BOT_SEND_TEXT, 
            "room_id": chat_id, 
            "name": name, 
            "text": text,
            "props": bot_props,
            "data": data
            }))

    def send_joined(self, chat_id, name='Bot', avatar_url=None):
        self.ws.send(json.dumps({
            "type": BOT_SEND_JOINED, 
            "room_id": chat_id, 
            "name": name, 
            "avatar": avatar_url
            }))

    def send_leave(self, chat_id, name='Bot',):
        self.ws.send(json.dumps({
            "type": BOT_SEND_LEAVE,
            "room_id": chat_id,
            "name": name
            }))
        
    def show_tooltip(self, chat_id, text, answer, command):
        self.ws.send(json.dumps({
            "type": BOT_SEND_TECH,
            "room_id": chat_id,
            "cmd": {
                "type": "TT_SHOW",
                "text": text,
                "answer": answer,
                "command": command,
            }
        }))
        
    def hide_tooltip(self, chat_id):
        self.ws.send(json.dumps({
            "type": BOT_SEND_TECH,
            "room_id": chat_id,
            "cmd": {
                "type": "TT_HIDE",
            }
        }))

    def get_custom_forms(self):
        self.ws.send(json.dumps({
            "type": BOT_LIST_FORMS,
        }))

    def send_form(self, chat_id, predefined=None, custom_id=None):
        if predefined:
            self.ws.send(json.dumps({
                "type": BOT_SEND_FORM,
                "room_id": chat_id,
                "predefined": predefined
            }))
        elif custom_id:
            self.ws.send(json.dumps({
                "type": BOT_SEND_FORM,
                "room_id": chat_id,
                "custom_id": custom_id,
            }))
        else:
            print('Please specify predefined name or id of custom form')

    def run(self):
        while True:
            result = json.loads(self.ws.recv())
            _type = result.get('type')
            if _type == BOT_ON_MESSAGE:
                data = result.get('data', {})
                if data['type'] == 'cli':
                    self.on_receive(result['chat_id'], data)
                else:
                    self.on_receive_tech(result['chat_id'], data)
            elif _type == BOT_GET_WEBSITE:
                self.website = result.get('data', {})
            elif _type == BOT_LIST_ROOMS:
                for room in result.get('data', []):
                    self.on_exists_room(room['id'], room)
            elif _type == BOT_LIST_FORMS:
                self.on_forms_list(result.get('data', []))
            elif _type == BOT_GET_ROOM:
                room = result.get('data', {})
                self.on_get_room(room['id'], room)
            elif _type == BOT_ON_ENDROOM:
                self._clean_session(result['chat_id'])
                self.on_end_room(result['chat_id'], result.get('data', {}))
            elif _type == BOT_ON_NEWROOM:
                self.on_new_room(result['chat_id'], result.get('data', {}))
            elif _type == BOT_ON_EDTROOM:
                self.on_exists_room(result['chat_id'], result.get('data', {}))
