import json
from datetime import datetime
from nicechat.utils.constants import WebsocketEventTypes
from apps.chat.models import Message, Chat, Visitor, VisitedPage
from apps.website.models import Website, Form
from apps.website.serializers import WebsiteSmallSerializer, FormForWidgetSerializer
from apps.chat.serializers import ChatBotSerializer, MessageSerializer

DEF_BOT_NAME='TheBot'

def bot_monitor_packet(dt, kwargs, ch_send):
    if isinstance(dt, dict):
        _type = dt.get('type', None)
        site_id = kwargs.get('site_id')
        room_id = dt.get('room_id')
        if room_id:
            room = Chat.objects.filter(website_id=site_id, is_open=True, pk=room_id).first()
        else:
            room = None

        if _type == WebsocketEventTypes.BOT_SEND_TEXT and room:
            ws_bot_send_text(room, dt, site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_GET_ROOM and room:
            ws_get_room(room, site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_SEND_JOINED and room:
            ws_bot_join_room(room, dt, site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_SEND_LEAVE and room:
            ws_bot_leave_room(room, dt, site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_GET_WEBSITE:
            ws_get_website(site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_SEND_FORM:
            ws_send_form(room, dt, site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_LIST_FORMS:
            ws_list_forms(site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_LIST_ROOMS:
            ws_list_rooms(site_id, ch_send)
        elif _type == WebsocketEventTypes.BOT_SEND_TECH:
            ws_bot_send_tech(room, dt, site_id, ch_send)
                

def ws_get_website(site_id, ch_send):
    ws = Website.objects.filter(pk=site_id).first()
    ch_send(text_data=json.dumps({
        'type': WebsocketEventTypes.BOT_GET_WEBSITE, 
        'data': WebsiteSmallSerializer(ws).to_representation(ws)
    }))


def ws_list_forms(site_id, ch_send):
    data = [{
                "custom_id": f.pk,
                "name": f.label
            } if f.mode == Form.MODE_SITE_REGULAR_FORM else {
                "predefined": f.mode,
                "name": f.label
            } for f in Form.objects.filter(website_id=site_id)]
    ch_send(text_data=json.dumps({
        'type': WebsocketEventTypes.BOT_LIST_FORMS, 
        'data': data
    }))


def ws_list_rooms(site_id, ch_send):
    rooms = Chat.objects.filter(website_id=site_id, is_open=True)
    ch_send(text_data=json.dumps({
        'type': WebsocketEventTypes.BOT_LIST_ROOMS, 
        'data': ChatBotSerializer(rooms, many=True).to_representation(rooms)
    }))

def ws_get_room(room, site_id, ch_send):
    data = ChatBotSerializer(room).to_representation(room)
    msgs = room.message_set.all()
    msgs = MessageSerializer(msgs, many=True).to_representation(msgs)
    for msg in msgs:
        if msg.get('type') == Message.TYPE_CLIENT:
            msg['sender_id'] = 'client'

    data['messages'] = msgs
    ch_send(text_data=json.dumps({
        'type': WebsocketEventTypes.BOT_GET_ROOM, 
        'data': data
    }))

def ws_send_form(room, dt, site_id, ch_send):
    predefined=dt.get('predefined', None)
    custom_id=dt.get('custom_id', None)

    form = Form.objects.filter(website_id=site_id)
    if predefined:
        form = form.filter(mode=predefined)
    else:
        form = form.filter(pk=custom_id)

    form = form.first()

    if form:
        data = dict(FormForWidgetSerializer(form).to_representation(form))
        Message.objects.create(
            chat=room,
            sender_id=0,
            sender_name=room.website.title,
            text=form.label,
            data=json.dumps(data),
            is_synthetic=True,
            data_type=Message.DATA_FORM,
            hidden=json.dumps({'form_id': form.pk}),
            type=Message.TYPE_OPERATOR
        )

def ws_bot_join_room(room, dt, site_id, ch_send):
    name=dt.get('name', DEF_BOT_NAME)
    avatar=dt.get('avatar', None)
    txt = f"~!JOINED:{name}"
    room.put_tech_message_wdata(txt, {
        'name': name,
        'avatar': avatar
    }, is_synthetic=True)

def ws_bot_send_tech(room, dt, site_id, ch_send):
    name=dt.get('name', DEF_BOT_NAME)
    cmd=dt.get('cmd', None)
    if cmd:
        room.put_tech_message_wdata("", {
            'name': name,
            'type': 'BOT_CMD',
            'cmd': cmd,
        }, is_synthetic=True)


def ws_bot_leave_room(room, dt, site_id, ch_send):
    name=dt.get('name', DEF_BOT_NAME)
    txt = f"~!LEAVE:{name}"
    room.put_tech_message_wdata(txt, {
        'name': name
    }, is_synthetic=True)


def ws_bot_send_text(room, dt, site_id, ch_send):
    name=dt.get('name', DEF_BOT_NAME)
    txt=dt.get('text', '')
    data=dt.get('data', None)
    props=dt.get('props', None)
    msg = Message(text=str(txt), sender_id="0", sender_name=name, is_synthetic=True, chat=room, type=Message.TYPE_OPERATOR)
    if data:
        try:
            msg.data_type = Message.DATA_ANSWERS
            msg.data = json.dumps(data)
        except Exception as e:
            import traceback
            traceback.print_exception(e)
    if props:
        try:
            msg.hidden = json.dumps({"bot_props": props})
        except Exception as e:
            import traceback
            traceback.print_exception(e)
    msg.save()
