import json
from django.db import IntegrityError, transaction
from django.urls import reverse
from django.conf import settings
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django.contrib.auth import get_user_model

from allauth.account.models import EmailConfirmation
from allauth.account.utils import setup_user_email
from rest_framework import viewsets, status
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from apps.users.models import UserSiteRole
from nicechat.analytics import event_track, AGENT_INVITED, AGENT_REMOVED
from nicechat.utils.mail import notify_users
from nicechat.utils.permissions import IsCompanyOwner, IsCompanyOwnerOrAdmin
from nicechat.utils.serializers import raise_serializer_error
from ..website.models import Website
from .models import User, UserFiltering
from .serializers import UserSerializer, AgentInviteSerializer, UserSiteRoleSerializer, UserFilteringSerializer

SEND_PASSWORD_SUBJECT = _('Your credentials for use %s dashboard') % settings.BASE_CHAT_NAME
INVITATION_MESSAGE_SUBJECT = _('You have been invited to %s') % settings.BASE_CHAT_NAME
INVITATION_WEBSITE_ERROR = _('Error occurred when trying to find website with id: {}')
DUPLICATE_USER_EMAIL_ERROR = _('User with this email is already registered')
AGENTS_LIMIT_REACHED = _('Current company plan does not allow addind more agents')

SITE_NOT_FOUND_ERROR_TEXT = "There was a problem with finding the site object"


class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer

    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        if self.action == 'partial_update':
            # update only current user's data
            self.queryset = self.queryset.filter(pk=self.request.user.pk)
        elif self.request.user.company \
                and (self.request.user == self.request.user.company.owner
                     or self.request.user.is_site_admin(self.request.current_site)):
            # all the other actions can be applied to the users
            # of the same company and by the company owner or it's admins
            self.queryset = self.queryset.filter(company=self.request.user.company)
        else:
            self.queryset = self.queryset.none()

        return super(UserViewSet, self).get_queryset()

    def get_serializer_class(self):
        if self.action == 'change_role':
            return UserSiteRoleSerializer

        return super(UserViewSet, self).get_serializer_class()

    def get_permissions(self):
        if self.action == 'list':
            return [IsAuthenticated(), IsCompanyOwnerOrAdmin()]
        return super(UserViewSet, self).get_permissions()

    @action(detail=False, methods=['get'])
    def me(self, request):
        # if request.current_site is None:
        #     raise NotFound('You seem to have nothing to do with the current site')

        serializer = self.get_serializer(request.user)
        data = serializer.data
        data['map_key'] = settings.GOOGLE_MAP_KEY
        return Response(data)

    @action(detail=False, methods=['get', 'post', 'delete'])
    def tfa(self, request):

        if not settings.TFA_ENABLE:
            return Response({
                "on": False,
                "disabled": True
            })

        user = self.request.user
        try:
            tfa = json.loads(user.tfa_opts)
        except ValueError:
            tfa = {}

        is_tfa = True if settings.TFA_ENABLE and tfa.get('on') == True else False

        ret = {}
        otp_factor = getattr(User, 'otp_factor')

        ret['method'] = tfa.get('method')
        ret["otp-list"] = ','.join(otp_factor.methods())
        ret['lock'] = otp_factor.lock(ret['method'])

        if request.method == 'DELETE':
            if not is_tfa:
                return Response({'detail': 'Not enabled'}, status=402)
            otp = request.GET.get('otp', None)
            if not otp:
                opts = otp_factor.prepare(user, request)
                if opts and len(opts) > 1:
                    ret['method'] = opts[0]
                    ret['valid_sec'] = opts[1]
            else:
                if otp_factor.validate(otp, user, request):
                    otp_factor.disable(user)
                    is_tfa = False
                    ret['lock'] = []
                    ret['method'] = None

        if request.method == 'POST':
            if not is_tfa:
                otp = request.data.get('otp', None)
                if otp:
                    if otp_factor.validate(otp, user, request, True):
                        is_tfa=True
                else:
                    opts = otp_factor.prepare(user, request, True)
                    if opts and len(opts) > 1:
                        ret['method'] = opts[0]
                        ret['valid_sec'] = opts[1]

        ret['on'] = is_tfa

        return Response(ret)

    @action(detail=False, methods=['post'])
    def avatar(self, request):
        if len(request.FILES['file']) > 0:
            filename = request.FILES['file'].name
            self.request.user.avatar.save(filename, request.FILES['file'].file, True)
        return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=True, methods=['post', 'get'], permission_classes=[IsCompanyOwner], serializer_class=UserFilteringSerializer)
    def filtering(self, request, pk):
        user = self.get_object()
        try:
            uf = user.userfiltering
        except User.userfiltering.RelatedObjectDoesNotExist:
            uf = UserFiltering(user=user)
            uf.save()

        if request.method == 'POST':
            serializer = UserFilteringSerializer(uf, data=request.data)
            serializer.is_valid(raise_exception=True)
            serializer.update(uf, serializer.validated_data)
        
        return Response(UserFilteringSerializer(uf).to_representation(uf), status=status.HTTP_200_OK)

    @action(detail=False, methods=['post'], permission_classes=[IsCompanyOwner])
    def invite(self, request):
        agents_data = request.data.get('agents')
        result = []

        for index, agent_data in enumerate(agents_data):
            result.append({'success': False})

            # check if current company plan allows more agents
            if self.request.user.company.users.count() >= self.request.user.company.billing_plan.agents_count:
                result[index].setdefault('errors', {}).setdefault('__all__', []).append(AGENTS_LIMIT_REACHED)
                continue

            email_address = agent_data.get('email')
            site_id = agent_data.get('site_id')
            role = agent_data.get('role')

            agent_invite_serializer = AgentInviteSerializer(
                data={'email': email_address, 'site_id': site_id, 'role': role}
            )
            if not agent_invite_serializer.is_valid():
                result[index].setdefault('errors', agent_invite_serializer.errors)
                continue

            try:
                website = Website.objects.get(pk=site_id, company=request.user.company)
            except (ValueError, Website.DoesNotExist) as e:
                result[index].setdefault('errors', {}).setdefault('site_id', []).append(
                    INVITATION_WEBSITE_ERROR.format(site_id)
                )
                # todo report error to admins/sentry etc?
                continue

            with transaction.atomic():
                try:
                    user = get_user_model().objects.create(email=email_address,
                                                           company=self.request.user.company,
                                                           is_active=False)
                    UserSiteRole.objects.create(user=user, site=website, role=role)
                except IntegrityError:
                    result[index].setdefault('errors', {}).setdefault('email', []).append(
                        DUPLICATE_USER_EMAIL_ERROR
                    )
                    continue

            website.users.add(user)

            email_object = setup_user_email(request, user, [])
            confirmation = EmailConfirmation.create(email_object)
            confirmation_link = settings.AGENT_CONFIRM_URL.replace(":token", confirmation.key)

            event_track(AGENT_INVITED, request.user.id, {
                'email': user.email,
                'companyId': request.user.company_id,
                'companyName': request.user.company.name,
            })

            notify_users(
                    [user],
                    INVITATION_MESSAGE_SUBJECT,
                    'users/email/invite.txt',
                    'users/email/invite.html',
                    {'confirmation_link': confirmation_link}
                )

            # allauth.account.models.EmailConfirmation#send
            confirmation.sent = timezone.now()
            confirmation.save()

            result[index]['success'] = True

        return Response(status=status.HTTP_200_OK,
                        data={'result': result})

    def destroy(self, request, *args, **kwargs):
        # todo don't delete user, but only remove him from the company
        # todo see apps.website.api.AgentViewSet#destroy
        if self.get_object() == self.request.user:
            return Response(status=status.HTTP_400_BAD_REQUEST,
                            data={'message': 'You can not remove yourself from the company'})
        else:
            event_track(AGENT_REMOVED, self.request.user.id, {
                'company': self.request.user.company_id,
            })
            return super(UserViewSet, self).destroy(request, *args, **kwargs)

    @action(detail=True, methods=['post'], permission_classes=[IsCompanyOwnerOrAdmin], url_path='change-role')
    def change_role(self, request, pk=None):
        user = self.get_object()

        if request.current_site is None:
            raise_serializer_error(SITE_NOT_FOUND_ERROR_TEXT)

        user_error = None
        if user == request.user:
            user_error = 'yourself'
        elif user.is_company_owner:
            user_error = 'company owner'
        if user_error:
            raise_serializer_error({'user': 'You cannot change the role for {}'.format(user_error)})

        serializer = self.get_serializer(user, data={'role': request.data.get('role'),
                                                     'site_id': request.current_site.pk})
        serializer.is_valid(raise_exception=True)

        UserSiteRole.objects.update_or_create(user=user, site=request.current_site,
                                              defaults={'role': serializer.validated_data['role']})

        return Response(serializer.validated_data)
