import time
import hashlib
import ipaddress
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from django.contrib.auth.models import AbstractUser
from sorl.thumbnail import ImageField, get_thumbnail

from apps.website.models import Website
from nicechat.utils.constants import UserRoles


class User(AbstractUser):
    """
    Extending custom user EmailUser model
    """
    ENGLISH = 'en'
    UKRAINIAN = 'uk'

    USERNAME_FIELD = "email"
    REQUIRED_FIELDS = []

    LANGUAGES = (
        (ENGLISH, _('English')),
        (UKRAINIAN, _('Ukrainian')),
    )

    username = None
    email = models.EmailField(_('email address'), unique=True)

    first_name = models.CharField(_('first name'), max_length=50, blank=True)
    last_name = models.CharField(_('last name'), max_length=50, blank=True)
    # todo can 'company' field be blank at all?
    company = models.ForeignKey('company.Company', blank=True, null=True, related_name='users', on_delete=models.CASCADE)
    site_roles = models.ManyToManyField('website.Website', through='UserSiteRole')
    label = models.CharField(max_length=100, blank=True, null=True)
    contacts = models.TextField(blank=True, null=True)
    avatar = ImageField(upload_to='avatars', blank=True, null=True)
    position = models.CharField(max_length=50, blank=True, default="")
    phone_number = models.CharField(max_length=50, blank=True, default="")
    timezone = models.CharField(max_length=50, blank=True, null=True)
    interface_language = models.CharField(max_length=2, choices=LANGUAGES, default=ENGLISH)
    tfa_opts = models.TextField(default='{}')

    @cached_property
    def get_full_name(self):
        if self.first_name or self.last_name:
            return ' '.join([self.first_name, self.last_name])
        else:
            return super(User, self).get_full_name()
    get_full_name.short_description = 'Full name'

    @cached_property
    def user_tag(self):
        return str(self.pk)

    @cached_property
    def status(self):
        return self.pk

    def is_admin_or_owner(self, site_obj):
        return self.get_site_role(site_obj) in [UserRoles.SITE_ADMIN, UserRoles.COMPANY_OWNER]

    def get_site_role(self, site_obj):
        """
        Returns user role at specific site provided as a parameter
        :param site_obj: apps.website.models.Website instance
        :return: user's role or None

        """
        # todo cache the value and invalidate on m2m changed signal
        assert type(site_obj) is Website
        site_role = None

        role_qs = UserSiteRole.objects.filter(user=self, site=site_obj, site__company=self.company)
        if role_qs.exists():
            site_role = role_qs.first().role

        if not site_role and self.is_company_owner:
            return 'owner'
        
        if not site_role and self.website_set.filter(pk=site_obj.pk).exists():
            UserSiteRole.objects.get_or_create(
                user=self,
                site=site_obj,
            )

        return site_role

    @cached_property
    def is_company_owner(self):
        return self.company and self.company.owner == self or False

    def is_site_admin(self, site):
        return UserSiteRole.objects.filter(
            user=self,
            site=site,
            site__company=self.company,
            role=UserRoles.SITE_ADMIN
        ).exists()

    def is_site_agent(self, site):
        return UserSiteRole.objects.filter(
            user=self,
            site=site,
            site__company=self.company,
            role=UserRoles.SITE_AGENT
        ).exists()

    def get_avatar(self):
        # todo makes no sense after change in 2ed4cd0beca012faa00ec98a9c16f1948bd3a48d
        if self.avatar:
            return self.avatar
        else:
            return None

    def avatar_thumbs(self):
        thumb = None
        medium = None
        large = None

        try:
            thumb = get_thumbnail(self.get_avatar(), '43x43', crop='center', quality=99)
            medium = get_thumbnail(self.get_avatar(), '86x86', crop='center', quality=99)
            large = get_thumbnail(self.get_avatar(), '150x150', crop='center', quality=99)
        except (IOError, ValueError):
            pass

        return {
            'thumb': thumb,
            'medium': medium,
            'large': large
        }


class Otp(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    module = models.CharField(max_length=16)
    val = models.CharField(max_length=128)
    ts = models.IntegerField(db_index=True)

    @staticmethod
    def write(user, kind, val, ts=180):
        return Otp.objects.create(
            user=user,
            module=kind,
            val=hashlib.md5(val).hexdigest(),
            ts=time.time() + ts
        )

    @staticmethod
    def read(user, kind, val):
        Otp.objects.filter(ts__lt=time.time()).delete()

        return Otp.objects.filter(
            user=user,
            module=kind,
            val=hashlib.md5(val).hexdigest(),
        ).first()


class UserSiteRole(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    site = models.ForeignKey('website.Website', on_delete=models.CASCADE)
    role = models.CharField(max_length=16, default=UserRoles.SITE_AGENT, choices=UserRoles.CHOICES)


class WhitelistAddr(models.Model):
    addr = models.GenericIPAddressField(protocol='both')
    mask = models.IntegerField(default=32)
    req = models.CharField(max_length=50, db_index=True, unique=True)

    def __str__(self):
        return f'WhitelistAddr-{self.req}'


    def save(self, *args, **kwargs):
        self.req = f'{self.addr}/{self.mask}'

        if self.mask != 32:
            ipaddress.ip_network(self.req)
            
        super(WhitelistAddr, self).save(*args, **kwargs)


class UserFiltering(models.Model):
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    enable = models.BooleanField(default=False)
    addrs = models.JSONField(default=list)
    
    def __str__(self):
        addr = ", ".join(self.addrs) if self.addrs and isinstance(self.addrs, list) else '[]'
        return f'Filter: User {self.user_id} {"on" if self.enable else "off"} - {addr}'
