from datetime import timedelta
import random
import string

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.utils import timezone
from django.views.i18n import LANGUAGE_QUERY_PARAMETER
from django.utils.translation import (
    check_for_language, get_language, to_locale,
)

from django.contrib.auth.password_validation import validate_password
from django.conf import settings

from rest_framework import serializers, exceptions
from rest_framework.fields import ImageField, empty
from dj_rest_auth.registration.serializers import RegisterSerializer

from nicechat.utils.constants import UserRoles
from ..billing.models import CompanyPlan
from ..billing.utils import get_trial_plan
from ..company.models import Company
from ..website.models import AgentStatus, Website, DomainList, Status
from ..users.models import UserFiltering
from ..users.utils import validate_addr


class AgentStatusSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        if hasattr(self, 'context') and self.context:
            request = self.context.get('request', {})
            site_id = None
            if self.context.get('site_id', None):
                site_id = self.context.get('site_id', None)
            if not site_id and hasattr(request, 'current_site'):
                site_id = request.current_site.pk if request.current_site else None
            if not site_id and request and hasattr(request, 'query_params'):
                site_id = request.query_params.get('site_id')
                
            if site_id:
                active = AgentStatus.objects.filter(website_id=site_id, operator_id=obj).first()
                if active:
                    return active.state if active.is_active else AgentStatus.STATUS_AWAY

        return AgentStatus.STATUS_AWAY


class UserSerializer(serializers.ModelSerializer):
    avatar = serializers.DictField(source='avatar_thumbs', child=ImageField())
    company = serializers.PrimaryKeyRelatedField(read_only=True)
    status = AgentStatusSerializer(many=False, read_only=True, required=False)
    own_company = serializers.PrimaryKeyRelatedField(read_only=True)
    role = serializers.SerializerMethodField(read_only=True, required=False)
    locale = serializers.SerializerMethodField(read_only=True, required=False)
    billing = serializers.SerializerMethodField(read_only=True, required=False)
    allow_video = serializers.SerializerMethodField(read_only=True, required=False)
    is_on_site = serializers.SerializerMethodField(read_only=True, required=False)

    @staticmethod
    def get_allow_video(instance):
        if not settings.NC_VIDEO_ENABLE:
            return False
        try:
            return instance.company.allow_video if instance.company else False
        except Company.DoesNotExist:
            return False

    @staticmethod
    def get_billing(instance):
        try:
            return instance.own_company.billing_id if instance.own_company else None
        except Company.DoesNotExist:
            return None

    def get_locale(self, instance):
        request = self.context.get('request')

        if not request:
            return None
        language = request.GET.get(LANGUAGE_QUERY_PARAMETER)
        if not (language and check_for_language(language)):
            language = get_language()
        return to_locale(language)

    def get_role(self, instance):
        request = self.context.get('request')

        if not getattr(request, 'user', AnonymousUser()).is_authenticated:
            return None

        if hasattr(request, 'current_site') and request.current_site:
            return request.user.get_site_role(request.current_site)

    def get_is_on_site(self, instance):
        site_id = self.context.get('site_id')
        if site_id:
            ass = AgentStatus.objects.filter(website_id=site_id, operator=instance).first()
            return ass.is_current if ass else False
        request = self.context.get('request')
        if hasattr(request, 'current_site') and request.current_site:
            ass = AgentStatus.objects.filter(website=request.current_site, operator=instance).first()
            return ass.is_current if ass else False
        return False

    class Meta:
        model = get_user_model()
        fields = (
            'id', 'email', 'first_name', 'last_name', 'avatar', 'company', 'position', 'phone_number', 'get_full_name',
            'own_company', 'is_active', 'is_on_site', 'status', 'role', 'interface_language', 'locale', 'billing', 'allow_video'
        )
        read_only_fields = ('id', 'email', 'avatar', 'get_full_name', 'company', 'own_company', 'is_active', 'locale', 'billing', 'is_on_site')


class UserSiteRoleSerializer(serializers.ModelSerializer):
    role = serializers.ChoiceField(UserRoles.CHOICES)
    site_id = serializers.IntegerField()

    def validate(self, data):
        """
        Check that the user belongs to the site
        """
        if data['site_id'] not in self.instance.website_set.values_list('pk', flat=True):
            raise serializers.ValidationError("User does not belong to the site")
        return data

    class Meta(UserSerializer.Meta):
        model = get_user_model()
        fields = ('role', 'site_id',)


class WebsiteRegisterSerializer(RegisterSerializer):
    """
    Extends basic register serializer to add website-related fields

    """
    # first_name = serializers.CharField()
    # last_name = serializers.CharField()
    # website_name = serializers.CharField()
    autopassword = serializers.BooleanField(required=False)
    website_url = serializers.RegexField('^(https?:\/\/)?(www\.)?(((?!www\.)[a-zA-Z0-9-])+\.)+([a-zA-Z]{2,})(\:\d+)?\/?$', required=False)
    token = serializers.SerializerMethodField()

    def __init__(self, instance=None, data=empty, **kwargs):
        if data is not empty:
            if data.get('autopassword', False) == True:
                alphabet = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                data['password1'] = ''.join([alphabet[random.randrange(len(alphabet))] for j in range(10)])
                data['password2'] = data['password1']
        super(WebsiteRegisterSerializer, self).__init__(instance, data, **kwargs)

    def _validate_website_data(self, field_name, value):
        """
        Validates website title and domain uniqueness

        """
        if Website.objects.filter(**{field_name: value}).exists():
            raise serializers.ValidationError("Website with this {0} is already registered".format(field_name))
        return value

    # def validate_website_name(self, value):
    #     return self._validate_website_data('title', value)

    def validate_website_url(self, value):
        return self._validate_website_data('domain', value) if value else None

    def get_cleaned_data(self):
        return {
            # 'first_name': self.validated_data.get('first_name', ''),
            # 'last_name': self.validated_data.get('last_name', ''),
            'first_name': self.validated_data.get('website_url', ''),
            'last_name': 'admin',
            'password1': self.validated_data.get('password1', ''),
            'email': self.validated_data.get('email', '')
        }

    def get_token(self, instance):
        return getattr(self, 'nc_token', None)

    def custom_signup(self, request, user):
        """
        Creates corresponding db objects for data provided during registration for:
         * company
         * website

        + adds newly registered user to the list of agents of new website

        :param request:
        :param user:
        :return:
        """
        from apps.balancer.models import ArbiterNode, ArbiterNodeWebsite
        
        tgt = ArbiterNode.objects.filter(for_new=True).values_list('pk', flat=True)
        if not tgt and not settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
            raise serializers.ValidationError("Not exists active worker-node for new websites")

        tgt_pk = random.choice(tgt) if tgt else None
        tgt = ArbiterNode.objects.get(pk=tgt_pk) if tgt_pk else None

        # website_name = self.validated_data.get('website_name')
        website_name = self.validated_data.get('website_url')
        if not website_name:
            website_name = f'Website {user}'
        from apps.balancer.tasks import sync_website, sync_user
        sync_user(user, target=tgt)

        new_company = Company.objects.create(
            label=website_name,
            owner=user
        )

        # set up a trial plan for new company
        trial_plan = get_trial_plan()
        now = timezone.now()

        trial_plan_agents_count = 3
        trial_plan_duration_days = 14

        CompanyPlan.objects.create(
            company=new_company,
            plan=trial_plan,
            agents_count=trial_plan_agents_count,
            begin=now,
            end=now + timedelta(days=trial_plan_duration_days),
        )

        # associate newly registered user with company
        user.company = new_company
        user.save()


        # new website for the company
        domain=self.validated_data.get('website_url')
        website = Website.objects.create(title=website_name, company=new_company)

        if tgt:
            ArbiterNodeWebsite.objects.create(
                node=tgt,
                website=website
            )

        sync_website(website)

        website.users.add(user)
        Status.objects.create(
            website=website,
            server_code=200,
            widget_installed=True,
            allow_site_access=True
        )

        if domain:
            try:
                DomainList.objects.create(
                    kind=DomainList.DOMAIN_REGULAR,
                    domain=domain,
                    website=website
                )
            except Exception:
                pass
        letters = string.digits + string.ascii_lowercase
        for q in range(0, 50):
            tok = ''.join(random.choice(letters) for i in range(8))
            if not DomainList.objects.filter(domain=tok, kind=DomainList.DOMAIN_DIRECT).count():
                DomainList.objects.create(
                    kind=DomainList.DOMAIN_DIRECT,
                    domain=tok,
                    website=website
                )
                break
        # TODO add to sync
        sync_website(website)
        sync_user(user)

        if self.validated_data.get('autopassword', False) == True:
            from .api import notify_users, SEND_PASSWORD_SUBJECT
            notify_users(
                [user],
                SEND_PASSWORD_SUBJECT,
                'users/email/sent_password.txt',
                'users/email/sent_password.html',
                self.validated_data
            )


class AgentInviteSerializer(serializers.Serializer):
    email = serializers.EmailField()
    site_id = serializers.IntegerField()
    role = serializers.ChoiceField(UserRoles.CHOICES)


class CreateAgentSerializer(serializers.Serializer):
    first_name = serializers.CharField()
    last_name = serializers.CharField()
    password1 = serializers.CharField(validators=[validate_password])
    password2 = serializers.CharField()

    def validate(self, values):
        if values.get('password1') != values.get('password2'):
            raise exceptions.ValidationError("Mismatch passwords")
        return values


class UserFilteringSerializer(serializers.ModelSerializer):
   
    def validate(self, data):
        if not isinstance(data.get('addrs', None), list):
            raise serializers.ValidationError("'addrs' must be a list")
        if not data['addrs'] and data['enable']:
            raise serializers.ValidationError("You cannot enable filtering by empty Address List")
        err = []
        for addr in data['addrs']:
            if not validate_addr(addr):
                err.append(f'{addr}')
        if err:
            raise serializers.ValidationError("Wrong address %s" % ', '.join(err))
        return data

    class Meta(UserSerializer.Meta):
        model = UserFiltering
        fields = ('enable', 'addrs',)
