import threading
from django.dispatch import receiver
from django.conf import settings
from allauth.account.signals import user_signed_up, user_logged_in
from django.db.models.signals import post_save, post_delete
from django.utils.translation import gettext_lazy as _
from django.core.cache import caches, InvalidCacheBackendError

from nicechat.analytics import event_people_set, event_track, REGISTER_COMPLETED
from apps.company.models import Company
from apps.users.models import User, UserSiteRole, WhitelistAddr
from nicechat.utils.mail import notify_emails
from nicechat.utils.billing import execute_billing
from nicechat.utils.arbiter import update_data_workernode
from nicechat.utils.arbiter_sync import get_user_relations


@receiver(user_signed_up)
def user_signed_up_(request, user, **kwargs):
    request.is_allauth_sign_out = True

    websites_without_script = ', '.join(map(str, [w.domain for w in user.website_set.filter(
                                             status__widget_installed__exact=False)]))
    website_not_reachable = ', '.join(map(str, [w.domain for w in user.website_set.filter(status__isnull=True)]))

    if websites_without_script == '':
        websites_without_script = '---'

    if website_not_reachable == '':
        website_not_reachable = '---'

    event_track(REGISTER_COMPLETED, user.id, {
        '$first_name': user.first_name,
        '$last_name': user.last_name,
        '$email': user.email,
        'companyId': user.company_id,
        'companyName': user.company.label,
        'signed_up': user.date_joined.replace(microsecond=0).isoformat(),
        'language': request.LANGUAGE_CODE,
        'websites': ', '.join(map(str, [w.domain for w in user.website_set.all()])),
        'website_without_script': websites_without_script,
        'website_not_reachable': website_not_reachable,
        'is_owner': user.is_company_owner
    })

    notify_emails(
        settings.NEW_COMPANY_NOTIFY_TO,
        _('New company have been created on') + ' ' + settings.BASE_CHAT_NAME,
        'users/email/new_company.txt',
        'users/email/new_company.html',
        {
            "first_name": user.first_name,
            "last_name": user.last_name,
            "company": user.company.label,
            "email": user.email,
            "phone": user.phone_number,
        }
    )

    if not settings.BILLING_ENABLED:
        # skip billing integration
        return

    try:
        user.own_company
    except Company.DoesNotExist:
        pass
    else:
        execute_billing('on_user_create', args=[user])


@receiver(user_logged_in)
def user_signed_in_(request, user, **kwargs):
    track_user_sign_in(user, request.LANGUAGE_CODE)


def track_user_sign_in(user, language_code=""):
    websites_without_script = ', '.join(map(str, [w.domain for w in user.website_set.filter(
        status__widget_installed__exact=False)]))
    website_not_reachable = ', '.join(map(str, [w.domain for w in user.website_set.filter(status__isnull=True)]))

    if websites_without_script == '':
        websites_without_script = '---'

    if website_not_reachable == '':
        website_not_reachable = '---'

    event_people_set(user.id, {
        '$first_name': user.first_name,
        '$last_name': user.last_name,
        '$email': user.email,
        'companyId': user.company_id,
        'companyName': user.company.label,
        'signed_up': user.date_joined.replace(microsecond=0).isoformat(),
        'language': language_code,
        'websites': ', '.join(map(str, [w.domain for w in user.website_set.all()])),
        'website_without_script': websites_without_script,
        'website_not_reachable': website_not_reachable,
        'is_owner': user.is_company_owner
    })


def cls_cache():
    try:
        cache = caches['default']
    
        if cache.__class__.__qualname__ == 'RedisCache':
            cli = cache._cache.get_client()
            keys = cli.keys('*ipwhitelist_*')
            [cli.delete(k) for k in keys]
        else:
            cache.clear()
    except InvalidCacheBackendError:
        pass



@receiver(post_save, sender=WhitelistAddr)
def whitelist_post_save(sender, instance=None, created=False, **kwargs):
    # TODO sync
    cls_cache()


@receiver(post_delete, sender=WhitelistAddr)
def whitelist_post_delete(sender, instance=None, **kwargs):
    # TODO sync
    cls_cache()

@receiver(post_delete, sender=User)
def user_post_delete(sender, instance=None, created=False, **kwargs):
    if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
        return
    print("POST_DEL", instance)
    update_data_workernode(instance.website_set.values_list('node2website__node', flat=True), 
                           'userdel', {'pk': instance.pk,})


@receiver(post_save, sender=User)
def user_post_save(sender, instance=None, created=False, **kwargs):
    if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
        return
    update_data_workernode(
        instance.website_set.values_list('node2website__node', flat=True),
        'user',
        get_user_relations(instance)
    )


@receiver(post_delete, sender=UserSiteRole)
def usersiterole_post_delete(sender, instance=None, created=False, **kwargs):
    if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
        return
    update_data_workernode(
        instance.site.node2website.filter(status=0).values_list('node', flat=True),
        'user',
        get_user_relations(instance.user)
    )


@receiver(post_save, sender=UserSiteRole)
def usersiterole_post_save(sender, instance=None, created=False, **kwargs):
    if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
        return
    update_data_workernode(
        instance.site.node2website.filter(status=0).values_list('node', flat=True),
        'user',
        get_user_relations(instance.user)
    )
