from __future__ import unicode_literals

from django.conf import settings
from django.core import mail
from django.urls import reverse
from django.contrib import auth
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string
from django.contrib.sites.models import Site
from allauth.account.models import EmailConfirmation
from rest_framework import status

from nicechat.utils.tests import TestCaseWithSites
from nicechat.utils.constants import UserRoles
from ..users.api import INVITATION_MESSAGE_SUBJECT, INVITATION_WEBSITE_ERROR, \
    DUPLICATE_USER_EMAIL_ERROR, AGENTS_LIMIT_REACHED, SITE_NOT_FOUND_ERROR_TEXT
from ..company.models import Company
from ..company.forms import AgentDataForm
from ..website.models import Website


class UsersAPITest(TestCaseWithSites):
    def setUp(self):
        self.users_per_company = 4
        super(UsersAPITest, self).setUp()

        self.users_api_url = self.api_url + 'users/'
        self.users_me_api_url = self.users_api_url + 'me/'
        self.websites_api_url = self.api_url + 'websites/'
        self.companies_api_url = self.api_url + 'companies/'
        self.registration_api_url = '/rest-auth/registration/'
        self.agent_invite_api_url = self.users_api_url + 'invite/'

    def test_non_authenticated_error(self):
        response = self.client.get(self.users_me_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_user_roles(self):
        # user does not belong to self.company_0
        self.assertFalse(self.user_1_0.is_site_agent(self.website_0_0))
        # self.website_2_1 is a site of self.company_2
        self.assertFalse(self.user_1_1.is_site_admin(self.website_2_1))
        self.assertIsNone(self.user_1_0.get_site_role(self.website_0_0))

        self.assertFalse(self.user_owner_0.is_site_admin(self.website_1_0))
        self.assertIsNone(self.user_owner_0.get_site_role(self.website_1_0))
        self.assertFalse(self.user_owner_1.is_site_admin(self.website_0_1))
        self.assertIsNone(self.user_owner_1.get_site_role(self.website_0_1))

        # self.user_1_0 does not belong to self.website_1_1
        self.assertFalse(self.user_1_0.is_site_agent(self.website_1_1))
        # self.user_1_1 does not belong to self.website_1_0
        self.assertFalse(self.user_1_1.is_site_admin(self.website_1_0))

        #
        # correct roles for corresponding sites
        #

        # self.user_1_0 is the agent for even index sites
        self.assertTrue(self.user_1_0.is_site_agent(self.website_1_0))
        self.assertTrue(self.user_1_0.is_site_agent(self.website_1_2))

        # self.user_1_1 is the admin for odd index sites
        self.assertTrue(self.user_1_1.is_site_admin(self.website_1_1))

        # # todo should company owners be admins too?
        # self.assertTrue(self.user_owner_0.is_site_admin(self.website_0_1))
        # self.assertTrue(self.user_owner_0.is_site_admin(self.website_0_2))
        # self.assertTrue(self.user_owner_1.is_site_admin(self.website_1_1))
        # self.assertTrue(self.user_owner_1.is_site_admin(self.website_1_2))

    def test_me_endpoint_correct_role(self):
        # self.user_1_0 is the agent for even index sites
        self.assertTrue(self.user_1_0.is_site_agent(self.website_1_0))
        self.assertTrue(self.user_1_0.is_site_agent(self.website_1_2))
        # self.user_1_1 is the admin for odd index sites
        self.assertTrue(self.user_1_1.is_site_admin(self.website_1_1))

        self.assertFalse(self.user_owner_1.is_site_admin(self.website_1_0))
        self.assertFalse(self.user_owner_1.is_site_admin(self.website_1_1))
        self.assertFalse(self.user_owner_1.is_site_admin(self.website_1_2))
        self.assertEqual(self.user_owner_1.company, self.company1)
        self.assertTrue(self.user_owner_1.is_company_owner)

        self.login_user(self.user_owner_1)
        response = self.client.get(self.users_me_api_url + '?site_id={}'.format(self.website_1_1.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('role', response.data.keys())
        self.assertEqual(response.data['role'], None)

        self.login_user(self.user_1_0)
        response = self.client.get(self.users_me_api_url + '?site_id={}'.format(self.website_1_0.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('role', response.data.keys())
        self.assertEqual(response.data['role'], UserRoles.SITE_AGENT)

        self.login_user(self.user_1_1)
        response = self.client.get(self.users_me_api_url + '?site_id={}'.format(self.website_1_1.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('role', response.data.keys())
        self.assertEqual(response.data['role'], UserRoles.SITE_ADMIN)

        # wrong site id parameter (self.user_1_1 does not belong to self.website_1_0)
        response = self.client.get(self.users_me_api_url + '?site_id={}'.format(self.website_1_0.pk), format='json')
        # self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['role'], None)

        # no site id parameter at all
        response = self.client.get(self.users_me_api_url, format='json')
        # self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['role'], None)

    def test_registration(self):
        self.assertEqual(Company.objects.count(), self.num_of_companies)
        self.assertEqual(Website.objects.count(), self.num_of_companies * self.site_per_company)

        registration_data_invalid = {
            'email': 'some.test.mail.com',
            'password1': 'super_pass',
            'website_url': 'mysite',
            # 'last_name': ''
        }

        registration_data_pass_mismatch = {
            'email': 'some.test@mail.com',
            'password1': 'super_pass',
            'password2': 'super_pass_different',
            # 'website_name': 'New Website',
            'website_url': 'www.mysite.com',
            # 'first_name': 'Some Name',
            # 'last_name': 'Some Last Name'
        }

        registration_data = {
            'email': 'some.name@email.com',
            'password1': 'super_pass',
            'password2': 'super_pass',
            # 'website_name': 'New Website',
            'website_url': 'www.mysite.com',
            # 'first_name': 'Some Name',
            # 'last_name': 'Some Last Name'
        }

        registration_data_duplicate_email = {
            'email': 'some.name@email.com',
            'password1': 'super_pass',
            'password2': 'super_pass',
            # 'website_name': 'New Website New Title',
            'website_url': 'www.new-mysite.com',
            # 'first_name': 'Some Name',
            # 'last_name': 'Some Last Name'
        }

        registration_data_duplicate_website = {
            'email': 'some.new.name@email.com',
            'password1': 'super_pass_1',
            'password2': 'super_pass_1',
            # 'website_name': 'New Website',
            'website_url': 'www.mysite.com',
            # 'first_name': 'Some Name',
            # 'last_name': 'Some Last Name'
        }

        # invalid data
        response = self.client.post(self.registration_api_url, registration_data_invalid, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        # self.assertItemsEqual(response.data.keys(), ['email', 'password2', 'website_name', 'website_url',
        #                                             'first_name', 'last_name'])
        self.assertItemsEqual(response.data.keys(), ['email', 'password2', 'website_url'])

        # password mismatch
        response = self.client.post(self.registration_api_url, registration_data_pass_mismatch, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['non_field_errors'])
        self.assertEqual(response.data['non_field_errors'], [u"The two password fields didn't match."])

        # successful registration
        response = self.client.post(self.registration_api_url, registration_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertTrue(get_user_model().objects.filter(email=registration_data['email']).exists())

        # new company and site were created
        self.assertEqual(Company.objects.count(), self.num_of_companies + 1)
        self.assertEqual(Website.objects.count(), self.num_of_companies * self.site_per_company + 1)

        new_user = get_user_model().objects.get(email=registration_data['email'])
        # self.assertEquals(new_user.first_name, registration_data['first_name'])
        # self.assertEquals(new_user.last_name, registration_data['last_name'])
        # self.assertEquals(new_user.company.label, registration_data['website_name'])
        self.assertEquals(new_user.first_name, registration_data['website_url'])
        self.assertEquals(new_user.last_name, 'admin')
        self.assertEquals(new_user.company.label, registration_data['website_url'])
        self.assertEquals(new_user.company.owner, new_user)
        self.assertEquals(new_user.company.websites.count(), 1)
        self.assertTrue(new_user.company.billing_plan.plan.is_trial)

        new_website = new_user.company.websites.first()
        # self.assertEquals(new_website.title, registration_data['website_name'])
        self.assertEquals(new_website.title, registration_data['website_url'])
        self.assertEquals(new_website.domain, registration_data['website_url'])
        self.assertEquals(new_website.users.all().count(), 1)
        self.assertEquals(new_website.users.first().email, registration_data['email'])

        # user is able to login using the password provided during registration
        login_result = self.client.login(username=registration_data['email'],
                                         password=registration_data['password1'])
        self.assertTrue(login_result)

        if settings.BILLING_ENABLED:
            # Billing customer account was created
            self.assertIsNotNone(Company.objects.get(pk=new_user.company.pk).billing_id)
            self.assertIsNotNone(Company.objects.get(pk=new_user.company.pk).bill)

        # duplicate email error
        response = self.client.post(self.registration_api_url, registration_data_duplicate_email, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['email'])

        # duplicate website title and domain errors
        response = self.client.post(self.registration_api_url, registration_data_duplicate_website, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        # self.assertItemsEqual(response.data.keys(), ['website_name', 'website_url'])
        self.assertItemsEqual(response.data.keys(), ['website_url'])

    def test_agent_invite(self):
        self.assertEqual(self.company2.users.count(), 5)  # owner + 2 agents + 2 admins
        self.assertEqual(self.company2.billing_plan.agents_count, 7)
        self.assertEqual(len(mail.outbox), 0)

        new_agent_email = 'some.new.email@mail.com'
        another_new_agent_email = 'another.new.email@mail.com'
        website1 = self.company2.websites.first()
        website2 = self.company2.websites.last()

        valid_invitation_data = {
            'agents': [
                {'email': new_agent_email, 'role': UserRoles.SITE_AGENT, 'site_id': website1.pk},
                {'email': another_new_agent_email, 'role': UserRoles.SITE_ADMIN, 'site_id': website2.pk},
            ]
        }

        invalid_site_id = '77777777777'
        invitation_data_invalid_site = {
            'agents': [
                {'email': 'a@b.com', 'role': UserRoles.SITE_AGENT, 'site_id': invalid_site_id},
            ]
        }

        invalid_email = '77777777777'
        invitation_data_invalid_email = {
            'agents': [
                {'email': invalid_email, 'role': UserRoles.SITE_AGENT, 'site_id': website1.pk},
            ]
        }

        invalid_role = 'asdasd'
        invitation_data_invalid_role = {
            'agents': [
                {'email': 'a@b.com', 'role': invalid_role, 'site_id': website1.pk},
            ]
        }
        invitation_data_no_role = {
            'agents': [
                {'email': 'a@b.com', 'site_id': website1.pk},
            ]
        }

        # unauthenticated access is forbidden
        response = self.client.post(self.agent_invite_api_url, valid_invitation_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # only company owners can invite agents
        self.login_user(self.user_2_1)
        response = self.client.post(self.agent_invite_api_url, valid_invitation_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.login_user(self.user_owner_2)

        # invalid website parameter error
        response = self.client.post(self.agent_invite_api_url, invitation_data_invalid_site, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 1)
        self.assertFalse(response.data['result'][0]['success'])
        self.assertEqual(response.data['result'][0]['errors'], {
            'site_id': [INVITATION_WEBSITE_ERROR.format(invalid_site_id)]
        })

        # email address validation
        response = self.client.post(self.agent_invite_api_url, invitation_data_invalid_email, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 1)
        self.assertFalse(response.data['result'][0]['success'])
        self.assertEqual(response.data['result'][0]['errors'], {
            'email': ['Enter a valid email address.']
        })

        # role field validation
        response = self.client.post(self.agent_invite_api_url, invitation_data_no_role, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 1)
        self.assertFalse(response.data['result'][0]['success'])
        self.assertEqual(response.data['result'][0]['errors'], {
            'role': ['This field may not be null.']
        })
        response = self.client.post(self.agent_invite_api_url, invitation_data_invalid_role, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 1)
        self.assertFalse(response.data['result'][0]['success'])
        self.assertEqual(response.data['result'][0]['errors'], {
            'role': ['"{}" is not a valid choice.'.format(invalid_role)]
        })

        # successful invitation
        response = self.client.post(self.agent_invite_api_url, valid_invitation_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 2)
        self.assertTrue(response.data['result'][0]['success'])
        self.assertTrue(response.data['result'][1]['success'])

        self.assertEqual(self.company2.users.count(), 7)  # owner + 6 agents now

        # Test that one message has been sent.
        self.assertEqual(len(mail.outbox), 2)

        # Verify that the subject of outbox messages are correct.
        self.assertEqual(mail.outbox[0].subject, INVITATION_MESSAGE_SUBJECT)
        self.assertEqual(mail.outbox[1].subject, INVITATION_MESSAGE_SUBJECT)
        self.assertEqual(mail.outbox[0].to, [new_agent_email])
        self.assertEqual(mail.outbox[1].to, [another_new_agent_email])

        new_agent = self.company2.users.get(email=new_agent_email)
        self.assertFalse(new_agent.is_active)
        self.assertEquals(new_agent.website_set.count(), 1)
        self.assertEquals(new_agent.website_set.first(), website1)
        self.assertEquals(new_agent.get_site_role(website1), UserRoles.SITE_AGENT)

        new_agent_2 = self.company2.users.get(email=another_new_agent_email)
        self.assertFalse(new_agent_2.is_active)
        self.assertEquals(new_agent_2.website_set.count(), 1)
        self.assertEquals(new_agent_2.website_set.first(), website2)
        self.assertEquals(new_agent_2.get_site_role(website2), UserRoles.SITE_ADMIN)

        # the key and url for specific agent invitation
        confirmation_key = EmailConfirmation.objects.get(email_address__email=new_agent_email).key
        confirmation_link = reverse('users:agent_confirm_email', kwargs={'key': confirmation_key})
        # invitation email contains correct data
        self.assertEqual(
            mail.outbox[0].body,
            render_to_string('users/email/invite.txt',
                             dict(user=new_agent,
                                  site=Site.objects.get_current(),
                                  confirmation_link=confirmation_link))
        )

        # billing plan agents limit error
        response = self.client.post(self.agent_invite_api_url, valid_invitation_data, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 2)
        self.assertFalse(response.data['result'][0]['success'])
        self.assertFalse(response.data['result'][1]['success'])
        self.assertEqual(response.data['result'][0]['errors'], {
            '__all__': [AGENTS_LIMIT_REACHED]
        })

        # duplicate email error

        self.company2.billing_plan.agents_count = 8  # allow additional agent seat first
        self.company2.billing_plan.save()

        response = self.client.post(self.agent_invite_api_url, valid_invitation_data, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEquals(len(response.data['result']), 2)
        self.assertFalse(response.data['result'][0]['success'])
        self.assertFalse(response.data['result'][1]['success'])
        self.assertEqual(response.data['result'][0]['errors'], {
            'email': [DUPLICATE_USER_EMAIL_ERROR]
        })

    def test_agent_invite_accepting(self):
        # Invitation

        new_agent_email = 'some.new.email@mail.com'
        website = self.company2.websites.first()
        valid_invitation_data = {
            'agents': [
                {'email': new_agent_email, 'role': UserRoles.SITE_ADMIN, 'site_id': website.pk},
            ]
        }
        invalid_agent_data_no_pass = {'first_name': 'SomeName', 'last_name': 'SomeSecondName'}
        invalid_agent_data_pass_mismatch = {'first_name': 'SomeName', 'last_name': 'SomeSecondName',
                                            'password1': '123456789q', 'password2': '987654321q'}
        valid_agent_data = {'first_name': 'SomeName', 'last_name': 'SomeSecondName',
                            'password1': 'somePassword', 'password2': 'somePassword'}

        self.assertEqual(self.company2.users.count(), 5)  # owner + 2 agents + 2 admins
        self.assertEqual(len(mail.outbox), 0)

        self.login_user(self.user_owner_2)

        # successful invitation sending
        response = self.client.post(self.agent_invite_api_url, valid_invitation_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(self.company2.users.count(), 6)  # owner + 3 agents + 2 admins now

        # Test that one message has been sent.
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to, [new_agent_email])

        # the key and url for specific agent invitation
        confirmation_key = EmailConfirmation.objects.get(email_address__email=new_agent_email).key
        confirmation_link = reverse('users:agent_confirm_email', kwargs={'key': confirmation_key})

        # ERRORS IN FORM DATA

        new_agent = get_user_model().objects.get(email=new_agent_email)

        # no data
        form = AgentDataForm({},
                             instance=new_agent,
                             email_required=False)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'first_name': [u'This field is required.'],
            'last_name': [u'This field is required.'],
            'password1': [u'This field is required.'],
            'password2': [u'This field is required.'],
        })

        # no password provided
        form = AgentDataForm(invalid_agent_data_no_pass,
                             instance=new_agent,
                             email_required=False)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {
            'password1': [u'This field is required.'],
            'password2': [u'This field is required.'],
        })

        # password repeat mismatch
        form = AgentDataForm(invalid_agent_data_pass_mismatch,
                             instance=new_agent,
                             email_required=False)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors, {'password2': [u'You must type the same password each time.']})

        # company owner is still logged in
        self.assertEquals(self.client.request().context['user'], self.user_owner_2)

        # check that view returns errors
        response = self.client.post(confirmation_link, invalid_agent_data_no_pass, format='json')
        # no redirect means that form didn't validate
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(response.context_data['form'].errors)

        # previously logged in user owner should be logged out
        self.assertEquals(response.context['user'], AnonymousUser())
        user = auth.get_user(self.client)
        self.assertFalse(user.is_authenticated())

        # successful invitation accepting
        response = self.client.post(confirmation_link, valid_agent_data)
        # redirect to success url was performed
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)

        # no new users created (just in case)
        self.assertEqual(self.company2.users.count(), 6)

        new_agent = self.company2.users.get(email=new_agent_email)

        # agent became active and has first and last names are set properly
        self.assertTrue(new_agent.is_active)
        self.assertEqual(new_agent.first_name, valid_agent_data['first_name'])
        self.assertEquals(new_agent.last_name, valid_agent_data['last_name'])
        self.assertEquals(new_agent.website_set.count(), 1)
        self.assertEquals(new_agent.website_set.first(), website)
        self.assertEquals(new_agent.get_site_role(website), UserRoles.SITE_ADMIN)

        # new agent gets logged in after successful form submission
        user = auth.get_user(self.client)
        self.assertTrue(user.is_authenticated())
        self.assertEqual(user, new_agent)

        # new agent can now sign in successfully using provided password

        self.client.logout()
        self.assertFalse(auth.get_user(self.client).is_authenticated())

        login_result = self.client.login(username=new_agent_email, password=valid_agent_data['password1'])
        self.assertTrue(login_result)

        self.assertEquals(self.client.request().context['user'], new_agent)

    def test_users_list(self):
        self.assertTrue(self.user_owner_0.is_company_owner)
        self.assertEqual(self.user_owner_0.company, self.company0)
        self.assertEqual(self.user_owner_0.own_company, self.company0)

        self.assertEqual(self.user_0_0.get_site_role(self.website_0_0), UserRoles.SITE_AGENT)
        self.assertTrue(self.user_0_0.is_site_agent(self.website_0_0))
        self.assertFalse(self.user_0_0.is_site_admin(self.website_0_0))
        self.assertEqual(self.user_0_0.company, self.company0)

        self.assertEqual(self.user_0_1.get_site_role(self.website_0_1), UserRoles.SITE_ADMIN)
        self.assertFalse(self.user_0_1.is_site_agent(self.website_0_1))
        self.assertTrue(self.user_0_1.is_site_admin(self.website_0_1))
        self.assertEqual(self.user_0_1.company, self.company0)

        self.login_user(self.user_owner_0)
        response = self.client.get(self.users_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 5)  # owner + 4 users
        for user in response.data['results']:
            self.assertIn(user['id'], [self.user_owner_0.pk, self.user_0_0.pk, self.user_0_1.pk,
                                       self.user_0_2.pk, self.user_0_3.pk])
        # incorrect site id specified
        # todo - current permissions class doesn't check if user is owner of the site's company
        # self.assertNotIn(self.user_owner_0, self.website_1_0.users.all())
        # response = self.client.get(self.users_api_url + '?site_id={}'.format(self.website_1_0.pk), format='json')
        # self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.login_user(self.user_0_1)  # log in the company admin
        # same response for company admin
        response = self.client.get(self.users_api_url + '?site_id={}'.format(self.website_0_1.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 5)  # owner + 4 users
        for user in response.data['results']:
            self.assertIn(user['id'], [self.user_owner_0.pk, self.user_0_0.pk, self.user_0_1.pk,
                                       self.user_0_2.pk, self.user_0_3.pk])
        # incorrect site id specified
        response = self.client.get(self.users_api_url + '?site_id={}'.format(self.website_1_0.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.login_user(self.user_0_0)  # log in the company employee
        # accessing users list is forbidden for non-admin and non-company owner
        response = self.client.get(self.users_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        # even if they specify the correct site id
        response = self.client.get(self.users_api_url + '?site_id={}'.format(self.website_0_0.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_change_role_endpoint(self):
        change_role_api_url = self.users_api_url + '{user_pk}/change-role/'

        self.assertTrue(self.user_owner_0.is_company_owner)
        self.assertEqual(self.user_owner_0.company, self.company0)
        self.assertEqual(self.user_owner_0.own_company, self.company0)

        self.assertIn(self.user_0_0, self.website_0_0.users.all())
        self.assertEqual(self.user_0_0.get_site_role(self.website_0_0), UserRoles.SITE_AGENT)
        self.assertTrue(self.user_0_0.is_site_agent(self.website_0_0))
        self.assertFalse(self.user_0_0.is_site_admin(self.website_0_0))
        self.assertEqual(self.user_0_0.company, self.company0)

        self.assertEqual(self.user_0_1.get_site_role(self.website_0_1), UserRoles.SITE_ADMIN)
        self.assertFalse(self.user_0_1.is_site_agent(self.website_0_1))
        self.assertTrue(self.user_0_1.is_site_admin(self.website_0_1))
        self.assertEqual(self.user_0_1.company, self.company0)

        self.login_user(self.user_0_0)  # log in the company employee
        # accessing change role endpoint is forbidden for non-admin and non-company owner
        response = self.client.post(change_role_api_url.format(user_pk=self.user_0_1.pk), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        #
        # processing for company owner
        #

        # Errors

        self.login_user(self.user_owner_0)
        # no data in request body
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_0_0.pk) + '?site_id={}'.format(self.website_0_0.pk),
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        # error in response data
        self.assertEqual(response.data.keys(), ['role'])

        # invalid role value
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_0_0.pk) + '?site_id={}'.format(self.website_0_0.pk),
            data={'role': '123'},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data.keys(), ['role'])

        # # todo invalid site id value. Currently breaks `nicechat.middleware.CheckOriginMiddleware#process_response`
        # response = self.client.post(
        #     change_role_api_url.format(user_pk=self.user_0_0.pk) + '?site_id=asdsadfasdf',
        #     data={'role': UserRoles.SITE_ADMIN},
        #     format='json'
        # )
        # self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        # self.assertEqual(response.data.keys(), ['site_id'])

        # another company's site id value
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_0_0.pk) + '?site_id={}'.format(self.website_1_1.pk),
            data={'role': UserRoles.SITE_ADMIN},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data, {u'non_field_errors': [SITE_NOT_FOUND_ERROR_TEXT]})

        # the user can't change the role for himself
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_owner_0.pk) + '?site_id={}'.format(self.website_0_0.pk),
            data={'role': UserRoles.SITE_AGENT},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data, {u'user': ['You cannot change the role for yourself']})

        # check that nothing was changed in database after invalid calls
        self.assertEqual(self.user_0_0.get_site_role(self.website_0_0), UserRoles.SITE_AGENT)

        #
        # correct processing
        #

        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_0_0.pk) + '?site_id={}'.format(self.website_0_0.pk),
            data={'role': UserRoles.SITE_ADMIN},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # check the new user account type value in response
        self.assertEqual(response.data['role'], UserRoles.SITE_ADMIN)

        # check the updated value in database
        self.assertEqual(self.user_0_0.get_site_role(self.website_0_0), UserRoles.SITE_ADMIN)

        #
        # Make sure that admin users can change user roles too
        #

        self.login_user(self.user_0_1)  # log in the admin of self.website_0_1

        self.assertIn(self.user_0_3, self.website_0_1.users.all())
        self.assertEqual(self.user_0_3.get_site_role(self.website_0_1), UserRoles.SITE_ADMIN)

        # correct response for site admin
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_0_3.pk) + '?site_id={}'.format(self.website_0_1.pk),
            data={'role': UserRoles.SITE_AGENT},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['role'], UserRoles.SITE_AGENT)

        # check the updated value in database
        self.assertEqual(self.user_0_3.get_site_role(self.website_0_1), UserRoles.SITE_AGENT)

        # access to other sites is forbidden
        self.assertNotIn(self.user_0_1, self.website_0_0.users.all())
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_0_0.pk) + '?site_id={}'.format(self.website_0_0.pk),
            data={'role': UserRoles.SITE_AGENT},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        # self.user_0_0 is still an agent
        self.assertEqual(self.user_0_0.get_site_role(self.website_0_0), UserRoles.SITE_ADMIN)

        # the user can't change the role for company owner
        response = self.client.post(
            change_role_api_url.format(user_pk=self.user_owner_0.pk) + '?site_id={}'.format(self.website_0_1.pk),
            data={'role': UserRoles.SITE_AGENT},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data, {u'user': ['You cannot change the role for company owner']})
