import re
import ipaddress
from apps.users.models import WhitelistAddr
from django.core.cache import caches, InvalidCacheBackendError


from django.db.models import CharField
from django.db.models import Lookup

@CharField.register_lookup
class ContainedIn(Lookup):
    lookup_name = 'iscontainedin'

    def as_postgresql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = lhs_params + rhs_params
        return 'CAST(%s as cidr) >> inet %s' % (lhs, rhs), params


def is_ip_whitelisted(headers, addrs=None):
    addr1 = headers.get('HTTP_CF_CONNECTING_IP', None)
    addr2 = headers.get('HTTP_X_REAL_IP', None)
    addr3 = headers.get('HTTP_X_FORWARDED_FOR', None)
    addr4 = headers.get('REMOTE_ADDR', None)

    if isinstance(addrs, list):
        if not addrs:
            return True
        for addr in addrs:
            if match_addr(addr, addr1) or match_addr(addr, addr2) or match_addr(addr, addr3) or match_addr(addr, addr4):
                return True
        return False

    return _check_addr(addr1) or _check_addr(addr2) or _check_addr(addr3) or _check_addr(addr4)


def _check_addr(addr):
    if not addr:
        return False

    try:
        cache = caches['default']
        cache_key = f'ipwhitelist_{addr}'
        res = cache.get(cache_key)

        if res is not None:
            return res

        if WhitelistAddr.objects.count() == 0:
            cache.set(cache_key, True)
            return True
        
        if WhitelistAddr.objects.filter(mask=32, addr = addr).count() > 0:
            cache.set(cache_key, True)
            return True
        
        if WhitelistAddr.objects.exclude(mask=32).filter(req__iscontainedin = addr).count() > 0:
            cache.set(cache_key, True)
            return True

        cache.set(cache_key, False)
        return False
    except InvalidCacheBackendError:
        return True

R_IP4 = re.compile('^\d+\.\d+\.\d+\.\d+$')
R_IP4R = re.compile('^\d+\.\d+\.\d+\.\d+\/\d+$')
R_IP4MR = re.compile('^(\d+|\d+-\d+)\.(\d+|\d+-\d+)\.(\d+|\d+-\d+)\.(\d+|\d+-\d+)$')


def validate_addr(addr):
    try:
        if R_IP4.match(addr):
            ipaddress.ip_address(addr)
            return True
        if R_IP4R.match(addr):
            ipaddress.ip_network(addr)
            return True
        if R_IP4MR.match(addr):
            if _parse_range(addr):
                return True
    except ValueError as e:
        # print(e)
        return False


def match_addr(addr, tgt):
    if not tgt:
        return False
    try:
        if R_IP4.match(tgt):
            _tgt = ipaddress.ip_address(tgt)
        if R_IP4.match(addr):            
            return ipaddress.ip_address(addr) == _tgt
        if R_IP4R.match(addr):
            return _tgt in ipaddress.ip_network(addr)
        if R_IP4MR.match(addr):
            return _match_range(_parse_range(addr), tgt)
    except ValueError as e:
        # print(e)
        return False
    
   

def _parse_range(addr):
    opts = addr.split('.')
    if len(opts) != 4:
        return None
    ret = []
    for op in opts:
        if '-' in op:
            op = op.split('-')
            if len(op) == 2:
                if op[0].isdigit() and op[1].isdigit():
                    op1 = int(op[0])
                    op2 = int(op[1])
                    if op1 >= 0 and op1 <= 255 and op1 >= 0 and op1 <= 255:
                        ret.append((op1, op2))
        else:
            if op.isdigit():
                op = int(op)
                if op >= 0 and op <= 255:
                    ret.append(op)
    if len(ret) != 4:
        return None
    # print(ret)
    return ret

def _match_range(rng, tgt):
    if not rng and len(rng) != 4:
        return False
    tgt = tgt.split('.')
    _tgt = []
    for t in tgt:
        if t.isdigit():
            t = int(t)
            if t >= 0 and t <= 255:
                _tgt.append(t)
    if len(_tgt) != 4:
        return False

    for j in range(0, 4):
        if isinstance(rng[j], tuple):
            if rng[j][0] > _tgt[j] or _tgt[j] > rng[j][1]:
                return False
        else:
            if rng[j] != _tgt[j]:
                return False
    return True
