import hashlib
import json
import time
import threading
from importlib import import_module
from django.conf import settings
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.contrib.auth import logout, login, authenticate, get_user_model
from django.views.generic.edit import FormView
from django.views.generic import View
from allauth.account.models import EmailConfirmation
from django.views.decorators.csrf import csrf_exempt

from allauth.account.views import ConfirmEmailView as ConfirmEmailViewBase
from dj_rest_auth.serializers import PasswordResetSerializer
from dj_rest_auth.views import PasswordResetView as PasswordResetViewBase, \
    PasswordResetConfirmView as PasswordResetConfirmViewBase
from dj_rest_auth.serializers import PasswordResetConfirmSerializer as PasswordResetConfirmSerializerBase
from rest_framework import serializers, response, generics, exceptions
from rest_framework_simplejwt.views import TokenObtainPairView
from rest_framework_simplejwt.serializers import TokenObtainPairSerializer
from dj_rest_auth.registration.views import RegisterView
from dj_rest_auth.utils import jwt_encode
from .forms import AuthPasswordResetForm
from .models import User
from .serializers import CreateAgentSerializer, WebsiteRegisterSerializer
from .utils import is_ip_whitelisted, validate_addr

from nicechat.analytics import event_track, AGENT_FINISHED_REGISTRATION


class AgentEmailConfirmationView(generics.CreateAPIView):
    serializer_class = CreateAgentSerializer

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        user = self.confirmation_obj.email_address.user
        user.first_name = serializer.data['first_name']
        user.last_name = serializer.data['last_name']
        user.is_active = True
        user.set_password(serializer.data['password1'])
        user.save()

        event_track(AGENT_FINISHED_REGISTRATION, user.id, {
            '$first_name': user.first_name,
            '$last_name': user.last_name,
            '$email': user.email,
            'company': user.company_id,
        })

        access_token, refresh_token = jwt_encode(user)

        return response.Response({
            "access": f'{access_token}',
            "refresh": f'{refresh_token}',
            "user": user.pk
        })

    def dispatch(self, request, *args, **kwargs):
        key = self.kwargs['key']
        try:
            self.confirmation_obj = EmailConfirmation.objects.all_valid().get(key=key.lower())
        except EmailConfirmation.DoesNotExist:
            raise Http404()

        # log out any previously signed in user
        logout(request)

        return super(AgentEmailConfirmationView, self).dispatch(request, *args, **kwargs)


class ServerServerQuery(View):
    http_method_names = ['post', 'options']

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(ServerServerQuery, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):

        session_key = request.POST.get('session')
        hsh = request.POST.get('hsh')

        engine = import_module(settings.SESSION_ENGINE)
        session_store = engine.SessionStore
        session_data = session_store(session_key)

        j_hsh = hashlib.sha1("%s-%s-%s" % (session_data.get('RND'), session_data.get('SIGN_USER_ID'),
                                           session_key)).hexdigest()

        if j_hsh == hsh:
            user = get_user_model().objects.filter(pk=session_data.get('SIGN_USER_ID')).first()

            icon = user.avatar_thumbs().get('thumb', None)
            icon_url = icon.url if icon and icon.exists() else None

            ret = {
                'state': 'ok',
                'id': user.pk,
                'email': user.email,
                'company': user.company_id,
                'fullname': user.get_full_name,
                'avatar': icon_url,
            }
        else:
            ret = {'state': 'err'}
            time.sleep(1)

        return HttpResponse(json.dumps(ret), content_type="application/json")


class AuthPasswordResetSerializer(PasswordResetSerializer):
    password_reset_form_class = AuthPasswordResetForm


class PasswordResetView(PasswordResetViewBase):
    authentication_classes = []
    serializer_class = AuthPasswordResetSerializer


class JSONWebTokenSerializer(TokenObtainPairSerializer):

    def __init__(self, *args, **kwargs):
        super(JSONWebTokenSerializer, self).__init__(*args, **kwargs)
        self.fields['otp'] = serializers.CharField(write_only=True, required=False)

    def validate(self, attrs):
        username = attrs.get(self.username_field)
        user = User.objects.filter(**{User.USERNAME_FIELD:username}).first()
        try:
            tfa = json.loads(user.tfa_opts) if user else {}
        except ValueError:
            tfa = {}
        is_tfa = True if settings.TFA_ENABLE and tfa.get('on') == True else False
        request = self.context.get('request', None)
        if is_tfa:
            otp = attrs.get('otp', None)
            otp_factor = getattr(User, 'otp_factor')
            if not otp:
                opts = otp_factor.prepare(user, request)
                err = LookupError()
                if opts and len(opts) > 1:
                    setattr(err, 'method', opts[0])
                    setattr(err, 'valid_sec', opts[1])
                raise err
            else:
                if not otp_factor.validate(otp, user, request):
                    err = LookupError()
                    setattr(err, 'msg', 'Wrong OTP')
                    raise err
                else:
                    pass # OTP correct

        return super(JSONWebTokenSerializer, self).validate(attrs)

class TokenObtainPairSerializerReq(TokenObtainPairSerializer):
    def validate(self, attrs):
        data = super().validate(attrs)
        request = self.context.get('request')
        if request and self.user:
            try:
                uf = self.user.userfiltering
                if uf.enable and not is_ip_whitelisted(request.META, uf.addrs):
                    raise exceptions.AuthenticationFailed('You cannot login here')
            except User.userfiltering.RelatedObjectDoesNotExist:
                pass
        return data


class GetJWTToken(TokenObtainPairView):
    serializer_class = TokenObtainPairSerializerReq
    def post(self, request, *args, **kwargs):

        try:
            return super(GetJWTToken, self).post(request, *args, **kwargs)
        except exceptions.AuthenticationFailed as erra:
            raise exceptions.ValidationError(erra.detail)
        except LookupError as err:
            info = {
                "detail": getattr(err, 'msg', "Otp required"),
                "method": getattr(err, 'method', None),
                "valid_sec": getattr(err, 'valid_sec', None)
            }
            return response.Response(info, status=202, content_type="application/json")


class NcRegisterView(RegisterView):
    authentication_classes = []
    serializer_class = WebsiteRegisterSerializer

    def get_response_data(self, user):
        data = super(NcRegisterView, self).get_response_data(user)

        data['billing_id'] = user.own_company.billing_id if user.own_company else None
        data['user'] = user.pk
        return data


class ConfirmEmailView(ConfirmEmailViewBase):
    def get_redirect_url(self):
        return settings.DOMAIN_ACC

    def get(self, *args, **kwargs):
        try:
            self.post(*args, **kwargs)
        except Http404:
            self.object = None
        return HttpResponseRedirect('https://%s/confirmed' % settings.DOMAIN_ACC)


class BlankView(View):
    def get(self, request, *args, **kwargs):
        return HttpResponse("chat.api")


class PasswordResetConfirmSerializer(PasswordResetConfirmSerializerBase):
    def __init__(self, *args, **kwargs):
        super(PasswordResetConfirmSerializer, self).__init__(*args, **kwargs)
        self.fields['otp'] = serializers.CharField(write_only=True, required=False)

    def custom_validation(self, attrs):
        try:
            tfa = json.loads(self.user.tfa_opts) if self.user else {}
        except ValueError:
            tfa = {}
        is_tfa = True if settings.TFA_ENABLE and tfa.get('on') == True else False
        request = self.context.get('request', None)

        if is_tfa:
            otp = attrs.get('otp', None)
            otp_factor = getattr(User, 'otp_factor')
            if not otp:
                opts = otp_factor.prepare(self.user, request, exclude=['EMAIL'])
                if not opts == False:
                    err = LookupError()
                    if opts and len(opts) > 1:
                        setattr(err, 'method', opts[0])
                        setattr(err, 'valid_sec', opts[1])
                    raise err
            else:
                if not otp_factor.validate(otp, self.user, request):
                    err = LookupError()
                    setattr(err, 'msg', 'Wrong OTP')
                    raise err
                else:
                    pass  # OTP correct


class PasswordResetConfirmView(PasswordResetConfirmViewBase):
    authentication_classes = []
    serializer_class = PasswordResetConfirmSerializer

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(PasswordResetConfirmView, self).dispatch(request, *args, **kwargs)

    def post(self, request):

        try:
            return super(PasswordResetConfirmView, self).post(request)
        except LookupError as err:
            info = {
                "detail": getattr(err, 'msg', "Otp required"),
                "method": getattr(err, 'method', None),
                "valid_sec": getattr(err, 'valid_sec', None)
            }
            return response.Response(info, status=202, content_type="application/json")
