from rest_framework import viewsets, mixins, pagination

from apps.visitors.models import Statistics, Order, Browser, OperatingSystem
from apps.website.api import FilterByWebsiteMixin
from django.db.models import Q
from django.http import Http404
from django.core.exceptions import FieldDoesNotExist
from .serializers import VisitorSerializer, StatisticsSerializer, VisitorsCountSerializer, OrdersSerializer, \
    BrowsersSerializer, OperationSystemsSerializer, BatamStatisticsSerializer
from ..chat.models import Visitor, BatamStatistics


class VisitorViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    serializer_class = VisitorSerializer
    pagination_class = None
    queryset = Visitor.objects.all()


class StatisticsViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    serializer_class = StatisticsSerializer
    pagination_class = None
    queryset = Statistics.objects.all().order_by('-ts')

    def get_serializer_class(self):
        if self.request.query_params.get('type') == 'short':
            return VisitorsCountSerializer
        return super(StatisticsViewSet, self).get_serializer_class()

    def get_queryset(self):
        qs = super(StatisticsViewSet, self).get_queryset()
        period = self.request.query_params.get('period')
        number = self.request.query_params.get('number')
        ts_from = self.request.query_params.get('from')
        ts_to = self.request.query_params.get('to')
        if ts_from:
            qs = qs.filter(ts__gte=ts_from)
        if ts_to:
            qs = qs.filter(ts__lte=ts_to)
        if period:
            qs = qs.filter(period=period)
        if number:
            qs = qs[0:int(number)]
        return qs


class BatamStatisticsPagination(pagination.LimitOffsetPagination):
    page_size=50

class BatamStatisticsViewSet(viewsets.ModelViewSet):
    serializer_class = BatamStatisticsSerializer
    pagination_class = BatamStatisticsPagination
    queryset = BatamStatistics.objects.all().order_by('-pk')


    def get_queryset(self):
        qs = super(BatamStatisticsViewSet, self).get_queryset()
        if not self.request.current_site:
            raise Http404()
        qs = qs.filter(website=self.request.current_site.pk)

        filter = self.request.query_params.get('filter', None)
        order = self.request.query_params.get('order', None)
        order_dir = self.request.query_params.get('order_dir', 'asc')

        if order:
            try:
                BatamStatistics._meta.get_field(order)
                qs = qs.order_by(f"-{order}" if order_dir == 'desc' else order)
            except FieldDoesNotExist:
                pass

        if filter:
            qs = qs.filter(
                Q(name__icontains=filter) |
                Q(email__icontains=filter) |
                Q(latest_page_url__icontains=filter) |
                Q(latest_page_title__icontains=filter) |
                Q(sys_browser__icontains=filter) |
                Q(sys_os__icontains=filter) |
                Q(sys_device__icontains=filter) |
                Q(geo_country__icontains=filter) |
                Q(geo_city__icontains=filter) |
                Q(page_ref__icontains=filter)
            )
        return qs


class OrdersViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    serializer_class = OrdersSerializer
    queryset = Order.objects.all()




class BrowsersViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = BrowsersSerializer
    queryset = Browser.objects.all()


class OperationSystemsViewSet(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = OperationSystemsSerializer
    queryset = OperatingSystem.objects.all()
