from django.db import models
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.utils.functional import cached_property

from apps.website.models import Website, Product
from apps.chat.models import ArchivedChat, Idempotency, IdempotencyPeer
from nicechat.utils.helpers import nice_address

from nicechat.utils.client_token import nc_gen_rnd_token
MOBILE = 'm'
DESKTOP = 'd'
PLATFORM_TYPE = (
    (MOBILE, 'mobile'),
    (DESKTOP, 'desktop'),
)


class Browser(models.Model):
    name = models.CharField(max_length=25)
    type = models.CharField(max_length=1, default=DESKTOP, choices=PLATFORM_TYPE)
    icon = models.ImageField(blank=True, null=True)

    def __str__(self):
        return ': '.join([self.type, self.name])


class OperatingSystem(models.Model):
    name = models.CharField(max_length=25)
    type = models.CharField(max_length=1, default=DESKTOP, choices=PLATFORM_TYPE)
    icon = models.ImageField(blank=True, null=True)

    def __str__(self):
        return ': '.join([self.type, self.name])


class Statistics(models.Model):
    QUARTER = 'q'
    HOUR = 'h'
    DAY = 'd'
    WEEK = 'w'
    MONTH = 'm'
    PERIODS = (
        (QUARTER, '15 minutes'),
        (HOUR, 'hour'),
        (DAY, 'day'),
        (WEEK, 'week'),
        (MONTH, 'month'),
    )
    website = models.ForeignKey(Website, related_name='statistics', on_delete=models.CASCADE)
    period = models.CharField(max_length=1, choices=PERIODS)
    ts = models.DateTimeField()
    visitors_count = models.IntegerField(default=0)
    browsers = models.JSONField(default=dict)
    desktop_os = models.JSONField(default=dict)
    mobile_os = models.JSONField(default=dict)
    cities = models.JSONField(default=dict)
    pages = models.JSONField(default=dict)
    referrers = models.JSONField(default=dict)
    utm_tags = models.JSONField(default=dict)
    aggregated = models.BooleanField(default=False)

    def __str__(self):
        return ': '.join([self.website.title, self.period, self.ts.isoformat()])

    class Meta:
        unique_together = ('website', 'period', 'ts',)


class Client(models.Model):
    client_id = models.CharField(max_length=100, db_index=True)
    client_name = models.CharField(max_length=100, blank=True, null=True)
    website = models.ForeignKey(Website, related_name='clients', on_delete=models.CASCADE)
    first_ts = models.DateTimeField(auto_now_add=True)
    last_activity_ts = models.DateTimeField(auto_now_add=True)
    visit_count = models.IntegerField(default=0)
    chat_count = models.IntegerField(default=0)
    void_chat_count = models.IntegerField(default=0)
    unaccepted_chat_count = models.IntegerField(default=0)
    purchase_count = models.IntegerField(default=0)
    product_view_count = models.IntegerField(default=0)
    rate_avg = models.FloatField(default=0.0)
    rate_count = models.IntegerField(default=0)
    props = models.JSONField(default=dict)
    cart = models.JSONField(default=dict)

    class Meta:
        unique_together = ('client_id', 'website',)

    def __str__(self):
        return ': '.join([self.website.title, str(self.client_id)])

    @staticmethod
    def create_new_client(website_id, name='Guest'):
        nice_chat_client_id = nc_gen_rnd_token()

        pair = Idempotency.objects.create(
            pk=nice_chat_client_id,
            mode=Idempotency.MODE_CLASSIC
        )
        IdempotencyPeer.objects.create(peer=nice_chat_client_id, node=pair)

        client = Client.objects.create(
            website_id=website_id,
            client_id=nice_chat_client_id,
            client_name=name
        )
        return client

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None, is_not_autocreate=False):
        if settings.DISABLE_CLIENT_AUTOCREATE and self.pk is None and not is_not_autocreate:
            raise PermissionDenied()
        super(Client, self).save(force_insert, force_update, using, update_fields)


class ClientEntrance(models.Model):
    client = models.ForeignKey(Client, related_name='entrances', on_delete=models.CASCADE)
    ts = models.DateTimeField(blank=True, null=True)
    pages = models.IntegerField(default=0)
    duration = models.IntegerField(default=0)


class ClientPromotion(models.Model):
    client = models.ForeignKey(Client, related_name='promotions', on_delete=models.CASCADE)
    ts = models.DateTimeField(blank=True, null=True)
    props = models.JSONField(default=dict)


class ClientCarts(models.Model):
    client = models.ForeignKey(Client, related_name='carts', on_delete=models.CASCADE)
    ts = models.DateTimeField(blank=True, null=True)
    cart = models.JSONField(default=dict)


class Order(models.Model):
    client = models.ForeignKey(Client, related_name='orders', on_delete=models.CASCADE)
    chat_id = models.IntegerField(blank=True, null=True, db_index=True)
    website = models.ForeignKey(Website, related_name='orders', on_delete=models.CASCADE)
    order_num = models.CharField(max_length=50, default='n/a', db_index=True)
    order_date = models.DateTimeField(auto_now_add=True)
    products = models.ManyToManyField(Product, related_name='orders')
    qty = models.IntegerField(default=1)
    agents = models.ManyToManyField(settings.AUTH_USER_MODEL, related_name='order_agents')
    shipping_address = models.JSONField(default=dict)
    billing_address = models.JSONField(default=dict)
    product_detail = models.JSONField(default=dict)

    @cached_property
    def contacts(self):
        client = {
            'id': self.client.id,
            'name': self.client.client_name,
            'coming': self.client.first_ts,
            'latest': self.client.last_activity_ts,
            'visits': self.client.visit_count,
            'chats': self.client.chat_count,
        }
        return {
            'client': client,
            'billing_address': nice_address(self.billing_address),
            'shipping_address': nice_address(self.shipping_address if self.shipping_address else self.billing_address),
            'info': {k: self.product_detail[k] for k in self.product_detail.keys() if k and k.startswith('geo')}
        }

    @cached_property
    def deal(self):
        products = []
        agents = []

        for prod in self.products.all():
            products.append({
                'id': prod.original_id,
                'title': prod.title,
                'price': prod.price,
                'qty': 1,
                'currency': prod.currency,
                'url': prod.url
            })

        for agent in self.agents.all():
            agents.append(agent.email)

        return {
            'chat': self.chat_id,
            'website': self.website.domain,
            'num': self.order_num,
            'date': self.order_date,
            'products': products,
            'agents': agents
        }

    def __str__(self):
        return ': '.join([self.website.title, str(self.client_id)])


class FormResult(models.Model):
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    label = models.CharField(max_length=50)
    form_id = models.IntegerField(db_index=True)
    chat_id = models.IntegerField()
    client = models.ForeignKey(Client, on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)
    result = models.JSONField(default=dict)
