import json
import os
import pytz
import datetime
from collections import defaultdict, Counter

from django.conf import settings
from django.utils import timezone
from django.db import IntegrityError, DatabaseError, InterfaceError, models

from apps.chat.utils import is_list_has_value
from apps.chat.models import VisitorStatWide, Visitor, BatamStatistics, ArchivedChat
from apps.visitors.models import Statistics, Browser, OperatingSystem, Client, MOBILE, DESKTOP, ClientPromotion, \
    ClientEntrance
from nicechat.celery import app

VISITORS_LABEL = 'visitors'
VISITORS_KEY = 0
WEBSITE_KEY = 1
TS_KEY = 2

HOUR = 'hour'
DAY = 'day'
WEEK = 'week'
MONTH = 'month'


def round_timestamp(dt, period):
    if period == HOUR:
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour)
    elif period == DAY:
        return datetime.datetime(dt.year, dt.month, dt.day)
    elif period == WEEK:
        dz = dt - datetime.timedelta(days=(dt.isoweekday() - 1))
        return datetime.datetime(dz.year, dz.month, dz.day)
    elif period == MONTH:
        return datetime.datetime(dt.year, dt.month, 1)


def get_back_date(period):
    if period == HOUR:
        dt = datetime.datetime.now() - datetime.timedelta(days=2)
        return datetime.datetime(dt.year, dt.month, dt.day, 0)
    elif period == DAY:
        dt = datetime.datetime.now() - datetime.timedelta(days=10)
        return datetime.datetime(dt.year, dt.month, dt.day)
    elif period == WEEK:
        dt = datetime.datetime.now() - datetime.timedelta(days=15)
        return datetime.datetime(dt.year, dt.month, dt.day)
    elif period == MONTH:
        dt = datetime.datetime.now() - datetime.timedelta(days=50)
        return datetime.datetime(dt.year, dt.month, 1)

@app.task
def clean_visitors():
    for v in Visitor.objects.filter(state=Visitor.STATE_OFFLINE):
        if not v.ts_close or (timezone.now() - v.ts_close).total_seconds() > 900:
            v.delete()



@app.task
def prepare_batam_visitors_statistics():
    prepare_batam_visitors_statistics_do()


def prepare_batam_visitors_statistics_do():
    for vs in VisitorStatWide.objects.filter(batam_processed=False).order_by('ts_open'):
        cl = Client.objects.filter(client_id = vs.client, website_id=vs.website).first()
        if not cl:
            continue
        rec, _ = BatamStatistics.objects.get_or_create(
            client=cl.id, website = cl.website_id
        )
        for f_nm in ['page_ref', 'page_utm', 'geo_country', 'geo_city']:
            if not getattr(rec, f_nm, None) and getattr(vs, f_nm, None):
                setattr(rec, f_nm, getattr(vs, f_nm))
        
        rec.sys_os = f'{vs.sys_os}:{vs.sys_os_ver}'
        rec.sys_browser = f'{vs.sys_browser}:{vs.sys_browser_ver}'
        rec.sys_device = vs.sys_device or ''
        rec.name =  cl.client_name
        if rec.first_visit_ts > vs.ts_open or rec.first_visit_ts == 0:
            rec.first_visit_ts = vs.ts_open 
        if rec.last_visit_ts < vs.ts_close:
            rec.last_visit_ts = vs.ts_close
            for q in vs.visited_pages:
                rec.latest_page_title = q.get('page_title', '')
                rec.latest_page_url = q.get('page_url', '')
        rec.total_visits += 1
        rec.total_duration += vs.duration if vs.duration else 0
        rec.save()
        vs.batam_processed = True
        vs.save()

    for ac in ArchivedChat.objects.filter(batam_processed=False).order_by('ts_open'):
        cl = Client.objects.filter(client_id = ac.client_id, website=ac.website).first()
        if not cl:
            continue
        rec, _ = BatamStatistics.objects.get_or_create(
            client=cl.id, website = cl.website_id
        )
        rec.total_chats += 1
        if ac.invited_by_agent:
            rec.total_chats_agentinit += 1
        if ac.ts_open and rec.first_chat_ts > ac.ts_open.timestamp() or rec.first_chat_ts == 0:
            rec.first_chat_ts = ac.ts_open.timestamp()
        if ac.ts_close and rec.last_chat_ts < ac.ts_close.timestamp():
            rec.last_chat_ts = ac.ts_close.timestamp()
            rec.latest_chat_id = ac.pk
        rec.save()
        ac.batam_processed = True
        ac.save()


@app.task
def prepare_visitors_statistics():
    prepare_visitors_statistics_do()


def prepare_visitors_statistics_do(interactive=False):
    browsers = Browser.objects.values_list('name', flat=True)
    oss_m = OperatingSystem.objects.filter(type=MOBILE).values_list('name', flat=True)
    oss_d = OperatingSystem.objects.filter(type=DESKTOP).values_list('name', flat=True)

    data_from = VisitorStatWide.objects.filter(processed=False).aggregate(
        models.Min('ts_close'),
    )

    if not data_from['ts_close__min']:
        return
    
    tsc_s = int(data_from['ts_close__min']/settings.STATISTICS_ROTATE_INTERVAL)
    tsc_e = int(timezone.now().timestamp()/settings.STATISTICS_ROTATE_INTERVAL)

    for ts in range(tsc_s, tsc_e):
        range_start = ts*settings.STATISTICS_ROTATE_INTERVAL
        range_end = (ts + 1)*settings.STATISTICS_ROTATE_INTERVAL -1
        wss = VisitorStatWide.objects.filter(ts_close__gte=range_start, ts_close__lt=range_end).distinct()
        dtwss = {}
        for ws in wss:
            if not ws.website in dtwss:
                dtwss[ws.website] = []
            dtwss[ws.website].append(ws)

        for website_id, lines in dtwss.items():
            # print(range_start, website_id)
        
            if lines:
                browsers_count = defaultdict(int)
                oss_d_count = defaultdict(int)
                oss_m_count = defaultdict(int)
                cities_count = defaultdict(int)
                pages_count = defaultdict(int)
                referrers_count = defaultdict(int)
                utms_count = defaultdict(int)
                visitors_count = 0

                for visitors_obj in lines:

                    visitors_count += 1

                    os_value = is_list_has_value(oss_m, visitors_obj.sys_os)
                    if os_value:
                        oss_m_count[os_value] += 1
                    else:
                        os_value = is_list_has_value(oss_d, visitors_obj.sys_os)
                        if not os_value:
                            os_value = 'Other'
                        oss_d_count[os_value] += 1

                    browser_value = is_list_has_value(browsers, visitors_obj.sys_browser)
                    if not browser_value:
                        browser_value = 'Other'
                    browsers_count[browser_value] += 1

                    cities_count[visitors_obj.geo_city] += 1

                    for p in visitors_obj.visited_pages:
                        pages_count[p['page_url']] += 1
                    if visitors_obj.page_ref:
                        referrers_count[visitors_obj.page_ref] += 1
                    if visitors_obj.page_utm:
                        utms_count[json.dumps(visitors_obj.page_utm)] += 1

                try:
                    if settings.TRY_TO_UPDATE_EXISTS_STAT:
                        Statistics.objects.filter(
                            website_id=website_id,
                            period=Statistics.QUARTER,
                            ts=datetime.datetime.utcfromtimestamp(range_start).replace(tzinfo=pytz.UTC),
                        ).delete()
                    Statistics.objects.create(
                        website_id=website_id,
                        visitors_count=visitors_count,
                        period=Statistics.QUARTER,
                        ts=datetime.datetime.utcfromtimestamp(range_start).replace(tzinfo=pytz.UTC),
                        browsers=browsers_count,
                        desktop_os=oss_d_count,
                        mobile_os=oss_m_count,
                        cities=cities_count,
                        pages=pages_count,
                        referrers=referrers_count,
                        utm_tags=utms_count
                    )
                except IntegrityError:
                    if interactive:
                        print('integrity error')
                    continue

                for visitors_obj in lines:
                    if visitors_obj.client:
                        try:
                            client, created = Client.objects.get_or_create(client_id=visitors_obj.client,
                                                                           website_id=website_id)
                            visitor_ts = datetime.datetime.utcfromtimestamp(range_start).replace(tzinfo=pytz.UTC)
                            if created:
                                client.client_name = settings.DEFAULT_CLIENT_NAME
                                client.visit_count = 1
                                client.first_ts = visitor_ts
                            else:
                                client.visit_count += 1
                            client.last_activity_ts = visitor_ts
                            client.save()
                            utm_tags = visitors_obj.page_utm
                            referrer = visitors_obj.page_ref
                            if utm_tags or (referrer and referrer != '(direct)'):
                                cpm = ClientPromotion(client=client)
                                cpm.ts = visitor_ts
                                if utm_tags:
                                    cpm.props['utm'] = utm_tags
                                if referrer:
                                    cpm.props['referrer'] = referrer
                                cpm.save()
                            ce = ClientEntrance(client=client)
                            ce.ts = visitor_ts
                            ce.pages = len(visitors_obj.visited_pages)
                            ce.duration = visitors_obj.duration
                            ce.save()
                            visitors_obj.processed=True
                            visitors_obj.save()
                        except IntegrityError:
                            continue


def statistics_aggregating(period_from, period_to, round_to, dateback=True):
    websites = defaultdict(list)
    stats = Statistics.objects.filter(period=period_from)
    if dateback:
        stats = stats.filter(ts__gte=get_back_date(round_to))
    [websites[stat.website_id].append(stat) for stat in stats]
    for website_id in websites:
        if type(website_id) is int:
            websites_stats = defaultdict(dict)
            for stats in websites[website_id]:
                index_ts = round_timestamp(stats.ts, round_to)
                if len(websites_stats[index_ts].keys()) == 0:
                    websites_stats[index_ts] = {
                        'visitors_count': stats.visitors_count,
                        'browsers': stats.browsers,
                        'desktop_os': stats.desktop_os,
                        'mobile_os': stats.mobile_os,
                        'cities': stats.cities,
                        'pages': stats.pages,
                        'referrers': stats.referrers,
                        'utm_tags': stats.utm_tags,
                    }
                else:
                    websites_stats[index_ts]['visitors_count'] += stats.visitors_count
                    websites_stats[index_ts]['browsers'] = Counter(websites_stats[index_ts]['browsers']) + \
                                                           Counter(stats.browsers)
                    websites_stats[index_ts]['desktop_os'] = Counter(websites_stats[index_ts]['desktop_os']) + \
                                                             Counter(stats.desktop_os)
                    websites_stats[index_ts]['mobile_os'] = Counter(websites_stats[index_ts]['mobile_os']) + \
                                                            Counter(stats.mobile_os)
                    websites_stats[index_ts]['cities'] = Counter(websites_stats[index_ts]['cities']) + \
                                                         Counter(stats.cities)
                    websites_stats[index_ts]['pages'] = Counter(websites_stats[index_ts]['pages']) + \
                                                        Counter(stats.pages)
                    websites_stats[index_ts]['referrers'] = Counter(websites_stats[index_ts]['referrers']) + \
                                                        Counter(stats.referrers)
                    websites_stats[index_ts]['utm_tags'] = Counter(websites_stats[index_ts]['utm_tags']) + \
                                                            Counter(stats.utm_tags)
                # stats.aggregated = True
                # stats.save()
        for ts in websites_stats:
                Statistics.objects.update_or_create(
                    website_id=website_id,
                    ts=ts.replace(tzinfo=pytz.UTC),
                    period=period_to,
                    defaults={
                        'visitors_count': websites_stats[ts]['visitors_count'],
                        'browsers': websites_stats[ts]['browsers'],
                        'desktop_os': websites_stats[ts]['desktop_os'],
                        'mobile_os': websites_stats[ts]['mobile_os'],
                        'cities': websites_stats[ts]['cities'],
                        'pages': websites_stats[ts]['pages'],
                        'referrers': websites_stats[ts]['referrers'],
                        'utm_tags': websites_stats[ts]['utm_tags'],
                    }
                )


@app.task
def produce_hourly_statistics(dateback=True):
    statistics_aggregating(Statistics.QUARTER, Statistics.HOUR, HOUR, dateback)


@app.task
def produce_daily_statistics(dateback=True):
    statistics_aggregating(Statistics.HOUR, Statistics.DAY, DAY, dateback)


@app.task
def produce_weekly_statistics(dateback=True):
    statistics_aggregating(Statistics.DAY, Statistics.WEEK, WEEK, dateback)


@app.task
def produce_monthly_statistics(dateback=True):
    statistics_aggregating(Statistics.DAY, Statistics.MONTH, MONTH, dateback)


@app.task
def visitor_cleaner_worker_task(client_id, token):
    try:
        vs = Visitor.objects.filter(client_id=client_id, session=token).first()
        if vs and vs.state == Visitor.STATE_OFFLINE:
            print('clean visitor', vs, vs.pk)
            vs.delete()
    except (DatabaseError, InterfaceError, Visitor.DoesNotExist) as e:
        print(e)
