from unittest import skip

import pytz
import json
import datetime

from django.conf import settings
from django.test import override_settings
from channels.tests import apply_routes, Client
from channels import route_class
from channels.generic import websockets

from nicechat.utils.tests import TestCaseWithSites, TestCaseWithStatistics
from nicechat.utils.constants import WebsocketEventTypes
from apps.chat.models import Visitor
from apps.visitors.ws import cli_wait_connect, cli_wait_receive, cli_wait_disconnect
from apps.visitors.models import Statistics


class TestStatisticsAggregation(TestCaseWithStatistics):
    def test_aggregation(self):
        from apps.visitors.tasks import produce_hourly_statistics, produce_daily_statistics, \
            produce_weekly_statistics, produce_monthly_statistics

        self.assertEqual(Statistics.objects.filter(period=Statistics.QUARTER).count(), len(self.website_list)*5)

        produce_hourly_statistics(False)
        produce_daily_statistics(False)
        produce_weekly_statistics(False)
        produce_monthly_statistics(False)

        self.assertEqual(Statistics.objects.filter(period=Statistics.HOUR).count(), len(self.website_list)*4)
        self.assertEqual(Statistics.objects.filter(period=Statistics.DAY).count(), len(self.website_list)*3)
        self.assertEqual(Statistics.objects.filter(period=Statistics.WEEK).count(), len(self.website_list)*2)
        self.assertEqual(Statistics.objects.filter(period=Statistics.MONTH).count(), len(self.website_list)*2)

        st_hour = Statistics.objects.filter(period=Statistics.HOUR,
                                            ts=datetime.datetime(2017, 05, 01, 10, tzinfo=pytz.UTC)).first()
        st_week = Statistics.objects.filter(period=Statistics.WEEK,
                                            ts=datetime.datetime(2017, 05, 01, tzinfo=pytz.UTC)).first()

        self.assertEqual(st_hour.visitors_count, 4)

        self.assertEqual(st_hour.browsers.get(self.BROWSER_1), 3)
        self.assertEqual(st_hour.browsers.get(self.BROWSER_2), 1)

        self.assertEqual(st_hour.mobile_os.get(self.OSMOB_1), 1)
        self.assertEqual(st_hour.desktop_os.get(self.OS_1), 2)
        self.assertEqual(st_hour.desktop_os.get(self.OS_2), 1)

        self.assertEqual(st_hour.cities.get(self.CITY_1), 3)
        self.assertEqual(st_hour.cities.get(self.CITY_2), 1)

        self.assertEqual(st_hour.pages.get(self.PAGE_1), 3)
        self.assertEqual(st_hour.pages.get(self.PAGE_2), 1)

        self.assertEqual(st_hour.referrers.get(self.REF_1), 3)
        self.assertEqual(st_hour.referrers.get(self.REF_2), 1)

        self.assertEqual(st_week.visitors_count, 8)

        self.assertEqual(st_week.browsers.get(self.BROWSER_1), 6)
        self.assertEqual(st_week.browsers.get(self.BROWSER_2), 2)

        self.assertEqual(st_week.mobile_os.get(self.OSMOB_1), 2)
        self.assertEqual(st_week.desktop_os.get(self.OS_1), 4)
        self.assertEqual(st_week.desktop_os.get(self.OS_2), 2)

        for site in self.website_list:
            Statistics.objects.create(
                website=site,
                period=Statistics.QUARTER,
                ts=datetime.datetime(2017, 05, 01, 10, 45, 0, tzinfo=pytz.UTC),
                visitors_count=3,
                browsers={self.BROWSER_2: 3},
                desktop_os={self.OS_2: 2},
                mobile_os={self.OSMOB_1: 1},
                cities={self.CITY_2: 3},
                pages={self.PAGE_2: 3},
                referrers={},
            )

        produce_hourly_statistics(False)
        produce_daily_statistics(False)
        produce_weekly_statistics(False)
        produce_monthly_statistics(False)

        self.assertEqual(Statistics.objects.filter(period=Statistics.HOUR).count(), len(self.website_list) * 4)
        self.assertEqual(Statistics.objects.filter(period=Statistics.DAY).count(), len(self.website_list) * 3)
        self.assertEqual(Statistics.objects.filter(period=Statistics.WEEK).count(), len(self.website_list) * 2)
        self.assertEqual(Statistics.objects.filter(period=Statistics.MONTH).count(), len(self.website_list) * 2)

        st_hour2 = Statistics.objects.filter(period=Statistics.HOUR, ts=datetime.datetime(2017, 05, 01, 10,
                                                                                          tzinfo=pytz.UTC)).first()

        self.assertEqual(st_hour2.visitors_count, 7)

        self.assertEqual(st_hour2.browsers.get(self.BROWSER_1), 3)
        self.assertEqual(st_hour2.browsers.get(self.BROWSER_2), 4)

        self.assertEqual(st_hour2.mobile_os.get(self.OSMOB_1), 2)
        self.assertEqual(st_hour2.desktop_os.get(self.OS_1), 2)
        self.assertEqual(st_hour2.desktop_os.get(self.OS_2), 3)

        self.assertEqual(st_hour2.cities.get(self.CITY_1), 3)
        self.assertEqual(st_hour2.cities.get(self.CITY_2), 4)

        self.assertEqual(st_hour2.pages.get(self.PAGE_1), 3)
        self.assertEqual(st_hour2.pages.get(self.PAGE_2), 4)

        self.assertEqual(st_hour2.referrers.get(self.REF_1), 3)
        self.assertEqual(st_hour2.referrers.get(self.REF_2), 1)


@skip('Need to fix the tests')
@override_settings(SESSION_ENGINE="django.contrib.sessions.backends.cache", VISITOR_REMOVE_TIMEOUT=2)
class TestVisitorBinding_FunctionalTest(TestCaseWithSites):
    def test_websockets_consumers(self):
        work_website = getattr(self, 'website_1_1', None)
        self.assertIsNotNone(work_website, msg="Wrong test setUp!")

        with apply_routes([route_class(WebsocketConsumer, path='/widget/(?P<site_id>[A-Z0-9]+)/')]):

            # setup "as visitor" channel
            visitor = Client()
            visitor_path = f'/widget/{work_website.pk}/'

            # setup "as operator" channel and "bind" it to website-monitor websocket
            employee = Client()
            employee.join_group("adm-website_mon-%i" % work_website.pk)

            # raise "visitor-connected" event
            visitor.send_and_consume('websocket.connect', {'path': visitor_path})

            # receive data for operator and check:
            #   data present
            #   website is right
            #   and visitor is offline
            received = employee.receive()
            self.assertIsNotNone(received)
            self.assertTrue('text' in received)
            data = json.loads(received['text'])
            self.assertTrue('type' in data and 'data' in data)
            self.assertEqual(data['type'], WebsocketEventTypes.BACKEND_SITE_VISITOR_UPDATE)
            self.assertEqual(data['data']['website'], work_website.pk)
            self.assertEqual(data['data']['state'], Visitor.STATE_ONLINE)

            # # raise "visitor-send data" event
            # consumer = visitor.send_and_consume('websocket.receive', {'path': visitor_path, 'text': 'text'})
            # # self.assertEqual(consumer.text, 'text')

            # raise "visitor-disconnected" event
            visitor.send_and_consume('websocket.disconnect', {'path': visitor_path})

            # receive data for operator and check:
            #   data present
            #   website is right
            #   and visitor is offline
            received = employee.receive()
            self.assertIsNotNone(received)
            self.assertTrue('text' in received)
            data = json.loads(received['text'])
            self.assertTrue('type' in data and 'data' in data)
            self.assertEqual(data['type'], WebsocketEventTypes.BACKEND_SITE_VISITOR_UPDATE)
            self.assertEqual(data['data']['website'], work_website.pk)
            self.assertEqual(data['data']['state'], Visitor.STATE_OFFLINE)


class WebsocketConsumer(websockets.WebsocketConsumer):
    # map appropriate "events" to real dispatch functions
    def connect(self, message, **kwargs):
        cli_wait_connect(message, **kwargs)

    def disconnect(self, message, **kwargs):
        cli_wait_disconnect(message, **kwargs)

    def raw_receive(self, message, **kwargs):
        cli_wait_receive(message, **kwargs)
