from apps.chat.models import Visitor, Chat
from django.conf import settings
from channels.generic.websocket import WebsocketConsumer
from django.shortcuts import get_object_or_404
from apps.chat.models import Website, IdempotencyPeer, Idempotency
from django.utils import timezone
from nicechat.utils.ip_info import parse_info
import threading
import time

from nicechat.utils.ip_info import get_info_by_ip
from apps.visitors.tasks import visitor_cleaner_worker_task


def set_visitor_connection_info(site_id, token, ch, client_id, connector=None):
    """
    Set info about active visitor to cache

    :param site_id: site_id from request (after checking))
    :param token: session token hash
    :param message: Channel.Message object
    :param client_id: anonymous session token (stored in cookies)
    :return: None
    """

    ws = Website.objects.filter(pk=site_id).first()
    if not ws or not ws.online:
        return "disabled"
    vs = Visitor.objects.filter(client_id=client_id, session=token).first()
    if not vs and not connector:
        pair, is_created = Idempotency.objects.get_or_create(
            pk=client_id,
            mode=Idempotency.MODE_CLASSIC
        )
        IdempotencyPeer.objects.get_or_create(peer=client_id, node=pair)
    if not vs and connector:
        pair, is_created = Idempotency.objects.get_or_create(
            pk=client_id,
            mode=Idempotency.MODE_CLASSIC,
            connector_id = connector,
        )
        IdempotencyPeer.objects.get_or_create(peer=client_id, node=pair)

    if isinstance(ch, WebsocketConsumer):
        vs = parse_info(ch, vs)

    if isinstance(ch, dict):
        if not vs:
            vs = Visitor()
        for _k, _v in ch.items():
            if hasattr(vs, _k):
                setattr(vs, _k, _v)

    vs.client_id = client_id
    vs.session = token
    vs.website = ws
    chat = vs.active_chat.filter(is_open=True).first() if vs.pk else None

    from apps.visitors.models import Client
    if settings.DISABLE_CLIENT_AUTOCREATE:
        if Client.objects.filter(client_id=client_id).first() is None:
            return "disabled"

    if not chat or not chat.is_open:
        chat = Chat.objects.filter(client_id=client_id, website_id=site_id, is_open=True).first()

    if chat:
        if chat.operators_current.count() > 0:
            vs.state = Visitor.STATE_CHATING
        else:
            vs.state = Visitor.STATE_CALLING
    else:
        vs.state = Visitor.STATE_ONLINE
    vs.ts_close = None
    vs.save()

    if isinstance(ch, dict) and ch.get('client_label'):
        cl = Client.objects.filter(client_id = client_id, website_id=site_id).first()
        if cl:
            nm = ch.get('client_label')
            cl.client_name = nm
            cl.props["NAME"] = nm
            cl.save()

    if chat and not chat.visitor == vs:
        chat.visitor = vs
        chat.save(update_fields=('visitor',))

    ip_info = get_info_by_ip(vs.sys_ip)
    vs.geo_city = ip_info['city']
    vs.geo_country = ip_info['country']
    vs.geo_lat = ip_info['lat']
    vs.geo_lon = ip_info['lon']

    vs.geo_country = vs.geo_country[0:30] if vs.geo_country and isinstance(vs.geo_country, str) else None
    vs.geo_city = vs.geo_city[0:100] if vs.geo_city and isinstance(vs.geo_city, str) else None

    vs.save()

    return chat


def remove_visitor_connection(token, client_id):
    """
    Remove info for disconnected visitor with 15s delay
    (switch to "offline", wait 15s, if visitor not returned - remove)

    :param token: session token hash
    :param client_id:  anonymous session token (stored in cookies)
    :return: None
    """

    vs = Visitor.objects.filter(client_id=client_id, session=token).first()
    if vs:
        print('Connection close', vs.pk)
        vs.state = Visitor.STATE_OFFLINE
        vs.ts_close = timezone.now()
        vs.save()

        if settings.IS_DJANGO_DEBUG_RUN:
            t = threading.Thread(target=visitor_cleaner_worker, args=[client_id, token, time.time() + settings.VISITOR_REMOVE_TIMEOUT])
            t.daemon = True
            t.start()
        else:
            visitor_cleaner_worker_task.apply_async(args=[client_id, token], countdown=settings.VISITOR_REMOVE_TIMEOUT)

        # close visitor chats
        from channels.layers import get_channel_layer
        channel_layer = get_channel_layer()
        visitor_chats = vs.active_chat.filter(is_open=True)
        from apps.visitors.ws_utils.room import chat_ws_close
        for chat in visitor_chats:
            chat_ws_close({}, None, {'chat_id': chat.pk}, channel_layer.send)


def visitor_cleaner_worker(client_id, token, ts):

    print("Wait to close visitor ID:", client_id)
    while True:
        time.sleep(2)

        if time.time() > ts:
            visitor_cleaner_worker_task.run(client_id, token)
