import json
import base64
import hashlib
import re
import traceback
import time
from datetime import datetime
from nicechat.utils.constants import WebsocketEventTypes
from channels.generic.websocket import WebsocketConsumer
from apps.chat.models import Message, Chat, Visitor, VisitedPage
from apps.website.models import Website, Form
from apps.website.serializers import FormForWidgetSerializer
from django.db.models import signals
from django.http import QueryDict
from django.http import parse_cookie
from django.conf import settings
from apps.chat.utils import fill_info_by_tags
from apps.chat.tasks import celery_chat_call_wdt
from apps.video.models import Call, CallPeer
from rest_framework.generics import get_object_or_404
from nicechat.analytics import event_track, VISITOR_CHAT
from django.core.cache import caches, InvalidCacheBackendError
from apps.website.models import Form, Product
from apps.visitors.models import FormResult, Client
from apps.website.utils.form import greet_form_validator
from nicechat.utils.client_token import nc_validate_2key
from nicechat.utils.video import get_video_module, get_video_provider, video_stop_call, video_start_call

GREET_FORM_NONE = 'none'
GREET_FORM_MODAL = 'modal'
GREET_FORM_STD = 'std'
GREET_FORM_STD_MSG = 'std_msg'


def visitor_monitor_packet(dt, kwargs, ch):
    if isinstance(dt, dict):
        _type = dt.get('type', None)
        if _type and _type in (WebsocketEventTypes.FRONTEND_CHAT_SHOW_WINDOW,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_CLIOPEN,
                               WebsocketEventTypes.FRONTEND_CHAT_HIDE_WINDOW,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_MONITOR,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_CALL,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_CALL_STOP,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_AUTO_INIT,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_GREETEING,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_CART,
                               WebsocketEventTypes.FRONTEND_CHAT_WS_GOODS,
                               WebsocketEventTypes.FRONTEND_CHAT_SET_INFO):
            session = ch.scope['session'].get('cli-session', None)
            site_id = kwargs.get('site_id')
            visitor = Visitor.objects.filter(website_id=site_id, client_id=session).first()
            if visitor:
                if _type == WebsocketEventTypes.FRONTEND_CHAT_SET_INFO:
                    # visitor push chat_state packet
                    monitor_do_set_info(dt, visitor)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_CLIOPEN:
                    # receive cliopen packet
                    monitor_do_cliopen(site_id, session, ch.send)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_GREETEING:
                    monitor_pushform(dt, session, kwargs, ch)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_GOODS:
                    monitor_get_goods(dt, session, kwargs, ch)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_CART:
                    monitor_cart(dt, session, kwargs, ch, visitor)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_CALL:
                    call_start(dt, session, kwargs, ch, visitor)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_CALL_STOP:
                    call_stop(dt, session, kwargs, ch, visitor)
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_MONITOR:
                    # receive visitor state
                    token = "visitor-state-%s-%s" % (site_id, session)
                    try:
                        cache = caches['default']
                        cache.set(token, dt.get('data', 'away'), 600)
                    except InvalidCacheBackendError:
                        pass
                    signals.post_save.send(sender=Visitor, instance=visitor, created=False, update_fields=[])
                elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_AUTO_INIT:
                    monitor_do_auto_init(dt, site_id, session)
        else:
            ws_cli_routes = getattr(WebsocketConsumer, 'ws_cli_routes', None)
            if ws_cli_routes:
                _call = ws_cli_routes.get(_type)
                if callable(_call):
                    session = ch.scope['session'].get('cli-session', None)
                    site_id = kwargs.get('site_id')
                    visitor = Visitor.objects.filter(website_id=site_id, client_id=session).first()
                    if visitor:
                        try:
                            _call(dt, session, kwargs, ch, visitor)
                        except Exception as x:
                            print(x)
                            traceback.print_exc()


def monitor_do_cliopen(site_id, session, ch_send, connector = None):

    chat = Chat.objects.filter(client_id=session,  website_id=site_id, is_open=True).first()
    visitor = Visitor.objects.filter(client_id=session, website_id=site_id).first()

    if chat:
        chat.visitor = visitor
        chat.save(update_fields=['visitor'])
        ch_send(text_data= json.dumps({'type': WebsocketEventTypes.FRONTEND_CHAT_WS_CLIOPEN,
                                                        'created': False, 'id': chat.pk}))
        return chat.pk
    else:
        website = Website.objects.filter(pk=site_id).first()
        if not website:
            return
        greet_form = website.props.get('greet_form', 'none')

        if greet_form == GREET_FORM_MODAL and not connector:
            form = Form.objects.filter(website=site_id, mode=Form.MODE_SITE_GREET_FORM).first()
            if form and form:
                data = dict(FormForWidgetSerializer(form).to_representation(form))

                ch_send(text_data= json.dumps({
                    'type': WebsocketEventTypes.FRONTEND_CHAT_WS_GREETEING,
                    'label': form.label,
                    'id': 'greet',
                    'data': json.dumps(data)
                }))
                return
            else:
                greet_form = GREET_FORM_NONE

        chat = Chat.create(client_id=session, ts_open=datetime.now(), website=website)
        if connector:
            chat.is_connector = True
            chat.connector_name = connector
            chat.save()
        if greet_form == GREET_FORM_NONE or greet_form == GREET_FORM_STD_MSG:
            send_welcome(chat, 0, website.title, website)

        if greet_form == GREET_FORM_STD or greet_form == GREET_FORM_STD_MSG:
            form = Form.objects.filter(website=chat.website, mode=Form.MODE_SITE_GREET_FORM).first()
            if form and form:
                data = dict(FormForWidgetSerializer(form).to_representation(form))

                Message.objects.create(
                    chat=chat,
                    sender_id=0,
                    sender_name=chat.website.title,
                    text=form.label,
                    data=json.dumps(data),
                    data_type=Message.DATA_FORM,
                    hidden=json.dumps({'form_id': form.pk}),
                    type=Message.TYPE_OPERATOR
                )

        chat.visitor = visitor
        chat.save(update_fields=['visitor'])

        # BotChat.objects.create(
        #     chat=chat,
        #     website=website,
        #     botname=settings.AI_BOT
        # )
        #
        if visitor:
            visitor.state = Visitor.STATE_CALLING
            visitor.save(update_fields=('state',))

    event_track(VISITOR_CHAT, None, {
        'chatId': chat.id,
        'clientID': chat.client_id,
        'mp_country_code': visitor.geo_country if visitor else '',
        '$city': visitor.geo_city if visitor else '',
        'os': visitor.sys_os if visitor else '',
        '$browser': visitor.sys_browser if visitor else '',
    })

    ch_send(text_data= json.dumps({'type': WebsocketEventTypes.FRONTEND_CHAT_WS_CLIOPEN,
                                                    'created': True, 'id': chat.pk}))
    
    return chat.pk


def send_welcome(chat, sender_id, sender_name, website):

    welcome = website.props.get(website.WELCOME_MESSAGE, None)
    if not welcome:
        # because if props have '', function .get() return '', but we need settings value
        welcome = settings.GLOBAL_WELCOME_MESSAGE
    if isinstance(welcome, list) or isinstance(welcome, tuple):
        for msg in welcome:
            Message.objects.create(chat=chat, sender_id=sender_id, sender_name=sender_name,
                                   text=msg, type=Message.TYPE_OPERATOR)
    elif isinstance(welcome, str):
        Message.objects.create(chat=chat, sender_id=sender_id, sender_name=sender_name,
                               text=welcome, type=Message.TYPE_OPERATOR)


def monitor_get_goods(dt, session, kwargs, ch):
    result = {}
    for prod in Product.objects.filter(website_id=kwargs.get('site_id'), original_id__in=dt.get('task')):
        result[prod.original_id] = {
            'id': prod.original_id,
            'title': prod.title,
            'image': prod.image_url
        }
    ch.send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_WS_GOODS,
        'data': result
    }))


def monitor_do_set_info(dt, visitor):
    # on receive chat-state packet (title, url, utm, referrer, timezone)
    try:
        visitor.sys_time = int(dt.get('tzone', 0))
    except ValueError:
        pass
    try:
        visitor.page_url = dt.get('href', '')[:200]
        visitor.page_title = dt.get('title', '')[:200]
        p_ref = dt.get('ref', '')[:200]
        try:
            p_utm = json.loads(dt.get('utm', '{}'))
        except ValueError:
            p_utm = {}
        if p_ref and not visitor.page_ref:
            visitor.page_ref = p_ref
        if p_utm and not visitor.page_utm:
            visitor.page_utm = p_utm
    except (ValueError, IndexError):
        p_ref = ''
        p_utm = ''
    visitor.save(update_fields=('sys_time', 'page_url', 'page_title', 'page_ref', 'page_utm'))
    VisitedPage.objects.create(visitor=visitor, page_url=visitor.page_url,
                               page_title=visitor.page_title)
    if p_ref or p_utm:
        chat = visitor.active_chat.filter(is_open=True).first()
        if chat:
            if p_ref and not chat.page_ref:
                chat.page_ref = p_ref
            if p_utm and not chat.page_utm:
                chat.page_utm = p_utm
            chat.save(update_fields=('page_ref', 'page_utm'))


def monitor_do_auto_init(dt, site_id, session):
    # receive auto-init packet (by client website URL nc-auto-init parameter)
    tag = dt.get('tag', None)
    if isinstance(tag, str):
        s_tag = tag.split("-", 3)
        try:
            site_id = int(site_id)
        except ValueError:
            return
        if len(s_tag) == 3:
            chat = Chat.objects.filter(website_id=site_id, client_id=session, is_open=True).first()
            if chat:
                return
            if s_tag[0] == 'c':
                monitor_do_auto_init_chat(s_tag, site_id, session)
            elif s_tag[0] == 'f':
                monitor_do_auto_init_form(s_tag, site_id, session)
        if len(s_tag) == 4:
            chat = Chat.objects.filter(website_id=site_id, client_id=session, is_open=True).first()
            if not chat and s_tag[0] == 'f':
                monitor_do_auto_init_form(s_tag, site_id, session)


def monitor_do_auto_init_chat(s_tag, site_id, session):
    # nc-auto-init starts from "c-" - need to open chat and push welcome message
    msg = s_tag[1]
    sign = s_tag[2]
    j_sign = base64.b64encode(hashlib.sha256(
        ("%s-%s-%s" % (site_id, msg, settings.AUTO_OPEN_LINK_SECRET)).encode('utf8')).digest(),
                              altchars=b"_.").rstrip(b"=").decode()
    if j_sign == sign:
        if msg:
            msg += '=' * (-len(msg) % 4)
            msg = msg.replace("_", "+").replace(".", "/")
            message = base64.b64decode(msg)
        else:
            message = None
        chat = Chat.create(website_id=site_id, ts_open = datetime.now(),
                           client_id=session)
        website = Website.objects.filter(pk=site_id).first()

        if message:
            welcome = re.sub("[^\d\s\w\W\-\,\.\:\?\!]", "", message.decode('utf8'), flags=re.UNICODE)
        elif website:
            welcome = website.props.get(website.WELCOME_MESSAGE, settings.GLOBAL_WELCOME_MESSAGE)
        else:
            welcome = settings.GLOBAL_WELCOME_MESSAGE

        if website:
            sender_name = website.title
        else:
            sender_name = ''

        if isinstance(welcome, list) or isinstance(welcome, tuple):
            for msg in welcome:
                Message.objects.create(chat=chat, sender_id=0, sender_name=sender_name,
                                       text=msg, type=Message.TYPE_OPERATOR)
        elif isinstance(welcome, str):
            Message.objects.create(chat=chat, sender_id=0, sender_name=sender_name,
                                   text=welcome, type=Message.TYPE_OPERATOR)


def monitor_do_auto_init_form(s_tag, site_id, session):
    # nc-auto-init starts with "f-" - neet to open form in new chat
    try:
        form = int(s_tag[1])
    except ValueError:
        return
    sign = s_tag[2]

    msg = s_tag[3] if len(s_tag) == 4 else ''


    website = Website.objects.filter(pk=site_id).first()
    if not website:
        return

    j_sign = base64.b64encode(
        hashlib.sha256("%s-%s-%s" % (form, website.domain,
                                     settings.AUTO_OPEN_LINK_SECRET)).digest(),
        altchars="_."
    ).rstrip("=")

    if msg:
        j_sign = base64.b64encode(
            hashlib.sha256("%s-%s-%s-%s" % (form, website.domain, msg,
                                            settings.AUTO_OPEN_LINK_SECRET)).digest(),
            altchars="_.").rstrip("=")

    print(sign, j_sign)

    if sign == j_sign:
        form = Form.objects.filter(pk=form, website=website, is_enabled=True).first()
        if form:
            chat = Chat.create(website_id=site_id, ts_open = datetime.now(),
                               client_id=session)
            data = dict(FormForWidgetSerializer(form).to_representation(form))

            if msg:
                msg += '=' * (-len(msg) % 4)
                msg = msg.replace("_", "+").replace(".", "/")
                xmessage = base64.b64decode(msg)

                Message.objects.create(
                    chat=chat,
                    sender_id=0,
                    sender_name=chat.website.title,
                    text=xmessage,
                    type=Message.TYPE_OPERATOR
                )

            Message.objects.create(
                chat=chat,
                sender_id=0,
                sender_name=chat.website.title,
                text=form.label,
                data=json.dumps(data),
                data_type=Message.DATA_FORM,
                hidden=json.dumps({'form_id': form.pk}),
                type=Message.TYPE_OPERATOR
            )


def monitor_pushform(dt, session, kwargs, ch):
    site_id = kwargs.get('site_id', 0)
    query_data = dt.get('data')

    form = Form.objects.filter(website=site_id, mode=Form.MODE_SITE_GREET_FORM).first()
    website = Website.objects.filter(pk=site_id).first()

    if form and website:
        form_data = dict(FormForWidgetSerializer(form).to_representation(form))

        data, is_valid, tags = greet_form_validator(query_data, form_data)

        if is_valid:

            visitor = Visitor.objects.filter(client_id=session, website_id=site_id).first()
            chat = Chat.create(client_id=session, ts_open=datetime.now(),
                               visitor=visitor, website=website)

            send_welcome(chat, 0, website.title, website)
            Message.objects.create(
                chat=chat,
                sender_id=chat.client_id,
                sender_name=tags.get('PRIMARY_NAME'),
                text=form.label,
                data=json.dumps(data),
                hidden=tags,
                data_type=Message.DATA_FORM_FILLED,
                type=Message.TYPE_EVENT_CLI
            )
            cl = Client.objects.filter(client_id=chat.client_id).first()
            FormResult.objects.create(
                website_id=site_id,
                chat_id=chat.pk,
                form_id=form.pk,
                label=form.label,
                result=data,
                client=cl
            )

            if tags:
                fill_info_by_tags(tags, chat)

            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_GREETEING,
                'cool': 'yes'
            }))
        else:
            # is form not valid - resend it
            print(form_data)
            print(query_data)
            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_GREETEING,
                'label': form.label,
                'id': 'greet',
                'data': json.dumps(form_data),
                'errors': json.dumps(data)
            }))


def monitor_cart(dt, session, kwargs, ch, visitor):
    cl = visitor.get_client()
    tmp_cart = cl.cart.get('cart', []) if cl else []
    new_cart = []

    if dt.get('get', '-') == 'get':
        ch.send(text_data= json.dumps({
            'type': WebsocketEventTypes.FRONTEND_CHAT_WS_CART,
            'data': json.dumps(tmp_cart),
        }))
        return

    for item in dt.get('data', []):
        itm = prepare_item(item.get('id', None), kwargs.get('site_id', None), tmp_cart)
        if itm:
            if item.get('curr'):
                try:
                    itm['currency'] = item.get('curr').upper()
                except (ValueError, UnicodeError):
                    pass
            try:
                itm['qty'] = int(item.get('qty', '1'))
            except ValueError:
                itm['qty'] = 1
            try:
                if item.get('price'):
                    itm['price'] = float(item.get('price'))
            except ValueError:
                pass
            if itm['price']:
                new_cart.append(itm)

    cl.cart['cart'] = new_cart
    cl.cart['mode'] = dt.get('mode', 'none')
    cl.cart['ts'] = datetime.now().isoformat()
    cl.save(update_fields=('cart',))


def prepare_item(item_id, site_id, tmp_cart):
    if item_id:
        for item in tmp_cart:
            if item.get('product_id') == item_id:
                return {
                    'product_id': item_id,
                    'title': item['title'],
                    'currency': item['currency'],
                    'price': item['price'],
                    'image_url': item['image_url']
                }
    if item_id and site_id:
        prod = Product.objects.filter(original_id=item_id, website_id=site_id).first()
        if prod:
            return {
                'product_id': prod.original_id,
                'title': prod.title,
                'currency': prod.currency,
                'price': prod.price,
                'image_url': prod.image_url
            }

    return None


def get_session_id_from_ws(ch):
    #print(get_session_if_from_wa_inner(ch))
    return nc_validate_2key(get_session_if_from_wa_inner(ch))


def get_session_if_from_wa_inner(ch):
    headers = ch.scope['headers']
    cookies = {}
    try:
        for item in headers:
            if item and len(item) > 1 and item[0] == 'cookie':
                cookies = parse_cookie(item[1])
                break
    except (AttributeError, ValueError, IndexError):
        pass

    if cookies and settings.VISITOR_COOKIE_NAME in cookies:
        return cookies.get(settings.VISITOR_COOKIE_NAME)

    qs = ch.scope.get('query_string').decode() if ch.scope and ch.scope.get('query_string') else None
    get_dict = QueryDict(qs)
    if settings.WIDGET_SESSION_IN_GET_PARAM in get_dict:
        return get_dict.get(settings.WIDGET_SESSION_IN_GET_PARAM)

    return None


def call_start(dt, session, kwargs, message, visitor):
    room_id = dt.get('room')
    if room_id:
        chat = Chat.objects.filter(pk=room_id, client__peers__peer=session).first()
        if chat:
            cl, _ = Call.objects.get_or_create(
                chat=chat,
                is_active=True,
                provider=get_video_provider(),
            )
            cl.client_accept = True
            cl.save()
            client = visitor.get_client()
            peer, is_peercreated = CallPeer.objects.get_or_create(
                call=cl,
                peer=client.client_id,
                defaults={
                    'props': json.dumps({
                        'session': session
                    })
                }
            )
            if CallPeer.objects.filter(call=cl).exclude(peer=client.client_id).count() > 0:
                video_start_call(cl, chat)
                return
            if not is_peercreated:
                return
            chat.clear_call_messages()
            chat.put_tech_message_wdata("~!CALL_CLI_BEGIN:1", {
                # 'name': user.get_full_name or str(user),
                # 'avatar': icon_url,
                'provider': get_video_provider(),
                'time': time.time()
            })
            celery_chat_call_wdt.apply_async(args=[chat.pk], countdown=20)


def call_stop(dt, session, kwargs, message, visitor):
    for cl in Call.objects.filter(chat__client__peers__peer=session, is_active=True):
        video_stop_call(cl, visitor.get_client(), session)

