import json
import re
from datetime import datetime
from nicechat.utils.constants import WebsocketEventTypes
from apps.chat.models import ArchivedChat, Message, Chat, Visitor, VisitedPage
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()
from django.conf import settings
from nicechat.analytics import event_track, VISITOR_FIRST_MESSAGE, VISITOR_PURCHASE_STEP_1, VISITOR_PURCHASE_STEP_2
from nicechat.utils.callback import make_callback, EVENT_FIRST_MESSAGE
from django.core.cache import caches, InvalidCacheBackendError
from django.utils import timezone
from apps.chat.utils import move_chat_to_archive, fill_order_by_product, fill_info_by_tags
from apps.chat.serializers import MessageSerializer
from apps.website.models import Form, Product, Website
from apps.visitors.models import FormResult, Client
from apps.website.utils.form import form_validator, final_form_validator
from django.utils.translation import gettext_lazy as _
from rest_framework.generics import get_object_or_404
from apps.website.serializers import FormForWidgetSerializer
from django.template.loader import get_template
from django.core.mail import EmailMultiAlternatives, EmailMessage


def visitor_chat_packet(dt, kwargs, ch, session):
    if isinstance(dt, dict):

        _type = dt.get('type', None)
        if _type == WebsocketEventTypes.FRONTEND_CHAT_SET_DELIVERED:
            ids = dt.get('ids', [])
            chat_id = kwargs.get('chat_id', 0)
            if isinstance(ids, list) and chat_id:
                for msg in Message.objects.filter(chat_id=chat_id, pk__in=ids):
                    msg.state = Message.STATE_DELIVERED
                    msg.save(update_fields=('state',))
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_NOTIFY:
            chat_ws_notify(dt, session, kwargs)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_CLOSE:
            chat_ws_close(dt, session, kwargs, ch.send)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_RATE:
            chat_ws_rate(dt, session, kwargs, ch)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_SET_TYPE:
            chat_ws_set_type(dt, session, kwargs, ch)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_MESSAGE:
            chat_ws_message(dt, session, kwargs, ch.send)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_PUSHFORM:
            chat_ws_pushform(dt, session, kwargs, ch)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_FINAL:
            chat_ws_finform(dt, session, kwargs, ch)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_PURCHASE:
            chat_ws_purchase(dt, session, kwargs, ch)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_TYPING_START:
            chat_typing_start(dt, session, kwargs)
        elif _type == WebsocketEventTypes.FRONTEND_CHAT_WS_TYPING_END:
            chat_typing_end(session, kwargs)


def chat_ws_close(dt, session, kwargs, ch_send):
    chat_id = kwargs.get('chat_id', 0)
    if chat_id:
        chat = Chat.objects.filter(pk=chat_id).first()
        if chat.ts_close is None and not chat.is_serialized:
            chat.ts_close = timezone.now()
            chat.is_open = False
            chat.put_tech_message("~!COMPLETE:")
            chat.save(update_fields=['ts_close', 'is_open'])
            if chat.visitor:
                chat.visitor.state = Visitor.STATE_ONLINE
                chat.visitor.ts_close = None
                chat.visitor.save(update_fields=('state', 'ts_close',))
            chat.visitor = None
            chat.save(update_fields=['visitor'])

            if ch_send:
                ch_send(text_data= json.dumps({
                    'type': WebsocketEventTypes.FRONTEND_CHAT_CLOSE, 'status': 'ok'
                }))

            from apps.chat.tasks import move_chat_to_archive_delayed
            move_chat_to_archive_delayed.apply_async(args=[chat.pk], kwargs={'force_del': True}, countdown=1)

            return
    ch_send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_CLOSE, 'status': 'fail'
    }))


def parse_message(msg):
    try:
        if msg and msg.data:
            setattr(msg, 'data', json.loads(msg.data))
    except ValueError:
        pass
    return msg


def chat_ws_rate(dt, session, kwargs, ch):
    chat_id = kwargs.get('chat_id', 0)
    if chat_id:
        chat = Chat.objects.filter(pk=chat_id).first()
        useful = dt.get('useful', None)
        email = dt.get('email', None)
        no_send = dt.get('no_send', False)
        sent_mode = dt.get('mode', "save")
        try:
            vote = int(dt.get('vote', '0'))
        except ValueError:
            vote = 0
        if useful:
            chat.of_client_description = _("Is helpful: %s") % useful
        if vote:
            chat.of_client_score = vote
        if useful or vote:
            print ('>>>', chat.__dict__)
            chat.save(update_fields=('of_client_description', 'of_client_score',))

        if (email and re.match(r"^[a-zA-Z0-9_.-]+@[a-zA-Z0-9-]+.[a-zA-Z0-9-.]+$", email) and not no_send) or sent_mode == "save3":
            template = get_template('chat/email/chat_history.html')
            text = get_template('chat/email/chat_history.txt')
            subject = get_template('chat/email/chat_history.subject')
            # headers = ch.scope['headers']
            # headers = {v[0].decode(): v[1].decode() for v in headers if isinstance(headers, list) and len(headers) > 2}
            messages = []
            m_ts = None

            for msg in chat.message_set.all().order_by('id'):
                _m_ts = msg.timestamp.date()
                if m_ts != _m_ts:
                    messages.append({
                        'type': 'dateline',
                        'timestamp': _m_ts
                    })
                    m_ts = _m_ts
                messages.append(parse_message(msg))

            d = {
                'num': chat.pk,
                'site': chat.website.title,
                'messages': messages
            }
            text_content = text.render(d)
            html_content = template.render(d)
            subject_content = subject.render(d)

            if sent_mode == "save2":
                msg = EmailMessage(
                    subject_content, "Your chat history:",
                    settings.DEFAULT_FROM_EMAIL,
                    [email])

                msg.attach("chat-history.txt", text_content)
                if msg.send(fail_silently=True):
                    ch.send(text_data= json.dumps({
                        'type': WebsocketEventTypes.FRONTEND_CHAT_RATE, 'status': 'ok', 'email': 'ok', "mode": "astxt"
                    }))
                    return
            elif sent_mode == "save3":
                ch.send(text_data= json.dumps({
                    'type': WebsocketEventTypes.FRONTEND_CHAT_RATE, 'status': 'ok', 'email': 'ok', "mode": "download",
                    "content": text_content
                }))
                return
    
            msg = EmailMultiAlternatives(
                subject_content, text_content,
                settings.DEFAULT_FROM_EMAIL,
                [email])

            msg.attach_alternative(html_content, "text/html")
            if msg.send(fail_silently=True):
                ch.send(text_data= json.dumps({
                    'type': WebsocketEventTypes.FRONTEND_CHAT_RATE, 'status': 'ok', 'email': 'ok', "mode": "asemail"
                }))
                return

        else:
            stt = 'ok' if useful or vote else 'noact'
            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_RATE, 'status': stt, 'email': 'no'
            }))
            return


    ch.send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_RATE, 'status': 'fail'
    }))


def chat_ws_set_type(dt, session, kwargs, ch):
    msg = Message.objects.filter(pk=dt.get('message_id')).first()
    if 'data_type' in dt and msg:
        data_type = dt.get('data_type')
        if (msg.data_type == Message.DATA_SUGGEST and data_type in (Message.DATA_SUGGEST_CANCEL,
                                                                    Message.DATA_SUGGEST_DONE)) or \
                (msg.data_type == Message.DATA_FORM and data_type in (Message.DATA_FORM_CANCEL,)):
            msg.data_type = data_type
            msg.save(update_fields=('data_type',))
            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_SET_TYPE, 'status': 'ok'
            }))
            return

    ch.send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_SET_TYPE, 'status': 'fail'
    }))


def chat_ws_message(dt, session, kwargs, ch_send):
    try:
        chat_id = int(kwargs.get('chat_id', '0'))
    except ValueError:
        chat_id = None

    chat_obj = Chat.objects.filter(id=chat_id).first()

    # TODO Stas M. add in widget message resend button
    if chat_obj and chat_obj.is_open and not chat_obj.is_serialized:

        save_kwargs = dt

        save_kwargs['chat_id'] = chat_id
        save_kwargs['type'] = Message.TYPE_CLIENT
        x_tag = dt.get('x_tag')
        if 'x_tag' in save_kwargs:
            del save_kwargs['x_tag']

        if dt.get('as_event', False):
            save_kwargs['type'] = Message.TYPE_EVENT_CLI

        if settings.WIDGET_SESSION_TOKEN_KEY:
            dt['sender_id'] = dt.get('sender_id', '').split('.')[0]

        if 'as_event' in dt.keys():
            del dt['as_event']
        if chat_obj.client_id != dt.get('sender_id'):
            ch_send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_MESSAGE, 'status': 'fail',
                'x_tag': x_tag, "text": 401
            }))
            return
        if dt.get('data_type') == Message.DATA_FORM_FILLED:
            ch_send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_MESSAGE, 'status': 'fail', 'x_tag': x_tag,
                'text': _("User can not write form data without validation")
            }))
            return
        if dt.get('data_type') == Message.DATA_ANSWERS_REPLY:
            if isinstance(dt.get('data', None), str):
                try:
                    data = json.loads(dt.get('data', '{}'))
                except json.JSONDecodeError:
                    data = {}
            else:
                data = {}                  
            msg_p = chat_obj.message_set.filter(pk=data.get('close_ans', 0), data_type=Message.DATA_ANSWERS).first()
            try:
                data = json.loads(msg_p.data) if msg_p.data else {}
            except json.JSONDecodeError:
                data = {}  
            data['done'] = True
            msg_p.data = json.dumps(data)
            msg_p.save()            

        msg = Message.objects.create(**save_kwargs)
        if msg:      
            ch_send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_MESSAGE, 'status': 'ok', 'x_tag': x_tag,
                'message_id': msg.pk
            }))
            ch_send(text_data=json.dumps({
                "type": WebsocketEventTypes.FRONTEND_CHAT_ID_UPDATE,
                "new_id": msg.pk,
                "x_id": x_tag,
                "chat": msg.chat_id
                }))

            return msg.pk

    ch_send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_MESSAGE, 'status': 'fail', 'x_tag': x_tag
    }))


def chat_ws_pushform(dt, session, kwargs, ch):
    site_id = kwargs.get('site_id', 0)
    chat_id = kwargs.get('chat_id', 0)
    chat = Chat.objects.filter(client_id=session,
                               website_id=site_id, pk=chat_id, is_open=True).first()
    j_data = dt.get('data')
    data = {}
    to_msg_id = None
    if chat:
        to_message = Message.objects.filter(pk=j_data.get('to_message'), chat=chat, data_type=Message.DATA_FORM).first()
        if to_message:
            to_msg_id = to_message.pk
            data, is_valid, tags = form_validator(to_message, j_data)

            if is_valid:
                Message.objects.create(
                    chat=chat,
                    sender_id=chat.client_id,
                    sender_name=dt.get('sender_name'),
                    text=to_message.text,
                    data=json.dumps(data),
                    hidden=tags,
                    data_type=Message.DATA_FORM_FILLED,
                    type=Message.TYPE_EVENT_CLI
                )
                to_message.data_type = Message.DATA_FORM_DONE
                to_message.save(update_fields=['data_type'])

                form_id = None
                correspond = None
                try:
                    if to_message.hidden:
                        hidden = json.loads(to_message.hidden)
                        if isinstance(hidden, dict):
                            form_id = hidden.get('form_id')
                            correspond = hidden.get('correspond')
                except ValueError:
                    pass

                if tags:
                    fill_info_by_tags(tags, chat)
                
                client = Client.objects.filter(client_id = chat.client_id).first()

                if form_id and client:
                    FormResult.objects.create(
                        website_id=site_id,
                        chat_id=chat.pk,
                        form_id=form_id,
                        label=to_message.text,
                        result=data,
                        client=client
                    )

                product_id = j_data.get('PRODUCTS')

                if product_id:
                    event_track(VISITOR_PURCHASE_STEP_2, params={
                        'chatId': chat.id,
                        'siteId': site_id,
                        'clientID': chat.client_id,
                        'productId': product_id
                    })

                    fill_order_by_product(data, chat, product_id)

                if correspond:
                    suggest_msg = Message.objects.filter(pk=correspond, data_type=Message.DATA_SUGGEST).first()
                    if suggest_msg:
                        suggest_msg.data_type = Message.DATA_SUGGEST_DONE
                        suggest_msg.save(update_fields=['data_type'])
                try:
                    mode = json.loads(to_message.data).get('mode', None)
                except (ValueError, IndexError):
                    mode = None
                if mode == Form.MODE_SITE_CATCH_FORM:
                    template = get_template('chat/email/unanswered.html')
                    text = get_template('chat/email/unanswered.txt')
                    subject = get_template('chat/email/unanswered.subject')
                    context = {
                        'data': data,
                        'site': chat.website,
                        'chat': chat
                    }
                    text_content = text.render(context)
                    html_content = template.render(context)
                    subject_content = subject.render(context)
                    emails = [user.email for user in chat.website.users.all()]

                    msg = EmailMultiAlternatives(
                        subject_content, text_content,
                        settings.DEFAULT_FROM_EMAIL,
                        emails)

                    msg.attach_alternative(html_content, "text/html")
                    msg.send(fail_silently=True)

                ch.send(text_data= json.dumps({
                    'type': WebsocketEventTypes.FRONTEND_CHAT_PUSHFORM, 'status': 'ok', 'to_message': to_msg_id
                }))
                return
    ch.send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_PUSHFORM, 'status': 'fail', 'data': data, 'to_message': to_msg_id
    }))


def visitor_send_finform(chat, ch):
    final_form = chat.website.props.get('final_form', 'std')

    if final_form != 'std':
        form = Form.objects.filter(website=chat.website, mode=Form.MODE_SITE_FINAL_FORM).first()
        if form and form:
            data = dict(FormForWidgetSerializer(form).to_representation(form))

            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_FINAL,
                'mode': 'form',
                'label': form.label,
                'id': 'final',
                'data': json.dumps(data)
            }))
            return
    
    ch.send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_WS_FINAL,
        'mode': 'std'
    }))


def chat_ws_finform(dt, session, kwargs, ch):
    site_id = kwargs.get('site_id', 0)
    chat_id = kwargs.get('chat_id', 0)
    query_data = dt.get('data')
    md = Form.MODE_SITE_FINAL_FORM
    form = Form.objects.filter(website=site_id, mode=md).first()
    website = Website.objects.filter(pk=site_id).first()
    chat = Chat.objects.filter(client_id=session, website_id=site_id, pk=chat_id, is_open=True).first()

    if chat and form and website:
        form_data = dict(FormForWidgetSerializer(form).to_representation(form))

        data, is_valid, tags = final_form_validator(query_data, form_data)

        if is_valid:

            Message.objects.create(
                chat=chat,
                sender_id=chat.client_id,
                sender_name=dt.get('sender_name', 'Guest'),
                text=form.label,
                data=json.dumps(data),
                hidden=tags,
                data_type=Message.DATA_FORM_FILLED,
                type=Message.TYPE_EVENT_CLI
            )
            cl = Client.objects.filter(client_id=chat.client_id).first()
            FormResult.objects.create(
                website_id=site_id,
                chat_id=chat.pk,
                form_id=form.pk,
                label=form.label,
                result=data,
                client=cl
            )

            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_FINAL,
                'cool': 'yes'
            }))
            chat_ws_close(dt, session, kwargs, ch)

            try:
                vote = int(tags.get('SCORE', '0'))
            except ValueError:
                vote = 0
            chat.of_client_score = vote
            chat.save(update_fields=('of_client_score',))
                
        else:
            # is form not valid - resend it
            print(form_data)
            print(query_data)
            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_FINAL,
                'label': form.label,
                'id': 'final',
                'data': json.dumps(form_data),
                'errors': json.dumps(data)
            }))


def chat_ws_purchase(dt, session, kwargs, ch):
    # dt product_id, message_id
    chat_id = kwargs.get('chat_id', 0)
    site_id = kwargs.get('site_id', 0)
    chat = Chat.objects.filter(client_id=session, website_id=site_id, pk=chat_id, is_open=True).first()
    product_id = dt.get('product_id')
    if chat and product_id:
        form = Form.objects.filter(website=chat.website, mode=Form.MODE_SITE_PURCHASE_FORM).first()
        if form and form:
            data = dict(FormForWidgetSerializer(form).to_representation(form))
            if not isinstance(product_id, list):
                product_id = [product_id]

            try:
                info = {j['id']: j for j in json.loads(dt.get('info', '[]'))}
            except ValueError:
                info = {}

            products = []
            for prod in Product.objects.filter(original_id__in=product_id, website_id=site_id):
                t_info = info.get(prod.original_id, {})
                products.append({
                        'pk': prod.pk,
                        'id': prod.original_id,
                        'name': prod.title,
                        'image': f'{settings.MEDIA_URL}{prod.image_url}' if prod.image_url else None,
                        'currency': t_info.get('currency', prod.currency),
                        'qty': t_info.get('qty', 1),
                        'url': prod.url,
                        'price': t_info.get('price', float(prod.price))
                    })

            data['PRODUCTS'] = products

            Message.objects.create(
                chat=chat,
                sender_id=0,
                sender_name=chat.website.title,
                text=form.label,
                data=json.dumps(data),
                data_type=Message.DATA_FORM,
                hidden=json.dumps({'form_id': form.pk, 'correspond': dt.get('message_id')}),
                type=Message.TYPE_OPERATOR
            )

            event_track(VISITOR_PURCHASE_STEP_1, params={
                'chatId': chat.id,
                'siteId': site_id,
                'clientID': chat.client_id,
                'productId': products
            })

            ch.send(text_data= json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_PURCHASE, 'status': 'ok'
            }))
            return

    ch.send(text_data= json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_PURCHASE, 'status': 'fail'
    }))


def chat_ws_notify(dt, session, kwargs):
    _call = dt.get('call', None)
    if _call == 'ClientFirstMessage':
        # remove if ... for track agent always (with "if" - only if agent create chat)
        agent_id = dt.get('agentId', None) if dt.get('chatInitedBy', None) == 'agent' else None
        event_track(VISITOR_FIRST_MESSAGE, user_id=agent_id, params={
            'clientID': session,
            'siteId': kwargs.get('site_id', 0),
            'chatId': kwargs.get('chat_id', 0),
            'chatInitedBy': dt.get('chatInitedBy', None)
        })
        chat_id = kwargs.get('chat_id', 0)
        if chat_id:
            chat = Chat.objects.filter(pk=chat_id).first()
            if chat:
                make_callback.apply_async(args=(EVENT_FIRST_MESSAGE, chat.website_id, "Chat", chat.pk,),
                                          countdown=2)


def chat_typing_start(dt, session, kwargs):
    chat_id = kwargs.get('chat_id', 0)
    token = "chat-typing-%s-client" % chat_id
    token_s = "chat-typing-s-%s-client" % session
    try:
        cache = caches['default']
        cache.set(token, dt.get('txt', '...'), 20)
        cache.set(token_s, dt.get('txt', '...'), 20)
    except InvalidCacheBackendError:
        pass
    async_to_sync(channel_layer.group_send)('chat-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
        'type': WebsocketEventTypes.BACKEND_SITE_START_TYPING,
        'name': 'client',
        'id': 0,
        'text': dt.get('txt', '...')
    })})
    site_id = kwargs.get('site_id', 0),
    if site_id:
        async_to_sync(channel_layer.group_send)("adm-website_mon-%s" % site_id, {"type": "websocket.send", 'text':  json.dumps({
            'type': WebsocketEventTypes.BACKEND_SITE_START_TYPING,
            'data': {
                'state': 'T',
                'chat_id': chat_id,
                'session': session,
                'type': WebsocketEventTypes.BACKEND_SITE_START_TYPING,
            }
        })})


def chat_typing_end(session, kwargs):
    chat_id = kwargs.get('chat_id', 0),
    token = "chat-typing-%s-client" % chat_id
    token_s = "chat-typing-s-%s-client" % session
    try:
        cache = caches['default']
        cache.delete(token)
        cache.delete(token_s)
    except InvalidCacheBackendError:
        pass
    async_to_sync(channel_layer.group_send)('chat-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
        'type': WebsocketEventTypes.BACKEND_SITE_STOP_TYPING,
        'name': 'client',
        'id': 0,
    })})
    site_id = kwargs.get('site_id', 0),
    if site_id:
        async_to_sync(channel_layer.group_send)("adm-website_mon-%s" % site_id, {"type": "websocket.send", 'text':  json.dumps({
            'type': WebsocketEventTypes.BACKEND_SITE_STOP_TYPING,
            'data': {
                'state': 'F',
                'session': session,
                'chat_id': '%s' % chat_id,
            }
        })})


def visitor_chat_close(kwargs, session):
    chat_id = kwargs.get('chat_id', 0),
    token = "chat-typing-%s-client" % chat_id
    token_s = "chat-typing-s-%s-client" % session
    try:
        cache = caches['default']
        cache.delete(token)
        cache.delete(token_s)
    except InvalidCacheBackendError:
        pass
    async_to_sync(channel_layer.group_send)('chat-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
        'type': WebsocketEventTypes.BACKEND_SITE_STOP_TYPING,
        'name': 'client',
        'id': 0,
    })})
    site_id = kwargs.get('site_id', 0),
    if site_id:
        async_to_sync(channel_layer.group_send)("adm-website_mon-%s" % site_id, {"type": "websocket.send", 'text':  json.dumps({
            'type': WebsocketEventTypes.BACKEND_SITE_STOP_TYPING,
            'data': {
                'state': 'T',
                'session': session,
                'chat_id': '%s' % chat_id,
            }
        })})


def visitor_send_typing(chat, ch):
    try:
        cache = caches['default']
    except InvalidCacheBackendError:
        cache = None
    if cache:
        typing = []
        for user in chat.operators_current.all():
            text = cache.get('chat-typing-%s-%s' % (chat.pk, user.pk), None)
            if text:
                typing.append(user.get_full_name or str(user))

        ch.send(text_data= json.dumps({
            'type': WebsocketEventTypes.FRONTEND_CHAT_WS_TYPING_DUMP,
            'data': typing
        }))


def visitor_send_messages(chat, ch):
    v_q = chat.message_set.all()
    ch.send(text_data=json.dumps({
        'type': WebsocketEventTypes.FRONTEND_CHAT_MESSAGES_LIST,
        'data': MessageSerializer(v_q, many=True).to_representation(v_q)
    }))


def chat_call_req(dt, session, kwargs):
    pass


def chat_call_acc(dt, session, kwargs):
    pass


def chat_call_stop(dt, session, kwargs):
    pass


def chat_call_ping(dt, session, kwargs):
    pass

