# -*- coding: utf-8 -*-

import csv
from datetime import datetime
import os
import re
import sys
import json
import base64
import random
import tempfile
import hashlib

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.db.models import Q
from django.db import IntegrityError
from django.http import HttpResponse, Http404
from django.core.exceptions import PermissionDenied
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from rest_framework import viewsets
from rest_framework import exceptions
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated, AllowAny
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework import status, serializers

from nicechat.utils.images import image_post_create, image_post_processing
from nicechat.utils.callback import EVENT_CHOICES

from nicechat.analytics import event_track, QA_ADDED, INSTRUCTIONS_SENT, FORM_ADDED, PRODUCT_ADDED
from ..billing.utils import FeatureAccessAPIView, HasAccessToFeatures
from .utils.parsers import do_process_import
from .tasks import process_import
from .models import Website, QuickReply, ProductsCategory, Product, Form, FormRow, Status, \
    AccessToken, Callbacks, WidgetTheme, WidgetSettings, DomainList
from .serializers import WebsiteSerializer, QuickReplySerializer, ProductsCategorySerializer, \
    ProductSerializer, AgentSerializer, FormSerializer, FormRowSerializer, FormForWidgetSerializer, \
    StatusSerializer, CallbacksSerializer, WidgetThemesSerializer, WidgetSettingsSerializer, \
    DomainListSerializer, AccessTokenSerializer, AccessTokenSerializerList, WidgetIconSerializer

TESTING = sys.argv[1:2] == ['test']

TESTING = sys.argv[1:2] == ['test']


class FilterByWebsiteMixin(object):
    # todo move to global utils package
    def get_queryset(self):
        self.queryset = self.queryset.filter(website=self.request.current_site)

        return super(FilterByWebsiteMixin, self).get_queryset()


class FilterByWebsiteOrCompanyMixin(object):
    def get_queryset(self):
        queryset = super(FilterByWebsiteOrCompanyMixin, self).get_queryset()

        if self.request.user and self.request.user.is_authenticated and \
                (self.request.user.own_company or self.request.user.is_superuser):
            return self.queryset.filter(website=self.request.current_site)
        elif self.request.user and self.request.user.is_authenticated:
            return queryset.none()
        else:
            raise PermissionDenied()


class LargeResultsSetPagination(PageNumberPagination):
    page_size = 1000
    page_size_query_param = 'page_size'
    max_page_size = 10000


class DomainListViewSet(viewsets.ModelViewSet):
    queryset = DomainList.objects.all()
    serializer_class = DomainListSerializer
    permission_classes = [IsAuthenticated]
    SIMPLE_REG_CHECK = re.compile('\*\.([\w\d\.]{0,8}|[\w\d]+)$')

    def get_queryset(self):
        if not self.request.user.is_superuser:
            return self.queryset.filter(website__company=self.request.user.company, website_id=self.kwargs['site_id'])
        else:
            return self.queryset.filter(website_id=self.kwargs['site_id'])

    def perform_create(self, serializer):
        if not self.request.user.is_superuser:
            ws = Website.objects.get(company=self.request.user.company, pk=self.kwargs['site_id'])
        else:
            ws = Website.objects.get(pk=self.kwargs['site_id'])

        kwa = {
            'website': ws
        }    
        kind = serializer.validated_data.get('kind')
        if not kind or kind in [DomainList.DOMAIN_WILDCARD,  DomainList.DOMAIN_REGULAR]:
            if '*' in serializer.validated_data['domain']:
                kwa['kind'] = DomainList.DOMAIN_WILDCARD
                if self.SIMPLE_REG_CHECK.match(serializer.validated_data['domain']):
                   raise exceptions.ValidationError({'domain': 'Domain mask is to wide. Please set more specific'}) 
            else:
                kwa['kind'] = DomainList.DOMAIN_REGULAR

        serializer.save(**kwa)

    def perform_update(self, serializer):
        if serializer.validated_data.get('kind') and serializer.validated_data.get('kind') != serializer.instance.kind:
            raise exceptions.ValidationError({'kind': 'You cannot change kind of existed records'})

        if serializer.instance.kind == DomainList.DOMAIN_WILDCARD and not '*' in serializer.validated_data['domain']:
            raise exceptions.ValidationError({'domain': 'Wildcard must contain asterisk'})

        if serializer.instance.kind == DomainList.DOMAIN_WILDCARD and self.SIMPLE_REG_CHECK.match(serializer.validated_data['domain']):
            raise exceptions.ValidationError({'domain': 'Domain mask is to wide. Please set more specific'}) 

        serializer.save()


class WebsiteViewSet(viewsets.ModelViewSet):
    queryset = Website.objects.all().order_by('id')
    serializer_class = WebsiteSerializer

    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        if self.action == 'get_current':
            return self.queryset.filter(users=self.request.user)
        else:
            if not self.request.user.is_superuser:
                return self.queryset.filter(company=self.request.user.company)
            else:
                return super(WebsiteViewSet, self).get_queryset()

    def get_serializer_class(self):
        if self.action == 'get_current':
            return WebsiteSerializer

        return super(WebsiteViewSet, self).get_serializer_class()

    @action(detail=False, methods=['get'], permission_classes=[IsAuthenticated])
    def get_current(self, request, **kwargs):
        return self.list(request)

    @action(detail=True, methods=['get'],  permission_classes=[AllowAny])
    def test(self, request, **kwargs):
        return HttpResponse('ok')
        from django.template.loader import get_template
        messages = [
            {
                'text': 'text',
                'sender_name': 'Sender',
                'timestamp': datetime.now(),
                't': '11:22',
                'd': '22 Jul'
            },{
                'text': 'text1',
                'sender_name': '',
                'timestamp': datetime.now(),
                't': '',
                'd': ''
            },{
                'text': 'text2',
                'sender_name': 'Sender',
                'timestamp': datetime.now(),
                't': '11:23',
                'd': '23 Jul'
            }
        ]
        cr_key = 'qweqweqweqweqwe'
        context = {
            'chat_name': 'Macari’s',
            'chat_mode': 'P2P',
            'key': cr_key,
            'messages': messages
        }
        template = get_template('chat/email/notification.html')
        html_content = template.render(context)
        text = get_template('chat/email/notification.txt')
        subject = get_template('chat/email/notification.subject')
        text_content = text.render(context)
        subject_content = subject.render(context)

        reply_to = settings.NOTIFICATOR_REPLY if settings.NOTIFICATOR_REPLY else settings.DEFAULT_FROM_EMAIL

        headers = {
            'Reply-To': reply_to.replace('@', '+%s@' % cr_key) if settings.NOTIFICATOR_PLUS_REPLY else reply_to, 
            'X-Chat-ID': cr_key
        }
        msg = EmailMultiAlternatives(
            subject_content, text_content,
            settings.DEFAULT_FROM_EMAIL,
            ['tmp@silverice.biz'], headers=headers)
        msg.attach_alternative(html_content, "text/html")
        try:
            msg.send()
            return HttpResponse(html_content)
        except Exception as ex:
            return HttpResponse(str(ex))

    @action(detail=True, methods=['post'])
    def open_token(self, request, **kwargs):
        site = self.get_object()
        text = base64.b64encode(request.data.get('text').encode('utf8'), altchars=b"_.").rstrip(b"=").decode()
        token = base64.b64encode(hashlib.sha256(("%s-%s-%s" % (site.pk, text,
                                             settings.AUTO_OPEN_LINK_SECRET)).encode('utf8')).digest(), altchars=b"_.").rstrip(b"=").decode()
        return Response(data="#nc-auto-open=c-%s-%s" % (text, token), status=status.HTTP_200_OK)

    @action(detail=True, methods=['post'], permission_classes=[IsAuthenticated])
    def icon(self, request, pk, **kwargs):
        if len(request.FILES['file']) > 0:
            filename = request.FILES['file'].name
            self.get_object().icon.save(filename, request.FILES['file'].file, True)
        return Response(status=204)

    @action(detail=True, methods=['get'], permission_classes=[AllowAny])
    def site_logo(self, request, *args, **kwargs):
        website = Website.objects.filter(pk=kwargs.get('pk', None)).first()
        if website:
            icon = website.icon_thumbs().get('thumb', None)
            if icon and icon.exists():
                return Response(status=status.HTTP_302_FOUND, headers={'Location': icon.url})
        return Response(status=status.HTTP_302_FOUND, headers={'Location': '/media/img/logo-notify.png'})

    @action(detail=False, methods=['get'], permission_classes=[IsAuthenticated])
    def def_site_logo(self, request, *args, **kwargs):
        return Response(status=status.HTTP_301_MOVED_PERMANENTLY, headers={'Location': '/media/img/logo-notify.png'})

    @action(detail=True, methods=['post'], permission_classes=[IsAuthenticated])
    def add_agents(self, request, pk, **kwargs):
        ids = []
        if len(request.data.get('ids')) > 0:
            ids = request.data.get('ids')
        website = self.get_object()
        try:
            [website.users.add(get_user_model().objects.get(id=uid)) for uid in ids]
        except ObjectDoesNotExist:
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            data={'message': 'The user you are trying to add does not exist'})
        return Response(status=status.HTTP_204_NO_CONTENT)

    def create(self, request, *args, **kwargs):
        result = super(WebsiteViewSet, self).create(request, *args, **kwargs)
        from apps.balancer.models import ArbiterNode, ArbiterNodeWebsite
        tgt = ArbiterNode.objects.filter(for_new=True).values_list('pk', flat=True)
        if tgt:
            tgt_pk = random.choice(tgt)
            tgt = ArbiterNode.objects.get(pk=tgt_pk)
        else:
            tgt = None
        if not tgt and not settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
            raise serializers.ValidationError("Not exists active worker-node for new websites")

        website = Website.objects.get(pk=result.data['id'])

        if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
            website.users.add(self.request.user)
            return result

        ArbiterNodeWebsite.objects.create(
            node=tgt,
            website=website
        )
        from apps.balancer.tasks import sync_website, sync_user
        sync_website(website)
        website.users.add(self.request.user)
        sync_user(self.request.user)
        return result

class WebsiteTokensViewSet(viewsets.ModelViewSet):
    queryset = AccessToken.objects.all()
    serializer_class = AccessTokenSerializer
    pagination_class = None
    permission_classes = [IsAuthenticated]


    def get_object(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()

        pk = self.kwargs.get('pk')
        if pk is not None:
            queryset = queryset.filter(token=pk)

        try:
            obj = queryset.get()
        except queryset.model.DoesNotExist:
            raise Http404(
                _("No %(verbose_name)s found matching the query")
                % {"verbose_name": queryset.model._meta.verbose_name}
            )
        return obj

    def get_queryset(self):
        if not self.request.user.is_superuser:
            return AccessToken.objects.filter(website_id=self.kwargs['site_id'], 
                                              website__company=self.request.user.company)
        else:
            return AccessToken.objects.filter(website_id=self.kwargs['site_id'])

    def get_serializer_class(self):
        if self.action == 'list':
            return AccessTokenSerializerList
        else:
            return AccessTokenSerializer

    @action(detail=False, methods=['post'], permission_classes=[IsAuthenticated])
    def add(self, request, **kwargs):
        alphabet = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        len_alphabet = len(alphabet)
        alphabet_upper = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        len_alphabet_upper = len(alphabet_upper)

        if not self.request.user.is_superuser:
            ws =  Website.objects.filter(pk=self.kwargs['site_id'], 
                                              company=self.request.user.company).first()
        else:
            ws = Website.objects.filter(pk=self.kwargs['site_id']).first()

        if not ws:
            return Response(status=status.HTTP_404_NOT_FOUND)

        at = None
        for t in range(1, 10):
            try:
                token = u''.join([alphabet_upper[random.randrange(len_alphabet_upper)] for j in range(16)])
                secret = u''.join([alphabet[random.randrange(len_alphabet)] for j in range(32)])
                at = AccessToken.objects.create(
                    website=ws,
                    token=token,
                    secret=secret,
                    gen_timestamp=timezone.now()
                )
                break
            except IntegrityError:
                continue
        if at:
            return Response(status=status.HTTP_201_CREATED, data={
                'token': at.token,
                'secret': at.secret,
                'timestamp': str(at.gen_timestamp),
                "isNew": True
            })
        else:
            return Response(status=status.HTTP_204_NO_CONTENT)


class QuickReplyViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    queryset = QuickReply.objects.all().order_by('text')
    serializer_class = QuickReplySerializer
    pagination_class = LargeResultsSetPagination
    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        self.queryset = self.queryset.filter(Q(is_private=False) | Q(created_by=self.request.user))

        search_value = self.request.query_params.get('text')

        if search_value is not None:
            self.queryset = self.queryset.filter(text__icontains=search_value)

        return super(QuickReplyViewSet, self).get_queryset()

    def perform_create(self, serializer):
        """Always set a current site on created new object"""
        if self.get_queryset().all().count() == 0:
            event_track(QA_ADDED, self.request.user.id, {
                'siteId': self.request.current_site.id,
                'companyID': self.request.user.company_id,
                'companyName': self.request.user.company.label,
            })
        serializer.save(website=self.request.current_site)

    @action(detail=False, methods=['DELETE'])
    def delete_replies(self, request, **kwargs):
        ids = self.request.query_params.getlist('ids')
        if not ids:
            return Response(status=status.HTTP_404_NOT_FOUND)
        else:
            self.queryset.filter(pk__in=ids, created_by=self.request.user).delete()
            self.queryset.filter(pk__in=ids, created_by__isnull=True).delete()
            return Response(status=status.HTTP_204_NO_CONTENT)


class StatusViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    queryset = Status.objects.all()
    serializer_class = StatusSerializer
    permission_classes = [IsAuthenticated]

    @action(detail=False, methods=['post'], permission_classes=[IsAuthenticated])
    def enable_interface(self, request, **kwargs):
        if not request.current_site:
            raise exceptions.ValidationError("You have not access to force unblock for website")
        Status.objects.update_or_create(website=request.current_site,
                                        defaults={'server_code': 0,
                                                  'allow_site_access': True})
        return Response(status=status.HTTP_202_ACCEPTED)


class ProductsCategoryViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    queryset = ProductsCategory.objects.all()
    serializer_class = ProductsCategorySerializer


class ProductViewSet(FeatureAccessAPIView, FilterByWebsiteMixin, viewsets.ModelViewSet):
    queryset = Product.objects.all()
    serializer_class = ProductSerializer

    # todo permissions check for products suggestion (not available in free plan)

    def perform_create(self, serializer):
        """Always set a current site on created new object"""
        if self.get_queryset().all().count() == 0:
            event_track(PRODUCT_ADDED, self.request.user.id, {
                'siteId': self.request.current_site.id,
                'companyID': self.request.user.company_id,
                'companyName': self.request.user.company.label,
            })
        serializer.save()

    def get_permissions(self):
        if self.action == 'create':
            # permission for product creation/editing is required (not available in "free" plan)
            self.permission_classes = [IsAuthenticated, HasAccessToFeatures(['products_add'])]
        return super(self.__class__, self).get_permissions()

    def get_queryset(self):
        search_text = self.request.query_params.get('q')
        category_id = self.request.query_params.get('category_id')
        if category_id:
            category = ProductsCategory.objects.filter(pk=category_id)
            self.queryset = self.queryset.filter(category=category)
        if search_text:
            self.queryset = self.queryset.filter(
                Q(title__icontains=search_text) |
                Q(description__icontains=search_text)
            )

        return super(ProductViewSet, self).get_queryset()

    @action(detail=False, methods=['DELETE'])
    def products_delete(self, request, **kwargs):
        ids = self.request.query_params.getlist('ids')
        if not ids:
            return Response(status=status.HTTP_404_NOT_FOUND)
        else:
            self.queryset.filter(pk__in=ids).delete()
            return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=False, methods=['post'])
    def photo(self, request, **kwargs):

        try:
            if request.META.get('CONTENT_LENGTH'):
                size = int(request.META['CONTENT_LENGTH'])
            elif request.META.get('HTTP_X_FILE_SIZE'):
                size = int(request.META['HTTP_X_FILE_SIZE'])
            else:
                size = None
        except ValueError:
            size = None

        if size and size > settings.AGENT_ATTACH_LIMIT * 1048576:
            return Response(status=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE)

        if len(request.FILES['file']) > 0 and request.FILES['file'].file and hasattr(request.FILES['file'].file, 'getvalue'):
            image, image_url = image_post_create(self.request.current_site, request.FILES['file'].file.getvalue(),
                                                 request.FILES['file'].name)
            if image:
                image_post_processing(image)
                return Response(status=status.HTTP_200_OK,
                                data={'image_url': image_url})
        return Response(status=status.HTTP_400_BAD_REQUEST)


class AgentViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    serializer_class = AgentSerializer
    queryset = get_user_model().objects.filter()
    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        if self.request.query_params.get('free'):
            site_id = self.request.query_params.get('site_id')
            self.queryset = get_user_model().objects \
                .filter(company=self.request.user.company) \
                .exclude(website__in=[site_id])
            return self.queryset
        return super(AgentViewSet, self).get_queryset()

    # todo permission class for owner or admin
    def destroy(self, request, pk=None):
        instance = self.get_object()
        # todo company owner cannot be deleted from the site
        if instance != self.request.user:
            self.request.current_site.users.remove(instance)

            return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            return Response(status=status.HTTP_400_BAD_REQUEST,
                            data={'message': 'You can remove yourself from the website'})


class ImportCsvFileUpload(FeatureAccessAPIView, APIView):
    """
    Upload csv file for manual products uploading
    """
    permission_classes = [IsAuthenticated, HasAccessToFeatures(['products_import'])]

    def post(self, request, *args, **kwargs):
        request_file = request.FILES.get('file')
        if request_file:
            filename = request_file.name
            try:
                demo_rows = []
                file_content = request_file.file.read()
                dialect = csv.Sniffer().sniff(file_content)
                reader = csv.reader(file_content.splitlines(), dialect=dialect)
                for row in reader:
                    demo_rows.append(row)

                has_header = csv.Sniffer().has_header(file_content)
                return Response(status=status.HTTP_200_OK, data={
                    'has_header': has_header,
                    'rows': demo_rows[:4],
                    'delimeter': dialect.delimiter,
                    'quotechar': dialect.quotechar,
                    'filename': filename
                })
            except csv.Error:
                return Response(status=status.HTTP_400_BAD_REQUEST,
                                data={'message': _('Uploaded file is not a csv')})
        return Response(status=status.HTTP_400_BAD_REQUEST,
                        data={'message': _('There is no file uploaded')})


class ProductImport(FeatureAccessAPIView, APIView):
    """
    Import products from google merchant or plain csv
    """

    permission_classes = [IsAuthenticated, HasAccessToFeatures(['products_import'])]

    def post(self, request, *args, **kwargs):
        site_id = request.data.get('site_id')
        google_merchant_csv = request.data.get('google_merchant_csv')
        csv = request.data.get('csv')
        current_site = self.request.current_site

        if not settings.DEBUG and not TESTING:
            process_import.delay(site_id, request.user.pk, current_site.pk, request.data, csv=csv,
                                 google_merchant_csv=google_merchant_csv)
        else:
            do_process_import(site_id, request.user.pk, current_site.pk, request.data, csv=csv,
                              google_merchant_csv=google_merchant_csv)
        return Response(status=status.HTTP_204_NO_CONTENT)


class SendInstallInstructions(APIView):
    """
    Send chat installing instructions to the client's webmaster
    """

    def post(self, request, *args, **kwargs):
        email = request.data.get('email')
        message = request.data.get('message')
        if email:
            context = {}
            if message is not None:
                context.update({'message': message})
            text_content = render_to_string('website/email/installation_instructions.txt', context)
            html_content = render_to_string('website/email/installation_instructions.html', context)
            msg = EmailMultiAlternatives('Instructions for %s widget installation' % settings.BASE_CHAT_NAME,
                                         text_content,
                                         settings.DEFAULT_FROM_EMAIL,
                                         [email])
            event_track(INSTRUCTIONS_SENT, request.user.id, {
                'siteId': self.request.current_site.id,
                'email': email
            })
            msg.attach_alternative(html_content, "text/html")
            msg.send()
            return Response(status=status.HTTP_200_OK,
                            data={'message': 'The email was sent to your webmaster',
                                  'status': 'success'})
        return Response(status=status.HTTP_400_BAD_REQUEST,
                        data={'message': 'Please specify the email'})


class FormsViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    serializer_class = FormSerializer
    queryset = Form.objects.all()
    pagination_class = LargeResultsSetPagination
    permission_classes = [IsAuthenticated]

    def perform_create(self, serializer):

        access_write = self.request.data.get('access_write', Form.PERM_WRITE_AGENT)
        access_read = self.request.data.get('access_read', Form.PERM_READ_WEBSITE)

        if access_read == Form.PERM_READ_PUBLIC or access_write == Form.PERM_WRITE_SUPER:
            # nobody can make public forms over API
            raise PermissionDenied()

        if not (self.request.user and self.request.user.is_authenticated):
            raise PermissionDenied()

        save_kwargs = {
            'access_write': access_write,
            'access_read': access_read,
            'created_by': self.request.user,
            'company': self.request.user.company,
            'mode': Form.MODE_SITE_REGULAR_FORM
            # on API you can create only regular forms
        }

        if self.request.data.get('website'):
            if not Website.objects.filter(pk=self.request.data.get('website'),
                                          pk__in=self.request.user.website_set.values_list('pk', flat=True)).exists():
                # can not assign foreign Website
                save_kwargs['website'] = None
        if self.get_queryset().all().count() == 0:
            event_track(FORM_ADDED, self.request.user.id, {
                'siteId': self.request.current_site.id,
                'companyId': self.request.user.company_id,
                'companyName': self.request.user.company.label,
            })
        serializer.save(**save_kwargs)

    @action(detail=False, methods=['DELETE'], permission_classes=[IsAuthenticated])
    def list_delete(self, request, **kwargs):
        ids = self.request.query_params.getlist('ids')
        if not ids:
            return Response(status=status.HTTP_404_NOT_FOUND)
        else:
            self.queryset.filter(pk__in=ids).delete()
            return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=False, methods=['GET'])
    def fast(self, request, **kwargs):

        queryset = self.client_ready_forms()

        result = {item.pk: item.label for item in queryset}

        return Response(data=result)

    @action(detail=False, methods=['GET'])
    def params(self, request, **kwargs):

        access_write = {item[0]: item[1] for item in Form.WRITE_PERM_CHOICES
                        if
                        item[0] >= (Form.PERM_WRITE_ADMIN if self.request.user.own_company else Form.PERM_WRITE_AGENT)}
        access_read = {item[0]: item[1] for item in Form.READ_PERM_CHOICES if item[0] > Form.PERM_READ_PUBLIC}

        field_kinds = [
            {
                'id': field[0],
                'label': field[1],
                'have_choices': True if field[0] in FormRow.TYPES_REQUIRE_CHOICE else False
            }
            for field in FormRow.FIELD_TYPES
        ]

        result = {
            'access_write': access_write,
            'access_read': access_read,
            'field_kinds': field_kinds,
            'auto_fill_tags': FormRow.ALLOWED_AUTO_FILL_TAGS
        }

        return Response(data=result)

    def perform_destroy(self, instance):
        if instance.mode != Form.MODE_SITE_REGULAR_FORM:
            raise PermissionDenied
        super(FormsViewSet, self).perform_destroy(instance)

    @action(detail=True, methods=['GET', 'POST'])
    def open_token(self, request, **kwargs):
        if request.method == 'POST' and request.data.get('message'):
            form = self.get_object()
            text = base64.b64encode(request.data.get('message').encode('utf8'), altchars="_.").rstrip("=")
            token = base64.b64encode(hashlib.sha256("%s-%s-%s-%s" % (form.pk, form.website.domain, text,
                                                                  settings.AUTO_OPEN_LINK_SECRET)).digest(),
                                     altchars="_.").rstrip("=")
            return Response(data="#nc-auto-open=f-%s-%s-%s" % (form.pk, token, text), status=status.HTTP_200_OK)

        form = self.get_object()
        token = base64.b64encode(hashlib.sha256("%s-%s-%s" % (form.pk, form.website.domain,
                                             settings.AUTO_OPEN_LINK_SECRET)).digest(), altchars="_.").rstrip("=")
        return Response(data="nc-auto-open=f-%s-%s" % (form.pk, token), status=status.HTTP_200_OK)

    @action(detail=False, methods=['POST'])
    def put_form_to_chat(self, request, **kwargs):
        """
            POST
            :chat_id - The ID of Chat to which you want to send the form
            :form_id - ID of Form to send
        """

        from ..chat.models import Chat, Message

        queryset = self.client_ready_forms()
        form = queryset.filter(pk=self.request.data.get('form_id')).first()

        chat = Chat.objects.filter(pk=self.request.data.get('chat_id'),
                                   is_open=True,
                                   operators_current=self.request.user).first()

        if chat and form:
            data = dict(FormForWidgetSerializer(instance=form).to_representation(instance=form))

            Message.objects.create(
                chat=chat,
                sender_id=self.request.user.user_tag,
                sender_name=self.request.user.get_full_name or str(self.request.user),
                text=form.label,
                data=json.dumps(data),
                hidden=json.dumps({'form_id': form.pk}),
                data_type=Message.DATA_FORM,
                type=Message.TYPE_OPERATOR
            )

            return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            return Response(status=status.HTTP_406_NOT_ACCEPTABLE)

    def client_ready_forms(self):
        return self.queryset.filter(website=self.request.current_site,
                                    mode=Form.MODE_SITE_REGULAR_FORM, is_enabled=True)


class FormRowsViewSet(FilterByWebsiteOrCompanyMixin, viewsets.ModelViewSet):
    serializer_class = FormRowSerializer
    queryset = Form.objects.all()
    pagination_class = None
    permission_classes = [IsAuthenticated]

    def perform_create(self, serializer):
        form = self.get_queryset(False)
        save_kwargs = {'form': form}

        try:
            # params must be valid json
            params = self.request.data.get('params')
            if params:
                params = json.loads(params)
            else:
                save_kwargs['params'] = '{}'
                params = {}
        except ValueError:
            params = {}
            save_kwargs['params'] = '{}'

        if not self.request.data.get('order_num'):
            save_kwargs['order_num'] = form.formrow_set.count() + 1

        if (self.request.data.get('kind') == FormRow.FIELD_TYPE_HIDDEN or self.request.data.get('is_read_only')) \
                and not 'auto-fill-tag' in params.keys():
            raise exceptions.ValidationError("'auto-fill-tag' must be specified for hidden and r/o fields.")

        serializer.save(**save_kwargs)

    def perform_update(self, serializer):
        super(FormRowsViewSet, self).perform_update(serializer)

    def perform_destroy(self, instance):
        super(FormRowsViewSet, self).perform_destroy(instance)

    def get_queryset(self, return_rows=True):

        queryset = super(FormRowsViewSet, self).get_queryset()
        queryset = queryset.filter(pk=self.kwargs.get('form_id'))
        queryset = queryset.filter(website=self.request.current_site)
        queryset = queryset.first()

        if queryset:
            return queryset.formrow_set.all().order_by('order_num') if return_rows else queryset
        else:
            raise PermissionDenied()


class CallbacksViewSet(FilterByWebsiteMixin, viewsets.ModelViewSet):
    queryset = Callbacks.objects.all()
    serializer_class = CallbacksSerializer
    pagination_class = None
    permission_classes = [IsAuthenticated]

    def perform_create(self, serializer):
        serializer.save(website=self.request.current_site)

    @action(detail=False, methods=['GET'])
    def options(self, request, **kwargs):
        items = [
            {
                'id': event_id,
                'text': label,
                'value': event_id,
                'description': description
            } for event_id, label, description in EVENT_CHOICES]
        return Response(data=items)


class WidgetThemesViewSet(viewsets.ModelViewSet):
    queryset = WidgetTheme.objects.all()
    serializer_class = WidgetThemesSerializer
    http_method_names = ['get', 'options']


class WidgetSettingsViewSet(viewsets.ModelViewSet):
    queryset = WidgetSettings.objects.all()
    serializer_class = WidgetSettingsSerializer

    @action(detail=True, methods=['post', 'delete'])
    def icon(self, request, **kwargs):
        instance = self.get_object()
        if request.method == 'POST':
            from apps.balancer.utils import save_site_icon
            serializer = WidgetIconSerializer(data=request.data)
            serializer.is_valid(raise_exception=True)
            img_s, url = save_site_icon(serializer.validated_data['widget_icon'].read())
            if img_s:
                website = instance.website
                website.props['widget_icon'] = url
                website.save()
                return Response({'url': website.props.get('widget_icon')})
            else:
                return Response({'error': 'Not updated', 'old_url': instance.website.props.get('widget_icon')}, status=400)


        if request.method == 'DELETE':
            if instance.website.props.get('widget_icon'):
                website = instance.website
                del website.props['widget_icon']
                website.save()
                return Response(status=204)
            else:
                return Response(status=208)

        return Response(status=404)


    def get_queryset(self):
        queryset = self.queryset.filter(website=self.request.current_site)

        if queryset.count() == 0 and self.action=='list' and self.request.current_site:
            from .utils.form import check_website_forms
            from .utils.service import check_website_settings

            website=self.request.current_site
            check_website_forms(website)
            check_website_settings(website)
            queryset = self.queryset.filter(website=self.request.current_site)
        return queryset
    
    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        serializer = self.get_serializer(instance)
        data = serializer.data
        if hasattr(instance, 'website'):
            data['widget_icon'] = instance.website.props.get('widget_icon')
            data['copy_text'] = instance.website.props.get('widget_text')
            data['copy_link'] = instance.website.props.get('widget_link')

        return Response(data)

    def perform_create(self, serializer):
        super(WidgetSettingsViewSet, self).perform_create(serializer)
        self.update_vals(serializer)

    def perform_update(self, serializer):
        super(WidgetSettingsViewSet, self).perform_update(serializer)
        self.update_vals(serializer)
    
    def update_vals(self, serializer):
        if instance := getattr(serializer, 'instance', None):
            website = instance.website
            _save = False
            if copy_text := serializer.data.get('copy_text'):
                website.props['widget_text'] = copy_text
                _save = True
            if copy_link := serializer.data.get('copy_link'):
                website.props['widget_link'] = copy_link
                _save = True
            if _save:
                website.save()
