from django.conf import settings
from django.db import models
from django.db.models import ImageField
from sorl.thumbnail import get_thumbnail
import re
from nicechat.utils.currencies import CURRENCY_CHOICES, USD


class Website(models.Model):
    WELCOME_MESSAGE = 'welcome_msg'
    ASK_VISITOR_NAME = 'ask_name'
    PROPS_KEY_CHOICES = (
        (WELCOME_MESSAGE, 'Welcome messages', 'text',),
        (ASK_VISITOR_NAME, 'Ask name of the visitor', 'bool',),
    )
    company = models.ForeignKey('company.Company', related_name='websites', on_delete=models.CASCADE)
    title = models.CharField(max_length=100, db_index=True)  # primary domain name (without www.)
    domain = models.CharField(max_length=512, default='')
    users = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True)
    icon = ImageField(upload_to='website-icons', blank=True, null=True)
    cart_callback = models.CharField(max_length=512, blank=True, null=True)
    error_count = models.IntegerField(default=0)
    props = models.JSONField(default=dict)
    online = models.BooleanField(default=True)

    auto_assign_enable = models.BooleanField(default=True)
    auto_assign_limit = models.IntegerField(default=20)
    auto_assign_away_tout = models.IntegerField(default=60)
    auto_assign_noreply_tout = models.IntegerField(default=120)

    offline_agents = {}

    def __str__(self):
        return self.title

    def get_icon(self):
        if self.icon:
            return self.icon
        else:
            return None

    def icon_thumbs(self):
        thumb = None
        medium = None
        large = None

        try:
            thumb = get_thumbnail(self.get_icon(), '43x43', crop='center', quality=99)
            medium = get_thumbnail(self.get_icon(), '86x86', crop='center', quality=99)
            large = get_thumbnail(self.get_icon(), '150x150', crop='center', quality=99)
        except (IOError, ValueError):
            pass

        return {
            'thumb': thumb,
            'medium': medium,
            'large': large
        }

    @property
    def agents_count(self):
        return self.users.count()

    @property
    def unread(self):
        return self.pk

    def update_domain_label(self):
        dl = ''
        dlsc = self.domainlist_set.count()
        for dls in self.domainlist_set.order_by('-kind'):
            dlsc += -1
            dl += f' {dls},'
            if len(dl) > 50 and dlsc > 0:
                dl += f'.. and {dlsc} more'
                break
        dl = dl.rstrip(',').lstrip()
        self.domain = dl
        self.save(update_fields=['domain'])


class DomainList(models.Model):
    DOMAIN_REGULAR = 'reg'
    DOMAIN_WILDCARD= 'wld'
    DOMAIN_DIRECT = 'dir'
    DOMAIN_CHOICES = (
        (DOMAIN_REGULAR, 'Regular',),
        (DOMAIN_WILDCARD, 'Wildcard',),
        (DOMAIN_DIRECT, 'Direct',),
    )
    CLR = re.compile('[^a-zA-Z0-9\-\.\*]') 
    # RFC952+RFC1123+RFC2181 +IDNA - hostname can have only letter, digits, dot and hyphen.
    kind = models.CharField(max_length=5, choices=DOMAIN_CHOICES, default=DOMAIN_REGULAR, blank=True, null=True)
    domain = models.CharField(max_length=512, unique=True)
    website = models.ForeignKey(Website, on_delete=models.CASCADE)

    @property
    def dl(self):
        return f'https://{settings.DOMAIN_DIR}/' if self.kind == DomainList.DOMAIN_DIRECT else ''
    
    def match(self, target):
        _pattern = DomainList.CLR.sub('', self.domain).lower()
        _pattern = _pattern.replace('_', '').replace('.', '\.').replace('*', '.*')
        return True if re.match(_pattern, target) else False

    def save(self, *args, **kwargs):
        super(DomainList, self).save(*args, **kwargs)
        if self.kind in [DomainList.DOMAIN_DIRECT, DomainList.DOMAIN_REGULAR]:
            da = None
            for da in DomainListApplied.objects.filter(parent=self):
                da.domain = self.domain
                da.save()
            if da is None:
                DomainListApplied.objects.create(
                    domain = self.domain,
                    parent = self,
                    website = self.website
                )
        else: # Domain.DOMAIN_WILDCARD
            for da in DomainListApplied.objects.filter(parent=self):
                if not self.match(da.domain):
                    da.delete()
        self.website.update_domain_label()

    def __str__(self):
        k = 'Link:' if self.kind == DomainList.DOMAIN_DIRECT else ''
        return f'{k}{self.domain}'


class DomainListApplied(models.Model):
    domain = models.CharField(max_length=512, unique=True)
    parent = models.ForeignKey(DomainList, on_delete=models.CASCADE)
    website = models.ForeignKey(Website, on_delete=models.CASCADE)


class Status(models.Model):
    website = models.OneToOneField(Website, on_delete=models.CASCADE)
    server_code = models.IntegerField()
    widget_installed = models.BooleanField(default=False)
    allow_site_access = models.BooleanField(default=False)
    partner_id = models.CharField(max_length=64, blank=True, null=True)
    timestamp = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name_plural = 'statuses'

    def __str__(self):
        return ': '.join([self.website.domain, str(self.server_code), str(self.timestamp)])


class AccessToken(models.Model):
    token = models.CharField(max_length=50, unique=True)
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    description = models.CharField(max_length=255, blank=True, null=True)
    secret = models.CharField(max_length=50)
    gen_timestamp = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return ' '.join(['Access token:', self.token, 'for', self.website.domain])


class QuickReply(models.Model):
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.SET_NULL)
    is_private = models.BooleanField(default=False)
    text = models.TextField()


class ProductsCategory(models.Model):
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    original_id = models.CharField(max_length=64)
    is_enabled = models.BooleanField(default=True)
    parent = models.ForeignKey("ProductsCategory", null=True, blank=True, on_delete=models.CASCADE)
    version = models.CharField(max_length=128, blank=True, null=True)

    class Meta:
        unique_together = ('website', 'original_id',)

    def __str__(self):
        return ' : '.join([self.website.domain, self.name])


class Product(models.Model):

    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    category = models.ForeignKey(ProductsCategory, blank=True, null=True, on_delete=models.CASCADE)
    original_id = models.CharField(max_length=64)
    title = models.CharField(max_length=100)
    price = models.DecimalField(max_digits=15, decimal_places=2, blank=True, null=True)
    currency = models.CharField(max_length=5, choices=CURRENCY_CHOICES, default=USD, blank=True, null=True)
    description = models.TextField(blank=True, null=True)
    url = models.CharField(max_length=512, blank=True, null=True)
    image_url = models.CharField(max_length=512, blank=True, null=True)
    in_stock = models.BooleanField(default=False)
    version = models.CharField(max_length=128, blank=True, null=True)
    props = models.JSONField(default=dict)

    class Meta:
        ordering = ('-id',)
        unique_together = ('website', 'original_id',)

    def __str__(self):
        return ' : '.join([self.website.domain, self.title])

#image_post_processing(self.image)


class Form(models.Model):
    PERM_READ_PUBLIC = 0
    PERM_READ_COMPANY = 1
    PERM_READ_WEBSITE = 2
    PERM_READ_PRIVATE = 3
    PERM_WRITE_SUPER = 0
    PERM_WRITE_ADMIN = 1
    PERM_WRITE_AGENT = 2
    WRITE_PERM_CHOICES = (
        (PERM_WRITE_AGENT, 'Can be written by agent and admin',),
        (PERM_WRITE_ADMIN, 'Can be written by admin',),
        (PERM_WRITE_SUPER, 'Can be written by superuser',),
    )
    READ_PERM_CHOICES = (
        (PERM_READ_PUBLIC, 'Public. But, not shown in list (choice by ID)',),
        (PERM_READ_COMPANY, 'Company. Visible for all sites of company.',),
        (PERM_READ_WEBSITE, 'Website. Visible to specified site only',),
        (PERM_READ_PRIVATE, 'Private. Visible only to the author',),
    )
    MODE_SITE_REGULAR_FORM = 'regular'
    MODE_SITE_PURCHASE_FORM = 'purchase'
    MODE_SITE_CATCH_FORM = 'catch'
    MODE_SITE_GREET_FORM = 'greet'
    MODE_SITE_FINAL_FORM = 'final'
    MODE_CHOICES = (
        (MODE_SITE_REGULAR_FORM, 'Regular form',),
        (MODE_SITE_PURCHASE_FORM, 'Purchase form',),
        (MODE_SITE_CATCH_FORM, 'Catch form',),
        (MODE_SITE_GREET_FORM, 'Greeting form',),
        (MODE_SITE_FINAL_FORM, 'Final form',),
    )

    website = models.ForeignKey('Website', null=True, blank=True, on_delete=models.CASCADE)
    company = models.ForeignKey('company.Company', null=True, blank=True, on_delete=models.CASCADE)
    access_write = models.IntegerField(choices=WRITE_PERM_CHOICES, default=PERM_WRITE_AGENT, db_index=True)
    access_read = models.IntegerField(choices=READ_PERM_CHOICES, default=PERM_READ_PRIVATE, db_index=True)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.SET_NULL)
    is_enabled = models.BooleanField(default=False)
    label = models.CharField(max_length=50)
    mode = models.CharField(max_length=10, choices=MODE_CHOICES, default=MODE_SITE_REGULAR_FORM, db_index=True)

    @property
    def rows(self):
        return self.formrow_set.all()

    @property
    def is_purchase(self):
        return self.mode == self.MODE_SITE_PURCHASE_FORM

    def __str__(self):
        return "Form #{}, \"{}\"".format(self.pk, self.label)


class FormRow(models.Model):
    FIELD_TYPE_TEXT = 'text'
    FIELD_TYPE_LINK = 'link'
    FIELD_TYPE_EMAIL = 'email'
    FIELD_TYPE_NUMBER = 'number'
    FIELD_TYPE_TEXTAREA = 'textarea'
    FIELD_TYPE_CHECKBOX = 'checkbox'
    FIELD_TYPE_MULTICHECKBOX = 'multicheckbox'
    FIELD_TYPE_RADIO = 'radio'
    FIELD_TYPE_RATE_STAR = 'ratestar'
    FIELD_TYPE_SELECT = 'choice'
    FIELD_TYPE_HIDDEN = 'hidden'
    FIELD_TYPE_COUNTRY = 'country'
    FIELD_TYPE_US_STATE = 'state'
    FIELD_TYPE_LABEL = 'label'
    FIELD_TYPE_ADDRESS = 'address'
    FIELD_TYPE_GROUP = 'group'
    # FIELD_TYPE_C_PRODUCT = 'c_product'
    # FIELD_TYPE_C_CATEGORY = 'c_category'

    FIELD_TYPES = (
        (FIELD_TYPE_TEXT, 'Input text'),
        (FIELD_TYPE_LINK, 'Link'),
        (FIELD_TYPE_EMAIL, 'Input Email'),
        (FIELD_TYPE_NUMBER, 'Input number'),
        (FIELD_TYPE_TEXTAREA, 'Textarea'),
        (FIELD_TYPE_CHECKBOX, 'Checkbox'),
        (FIELD_TYPE_RADIO, 'Radio choice'),
        (FIELD_TYPE_RATE_STAR, 'Radio choice'),
        (FIELD_TYPE_SELECT, 'Dropdown choice'),
        (FIELD_TYPE_MULTICHECKBOX, 'Multiple selection list'),
        (FIELD_TYPE_HIDDEN, 'Hidden'),
        (FIELD_TYPE_COUNTRY, 'Country (dropdown)'),
        (FIELD_TYPE_US_STATE, 'State of USA'),
        (FIELD_TYPE_LABEL, 'Label (text)'),
        (FIELD_TYPE_GROUP, 'Group begin'),
        (FIELD_TYPE_ADDRESS, 'Address (block)')
        # (FIELD_TYPE_C_PRODUCT, _('Product selector')),
        # (FIELD_TYPE_C_CATEGORY, _('Category selector')),
    )

    TYPES_REQUIRE_CHOICE = (FIELD_TYPE_MULTICHECKBOX, FIELD_TYPE_RADIO, FIELD_TYPE_SELECT,)

    ALLOWED_AUTO_FILL_TAGS = {
        'PRODUCT_ID': "ID of event.Product (purchases forms)",
        'PRODUCT_NAME': "Name of event.Product (purchases forms)",
        'CLIENT_ID': "Visitor trace-token",
        'CLIENT_NAME': "Visitor display name",
        'AGENT_ID': 'ID of agent which push the form',
        'AGENT_NAME': 'Name of agent which push the form'
    }

    form = models.ForeignKey(Form, on_delete=models.CASCADE)
    kind = models.CharField(max_length=15, choices=FIELD_TYPES)
    label = models.CharField(max_length=50)
    description = models.TextField(blank=True, null=True)
    is_read_only = models.BooleanField(default=False)
    is_required = models.BooleanField(default=False)
    protected = models.BooleanField(default=False)
    order_num = models.PositiveSmallIntegerField(blank=True, null=True)
    params = models.TextField(default="{}")


class AgentStatus(models.Model):
    STATUS_ONLINE = 'online'
    STATUS_BUSY = 'busy'
    STATUS_AWAY = 'away'
    STATUS_CHOICES = (
        (STATUS_ONLINE, 'ON-Line'),
        (STATUS_BUSY, 'Busy'),
        (STATUS_AWAY, 'Away'),
    )
    website = models.ForeignKey(Website, db_index=True, on_delete=models.CASCADE)
    operator = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='agent_status', on_delete=models.CASCADE)
    is_active = models.BooleanField(default=False, db_index=True)
    is_current = models.BooleanField(default=False, db_index=True)
    state = models.CharField(max_length=8, choices=STATUS_CHOICES, default=STATUS_AWAY)
    state_timestamp = models.BigIntegerField(default=0)


class Callbacks(models.Model):

    TARGET_EMAIL_TECH = 't_email'
    TARGET_EMAIL_HUMAN = 'h_email'
    TARGET_SLACK = 'slack'
    TARGET_HTTP = 'http'
    TARGET_CHOICES = (
        (TARGET_EMAIL_TECH, 'E-mail (JSON-data)'),
        (TARGET_EMAIL_HUMAN, 'E-mail (human-like data'),
        (TARGET_SLACK, 'Slack'),
        (TARGET_HTTP, 'HTTP'),
    )
    website = models.ForeignKey(Website, on_delete=models.CASCADE)
    event = models.CharField(max_length=20, db_index=True)
    target = models.CharField(max_length=8, choices=TARGET_CHOICES)
    is_active = models.BooleanField(default=True, db_index=True)
    params = models.JSONField(default=dict)


class WidgetTheme(models.Model):
    TO_WIDGET = [ ('icon_color', 'iconColor'),  ('buttons_color', 'buttonsColor'),  ('buttons_text_color', 'buttonsTextColor'),  ('system_text_color', 'systemTextColor'),  ('bg_color', 'bgColor'),  ('notifications_text_color', 'notificationsTextColor'),  ('border_color', 'borderColor'),  ('outchat_bg_color', 'outchatBgColor'),  ('intchat_bg_color', 'inchatBgColor'),  ('input_bg_color', 'inputBgColor'),  ('input_text_color', 'inputTextColor'),  ('input_buttons_color', 'inputButtonsColor'),  ('operator_icon_bg', 'operatorIconBg'), ('operator_border', 'operatorBorder')]

    name = models.CharField(max_length=50)
    icon_color = models.CharField(max_length=50, default='#656ab3')
    buttons_color = models.CharField(max_length=50, default='#656ab3')
    buttons_text_color = models.CharField(max_length=50, default='#ffffff')
    system_text_color = models.CharField(max_length=50, default='#4e5169')
    bg_color = models.CharField(max_length=50, default='#cacce8')
    notifications_text_color = models.CharField(max_length=50, default='#222222')
    border_color = models.CharField(max_length=50, default='#ffffff')
    outchat_bg_color = models.CharField(max_length=50, default='#686b89')
    intchat_bg_color = models.CharField(max_length=50, default='#ffffff')
    input_bg_color = models.CharField(max_length=50, default='#ffffff')
    input_text_color = models.CharField(max_length=50, default='#222222')
    input_buttons_color = models.CharField(max_length=50, default='#586778')
    operator_icon_bg = models.CharField(max_length=50, default='#4e5169')
    operator_border = models.CharField(max_length=50, default='#ffffff')

    def make_props(self):
        ret = ''
        for k, nm in self.TO_WIDGET:
            ret += "'theme_%s': '%s', " % (nm, getattr(self, k, ''))
        return ret;

    def __str__(self):
        return self.name


class WidgetSettings(models.Model):
    POSITION_LEFT = 'left'
    POSITION_RIGHT = 'right'
    POSITION_CHOICES = (
        (POSITION_LEFT, 'Left'),
        (POSITION_RIGHT, 'Right'),
    )

    SHAPE_ROUND = 'round'
    SHAPE_SQUARE = 'square'
    SHAPE_ROUNDED_SQUARE = 'rounded_square'
    SHAPE_RECTANGLE = 'rectangle'
    SHAPE_ROUNDED_RECTANGLE = 'rounded_rectangle'
    SHAPE_CHOICES = (
        (SHAPE_ROUND, 'Round'),
        (SHAPE_SQUARE, 'Square'),
        (SHAPE_ROUNDED_SQUARE, 'Rounded square'),
        (SHAPE_RECTANGLE, 'Rectangle'),
        (SHAPE_ROUNDED_RECTANGLE, 'Rounded rectangle'),
    )

    SIZE_SMALL = 's'
    SIZE_MEDIUM = 'm'
    SIZE_LARGE = 'l'
    SIZE_CHOICES = (
        (SIZE_SMALL, 'Small'),
        (SIZE_MEDIUM, 'Medium'),
        (SIZE_LARGE, 'Large'),
    )

    position = models.CharField(max_length=5, choices=POSITION_CHOICES, default=POSITION_RIGHT)
    shape = models.CharField(max_length=17, choices=SHAPE_CHOICES, default=SHAPE_ROUNDED_RECTANGLE)
    size = models.CharField(max_length=2, choices=SIZE_CHOICES, default=SIZE_MEDIUM)
    icon_img = models.ImageField(upload_to='widget-icons', blank=True, null=True)
    agent_avatar = models.BooleanField(default=False)
    text_shape_label = models.CharField(max_length=50, blank=True, null=True)
    text_title = models.CharField(max_length=50, blank=True, null=True)
    theme = models.ForeignKey(WidgetTheme, on_delete=models.CASCADE)
    website = models.OneToOneField(Website, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.website.domain} {self.__dict__}"
