from django.contrib.auth import get_user_model
from django.conf import settings
from rest_framework import serializers
from rest_framework.fields import ImageField, empty
import random
import string

from ..users.serializers import AgentStatusSerializer
from ..website.models import Website, QuickReply, Product, ProductsCategory, Form, FormRow, Status, Callbacks, \
    WidgetSettings, WidgetTheme, DomainList, AgentStatus, AccessToken
from ..chat.models import ChatReadStack, Visitor
from nicechat.serialization import ImageUrlSerializer
from nicechat.utils.callback import EVENT_CHOICES


class UnreadSerializer(serializers.BaseSerializer):
    def to_representation(self, obj):
        if hasattr(self, 'context') and isinstance(self.context, dict):
            request = self.context.get('request', None)
            if request and hasattr(request, 'user') and request.user.is_authenticated:
                user = request.user
                unread_list = []
                unread_count = 0
                unread_total = 0
                query = ChatReadStack.objects.filter(operator=request.user,
                                                     chat__website=obj
                                                     ).exclude(unread_count=0)
                for item in query:
                    unread_list.append({
                        'chat': item.chat_id,
                        'unread': item.unread_count,
                        'msg_total': item.chat_msg_count,
                        'msg_was_read': item.user_msg_read,
                    })
                    unread_count += 1
                    unread_total += item.unread_count

                return {
                    'unread_chat_count': unread_count,
                    'unread_msg_total': unread_total,
                    'unread_list': unread_list
                }

        return {
            'unread_chat_count': None,
            'unread_msg_total': None,
            'unread_list': None
        }


class StatusSerializer(serializers.ModelSerializer):

    class Meta:
        model = Status
        fields = '__all__'


class WidgetThemesSerializer(serializers.ModelSerializer):

    class Meta:
        model = WidgetTheme
        fields = '__all__'


class WidgetSettingsSerializer(serializers.ModelSerializer):
    copy_text = serializers.CharField(required=False)
    copy_link = serializers.CharField(required=False)
    widget_icon = serializers.ReadOnlyField()

    class Meta:
        model = WidgetSettings
        fields = '__all__'


class WidgetIconSerializer(serializers.Serializer):
    widget_icon = serializers.FileField()


class DomainListSerializer(serializers.ModelSerializer):
    dl = serializers.ReadOnlyField()

    class Meta:
        model = DomainList
        fields = ('pk', 'kind', 'domain', 'dl',)

    def run_validation(self, data=empty):

        if data.get('kind') == DomainList.DOMAIN_DIRECT and not data.get('domain'):
            letters = string.digits + string.ascii_lowercase
            for q in range(0, 50):
                tok = ''.join(random.choice(letters) for i in range(8))
                if not DomainList.objects.filter(domain=tok, kind=DomainList.DOMAIN_DIRECT).count():
                    data['domain'] = tok
                    break

        print(len(data.get('domain', '')))
        if data.get('kind') == DomainList.DOMAIN_DIRECT and len(data.get('domain', '')) < 3:
            raise serializers.ValidationError({'domain': "direct-anchor must be at least 3 chars length"})

        return super(DomainListSerializer, self).run_validation(data)


class WebsiteSerializer(serializers.ModelSerializer):
    icons = serializers.DictField(source='icon_thumbs', child=ImageField(), required=False, allow_null=True,
                                  read_only=True)
    unread = UnreadSerializer(many=False, read_only=True, required=False)
    status = StatusSerializer(many=False, read_only=True, required=False)
    path_prefix = serializers.SerializerMethodField()
    company_allow_whitelabel = serializers.SerializerMethodField() 

    def get_path_prefix(self, obj):
        if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
            return '/chatting/'
        anode = obj.node2website.filter(status=0).first()
        return f'/{anode.node.name}/' if anode else None

    def get_company_allow_whitelabel(self, obj):
        return True if obj.company and obj.company.allow_branding else False

    class Meta:
        model = Website
        fields = ('id', 'title', 'company', 'domain', 'agents_count', 'unread', 'path_prefix', 'icons', 'props', 'status', 
                  'auto_assign_enable', 'auto_assign_limit', 'auto_assign_away_tout', 'auto_assign_noreply_tout', 'company_allow_whitelabel')
        read_only_fields = ('domain',)

class WebsiteSmallSerializer(serializers.ModelSerializer):
    icons = serializers.DictField(source='icon_thumbs', child=ImageField(), required=False, allow_null=True,
                                  read_only=True)
    class Meta:
        model = Website
        fields = ('id', 'title', 'domain', 'agents_count', 'icons')

class QuickReplySerializer(serializers.ModelSerializer):
    created_by = serializers.PrimaryKeyRelatedField(read_only=True, default=serializers.CurrentUserDefault())

    class Meta:
        model = QuickReply
        fields = ('id', 'website', 'created_by', 'is_private', 'text')
        read_only_fields = ('website',)


class ProductsCategorySerializer(serializers.ModelSerializer):
    class Meta:
        model = ProductsCategory
        fields = ('id', 'website', 'name', 'is_enabled', 'parent')


class ProductSerializer(serializers.ModelSerializer):
    image_url = ImageUrlSerializer(required=False)

    class Meta:
        model = Product
        fields = ('id', 'original_id', 'website', 'category', 'title', 'props',
                  'price', 'currency', 'description', 'url', 'image_url', 'in_stock')


class AgentSerializer(serializers.ModelSerializer):
    avatar = serializers.DictField(source='avatar_thumbs', child=ImageField())
    status = AgentStatusSerializer(many=False, read_only=True, required=False)
    role = serializers.SerializerMethodField(read_only=True, required=False)
    is_on_site = serializers.SerializerMethodField(read_only=True, required=False)
    own_company = serializers.PrimaryKeyRelatedField(read_only=True)

    def get_role(self, instance):
        request = self.context.get('request')

        if hasattr(request, 'current_site') and request.current_site:
            return instance.get_site_role(request.current_site)

    def get_is_on_site(self, instance):
        site_id = self.context.get('site_id')
        if site_id:
            ass = AgentStatus.objects.filter(website_id=site_id, operator=instance).first()
            return ass.is_current if ass else False
        request = self.context.get('request')
        if hasattr(request, 'current_site') and request.current_site:
            ass = AgentStatus.objects.filter(website=request.current_site, operator=instance).first()
            return ass.is_current if ass else False
        return False

    class Meta:
        model = get_user_model()
        fields = ('id', 'first_name', 'last_name', 'phone_number', 'get_full_name', 'role', 'is_on_site',
                  'position', 'avatar', 'timezone', 'email', 'status', 'is_active', 'own_company')


class FormRowSerializer(serializers.ModelSerializer):
    class Meta:
        model = FormRow
        exclude = ('form',)
        read_only_fields = ('protected',)


class FormSerializer(serializers.ModelSerializer):
    is_purchase = serializers.ReadOnlyField()
    # rows = FormRowSerializer(many=True, read_only=True)

    class Meta:
        model = Form
        fields = '__all__'
        read_only_fields = ('id', 'company', 'mode' 'created_by')


class FormForWidgetSerializer(serializers.ModelSerializer):
    rows = FormRowSerializer(many=True, read_only=True)

    class Meta:
        model = Form
        fields = ('rows', 'label', 'mode')


class CallbacksSerializer(serializers.ModelSerializer):
    def validate_event(self, data):
        for item in EVENT_CHOICES:
            if item and data == item[0]:
                return data
        raise serializers.ValidationError('Invalid field value. Look /api/v1/hooks/options/')

    class Meta:
        model = Callbacks
        fields = '__all__'
        read_only_fields = ('website',)


class AccessTokenSerializer(serializers.ModelSerializer):

    class Meta:
        model = AccessToken
        fields = ['gen_timestamp', 'token', 'secret']
        read_only_fields = ('gen_timestamp', 'token', 'secret',)
        
class AccessTokenSerializerList(AccessTokenSerializer):
    secret = serializers.SerializerMethodField()

    def get_secret(self, obj):
        return '********'
