import hashlib
import json
import datetime
from django.db.models.signals import post_save, pre_save, post_delete, m2m_changed
from django.dispatch import receiver
from django.core.cache import caches, InvalidCacheBackendError
from django.conf import settings
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()

from nicechat.utils.constants import WebsocketEventTypes
from ..users.serializers import UserSerializer
from .tasks import create_website_celery
from .serializers import StatusSerializer
from .models import AgentStatus, Product, Status, Callbacks, Website
from nicechat.utils.arbiter import update_data_workernode
from nicechat.utils.arbiter_sync import get_user_relations


@receiver(post_save, sender=AgentStatus)
def operator_active_post_save(sender, instance=None, **kwargs):
    if instance:
        a_key = "adm-website_mon-%i" % instance.website_id
        async_to_sync(channel_layer.group_send)(a_key, {"type": "websocket.send",  "text":  json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_OPERATOR_UPDATE,
                                               "data": dict(UserSerializer(instance=instance.operator,
                                                                           context={'request': None,
                                                                                    'site_id': instance.website_id})
                                                            .to_representation(instance=instance.operator)),
                                               "pk": instance.pk
                                               })})


@receiver(pre_save, sender=Product)
def add_original_id(sender, instance=None, **kwargs):
    if instance and not instance.original_id:
        instance.original_id = hashlib.sha1(''.join([str(str(datetime.datetime.now()).split('.')[0]),
                                                     instance.title]).encode('utf-8')).hexdigest()


@receiver(post_save, sender=Status)
def set_company_partner_on_first_status(sender, instance, created, **kwargs):
    if instance.partner_id and not instance.website.company.partner_id:
        instance.website.company.partner_id = instance.partner_id
        instance.website.company.save()
    if instance:
        a_key = "adm-website_mon-%i" % instance.website_id
        async_to_sync(channel_layer.group_send)(a_key, {"type": "websocket.send",  "text":  json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_STATUS_UPDATE,
                                               "data": dict(StatusSerializer(instance=instance,
                                                                           context={'request': None,
                                                                                    'site_id': instance.website_id})
                                                            .to_representation(instance=instance)),
                                               "pk": instance.pk
                                               })})

@receiver(post_save, sender=Callbacks)
def callbacks_post_save(sender, instance, **kwargs):
    if instance:
        token = "callbacks-%s-%s" % (instance.event, instance.website_id)
        try:
            cache = caches['default']
            cache.delete(token)
        except InvalidCacheBackendError:
            pass


@receiver(post_delete, sender=Callbacks)
def callbacks_post_delete(sender, instance, **kwargs):
    if instance:
        token = "callbacks-%s-%s" % (instance.event, instance.website_id)
        try:
            cache = caches['default']
            cache.delete(token)
        except InvalidCacheBackendError:
            pass


@receiver(post_save, sender=Website)
def website_post_save(sender, instance, created, **kwargs):
    if instance and created:
        create_website_celery.run(instance.pk)
    from apps.balancer.tasks import sync_website
    sync_website(instance)


@receiver(m2m_changed, sender=Website.users.through)
def website_m2m_users_change(instance, action, **kwargs):
    if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
        return
    if action in ['post_add', 'post_remove']:
        User = kwargs['model']
        for u in User.objects.filter(pk__in=list(kwargs['pk_set'])):
            update_data_workernode(
                instance.node2website.filter(status=0).values_list('node', flat=True), 
                'user',
                get_user_relations(u)
            )

    if action == 'post_clear':
        update_data_workernode(
            instance.node2website.filter(status=0).values_list('node', flat=True), 
            f'w2userpostclear', 
            {'site': instance.pk}
        )


@receiver(post_delete, sender=Website)
def webiste_post_delete(sender, instance=None, created=False, **kwargs):
    if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
        return
    print("POST_DEL", instance)
    update_data_workernode(instance.node2website.filter(status=0).values_list('node_id', flat=True), 
                           'websitedel', {'pk': instance.pk,})

