import os
from rest_framework import status
from django.test.testcases import TestCase

from nicechat.utils.tests import TestCaseWithSites
from nicechat.utils.helpers import parse_price
from ..website.models import QuickReply, Product


class QuickReplyAPITest(TestCaseWithSites):
    class TestData:
        SITE_ID_PARAM = 'site_id={site_id}'
        PUBLIC_REPLY = {'is_private': '0', 'text': 'hello world'}
        PUBLIC_REPLY_2 = {'is_private': '0', 'text': 'bye'}
        PRIVATE_REPLY = {'is_private': '1', 'text': 'hello guys'}

    def setUp(self):
        super(QuickReplyAPITest, self).setUp()

        self.sites_api_url = self.api_url + 'websites/'
        self.companies_api_url = self.api_url + 'companies/'
        self.settings_detail_api_url = self.api_url + 'settings/{}/'
        self.quick_reply_api_url = self.api_url + 'quick-reply/?site_id={site_id}'
        self.quick_reply_api_url_detail = self.api_url + 'quick-reply/{pk}/'
        self.quick_reply_api_url_filter = self.api_url + 'quick-reply/?text={}'
        self.delete_reply_api_url = self.api_url + 'quick-reply/reply_delete/'
        self.delete_replies_api_url = self.api_url + 'quick-reply/delete_replies/'
        self.delete_replies_api_url_ids = self.delete_replies_api_url + '?ids={}&ids={}&site_id={site_id}'

    def _create_replies(self):
        site_0_1_id = self.website_0_1.pk
        site_1_1_id = self.website_1_1.pk

        self.assertEqual(QuickReply.objects.all().count(), 0)

        self.login_user(self.user_0_1)

        # new private reply
        response = self.client.post(self.quick_reply_api_url.format(site_id=site_0_1_id),
                                    self.TestData.PRIVATE_REPLY, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        site_0_1_qr_pk = response.data['id']

        # new public reply for current site
        response = self.client.post(self.quick_reply_api_url.format(site_id=site_0_1_id),
                                    self.TestData.PUBLIC_REPLY, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data['website'], site_0_1_id)
        # Check current user
        self.assertEqual(response.data['created_by'], self.user_0_1.pk)

        site_0_1_qr_2_pk = response.data['id']

        self.assertEqual(QuickReply.objects.all().count(), 2)

        self.login_user(self.user_1_1)

        # Create public reply
        response = self.client.post(self.quick_reply_api_url.format(site_id=site_1_1_id),
                                    self.TestData.PUBLIC_REPLY, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        site_1_1_qr_pk = response.data['id']

        # Create private reply
        response = self.client.post(self.quick_reply_api_url.format(site_id=site_1_1_id),
                                    self.TestData.PRIVATE_REPLY, format='json')
        site_1_1_qr_2_pk = response.data['id']

        quick_replies_total = 4

        self.assertEqual(QuickReply.objects.all().count(), quick_replies_total)

        # Get all private and public replies for site_1_1 and user_1_1
        response = self.client.get(self.quick_reply_api_url.format(site_id=site_1_1_id), format='json')

        self.assertEqual(response.data['count'], 2)
        self.assertEqual(QuickReply.objects.count(), quick_replies_total)

        return (site_0_1_qr_pk, site_0_1_qr_2_pk,
                site_1_1_qr_pk, site_1_1_qr_2_pk)

    def test_reply_create_update(self):
        site_0_1_id = self.website_0_1.pk
        site_1_1_id = self.website_1_1.pk

        site_0_1_qr_pk, site_0_1_qr_2_pk, site_1_1_qr_pk, site_1_1_qr_2_pk = self._create_replies()

        #
        # Update one of user replies
        #
        new_text = 'new hello'
        upd_data = {'is_private': False, 'text': new_text}
        update_url = self.quick_reply_api_url_detail.format(pk=site_1_1_qr_2_pk) \
            + '?' + self.TestData.SITE_ID_PARAM.format(site_id=site_1_1_id)

        self.client.put(update_url, upd_data, format='json')
        response = self.client.get(self.quick_reply_api_url.format(site_id=site_1_1_id), format='json')
        self.assertEqual(response.data['count'], 2)
        self.assertEqual(response.data['results'][0]['id'], site_1_1_qr_pk)
        self.assertEqual(response.data['results'][1]['id'], site_1_1_qr_2_pk)
        self.assertEqual(response.data['results'][1]['text'], new_text)

    def test_delete_replies(self):
        site_0_1_id = self.website_0_1.pk
        site_1_1_id = self.website_1_1.pk
        quick_replies_total = 4
        site_0_1_qr_pk, site_0_1_qr_2_pk, site_1_1_qr_pk, site_1_1_qr_2_pk = self._create_replies()

        #
        # Deleting one reply
        #

        # no access to the object belonging to another site
        self.assertNotEqual(QuickReply.objects.get(pk=site_0_1_qr_pk).website_id, site_1_1_id)
        delete_url = self.quick_reply_api_url_detail.format(pk=1) \
            + '?' + self.TestData.SITE_ID_PARAM.format(site_id=site_1_1_id)
        response = self.client.delete(delete_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        # nothing was changed in database
        self.assertEqual(QuickReply.objects.count(), quick_replies_total)

        # object belonging to current site
        self.assertEqual(QuickReply.objects.get(pk=site_1_1_qr_pk).website_id, site_1_1_id)
        delete_url = self.quick_reply_api_url_detail.format(pk=site_1_1_qr_pk) \
            + '?' + self.TestData.SITE_ID_PARAM.format(site_id=site_1_1_id)
        response = self.client.delete(delete_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(QuickReply.objects.count(), quick_replies_total - 1)

        #
        # bulk delete
        #

        # no params
        self.login_user(self.user_owner_0)
        response = self.client.delete(self.delete_replies_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.login_user(self.user_0_1)
        response = self.client.delete(self.delete_replies_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        # Delete replies
        delete_url = self.delete_replies_api_url_ids.format(site_0_1_qr_pk, site_0_1_qr_2_pk, site_id=site_0_1_id)
        response = self.client.delete(delete_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        # counts of replies
        self.assertEqual(QuickReply.objects.all().count(), 1)
        response = self.client.get(self.quick_reply_api_url.format(site_id=site_0_1_id), format='json')
        self.assertEqual(response.data['count'], 0)

    def test_filter(self):
        site_1_1_id = self.website_1_1.pk

        self._create_replies()

        # one more reply by self.user_1_1 with unusual text
        response = self.client.post(self.quick_reply_api_url.format(site_id=site_1_1_id),
                                    self.TestData.PUBLIC_REPLY_2, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        response = self.client.get(self.quick_reply_api_url.format(site_id=site_1_1_id), format='json')
        self.assertEqual(response.data['count'], 3)

        # Using filter by field text
        filter_url = self.quick_reply_api_url_filter.format('hello') \
            + '&' + self.TestData.SITE_ID_PARAM.format(site_id=site_1_1_id)
        response = self.client.get(filter_url, format='json')

        # Check filtered data
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 2)
        self.assertTrue('hello' in response.data['results'][0]['text'])
        self.assertTrue('hello' in response.data['results'][1]['text'])


# todo products and categories tests

class ProductsTestCase(TestCaseWithSites):
    def setUp(self):
        super(ProductsTestCase, self).setUp()

        self.site_id_param = 'site_id={site_id}'
        self.products_api_url = self.api_url + 'products/'
        self.products_import_api_url = self.api_url + 'product_import/'
        self.products_import_csv_api_url = self.api_url + 'product_import_csv/'

    def test_features_permissions(self):
        valid_product_data = {
            'original_id': 'ID-X',
            'title': 'Some Product',
            'website': '1',
        }
        prod_add_feature = 'products_add'
        prod_import_feature = 'products_import'

        # anonymous user has access to products list
        response = self.client.get(self.products_api_url, format='json')
        self.assertEquals(response.status_code, status.HTTP_200_OK)

        # anonymous user has no access to any action other then 'list'
        response = self.client.post(self.products_api_url, valid_product_data, format='json')
        self.assertEquals(response.status_code, status.HTTP_403_FORBIDDEN)

        # some assertions for user in a company with trial plan
        self.assertEquals(self.user_0_1.company, self.company0)
        self.assertTrue(self.company0.billing_plan.plan.is_trial)
        self.assertTrue(self.company0.has_access_to_feature(prod_add_feature))

        site_0_1_api_url = self.products_api_url + '?' + self.site_id_param.format(site_id=self.website_0_1.pk)

        self.login_user(self.user_0_1)

        # authenticated user has access to products list
        response = self.client.get(site_0_1_api_url, format='json')
        self.assertEquals(response.status_code, status.HTTP_200_OK)

        # trial plan company user has access to all the actions of viewset
        response = self.client.post(site_0_1_api_url, valid_product_data, format='json')
        print(response)
        self.assertEquals(response.status_code, status.HTTP_201_CREATED)

        # some assertions for user in a company with "Free" plan
        self.assertEquals(self.user_1_1.company, self.company1)
        self.assertFalse(self.company1.billing_plan.plan.is_trial)
        self.assertEquals(self.company1.billing_plan.plan.name, 'Free')
        self.assertFalse(self.company1.has_access_to_feature(prod_add_feature))

        site_1_1_api_url = self.products_api_url + '?' + self.site_id_param.format(site_id=self.website_1_1.pk)
        self.login_user(self.user_1_1)

        # free plan company user has no access to any action other then 'list'
        response = self.client.post(site_1_1_api_url, {}, format='json')
        self.assertEquals(response.status_code, status.HTTP_403_FORBIDDEN)

        #
        # Product importing views
        #

        self.client.logout()
        # anonymous user access is forbidden
        response = self.client.post(self.products_import_api_url, {}, format='json')
        self.assertEquals(response.status_code, status.HTTP_403_FORBIDDEN)
        response = self.client.post(self.products_import_csv_api_url, {}, format='json')
        self.assertEquals(response.status_code, status.HTTP_403_FORBIDDEN)

        # company1 has "Trial" plan
        self.assertTrue(self.company0.has_access_to_feature(prod_import_feature))
        self.login_user(self.user_0_1)

        # successful endpoints access by user
        site_0_1_api_url = self.products_import_api_url + '?' + self.site_id_param.format(site_id=self.website_0_1.pk)
        response = self.client.post(site_0_1_api_url, {}, format='json')
        self.assertEquals(response.status_code, status.HTTP_204_NO_CONTENT)

        site_0_1_api_url = self.products_import_csv_api_url + '?' + self.site_id_param.format(site_id=self.website_0_1.pk)
        response = self.client.post(site_0_1_api_url, {}, format='json')
        # response error due to no file data passed, but still it's not 403
        self.assertEquals(response.status_code, status.HTTP_400_BAD_REQUEST)

        # company1 has "Free" plan
        self.assertFalse(self.company1.has_access_to_feature(prod_import_feature))
        self.login_user(self.user_1_1)

        # access is forbidden for users of company with no "products_import" permission in billing plan
        site_1_1_api_url = self.products_import_api_url + '?' + self.site_id_param.format(site_id=self.website_1_1.pk)
        response = self.client.post(site_1_1_api_url, {}, format='json')
        self.assertEquals(response.status_code, status.HTTP_403_FORBIDDEN)

        site_1_1_api_url = self.products_import_csv_api_url + '?' + self.site_id_param.format(site_id=self.website_1_1.pk)
        response = self.client.post(site_1_1_api_url, {}, format='json')
        self.assertEquals(response.status_code, status.HTTP_403_FORBIDDEN)


# @TODO Finish product image uploading tests
class ProductImageCreationTest(TestCaseWithSites):
    class TestData:
        SITE_ID_PARAM = 'site_id={site_id}'

    def setUp(self):
        super(ProductImageCreationTest, self).setUp()

        self.product_image_upload_api_url = self.api_url + 'websites/'


class ImportTest(TestCaseWithSites):
    class TestData:
        SITE_ID_PARAM = 'site_id={site_id}'

    def setUp(self):
        super(ImportTest, self).setUp()

        self.product_import_url = self.api_url + 'product_import/' + '?' + self.TestData.SITE_ID_PARAM

    def test_product_import(self):

        site_id = None
        for prop in dir(self):
            if prop.startswith("website_"):
                item = getattr(self, prop)
                if item.company.billing_plan.plan.name != 'free':
                    user = item.users.all().exclude(pk=item.company.owner.pk).first()
                    if user:
                        site_id = item.pk
                        self.login_user(user)
                        break

        self.assertIsNotNone(site_id)

        c_dir = os.path.dirname(os.path.abspath(__file__))
        c_dir = os.path.join(c_dir, '..', '..', 'nicechat', 'data', 'for_tests')

        product_import_url = self.product_import_url.format(site_id=site_id)

        for f in os.listdir(c_dir):
            if f.endswith('.google.csv'):
                p = {'site_id': site_id, 'google_merchant_csv': 'file://%s' % os.path.abspath(os.path.join(c_dir, f))}
            else:
                continue

            response = self.client.post(product_import_url, p)
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

            self.assertNotEqual(Product.objects.all().count(), 0)


class PriceDecodeTest(TestCase):

    def test_decode_price(self):
        self.assertEquals(parse_price("$ 1 100.00"), (1100.0, 'usd'))
        self.assertEquals(parse_price("USD 1,200,000.00"), (1200000.0, 'usd'))
        self.assertEquals(parse_price("   USD     1 300.0   "), (1300.0, 'usd'))
        self.assertEquals(parse_price(" usd 1400,00 "), (1400.0, 'usd'))
        self.assertEquals(parse_price("USD 1 500"), (1500.0, 'usd'))
        self.assertEquals(parse_price("uSd 1600"), (1600.0, 'usd'))
        self.assertEquals(parse_price("1,700.00 USD "), (1700.0, 'usd'))
        self.assertEquals(parse_price(" 1 800.00 UAH"), (1800.0, 'uah'))
        self.assertEquals(parse_price("1900 000,00 CAD"), (1900000.0, 'cad'))
        self.assertEquals(parse_price("2 000 GBP"), (2000.0, 'gbp'))
        self.assertEquals(parse_price("21 AUD"), (21.0, 'aud'))
        self.assertEquals(parse_price("2200"), (2200.0, 'usd'))
        self.assertEquals(parse_price('2300.23'), (2300.23, 'usd'))
        self.assertEquals(parse_price(2400), (2400.0, 'usd'))
        self.assertEquals(parse_price(2500.0), (2500.0, 'usd'))
