import json
import re
from apps.website.models import FormRow, Product, Website, Form
from nicechat.utils.form_templates import FORMS_TEMPLATES

ADDRESS_TPL = (
    ('first', 'First name',),
    ('last', 'Last name',),
    ('addr_1', 'Address'),
    ('addr_2', 'Address',),
    ('city', 'City',),
    ('state', 'State',),
    ('country', 'Country',),
    ('post', 'POST/ZIP',)
)


def form_validator(to_message, data):
    try:
        form = json.loads(to_message.data)
    except ValueError:
        return {}, False
    f_data = prepare_data(data, to_message.pk)
    errors = validate_form_fields(form, f_data)
    is_purchase = data.get('PRODUCTS', None)
    if errors:
        return errors, False, None
    else:
        results, tags = to_display_dict(form, f_data, is_purchase)
        return results, True, tags


def greet_form_validator(data, form):

    f_data = prepare_data(data, 'greet')
    errors = validate_form_fields(form, f_data)
    is_purchase = None
    if errors:
        return errors, False, []
    else:
        results, tags = to_display_dict(form, f_data, is_purchase)
        return results, True, tags


def final_form_validator(data, form):

    f_data = prepare_data(data, 'final')
    errors = validate_form_fields(form, f_data)
    is_purchase = None
    if errors:
        return errors, False, []
    else:
        results, tags = to_display_dict(form, f_data, is_purchase)
        print(results)
        return results, True, tags


def validate_form_fields(form, data):
    errors = {}
    for row in form.get('rows', []):
        row_id = row.get('id')
        row_kind = row.get('kind')
        row_is_req = row.get('is_required')
        row_params = row.get('params', '{}')
        row_val = data.get(row_id)
        if row_is_req and not row_val:
            errors[row_id] = 'IS_REQUIRED'
        elif row_val and not validate_by_type(row_kind, row_val, row_params):
            errors[row_id] = 'BAD_TYPE'
    return errors


def validate_by_type(kind, value, params):
    if kind == FormRow.FIELD_TYPE_EMAIL:
        if not re.match(r"^[a-zA-Z0-9_.-]+@[a-zA-Z0-9-]+.[a-zA-Z0-9-.]+$", value):
            return False
    elif kind == FormRow.FIELD_TYPE_NUMBER or kind == FormRow.FIELD_TYPE_RATE_STAR:
        try:
            int(value)
        except ValueError:
            return False
    elif kind == FormRow.FIELD_TYPE_US_STATE or kind == FormRow.FIELD_TYPE_COUNTRY:
        if 2 < len(value) > 3:
            return False
    elif kind in FormRow.TYPES_REQUIRE_CHOICE:
        try:
            params = json.loads(params)
        except ValueError:
            params = {}
        values = params.get('values', [])
        if isinstance(value, list):
            for val in value:
                if val not in values:
                    return False
        else:
            if value not in values:
                return False
    return True


def prepare_data(data, pk):
    result = {}
    t_key = "row_%s-" % pk
    for d_key, d_rec in data.items():
        if d_key.startswith(t_key):
            key = d_key.replace(t_key, '')
            key = key.split('__')[0]
            sub_mode = None
            try:
                if '-' in key:
                    (key, sub_mode) = key.split('-')
                key = int(key)
            except ValueError:
                continue
            if sub_mode is not None:
                if not result.get(key):
                    result[key] = {}
                result[key][sub_mode] = d_rec
                continue
            if result.get(key):
                if not isinstance(result.get(key), list):
                    result[key] = [result.get(key)]
                result.get(key).append(d_rec)
            else:
                result[key] = d_rec
    return result


def to_display_dict(form, data, is_purchase):
    result = [{
        'label': form.get('label'),
        'type': 'title'
    }]
    group = {}
    row_keys = []
    sorted_row = {}
    for row in form.get('rows', []):
        row_id = row.get('id', 0)
        row_order = row.get('order_num', 0)
        row_key = "%010d-%010d" % (row_order, row_id)
        row_keys.append(row_key)
        sorted_row[row_key] = row

    is_group = 'default'
    was_product_set = False
    group['default'] = {
        'data': [],
        'label': 'default',
        'type': 'flat'
    }

    tags = {}

    for row_key in sorted(row_keys):
        row = sorted_row[row_key]
        row_id = row.get('id', 0)

        if row['kind'] == 'address':
            is_group = row_key
            rows = []
            addr_rows = data.get(row_id, {})
            for key, lbl in ADDRESS_TPL:
                if addr_rows.get(key, None):
                    rows.append({
                        'type': 'text',
                        'label': lbl,
                        'value': addr_rows.get(key)
                    })
            group[is_group] = {
                'data': rows,
                'label': row['label'],
                'type': 'address'
            }
            is_group += 'defaul'
            group[is_group] = {
                'data': [],
                'label': is_group,
                'type': 'group'
            }
        elif row['kind'] == 'group':
            is_group = row_key
            group[is_group] = {
                'data': [],
                'label': row['label'],
                'type': 'group'
            }
        else:
            try:
                params = json.loads(row.get('params', '{}'))
            except ValueError:
                params = {}
            if is_group in group and data.get(row_id, None):
                group[is_group]['data'].append({
                    'type': params.get('display', 'text'),
                    'label': row['label'],
                    'value': data.get(row_id, '')
                })
                if params.get('auto-fill-tag', None) == 'PRODUCT_ID':
                    group[is_group]['type'] = 'product'
                    fill_prod_form(group[is_group], is_purchase)
                    was_product_set = True
                if params.get('tag', None):
                    tags[params.get('tag')] = data.get(row_id, '')

    if is_purchase and not was_product_set:
        dct = {
            'data': [],
            'label': 'Product',
            'type': 'product',
        }
        fill_prod_form(dct, is_purchase)
        result.append(dct)

    for group_key in sorted(group.keys()):
        if group[group_key].get('data', []):
            result.append(group[group_key])

    return result, tags


def fill_prod_form(data, prod_id):
    data['PRODUCTS'] = prod_id


def check_all_websites_forms():
    for website in Website.objects.all():
        check_website_forms(website)


def check_website_forms(website):
    for form_tpl in FORMS_TEMPLATES:
        form = Form.objects.filter(mode=form_tpl.get('mode'), website=website).first()
        if not form:
            form = Form.objects.create(
                website=website,
                company=website.company,
                access_write=Form.PERM_WRITE_AGENT,
                access_read=Form.PERM_READ_WEBSITE,
                is_enabled=True,
                mode=form_tpl.get('mode'),
                label=form_tpl.get('label')
            )

            for row in form_tpl.get('rows', []):
                FormRow.objects.create(form=form, **row)
