import os
import json
import time
import requests
from datetime import datetime
import requests
import csv
from xml.etree import ElementTree
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()
from nicechat.utils.constants import WebsocketEventTypes
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.core.cache import caches, InvalidCacheBackendError
from django.conf import settings
from nicechat.analytics import event_track, PRODUCTS_IMPORTED, PRODUCTS_IMPORT_FAILURE

from nicechat.utils.mail import notify_users
from .import_handlers import create_or_update_product, create_or_update_category
from ..models import Product, ProductsCategory
from nicechat.utils.helpers import parse_price


def do_process_import(site_id, user_pk, current_site_pk, data, csv=None, google_merchant_csv=None,
                      token=None, callback=None):
    imported_elements_count = 0
    source = ''
    source_feed = ''

    if token:
        try:
            cache = caches['default']
            ret = cache.get(token)
            ret['updated'] = time.time()
            ret['state'] = 'started'
            cache.set(token, ret)
        except InvalidCacheBackendError:
            pass

    current_site = Site.objects.filter(pk=current_site_pk).first()
    user = get_user_model().objects.filter(pk=user_pk).first()

    try:
        if google_merchant_csv:
            source = 'Google'
            source_feed = google_merchant_csv
            imported_elements_count = parse_google_merchant(google_merchant_csv, site_id)
        elif csv:
            source = 'CSV'
            content = data.get('content')
            delimeter = data.get('delimeter')
            indexes = data.get('indexes')
            imported_elements_count = parse_csv(content,
                                                delimeter,
                                                indexes,
                                                site_id)
        if user and current_site:
            notify_users([user],
                     'Import success',
                     'website/email/import_success.txt',
                     'website/email/import_success.html',
                     dict(user=user,
                          current_site=current_site,
                          imported_elements_count=imported_elements_count, ))

        event_track(PRODUCTS_IMPORTED, user.id if user else None, {
            'siteID': site_id,
            'imported_elements_count': imported_elements_count,
            'source': source,
            'source_feed': source_feed
        })

        data = {"type": "websocket.send", "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_SERVER_NOTIFY,
                                    "data": {
                                        "text": "Feed %(feed)s successfully imported. %(count)s message added." % {
                                                    'feed': source_feed,
                                                    'count': imported_elements_count
                                                },
                                        "level": 'success'
                                    }}
               )}
        async_to_sync(channel_layer.group_send)("adm-website_mon-%s" % site_id, data)
        print("adm-website_mon-%s" % site_id, data)
        if token:
            try:
                cache = caches['default']
                ret = cache.get(token)
                ret['updated'] = time.time()
                ret['state'] = 'success'
                ret['imported elements count'] = imported_elements_count
                cache.set(token, ret)
            except InvalidCacheBackendError:
                pass

        if callback:
            make_callback(callback, "success", datetime.utcnow().isoformat(), imported_elements_count)

    except Exception as ex:
        event_track(PRODUCTS_IMPORT_FAILURE, user.id if user else None, {
            'siteId': site_id,
            'source': source,
            'source_feed': source_feed,
            'description': str(ex)
        })

        if callback:
            make_callback(callback, "failed", datetime.utcnow().isoformat(), 0)

        data = {"type": "websocket.send", "text": json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_SERVER_NOTIFY,
                                    "data": {
                                        "text": "Feed %(feed)s not imported." % {
                                                    'feed': source_feed
                                                },
                                        "level": 'error'
                                    }}
               )}
        if token:
            try:
                cache = caches['default']
                ret = cache.get(token)
                ret['updated'] = time.time()
                ret['state'] = 'failed'
                cache.set(token, ret)
            except InvalidCacheBackendError:
                pass

        async_to_sync(channel_layer.group_send)("adm-website_mon-%s" % site_id, data)
        print("adm-website_mon-%s" % site_id, data)


def make_callback(to_url, state, timestamp, updated):
    try:
        data = {'state': state, 'updated': timestamp}
        if updated:
            data['imported elements count'] = updated
        requests.post(to_url, data=data)
    except Exception:
        pass


def get_xml_from_url(url):
    if not url.startswith('file://'):
        user_agent = {'User-agent': 'Mozilla/5.0'}
        response = requests.get(url, headers=user_agent)
        tree = ElementTree.fromstring(response.content)
        return tree
    else:
        with open(url[5:]) as f:
            content = f.read()
            tree = ElementTree.fromstring(content)
            return tree


# @TODO add format checks and all the fields
def parse_google_merchant(url, site_id):
    if not url.startswith('file://'):
        user_agent = {'User-agent': 'Mozilla/5.0'}
        response = requests.get(url, headers=user_agent)
        rows = response.content.decode('utf-8').splitlines()
    else:
        with open(url[5:]) as f:
            content = f.read()
            rows = content.decode('utf-8').splitlines()

    delimiter = detect_delimiter(rows[0]) if len(rows) else False
    csv_rows = csv.reader(rows, delimiter=delimiter) if delimiter else csv.reader(rows)
    count = 0
    for index, row in enumerate(csv_rows):
        if index == 0:
            i_id = row.index('id')
            i_title = row.index('title')
            i_description = row.index('description')
            i_url = row.index('link')
            i_image_url = row.index('image link')
            i_price = row.index('price')
            i_availability = row.index('availability')
        else:
            price, currency = parse_price(row[i_price])
            count += 1
            create_or_update_product(title=row[i_title],
                                     price=price,
                                     currency=currency,
                                     description=row[i_description] if row[i_description] is not None else '',
                                     url=row[i_url],
                                     website_id=site_id,
                                     image_url=row[i_image_url],
                                     original_id=row[i_id],
                                     category_id=None
                                     )
    return count


def get_node_param(obj, prop="text", default=None):
    if obj is not None:
        return getattr(obj, prop, default)
    else:
        return default

def detect_delimiter(top_row):
    for delimiter in [',', ';', '\t']:
        c = csv.reader([top_row], delimiter=delimiter)
        for index, item in enumerate(c):
            if index == 0 and len(item) > 1:
                return delimiter
    return False


def parse_csv(file_content, delimeter, indexes, site_id):
    csv_rows = csv.reader(file_content.splitlines())

    for index, row in enumerate(csv_rows):
        if index != 0:
            Product.objects.create(title=row[0].split(delimeter)[indexes.get('title')],
                                    price=float(row[0].split(delimeter)[indexes.get('price')].strip('"')),
                                    # currency=row[0].split(delimeter)[indexes.get('currency')],
                                    # description=row[0].split(delimeter)[indexes.get('description')],
                                    # url=row[0].split(delimeter)[indexes.get('url')],
                                    # image_url=row[0].split(delimeter)[indexes.get('image_url')],
                                    website_id=site_id)
    return len(csv_rows)

