import json
import time
import threading
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()
from django.core.cache import caches, InvalidCacheBackendError
from django.conf import settings
from django.contrib.auth import get_user_model

from apps.chat.models import Visitor, Chat
from ..models import Website, AgentStatus, Status
from apps.visitors.serializers import VisitorSerializer
from apps.chat.serializers import ChatSerializer
from apps.website.serializers import AgentSerializer, StatusSerializer
from nicechat.analytics import event_track, AGENT_JOIN_SITE, AGENT_LEAVE_SITE, event_people_set

from nicechat.utils.constants import WebsocketEventTypes


def agent_send_lists(site_id, company, reply_channel):
    v_q = Visitor.objects.filter(website_id=site_id)
    reply_channel.send(text_data=json.dumps({
        'type': WebsocketEventTypes.BACKEND_SITE_VISITOR_LIST,
        'data': VisitorSerializer(v_q, many=True).to_representation(v_q)
    }))
    c_q = Chat.objects.filter(website_id=site_id)
    reply_channel.send(text_data=json.dumps({
        'type': WebsocketEventTypes.BACKEND_SITE_CHAT_LIST,
        'data': ChatSerializer(c_q, many=True).to_representation(c_q)
    }))
    a_q = get_user_model().objects.filter(company=company, is_active=True, website__id=site_id)
    reply_channel.send(text_data=json.dumps({
        'type': WebsocketEventTypes.BACKEND_SITE_OPERATOR_LIST,
        'data': AgentSerializer(a_q, many=True).to_representation(a_q)
    }))


def on_agent_connect(site_obj, user_id):
    token = "%s-%s" % (site_obj.pk, user_id)
    try:
        cache = caches['default']
        if not cache.get(token):
            event_track(AGENT_JOIN_SITE, user_id, {'siteId': site_obj.pk})
            user = get_user_model().objects.filter(pk=user_id).first()

            if user and user.is_site_admin(site_obj):
                websites_without_script = ', '.join(map(str, [w.domain for w in user.website_set.filter(
                    status__widget_installed__exact=False)]))
                website_not_reachable = ', '.join(
                    map(str, [w.domain for w in user.website_set.filter(status__isnull=True)]))

                if websites_without_script == '':
                    websites_without_script = '---'

                if website_not_reachable == '':
                    website_not_reachable = '---'

                event_people_set(user_id, {
                    '$first_name': user.first_name,
                    '$last_name': user.last_name,
                    '$email': user.email,
                    'companyId': user.company_id,
                    'companyName': user.company.label,
                    'signed_up': user.date_joined.replace(microsecond=0).isoformat(),
                    'websites': ', '.join(map(str, [w.domain for w in user.website_set.all()])),
                    'website_without_script': websites_without_script,
                    'website_not_reachable': website_not_reachable,
                    'is_owner': user.company.owner == user
                })
        cache.delete(token)
    except InvalidCacheBackendError:
        pass


def on_agent_disconnect(site_id, user_id):
    token = "%s-%s" % (site_id, user_id)
    try:
        cache = caches['default']
        cache.set(token, '1', settings.AGENT_LEAVE_TIMEOUT_S * 1.5)
    except InvalidCacheBackendError:
        pass

    if not hasattr(Website, 'monitor_thread'):
        t = threading.Thread(target=agent_cleaner_worker, )
        t.daemon = True
        t.start()
        setattr(Website, 'monitor_thread', t)

    Website.offline_agents[token] = time.time()


def agent_cleaner_worker():

    import time as time_l

    while True:
        time_l.sleep(2)

        close_time = time_l.time() - settings.AGENT_LEAVE_TIMEOUT_S

        for vs, ts in list(Website.offline_agents.items()):
            if ts < close_time:
                if vs in Website.offline_agents.keys():
                    del Website.offline_agents[vs]
                try:
                    cache = caches['default']
                    if cache.get(vs):
                        cache.delete(vs)
                        try:
                            user_id, site_id = vs.split('-')
                            event_track(AGENT_LEAVE_SITE, user_id, {'siteId': site_id})
                        except ValueError:
                            pass
                except InvalidCacheBackendError:
                    pass


def send_agent_status(site_id):
    active = True if AgentStatus.objects.filter(website_id=site_id, is_active=True).count() > 0 else False
    async_to_sync(channel_layer.group_send)('cli_site-%s' % site_id, {"type": "websocket.send", "text":  json.dumps({"type": "is_agents", "val": active})})


def send_site_status(site_id):
    instance = Status.objects.filter(website_id = site_id).first()
    if instance:
        a_key = "adm-website_mon-%i" % instance.website_id
        async_to_sync(channel_layer.group_send)(a_key, {"type": "websocket.send",  "text":  json.dumps({"type": WebsocketEventTypes.BACKEND_SITE_STATUS_UPDATE,
                                        "data": dict(StatusSerializer(instance=instance,
                                                                    context={'request': None,
                                                                            'site_id': instance.website_id})
                                                    .to_representation(instance=instance)),
                                        "pk": instance.pk
                                    })})
