import json
import time
import base64
import struct
import hashlib
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync
channel_layer = get_channel_layer()
from channels.generic.websocket import WebsocketConsumer
from django.conf import settings

from .models import Website, AgentStatus
from urllib.parse import parse_qs
from nicechat.utils.constants import WebsocketEventTypes
from .utils.ws_list import agent_send_lists, on_agent_connect, on_agent_disconnect, send_agent_status, send_site_status
from nicechat.utils.helpers import auth_jwt
from django.core.cache import caches, InvalidCacheBackendError
from nicechat.utils.video import get_video_provider, video_stop_call, video_start_call
from nicechat.utils.arbiter_sync import WatcherMiddleware


class EventStreamConsumer(WebsocketConsumer):
    method_mapping = {
        "websocket.connect": "connect",
        "websocket.receive": "receive",
        "websocket.send": "websocket_send",
        "websocket.disconnect": "disconnect",
    }

    def connect(self):
        try:
            site_id = int(self.scope['url_route'].get('kwargs', {}).get('site_id', None))
        except ValueError:
            site_id = None

        # if User allowed for this Website
        qs = self.scope.get('query_string').decode() if self.scope and self.scope.get('query_string') else None
        user = None
        is_current = False
        if qs:
            token = parse_qs(qs).get('token', []).pop()
            is_current = bool(parse_qs(qs).get('is_current', None))
            result, detail = auth_jwt(token)
            if result:
                user = detail

        if site_id and user:
            website = Website.objects.filter(pk=site_id, users__pk=user.pk).first()
            if website and website.online:
                self.accept()
                self.scope["session"]['adm-website_user'] = user.pk
                self.scope["session"]['adm-website_user_name'] = user.get_full_name or str(user)
                token = "adm-website_mon-%i" % site_id
                async_to_sync(self.channel_layer.group_add)(token, self.channel_name)
                token = "website_global-%i" % site_id
                async_to_sync(self.channel_layer.group_add)(token, self.channel_name)
                token = "adm-website_mon-u-%i-%i" % (site_id, user.id)
                async_to_sync(self.channel_layer.group_add)(token, self.channel_name)
                agent_send_lists(site_id, user.company, self)
                active, created = AgentStatus.objects.get_or_create(website=website, operator=user)
                active.is_active = True
                active.is_current = is_current
                active.state_timestamp = int(time.time())
                active.save()
                if is_current:
                    AgentStatus.objects.filter(operator=user, is_current=True).exclude(website=website).update(is_current=False)
                on_agent_connect(website, user.id)
                send_agent_status(site_id)
                send_site_status(site_id)
                return

        self.close()


    def receive(self, text_data="", bytes_data=None):
        try:
            dt = json.loads(text_data)
        except ValueError:
            return
        website_monitor_packet(dt, 
                               self.scope['url_route'].get('kwargs', {}),
                               self.scope["session"].get('adm-website_user', None),
                               self.scope["session"].get('adm-website_user_name', None))
        site_id = self.scope['url_route'].get('kwargs', {}).get('site_id')
        WatcherMiddleware.stats[f"ws:{site_id}"] = 1 + WatcherMiddleware.stats.get(f"ws:{site_id}", 0)

    def websocket_send(self, message):
        if message.get('text') == "setoffline":
            self.close()
            return
        self.send(text_data=message.get('text'))

    def disconnect(self, close_code):
        try:
            site_id = int(self.scope['url_route'].get('kwargs', {}).get('site_id', None))
        except ValueError:
            site_id = None
        user_id = self.scope["session"].get('adm-website_user', None)

        if site_id:
            token = "adm-website_mon-%s" % site_id
            async_to_sync(self.channel_layer.group_discard)(token, self.channel_name)
            token = "website_global-%i" % site_id
            async_to_sync(self.channel_layer.group_discard)(token, self.channel_name)

        if site_id and user_id:
            token = "adm-website_mon-u-%i-%i" % (site_id, user_id)
            async_to_sync(self.channel_layer.group_discard)(token, self.channel_name)
            active = AgentStatus.objects.filter(website_id=site_id, operator_id=user_id).first()
            if active:
                active.is_active = False
                active.state = AgentStatus.STATUS_AWAY
                active.state_timestamp = int(time.time())
                active.save(update_fields=('state', 'state_timestamp', 'is_active',))
                send_agent_status(site_id)
            on_agent_disconnect(site_id, user_id)


def website_monitor_packet(dt, kwargs, user_id, user_name):
    try:
        site_id = int(kwargs.get('site_id', None))
    except ValueError:
        return

    if site_id and isinstance(dt, dict) and user_id:
        _type = dt.get('type', None)
        if _type == WebsocketEventTypes.BACKEND_SITE_SET_OPERATOR_STATE:
            active = AgentStatus.objects.filter(website_id=site_id, operator_id=user_id).first()
            if active:
                active.is_active = True
                active.state = dt.get('state', AgentStatus.STATUS_ONLINE)
                active.state_timestamp = int(time.time())
                active.save(update_fields=('state', 'state_timestamp', 'is_active',))
        elif _type == WebsocketEventTypes.BACKEND_SITE_SEND_NOTIFY:
            _call = dt.get('call', None)
            # @TODO deactivated for now
            # if _call == 'VisitorSelected':
            #     event_track(BACKEND_VISITOR_SELECT, user_id, {
            #         'clientID': dt.get('client_id'),
            #         'chatId': site_id,
            #     })

        elif _type == WebsocketEventTypes.BACKEND_SITE_WS_CALL_PING:
            chat_call_ping(dt, kwargs, user_id, user_name)
        elif _type == WebsocketEventTypes.BACKEND_SITE_WS_CALL_STOP:
            chat_call_stop(dt, kwargs, user_id, user_name)
        elif _type == WebsocketEventTypes.BACKEND_SITE_WS_CALL_START:
            chat_call_start(dt, kwargs, user_id, user_name)
        elif _type == WebsocketEventTypes.BACKEND_SITE_START_TYPING:
            chat_id = dt.get('chatId', 0)
            token = "chat-typing-%s-%s" % (chat_id, user_id)
            try:
                cache = caches['default']
                cache.set(token, dt.get('text', '...'), 20)
            except InvalidCacheBackendError:
                pass
            async_to_sync(channel_layer.group_send)('chat-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
                'type': WebsocketEventTypes.BACKEND_SITE_START_TYPING,
                'id': user_id,
                'name': user_name,
                'text': dt.get('text', '...')
            })})
            async_to_sync(channel_layer.group_send)('cli_speak-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_TYPING_START,
                'name': user_name,
            })})
        elif _type == WebsocketEventTypes.BACKEND_SITE_STOP_TYPING:
            chat_id = dt.get('chatId', 0)
            token = "chat-typing-%s-%s" % (chat_id, user_id)
            try:
                cache = caches['default']
                cache.delete(token)
            except InvalidCacheBackendError:
                pass
            async_to_sync(channel_layer.group_send)('chat-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
                'type': WebsocketEventTypes.BACKEND_SITE_STOP_TYPING,
                'name': user_name,
                'id': user_id,
            })})
            async_to_sync(channel_layer.group_send)('cli_speak-%s' % chat_id, {"type": "websocket.send", 'text':  json.dumps({
                'type': WebsocketEventTypes.FRONTEND_CHAT_WS_TYPING_END,
                'name': user_name,
            })})


def chat_call_start(dt, kwargs, user_id, user_name):
    from ..chat.tasks import celery_chat_call_wdt
    from ..chat.models import Chat
    from ..video.models import Call, CallPeer
    room_id = dt.get('chat_id')
    if room_id:
        chat = Chat.objects.filter(pk=room_id).first()
        if chat:
            cl, _ = Call.objects.get_or_create(
                chat=chat,
                is_active=True,
                provider=get_video_provider(),
                defaults={'client_accept': False}
            )
            peer, is_peercreated = CallPeer.objects.get_or_create(
                call=cl,
                peer='ag_%s' % user_id,
                defaults={
                    'props': json.dumps({
                        'user_id': user_id,
                        'user_name': user_name
                    })
                }
            )
            if cl.client_accept:
                video_start_call(cl, chat)
                return
            if not is_peercreated:
                return
            chat.clear_call_messages()
            chat.put_tech_message_wdata("~!CALL_AGT_BEGIN:1", {
                'name': user_name,
                'provider': get_video_provider(),
                'time': time.time()
            })
            celery_chat_call_wdt.apply_async(args=[chat.pk], countdown=20)


def chat_call_stop(dt, kwargs, user_id, user_name):
    from ..chat.tasks import celery_chat_call_wdt
    from ..video.models import Call, CallPeer
    room_id = dt.get('chat_id')
    for cl in Call.objects.filter(chat__pk=room_id, is_active=True):
        video_stop_call(cl, 'ag_%s' % user_id, user_id)


def chat_call_ping(dt, kwargs, user_id, user_name):
    pass
