import braintree
import logging
from .utils import convert_braintree_status
from nicechat.analytics import event_track, SUBSCRIPTION_CANCELLED, PLAN_CHANGED, PLAN_CHANGE_FAIL
from braintree.exceptions.braintree_error import BraintreeError
from nicechat.utils.billing import BillingError

braintree_logger = logging.getLogger('billing')


def sync(company):
    from apps.billing.models import PaymentMethod, Plan, Transaction
    from apps.company.models import Company
    from . import PAYMENT_NAME

    assert type(company) is Company
    assert company.billing_id is not None

    braintree_logger.debug("Searching transactions for company "
                           "with billing id of \"{}\"".format(company.billing_id))
    transactions = braintree.Transaction.search([
        braintree.TransactionSearch.customer_id == company.billing_id
    ])
    braintree_logger.debug("Result of transactions search for company "
                           "with billing id of \"{}\":\n{}".format(company.billing_id, transactions.ids))

    for braintree_transaction in transactions:
        if braintree_transaction.payment_instrument_type == braintree.PaymentInstrumentType.CreditCard:
            try:
                payment_method = PaymentMethod.objects.get(
                    billing_token=braintree_transaction.credit_card_details.token,
                    billing_provider=PAYMENT_NAME
                )
            except PaymentMethod.DoesNotExist:
                payment_method = PaymentMethod.objects.create(
                    billing_token=braintree_transaction.credit_card_details.token,
                    billing_provider=PAYMENT_NAME,
                    company=company,
                    type=PaymentMethod.CREDIT_CARD,
                    card_number=braintree_transaction.credit_card_details.masked_number
                )
        else:
            braintree_logger.error('Braintree transaction ("{}") cannot be processed '
                                   'due to unsupported payment method type: "{}"'.format(
                braintree_transaction.payment_method_token,
                braintree_transaction.payment_instrument_type
            ))
            # todo mail_admins, event_track ?
            continue

        payment_plan, payment_period = Plan.get_by_billing_id(braintree_transaction.plan_id)

        Transaction.objects.update_or_create(
            billing_id=braintree_transaction.id,
            billing_provider=PAYMENT_NAME,
            defaults={
                'company': company,
                'plan': payment_plan,
                'plan_period': payment_period,
                'user': company.owner,
                'payment_method': payment_method,
                'timestamp': braintree_transaction.created_at,
                'result': braintree_transaction,
                'status': convert_braintree_status(braintree_transaction.status),
                'amount': braintree_transaction.amount,
            }
        )

        braintree_logger.debug('Database record for transaction "{}" has been created'.format(braintree_transaction.id))


def apply_payment(billing_id, validated_plan_id, company_plan_price, braintree_token, promo_code_obj, agents_count, request_user):
    # prepare the data of required format
    if not billing_id:
        raise BillingError('No billing plan ID')

    from . import PAYMENT_NAME

    braintree_request_data = {
        "payment_method_token": braintree_token,
        "plan_id": billing_id,
        "price": company_plan_price
    }

    if promo_code_obj:
        braintree_request_data['discounts'] = {
            'add': [{
                'amount': (company_plan_price * promo_code_obj.discount_percent) / 100,
                'inherited_from_id': promo_code_obj.billing_id,
                'number_of_billing_cycles': 1
            }]
        }

    print(braintree_request_data)


    braintree_logger.debug('Create subscription request: {}'.format(braintree_request_data))
    result = braintree.Subscription.create(braintree_request_data)
    braintree_logger.debug('Create subscription request result: {}'.format(result))

    result_code = False
    result_obj = {}

    if result.is_success:
        result_code = True
        result_obj['subscription_id'] = result.subscription.id
        result_obj['billing_provider'] = PAYMENT_NAME
        # todo result.subscription.transactions - list, where each transaction has the following structure:
        # https://developers.braintreepayments.com/reference/response/transaction/python
        # todo business_subscription.next_billing_date
        # todo business_subscription.next_billing_period_amount

        if promo_code_obj and len(result.subscription.discounts):
            result_obj['promo_code_obj'] = promo_code_obj


        event_track(PLAN_CHANGED, user_id=request_user.id, params={
            'companyId': request_user.own_company.pk,
            'plan': {
                'id': validated_plan_id,
                'billing_id': billing_id,
            }
        })
    else:
        event_track(PLAN_CHANGE_FAIL, user_id=request_user.id, params={
            'companyId': request_user.own_company.pk,
            'errorMessage': result.message,
            'plan': {
                'id': validated_plan_id,
                'billing_id': billing_id,
            }
        })
        raise BillingError({'Braintree error': result.message})

    return result_code, result_obj


def cancel_subscription(braintree_subscription_id, company_plan):
    braintree_logger.debug('Cancellation request for subscription with id="{}"'.format(
        braintree_subscription_id
    ))
    result = braintree.Subscription.cancel(braintree_subscription_id)
    braintree_logger.debug('Cancellation request for subscription with id="{}" result: {}'.format(
        braintree_subscription_id, result
    ))

    if not result.is_success:
        raise BraintreeError(result.message)

    event_track(SUBSCRIPTION_CANCELLED, params={
        'companyId': company_plan.company.pk,
        'plan': company_plan.plan.get_billing_id(company_plan.payment_period)
    })
    return result.is_success

# todo process Braintree webhooks
# (https://developers.braintreepayments.com/reference/general/webhooks/subscription/python)
