from __future__ import absolute_import

import decimal
import unittest
import braintree
from datetime import datetime

from django.conf import settings
from django.utils import timezone
from rest_framework import status
from braintree.test.nonces import Nonces

from nicechat.utils.tests import TestCaseWithSites
from .utils import get_trial_plan, sync_company_braintree_transactions
from .models import PaymentMethod, Plan, CompanyPlan, BillingCycles, PromoCode, PromoCodeUsage
from .utils import create_customers_braintree_accounts


class PaymentMethodsAPITestCase(TestCaseWithSites):
    def setUp(self):
        super(PaymentMethodsAPITestCase, self).setUp()

        self.pm_api_url = self.api_url + 'payment-methods/'

        if settings.BILLING_ENABLED:
            create_customers_braintree_accounts()
            self.refresh_companies_from_db()

    def test_non_authenticated_error(self):
        response = self.client.get(self.pm_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_payment_methods(self):
        self.assertEqual(PaymentMethod.objects.all().count(), 0)

        pm_data_invalid = {
            'card_number': 'some_invalid_string',
        }

        pm_data_valid = {
            'nonce': Nonces.TransactableDebit
        }

        # non-authenticated users access is forbidden
        response = self.client.post(self.pm_api_url, pm_data_valid, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # access is restricted to company owners only
        self.login_user(self.user_1_1)
        response = self.client.post(self.pm_api_url, pm_data_valid, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        response = self.client.get(self.pm_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.login_user(self.user_owner_0)

        # no payment methods yet
        response = self.client.get(self.pm_api_url, format='json')
        self.assertEqual(response.data['count'], 0)
        self.assertEqual(PaymentMethod.objects.count(), 0)

        # invalid data
        response = self.client.post(self.pm_api_url, pm_data_invalid, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['nonce'])

        if not settings.BILLING_ENABLED:
            raise unittest.SkipTest("No Braintree payment methods testing "
                                    "due to `settings.BILLING_ENABLED` set to False")

        # successful creation
        response = self.client.post(self.pm_api_url, pm_data_valid, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(response.data['id'] is None)
        self.assertEqual(PaymentMethod.objects.count(), 1)
        self.assertEqual(self.company0.payment_methods.count(), 1)
        card_pk = response.data['id']

        # payment method list now has one item
        response = self.client.get(self.pm_api_url, format='json')
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(PaymentMethod.objects.all().count(), 1)

        # payment method can only be created for current user's company
        new_payment_method = PaymentMethod.objects.get(pk=card_pk)
        self.assertEquals(new_payment_method.company, self.user_owner_0.own_company)
        # credit card type by default
        self.assertEquals(new_payment_method.type, PaymentMethod.CREDIT_CARD)
        self.assertIsNotNone(new_payment_method.braintree_token)

        # log in another company owner
        self.login_user(self.user_owner_1)

        response = self.client.get(self.pm_api_url, format='json')
        # payment method api endpoint lists only current user's company methods
        self.assertEqual(response.data['count'], 0)

        # todo {"Braintree error":"Duplicate card exists in the vault."}

        # successful creation of new payment method for another company
        response = self.client.post(self.pm_api_url, {'nonce': Nonces.ApplePayMasterCard}, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        another_card_pk = response.data['id']

        # two objects in database
        self.assertEqual(PaymentMethod.objects.all().count(), 2)

        # payment method list has only one item though - the one belonging to current user's company
        response = self.client.get(self.pm_api_url, format='json')
        self.assertEqual(response.data['count'], 1)

        # payment method can only be created for current user's company
        another_payment_method = PaymentMethod.objects.get(pk=another_card_pk)
        self.assertEquals(another_payment_method.company, self.user_owner_1.own_company)
        self.assertEquals(another_payment_method.type, PaymentMethod.CREDIT_CARD)

        # todo selecting/changing primary payment method functionality


class PlansAPITestCase(TestCaseWithSites):
    def setUp(self):
        super(PlansAPITestCase, self).setUp()

        if settings.BILLING_ENABLED:
            create_customers_braintree_accounts()
            self.refresh_companies_from_db()

        self.bp_api_url = self.api_url + 'plans/'
        self.company_plan_api_url = self.api_url + 'company-plan/'
        self.bp_detail_api_url = self.bp_api_url + '1/'

        self.pm_api_url = self.api_url + 'payment-methods/'
        self.promo_validation_api_url = self.api_url + 'validate-promo-code/'

        self.free_plan = Plan.objects.get(name='Free')
        self.business_plan = Plan.objects.get(name='Business')
        self.enterprise_plan = Plan.objects.get(name='Enterprise')

        self.fifty_off_code = '50off'
        self.fifty_percent_off_code = PromoCode.objects.create(code=self.fifty_off_code,
                                                               discount_percent=50,
                                                               billing_id='test_50off')

    def test_only_list_read(self):
        # reading list api endpoint is allowed
        response = self.client.get(self.bp_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # every other method is not allowed
        response = self.client.post(self.bp_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)
        response = self.client.get(self.bp_detail_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        response = self.client.put(self.bp_detail_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        response = self.client.delete(self.bp_detail_api_url, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    # todo test that self.company_plan_api_url returns only billing plan of a current company

    # see apps.website.tests.ProductsTestCase#test_features_permissions for products-related features tests

    # todo add more plan features tests + plans numeric restrictions tests

    def test_change_plan_errors(self):
        # unauthorized access is forbidden
        response = self.client.get(self.company_plan_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # access is restricted to company owners only
        self.login_user(self.user_1_1)
        response = self.client.put(self.company_plan_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # log in a company owner
        self.login_user(self.user_owner_1)

        self.assertEqual(self.company1.payment_methods.count(), 0)

        # can't change billing plan (to non-free) without any payment methods previously set up
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': 3,
                                    'payment_period': BillingCycles.YEAR},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['payment_method'])

        if not settings.BILLING_ENABLED:
            raise unittest.SkipTest("No Braintree plan changing testing "
                                    "due to `settings.BILLING_ENABLED` set to False")

        # create new payment method for self.company1
        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactableCommercial},
                                    format='json')

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.company1.payment_methods.count(), 1)

        payment_method_id = self.company1.payment_methods.first().pk

        # post method is not allowed
        response = self.client.post(self.company_plan_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

        #
        #  Changing plan
        #

        # no data
        response = self.client.put(self.company_plan_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['agents_count', 'plan'])

        # wrong plan_id
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': 'some_random_id_value', 'agents_count': 5,
                                    'payment_period': BillingCycles.MONTH},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['plan'])

        response = self.client.put(self.company_plan_api_url,
                                   {'plan': 123456789, 'agents_count': 5,
                                    'payment_period': BillingCycles.MONTH},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['plan'])

        # switching to trial plan is impossible
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': get_trial_plan().pk, 'agents_count': 5,
                                    'payment_period': BillingCycles.MONTH},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['plan'])

        # "0" as agents_num
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': 0,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': payment_method_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['agents_count'])

        # wrong value for payment_period
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': 3,
                                    'payment_period': 'some_random_str', 'payment_method': payment_method_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['payment_period'])

        # still can skip sending 'payment_period' and 'payment_method' for the "Free" plan
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.free_plan.pk, 'agents_count': 3},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # wrong value for payment_method parameter
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': 3,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': 'some_random_str'},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['payment_method'])

        #
        # another company payment method
        #
        # login company0 owner and create new payment method for self.company0
        self.login_user(self.user_owner_0)
        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactableMasterCard},
                                    format='json')
        company_0_pm_id = response.data['id']

        # log in the company1 owner back
        self.login_user(self.user_owner_1)

        # another company payment method connot be used
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': 3,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_0_pm_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['payment_method'])

    def test_change_plan(self):
        # log in the company owner
        self.login_user(self.user_owner_1)

        # initial assertions
        self.assertEqual(self.company1.billing_plan.plan, self.free_plan)
        self.assertEqual(self.company1.billing_plan.agents_count, 1)
        self.assertEqual(self.company1.billing_plan.payment_period, BillingCycles.MONTH)
        self.assertEqual(self.company1.payment_methods.count(), 0)

        if not settings.BILLING_ENABLED:
            raise unittest.SkipTest("No Braintree plan changing testing "
                                    "due to `settings.BILLING_ENABLED` set to False")

        # create new payment method for self.company1
        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactableMasterCard},
                                    format='json')
        company_1_pm_id = response.data['id']

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.company1.payment_methods.count(), 1)
        self.assertIsNotNone(self.company1.payment_methods.first().braintree_token)

        # change plan from "Free" to "Business"
        agents_needed = 5
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': agents_needed,
                                    'payment_period': BillingCycles.YEAR, 'payment_method': company_1_pm_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # correct updated values in response
        self.assertEqual(response.data['plan'], self.business_plan.pk)
        self.assertEqual(response.data['agents_count'], agents_needed)

        # related BillingPlan object was cached on Company instance, the old values are still there
        # self.assertEqual(self.company1.billing_plan.plan, self.business_plan)
        # self.assertEqual(self.company1.billing_plan.agents_count, 5)

        # re-fetching corresponding object from db
        updated_company_plan_obj = CompanyPlan.objects.get(pk=self.company1.billing_plan.pk)
        self.assertEqual(updated_company_plan_obj.plan, self.business_plan)
        self.assertEqual(updated_company_plan_obj.agents_count, agents_needed)
        self.assertEqual(updated_company_plan_obj.payment_period, BillingCycles.YEAR)

        business_subscription_id = updated_company_plan_obj.braintree_subscription_id
        self.assertIsNotNone(business_subscription_id)

        # a request to Braintree for the newly created subscription data
        business_subscription = braintree.Subscription.find(business_subscription_id)

        self.assertEqual(business_subscription.plan_id, self.business_plan.get_billing_id(BillingCycles.YEAR))
        self.assertEqual(business_subscription.status, braintree.Subscription.Status.Active)
        self.assertEqual(business_subscription.price, agents_needed * self.business_plan.annual_price * 12)

        company_1_pm_token = PaymentMethod.objects.get(pk=company_1_pm_id).braintree_token
        self.assertEqual(business_subscription.payment_method_token, company_1_pm_token)
        self.assertEqual(business_subscription.transactions[0].vault_credit_card.token, company_1_pm_token)

        #
        # Changing to free plan resets the braintree_subscription_id and cancels active braintree subscription (if any)
        #

        # change plan from "Business" to "Free"
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.free_plan.pk, 'agents_count': 50,
                                    'payment_period': BillingCycles.YEAR},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # correct updated values in response
        self.assertEqual(response.data['plan'], self.free_plan.pk)
        # when switching to free plan only certain number of agents is allowed
        # (regardless of what data was sent to "change plan" endpoint)
        self.assertEqual(response.data['agents_count'], self.free_plan.agents_num)

        # updated data assertions
        updated_company_plan_obj = CompanyPlan.objects.get(pk=self.company1.billing_plan.pk)
        self.assertEqual(updated_company_plan_obj.plan, self.free_plan)
        self.assertEqual(updated_company_plan_obj.agents_count, self.free_plan.agents_num)

        self.assertIsNone(updated_company_plan_obj.braintree_subscription_id)

        # a request to Braintree for the previously created subscription data
        business_subscription = braintree.Subscription.find(business_subscription_id)
        self.assertEqual(business_subscription.status, braintree.Subscription.Status.Canceled)

        # current number of active subscriptions
        search_results = braintree.Subscription.search(
            braintree.SubscriptionSearch.status == braintree.Subscription.Status.Active
        )
        num_of_active_subscriptions_init = search_results.maximum_size

        # change plan from "Free" to "Business" (one more time)
        agents_needed_updated = 15
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': agents_needed_updated,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_1_pm_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # correct updated values in response
        self.assertEqual(response.data['plan'], self.business_plan.pk)
        self.assertEqual(response.data['agents_count'], agents_needed_updated)

        # re-fetching corresponding object from db
        updated_company_plan_obj = CompanyPlan.objects.get(pk=self.company1.billing_plan.pk)
        self.assertEqual(updated_company_plan_obj.plan, self.business_plan)
        self.assertEqual(updated_company_plan_obj.agents_count, agents_needed_updated)
        self.assertEqual(updated_company_plan_obj.payment_period, BillingCycles.MONTH)

        business_subscription_id = updated_company_plan_obj.braintree_subscription_id
        self.assertIsNotNone(business_subscription_id)

        # a request to Braintree for the newly created subscription data
        business_subscription = braintree.Subscription.find(business_subscription_id)

        self.assertEqual(business_subscription.plan_id, self.business_plan.get_billing_id(BillingCycles.MONTH))
        self.assertEqual(business_subscription.status, braintree.Subscription.Status.Active)
        self.assertEqual(business_subscription.price, agents_needed_updated * self.business_plan.month_price)

        # current number of active subscriptions increased by one
        search_results = braintree.Subscription.search(
            braintree.SubscriptionSearch.status == braintree.Subscription.Status.Active
        )
        self.assertEqual(search_results.maximum_size, num_of_active_subscriptions_init + 1)

        #
        # switching from "business" to "enterprise" plan
        #
        agents_needed_updated_upd = 10
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.enterprise_plan.pk, 'agents_count': agents_needed_updated_upd,
                                    'payment_period': BillingCycles.YEAR, 'payment_method': company_1_pm_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        company_plan_obj_upd = CompanyPlan.objects.get(pk=self.company1.billing_plan.pk)
        enterprise_subscription_id = company_plan_obj_upd.braintree_subscription_id

        # newly created "Enterprise" subscription data
        enterprise_subscription = braintree.Subscription.find(enterprise_subscription_id)

        self.assertEqual(enterprise_subscription.plan_id, self.enterprise_plan.get_billing_id(BillingCycles.YEAR))
        self.assertEqual(enterprise_subscription.status, braintree.Subscription.Status.Active)
        self.assertEqual(enterprise_subscription.price,
                         agents_needed_updated_upd * self.enterprise_plan.annual_price * 12)

        # previous plan should be cancelled
        business_subscription_upd = braintree.Subscription.find(business_subscription_id)
        self.assertEqual(business_subscription_upd.status, braintree.Subscription.Status.Canceled)

        #
        # switching from "trial" to "business" plan
        #

        # log in the owner of self.company0, which is on a trial plan currently
        self.login_user(self.user_owner_0)

        self.assertEqual(self.company0.billing_plan.plan.name, 'Trial')
        self.assertEqual(self.company0.payment_methods.count(), 0)

        agents_num_trial_to_business = 7

        # create two new payment methods for self.company0
        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactablePrepaid},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.company0.payment_methods.count(), 1)
        self.assertIsNotNone(self.company0.payment_methods.get(pk=response.data['id']).braintree_token)
        company_0_pm_1_id = response.data['id']

        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactableDebit},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.company0.payment_methods.count(), 2)
        self.assertIsNotNone(self.company0.payment_methods.get(pk=response.data['id']).braintree_token)
        company_0_pm_2_id = response.data['id']

        # change the plan for the company0 using the payment method
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': agents_num_trial_to_business,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_0_pm_2_id},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        company_plan_obj_upd = CompanyPlan.objects.get(pk=self.company0.billing_plan.plan_id)
        new_business_subscription_id = company_plan_obj_upd.braintree_subscription_id

        # newly created "Business" subscription data
        trial_business_subscription = braintree.Subscription.find(new_business_subscription_id)

        self.assertEqual(trial_business_subscription.plan_id, self.business_plan.get_billing_id(BillingCycles.MONTH))
        self.assertEqual(trial_business_subscription.status, braintree.Subscription.Status.Active)
        self.assertEqual(trial_business_subscription.price,
                         agents_num_trial_to_business * self.business_plan.month_price)

        # make sure the secondly created payment method was used
        company_0_pm_2_token = PaymentMethod.objects.get(pk=company_0_pm_2_id).braintree_token
        self.assertEqual(trial_business_subscription.payment_method_token, company_0_pm_2_token)
        self.assertEqual(trial_business_subscription.transactions[0].vault_credit_card.token, company_0_pm_2_token)

        sync_company_braintree_transactions(self.company0)
        sync_company_braintree_transactions(self.company1)
        sync_company_braintree_transactions(self.company2)

        # transactions were synchronized successfully
        self.assertEqual(self.company0.transactions.count(), 1)
        self.assertEqual(self.company1.transactions.count(), 3)
        self.assertEqual(self.company2.transactions.count(), 0)

    def test_promo_code_validation(self):
        # no company restrictions available
        self.assertEqual(self.fifty_percent_off_code.companies.count(), 0)

        # unauthorized access is forbidden
        response = self.client.get(self.promo_validation_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # access is restricted to company owners only
        self.login_user(self.user_1_1)
        response = self.client.put(self.promo_validation_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        # log in a company owner
        self.login_user(self.user_owner_1)

        # no 'code' parameter
        response = self.client.post(self.promo_validation_api_url, {}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data.keys(), ['non_field_errors'])

        # wrong 'code' parameter values
        response = self.client.post(self.promo_validation_api_url, {'code': '123123123'}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        # correct 'code' parameter values
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': True,
                                         'discount_percent': self.fifty_percent_off_code.discount_percent})

        #
        #  company restriction. The code should be available for self.company0 only
        #
        self.fifty_percent_off_code.companies.add(self.company0)
        self.assertEqual(self.fifty_percent_off_code.companies.count(), 1)
        self.assertEqual(self.fifty_percent_off_code.companies.first(), self.company0)

        # The same code does not work for self.company1 anymore
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        # log in a self.company0 owner
        self.login_user(self.user_owner_0)

        # The code still works for self.company0
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': True,
                                         'discount_percent': self.fifty_percent_off_code.discount_percent})

        #
        #  Deactivating the code. It should become non-available for all the companies
        #
        self.fifty_percent_off_code.is_active = False
        self.fifty_percent_off_code.save()

        # The code does not work for companies anymore
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        self.login_user(self.user_owner_1)
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        #
        # Billing plan restriction
        #

        self.fifty_percent_off_code.is_active = True
        self.fifty_percent_off_code.save()
        self.fifty_percent_off_code.companies.clear()
        self.fifty_percent_off_code.plans.add(self.business_plan)

        self.assertEqual(self.fifty_percent_off_code.companies.count(), 0)
        self.assertEqual(self.fifty_percent_off_code.plans.count(), 1)

        # exception is raised if no plan argument is provided
        self.assertRaises(AssertionError, self.fifty_percent_off_code.is_valid_for_company, self.company0)

        self.assertFalse(self.fifty_percent_off_code.is_valid_for_company(self.company0, self.enterprise_plan))
        self.assertTrue(self.fifty_percent_off_code.is_valid_for_company(self.company0, self.business_plan))

        # API

        # The code does not work without plan specification anymore
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data.keys(), ['plan_id'])

        # The code does not work for enterprise plan
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.enterprise_plan.pk}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        # The code works for business plan
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.business_plan.pk}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': True,
                                         'discount_percent': self.fifty_percent_off_code.discount_percent})

        #
        # Payment period restrictions
        #

        self.fifty_percent_off_code.companies.clear()
        self.fifty_percent_off_code.plans.clear()
        # self.fifty_percent_off_code.plans.add(self.business_plan)

        self.fifty_percent_off_code.payment_period = BillingCycles.MONTH
        self.fifty_percent_off_code.save()

        self.assertEqual(self.fifty_percent_off_code.companies.count(), 0)
        self.assertEqual(self.fifty_percent_off_code.plans.count(), 0)

        # exception is raised if no plan payment_period argument provided
        self.assertRaises(AssertionError, self.fifty_percent_off_code.is_valid_for_company, self.company0)

        self.assertFalse(
            self.fifty_percent_off_code.is_valid_for_company(self.company0, payment_period=BillingCycles.YEAR))
        self.assertTrue(
            self.fifty_percent_off_code.is_valid_for_company(self.company0, payment_period=BillingCycles.MONTH))

        # API

        # The code does not work without payment period specification anymore
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.enterprise_plan.pk}, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.data.keys(), ['payment_period'])

        # The code does not work for annual plans
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.enterprise_plan.pk,
                                                                    'payment_period': BillingCycles.YEAR},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.business_plan.pk,
                                                                    'payment_period': BillingCycles.YEAR},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': False})

        # The code works for monthly plans
        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.business_plan.pk,
                                                                    'payment_period': BillingCycles.MONTH},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': True,
                                         'discount_percent': self.fifty_percent_off_code.discount_percent})

        response = self.client.post(self.promo_validation_api_url, {'code': self.fifty_off_code,
                                                                    'plan_id': self.enterprise_plan.pk,
                                                                    'payment_period': BillingCycles.MONTH},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data, {'valid': True,
                                         'discount_percent': self.fifty_percent_off_code.discount_percent})

    def test_billing_plan_change_promo_code(self):
        # log in the company owner
        self.login_user(self.user_owner_1)

        # initial assertions
        self.assertEqual(self.company1.billing_plan.plan, self.free_plan)
        self.assertEqual(self.company1.billing_plan.agents_count, 1)
        self.assertEqual(self.company1.billing_plan.payment_period, BillingCycles.MONTH)
        self.assertEqual(self.company1.payment_methods.count(), 0)

        if not settings.BILLING_ENABLED:
            raise unittest.SkipTest("No Braintree plan changing testing "
                                    "due to `settings.BILLING_ENABLED` set to False")

        # create new payment method for self.company1
        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactableDebit},
                                    format='json')

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.company1.payment_methods.count(), 1)
        self.assertIsNotNone(self.company1.payment_methods.first().braintree_token)
        company_1_pm_id = response.data['id']

        # change plan from "Free" to "Business"
        agents_needed = 5
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': agents_needed,
                                    'payment_period': BillingCycles.YEAR, 'payment_method': company_1_pm_id,
                                    'promo_code': self.fifty_off_code},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # correct updated values in response
        self.assertEqual(response.data['plan'], self.business_plan.pk)
        self.assertEqual(response.data['agents_count'], agents_needed)

        # re-fetching corresponding object from db
        updated_company_plan_obj = CompanyPlan.objects.get(pk=self.company1.billing_plan.pk)
        self.assertEqual(updated_company_plan_obj.plan, self.business_plan)
        self.assertEqual(updated_company_plan_obj.agents_count, agents_needed)
        self.assertEqual(updated_company_plan_obj.payment_period, BillingCycles.YEAR)

        business_subscription_id = updated_company_plan_obj.braintree_subscription_id
        self.assertIsNotNone(business_subscription_id)

        # a request to Braintree for the newly created subscription data
        business_subscription = braintree.Subscription.find(business_subscription_id)

        self.assertEqual(business_subscription.plan_id, self.business_plan.get_billing_id(BillingCycles.YEAR))
        self.assertEqual(business_subscription.status, braintree.Subscription.Status.Active)

        expected_subscription_price = agents_needed * self.business_plan.annual_price * 12

        # discount object should is applied
        self.assertEqual(len(business_subscription.transactions), 1)
        # the price should be 50 percent lower for the first transaction

        # the actually paid amount should be 50 percent lower now
        self.assertEqual(business_subscription.transactions[0].amount,
                         expected_subscription_price * decimal.Decimal('0.5'))
        # the price of the subscription is not changed, so discount is not applied for the next payment
        self.assertEqual(business_subscription.price, expected_subscription_price)
        self.assertEqual(business_subscription.next_billing_period_amount, expected_subscription_price)
        self.assertEqual(business_subscription.next_billing_date,
                         datetime.date(timezone.now()).replace(year=timezone.now().year + 1))

        # correct promo code usage tracking object was created
        self.assertTrue(PromoCodeUsage.objects.filter(code=self.fifty_percent_off_code,
                                                      company=self.company1).exists())
        code_usage_obj = PromoCodeUsage.objects.get(code=self.fifty_percent_off_code, company=self.company1)
        self.assertIsNotNone(code_usage_obj.date_used)

        #
        # Errors while providing invalid promo code
        #

        # log in the owner of self.company0, that uses a trial plan currently
        self.login_user(self.user_owner_0)

        self.assertEqual(self.company0.billing_plan.plan.name, 'Trial')
        self.assertEqual(self.company0.payment_methods.count(), 0)

        agents_num_trial_to_business = 7

        # create new payment method for self.company0
        response = self.client.post(self.pm_api_url,
                                    {'nonce': Nonces.TransactablePrepaid},
                                    format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.company0.payment_methods.count(), 1)
        self.assertIsNotNone(self.company0.payment_methods.first().braintree_token)
        company_0_pm_id = response.data['id']

        # Invalid promo code value
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': agents_num_trial_to_business,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_0_pm_id,
                                    'promo_code': 'some_invalid_string'},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['promo_code'])

        # no code restrictions set up yet
        self.assertEqual(self.fifty_percent_off_code.companies.count(), 0)
        self.assertEqual(self.fifty_percent_off_code.plans.count(), 0)

        # adding the billing plan restriction
        self.fifty_percent_off_code.plans.add(self.business_plan)

        # error for enterprise plan
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.enterprise_plan.pk, 'agents_count': agents_num_trial_to_business,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_0_pm_id,
                                    'promo_code': self.fifty_off_code},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['promo_code'])

        # correct processing for business plan
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.business_plan.pk, 'agents_count': agents_num_trial_to_business,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_0_pm_id,
                                    'promo_code': self.fifty_off_code},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # correct updated values in response
        self.assertEqual(response.data['plan'], self.business_plan.pk)
        self.assertEqual(response.data['agents_count'], agents_num_trial_to_business)

        # reset plan restrictions for promo code and add company restriction
        self.fifty_percent_off_code.plans.clear()
        self.fifty_percent_off_code.companies.add(self.company1)
        self.assertEqual(self.fifty_percent_off_code.plans.count(), 0)
        self.assertEqual(self.fifty_percent_off_code.plans.count(), 0)
        self.assertEqual(self.fifty_percent_off_code.companies.count(), 1)
        self.assertEqual(self.fifty_percent_off_code.companies.first(), self.company1)

        # error for company0
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.enterprise_plan.pk, 'agents_count': agents_num_trial_to_business,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_0_pm_id,
                                    'promo_code': self.fifty_off_code},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertItemsEqual(response.data.keys(), ['promo_code'])

        # correct processing for company1
        self.login_user(self.user_owner_1)
        response = self.client.put(self.company_plan_api_url,
                                   {'plan': self.enterprise_plan.pk, 'agents_count': agents_num_trial_to_business,
                                    'payment_period': BillingCycles.MONTH, 'payment_method': company_1_pm_id,
                                    'promo_code': self.fifty_off_code},
                                   format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # correct updated values in response
        self.assertEqual(response.data['plan'], self.enterprise_plan.pk)
        self.assertEqual(response.data['agents_count'], agents_num_trial_to_business)
