from django.conf import settings
from django.db import IntegrityError
from datetime import datetime

def convert_invoice_status(status):
    from apps.billing.models import Invoice
    STATUSES = {
        'draft': Invoice.DRAFT,
        'open': Invoice.PENDING,
        'paid': Invoice.SUCCESS,
        'void': Invoice.FAILED,
        'uncollectible': Invoice.PENDING,
    }
    return  STATUSES.get(status, Invoice.FAILED)


def conver_intent_status(status):
    from apps.billing.models import Transaction
    STATUSES = {
        'requires_payment_method': Transaction.FAILED,
        'requires_confirmation': Transaction.REQACT,
        'requires_action': Transaction.REQACT,
        'processing': Transaction.PENDING,
        'succeeded': Transaction.SUCCESS,
        'canceled': Transaction.FAILED,
    }
    return  STATUSES.get(status, Transaction.FAILED)


def convert_sub_status(status):
    STATUSES = {
        'active': True,
        'trialing': True,
        'incomplete': False,
        'incomplete_expired': False,
        'past_due': False,
        'canceled':  False,
        'unpaid':  False
    }
    return  STATUSES.get(status, False)


def to_stripe_amount(val):
    return int(val*100)


def do_sync_invoice(inv, products=None):
    from apps.billing.models import Invoice, Plan
    from apps.company.models import Company
    from . import stripe_cli, PAYMENT_NAME

    invt = Invoice.objects.filter(billing_id=inv.id, billing_provider=PAYMENT_NAME).first()
    if not invt:
        payment_plan = None
        payment_period = None
        for rec in inv.lines.data:
            if hasattr(rec, 'price') and hasattr(rec.price, 'product'):
                if products and rec.price.product in products:
                    payment_plan, payment_period = products.get(rec.price.product)
                    break
                product = stripe_cli.Product.retrieve(rec.price.product)
                payment_plan = Plan.objects.filter(billing_prefix=product.name).first()
                payment_period = rec.price.recurring.interval
                if products:
                    products[rec.price.product] = payment_plan, payment_period
                break
        if not payment_plan:
            return
        company = Company.objects.filter(billing_id=inv.customer, billing_provider=PAYMENT_NAME).first()
        if payment_plan and company:
            invt = Invoice(
                billing_id=inv.id,
                billing_provider=PAYMENT_NAME,
                link=inv.invoice_pdf,
                company=company,
                plan_period=payment_period,
                plan=payment_plan,
                number=inv.number,
                user=company.owner,
                timestamp=datetime.fromtimestamp(inv.created),
                amount=(inv.amount_due) / 100.0,
                result='',
                status='',
            )
            try:
                invt.save()
            except IntegrityError:
                invt = Invoice.objects.filter(billing_id=inv.id, billing_provider=PAYMENT_NAME).first()
                if not invt:
                    return
        else:
            return

    invt.result += "\n\n"
    invt.result += str(inv)
    invt.status = convert_invoice_status(inv.status)

    invt.save()


def do_sync_intent(tr, products=None):
    from apps.billing.models import Transaction, Plan, PaymentMethod
    from apps.company.models import Company
    from . import stripe_cli, PAYMENT_NAME
    trx = Transaction.objects.filter(billing_id=tr.id, billing_provider=PAYMENT_NAME).first()
    if not trx:
        payment_plan = None
        payment_period = None
        if not tr.invoice:
            return
        inv = stripe_cli.invoice.Invoice.retrieve(tr.invoice)
        if inv:
            for rec in inv.lines.data:
                if hasattr(rec, 'price') and hasattr(rec.price, 'product'):
                    if products and rec.price.product in products:
                        payment_plan, payment_period = products.get(rec.price.product)
                        break
                    product = stripe_cli.Product.retrieve(rec.price.product)
                    payment_plan = Plan.objects.filter(billing_prefix=product.name).first()
                    payment_period = rec.price.recurring.interval
                    if products:
                        products[rec.price.product] = payment_plan, payment_period
                    break
        if not payment_plan:
            return
        payment_method = PaymentMethod.objects.filter(billing_token=tr.payment_method, billing_provider=PAYMENT_NAME).first()
        if not payment_method:
            return

        company = Company.objects.filter(billing_id=tr.customer, billing_provider=PAYMENT_NAME).first()
        if payment_plan and company:
            trx = Transaction(
                billing_id=tr.id,
                billing_provider=PAYMENT_NAME,
                company=company,
                plan_period=payment_period,
                plan=payment_plan,
                user=company.owner,
                payment_method=payment_method,
                timestamp=datetime.fromtimestamp(tr.created),
                amount=(tr.amount) / 100.0,
                result='',
                status='',
            )
            try:
                trx.save()
            except IntegrityError:
                trx = Transaction.objects.filter(billing_id=tr.id, billing_provider=PAYMENT_NAME).first()
                if not trx:
                    return
        else:
            return

    trx.result += "\n\n"
    trx.result += str(tr)
    trx.status = conver_intent_status(tr.status)
    trx.save()


def do_sync_sub(sub):
    from apps.billing.models import CompanyPlan
    from . import stripe_cli, PAYMENT_NAME
    cp = CompanyPlan.objects.filter(billing_subscription_id=sub.id, billing_provider=PAYMENT_NAME).first()
    if not cp:
        return
    cp.agents_count = sub.quantity if convert_sub_status(sub.status) else 0
    cp.end = datetime.fromtimestamp(sub.current_period_end)
    cp.save()


def sync_invoices():
    from . import stripe_cli

    off = 0
    has_next = True
    products = {}

    while has_next:
        req = stripe_cli.Invoice.list(limit=100, offset=off)
        for inv in req.data:
            do_sync_invoice(inv, products)
        has_next = req.has_more
        off += 100


def sync_intents():
    from . import stripe_cli

    off = 0
    has_next = True
    products = {}

    while has_next:
        req = stripe_cli.PaymentIntent.list(limit=100, offset=off)
        for tr in req.data:
            do_sync_intent(tr, products)
        has_next = req.has_more
        off += 100


def sync_sub():
    from . import stripe_cli

    off = 0
    has_next = True

    while has_next:
        req = stripe_cli.Subscription.list(limit=100, offset=off)
        for sub in req.data:
            do_sync_sub(sub)
        has_next = req.has_more
        off += 100


def sync_products():
    from apps.billing.models import Plan
    from . import stripe_cli

    prods = {}

    for prod in Plan.objects.exclude(billing_prefix__isnull=True):
        prods[prod.billing_prefix] = [1, prod, None]

    for prod in stripe_cli.product.Product.list(limit=100).data:
        if not prod.name in prods.keys():
            prods[prod.name] = [2, None, prod]
        else:
            prods[prod.name][0] += 2
            prods[prod.name][2] = prod

    for key, item in prods.items():
        mode, new, old = item

        if mode == 1:
            stripe_cli.product.Product.create(name=key)
            print('Product %s created' % key)
        elif mode == 2:
            if old.active:
                old.active = False
                old.save()
                print('Product %s disabled' % key)
            else:
                print('Product %s no change' % key)
        elif mode == 3:
            if not old.active:
                old.active = True
                old.save()
                print('Product %s enabled' % key)
            else:
                print('Product %s no change' % key)


def sync_plans():
    from apps.billing.models import Plan
    from . import stripe_cli

    prods = {prod.name: prod.id
            for prod in stripe_cli.product.Product.list(limit=100, active=True).data
    }

    plans = {}

    for plan in Plan.objects.exclude(billing_prefix__isnull=True):
        plan_id_m = '%s-%s-m-%s' % (plan.billing_prefix, to_stripe_amount(plan.month_price), settings.STRIPE_CURRENCY)
        plan_id_y = '%s-%s-y-%s' % (plan.billing_prefix, to_stripe_amount(plan.annual_price), settings.STRIPE_CURRENCY)
        plans[plan_id_m] = [1, plan, None, [plan.billing_prefix, to_stripe_amount(plan.month_price), "month"]]
        plans[plan_id_y] = [1, plan, None, [plan.billing_prefix, to_stripe_amount(plan.annual_price*12), "year"]]

    for plan in stripe_cli.price.Price.list(limit=100, active=True).data:
        plan_name = plan.metadata.get('name')
        if plan_name and plan_name in plans:
            plans[plan_name][0] += 2
            plans[plan_name][2] = plan
        else:
            plans[plan_name] = [2, None, plan, []]

    for key, item in plans.items():
        mode, new, old, props = item

        if mode == 1:
            product, amount, interval = props
            product = prods[product]
            stripe_cli.price.Price.create(
                unit_amount=amount,
                recurring={'interval': interval},
                currency=settings.STRIPE_CURRENCY,
                metadata={'name': key},
                product=product,
            )
            print('Price/Plan %s created' % key)
        elif mode == 2:
            if old.active:
                old.active = False
                old.save()
                print('Price/Plan %s disabled' % key)
            else:
                print('Price/Plan %s no change' % key)
        elif mode == 3:
            if not old.active:
                old.active = True
                old.save()
                print('Price/Plan %s enabled' % key)
            else:
                print('Price/Plan %s no change' % key)


def sync_promos():
    pass

def create_hook():
    from . import stripe_cli
    from apps.billing.models import BillingProps

    for hook in stripe_cli.WebhookEndpoint.list(limit=100).data:
        if hook.metadata.get('nicechat_bill'):
            return

    hook = stripe_cli.WebhookEndpoint.create(
        url="https://api.%s/api/v1/billing-hook/" % settings.DOMAIN_BASE,
        metadata={'nicechat_bill': 1},
        enabled_events=[
            "*",
            # "charge.failed",
            # "charge.succeeded",
        ],
    )

    pr, _ = BillingProps.objects.get_or_create(k='stripe_hook')
    pr.v = hook.secret
    pr.save()
