from urllib.parse import urlparse
from django.http import HttpResponse

from django.utils import translation
from django.utils.deprecation import MiddlewareMixin
from django.conf import settings
from rest_framework_simplejwt.authentication import JWTAuthentication

from .settings import LANGUAGE_COOKIE_NAME, DEFAULT_LANGUAGE
from .utils.client_token import check_crc, nc_validate_2key
from .utils.helpers import get_current_site_from_request


class CheckOriginMiddleware(MiddlewareMixin):
    @staticmethod
    def process_response(request, response):
        is_rendered = getattr(response, 'is_rendered', False)
        context = getattr(response, 'renderer_context', {})
        origin = request.META.get('HTTP_ORIGIN', '')

        if origin and is_rendered:
            origin = origin.rstrip('/')
            if origin in settings.ALLOW_ORIGIN:
                if request.method == 'OPTIONS':
                    response = HttpResponse("", status=204)

                response['Access-Control-Allow-Origin'] = request.META.get('HTTP_ORIGIN')
                response['Access-Control-Allow-Methods'] = 'POST, GET, PUT, PATCH, DELETE, OPTIONS'
                response['Access-Control-Allow-Headers'] = 'DNT,User-Agent,X-Requested-With,Signature,If-Modified-Since,Cache-Control,Content-Type,Range,Authorization,X-Site-Id,X-Csrftoken'
                response['Access-Control-Allow-Credentials'] = 'true'
                return response

        if is_rendered and context and isinstance(context, dict):
            kwargs = context.get('kwargs', {})
            site_id = kwargs.get('site_id') or request.GET.get('site_id')
            if site_id:
                from apps.chat.models import Website
                website = Website.objects.filter(pk=site_id).first()
                if website:
                    try:
                        domain = urlparse(origin).hostname if origin else None
                    except (ValueError, AttributeError):
                        domain = None

                    if domain and domain.startswith('www.'):
                        domain = domain[4:]
                    if website.domain == domain:
                        response['Access-Control-Allow-Origin'] = request.META.get('HTTP_ORIGIN')

        return response

    @staticmethod
    def checkid(nice_chat_client_id):
        nice_chat_client_id = nc_validate_2key(nice_chat_client_id) if nice_chat_client_id else None
        return nice_chat_client_id if nice_chat_client_id and check_crc(nice_chat_client_id) else None

    @staticmethod
    def process_request(request):
        request.nice_chat_client_id_raw = None
        request.nice_chat_client_id = None
        if not request.nice_chat_client_id and settings.WIDGET_SESSION_IN_GET_PARAM in request.GET.keys():
            request.nice_chat_client_id_raw = request.GET.get(settings.WIDGET_SESSION_IN_GET_PARAM)
            request.nice_chat_client_id = CheckOriginMiddleware.checkid(request.nice_chat_client_id_raw)
        if not request.nice_chat_client_id and settings.VISITOR_COOKIE_NAME in request.COOKIES.keys():
            request.nice_chat_client_id_raw = request.COOKIES.get(settings.VISITOR_COOKIE_NAME)
            request.nice_chat_client_id = CheckOriginMiddleware.checkid(request.nice_chat_client_id_raw)
        if not request.nice_chat_client_id and request.POST and settings.WIDGET_SESSION_IN_GET_PARAM in request.POST.keys():
            request.nice_chat_client_id_raw = request.POST.get(settings.WIDGET_SESSION_IN_GET_PARAM)
            request.nice_chat_client_id = CheckOriginMiddleware.checkid(request.nice_chat_client_id_raw)


class LanguageCookieMiddleware(MiddlewareMixin):
    @staticmethod
    def process_request(request):
        """
        Sets language from the cookie value.
        """
        language = None
        if request.COOKIES.get(LANGUAGE_COOKIE_NAME):
            language = request.COOKIES.get(LANGUAGE_COOKIE_NAME)

        if not language:
            language = translation.get_language()

        if not language:
            language = DEFAULT_LANGUAGE
        translation.activate(language)
        request.LANGUAGE_CODE = translation.get_language()

    @staticmethod
    def process_response(request, response):
        """
        Create cookie if not there already.

        Also deactivates language.
        (See http://stackoverflow.com/a/13031239/388835 )
        """

        if not request.COOKIES.get(LANGUAGE_COOKIE_NAME):
            language = translation.get_language_from_request(request).split('-')[0]
            if not language:
                language = DEFAULT_LANGUAGE
            response.set_cookie(LANGUAGE_COOKIE_NAME,
                                language)
        if getattr(request, 'is_allauth_sign_in', False):
            response.set_cookie(LANGUAGE_COOKIE_NAME,
                                request.user.interface_language)
        return response


class CurrentSiteMiddleware(object):
    """
    Fetches the current site object through `nicechat.utils.helpers.get_current_site_from_request`
    and "caches" the site instance (if any) per request

    Should be placed after 'django.contrib.auth.middleware.AuthenticationMiddleware'
    """

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        request.current_site = get_current_site_from_request(request, raise_exception=False)
        # request.current_site = get_current_site_from_request(request)

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response

class NCJSONWebTokenAuthentication(JWTAuthentication):
    def authenticate(self, request):

        item = super(NCJSONWebTokenAuthentication, self).authenticate(request)

        if not item:
            return item
        user, jwt_value = item

        request.user = user
        if not request.current_site:
            request.current_site = get_current_site_from_request(request, raise_exception=False)

        return user, jwt_value
