import datetime
import random
import string
import base64
import json
import io
import urllib
from django.conf import settings
from django.template.loader import get_template
from django.core.mail import EmailMultiAlternatives

from apps.users.models import Otp


def get_random_string(length=8):
    letters = string.digits + string.ascii_uppercase
    return ''.join(random.choice(letters) for i in range(length))


def get_random_num(length=5):
    letters = string.digits
    return ''.join(random.choice(letters) for i in range(length))


class TfaBase(object):
    def enable_user_factor(self, user, request):
        pass

    def check_user_factor(self, code, user, request, first=False):
        pass

    def prepare_user_factor(self, user, request, init=False):
        pass

    def get_method(self):
        return None

    def get_lock(self):
        return []


class TfaDoLoop(object):

    def __init__(self, proc):
        super(TfaDoLoop, self).__init__()
        self.proc = proc

    def disable(self, user):
        user.tfa_opts = '{}'
        user.save(update_fields=['tfa_opts'])

    def validate(self, code, user, request, do_on = False):
        for p in self.proc:
            try:
                r = p.check_user_factor(code, user, request, do_on)
                if r == True:
                    if do_on:
                        p.enable_user_factor(user, request)
                    return r
            except Exception as e:
                 print(e)
        return False

    def prepare(self, user, request, init=False, exclude=None):
        if exclude is None:
            exclude = []
        for p in self.proc:
            if p.get_method() in exclude:
                continue
            try:
                r = p.prepare_user_factor(user, request, init)
                if r:
                    return r
            except Exception as e:
                print(e)
        return False

    def lock(self, method):
        for p in self.proc:
            if p.get_method() == method:
                return p.get_lock()
        return []

    def methods(self):
        arr = []
        for p in self.proc:
            arr.append(p.get_method())
        return arr


class TfaAwsSnS(TfaBase):
    class_kind = 'SMSOTP'
    class_ts = 180

    def enable_user_factor(self, user, request):
        user.tfa_opts = json.dumps({
            "on": True,
            "method": self.class_kind
        })
        user.save(update_fields=['tfa_opts'])

    def check_user_factor(self, code, user, request, first=False):
        o = Otp.read(user, self.class_kind, code)
        if o is not None:
            return True

    def prepare_user_factor(self, user, request, init=False):
        if user.phone_number and settings.AWS_SNS_ACCESS_KEY:
            val = get_random_num()
            Otp.write(user, self.class_kind, val, ts=self.class_ts)
            import boto3

            client = boto3.client(
                "sns",
                aws_access_key_id=settings.AWS_SNS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SNS_SECRET_KEY,
                region_name="us-east-1"
            )

            try:
                client.publish(
                    PhoneNumber=user.phone_number,
                    Message="OTP: %s at %s" % (val, settings.BASE_CHAT_NAME)
                )
            except Exception as x:
                print(x)
            print(user.phone_number)

            return self.class_kind, self.class_ts
        return None

    def get_lock(self):
        return ['phone_number']

    def get_method(self):
        return self.class_kind

class TfaTOtp(TfaBase):
    class_kind = 'TOTP'

    def enable_user_factor(self, user, request):
        try:
            tfa = json.loads(user.tfa_opts) if user else {}
        except ValueError:
            tfa = {}

        user.tfa_opts = json.dumps({
            "on": True,
            "token": tfa['pre'],
            "method": self.class_kind
        })
        user.save(update_fields=['tfa_opts'])

    def check_user_factor(self, code, user, request, first=False):
        try:
            tfa = json.loads(user.tfa_opts) if user else {}
        except ValueError:
            tfa = {}

        if first:
            token = tfa.get('pre', None)
        else:
            token = tfa.get('token', None)

        if not token:
            return False

        from pyotp import totp

        if not code.isdigit():
            return False

        tt = totp.TOTP(token)
        return tt.verify(int(code)) or tt.verify(int(code), for_time=datetime.datetime.now() + datetime.timedelta(seconds=-30)) or tt.verify(int(code), for_time=datetime.datetime.now() + datetime.timedelta(seconds=30))

    def prepare_user_factor(self, user, request, init=False):

        try:
            tfa = json.loads(user.tfa_opts) if user else {}
        except ValueError:
            tfa = {}

        token = tfa.get('token', None)

        from pyotp import totp, random_base32

        if token:
            return self.class_kind, ""

        if init:
            token = random_base32()
            un = getattr(user, user.USERNAME_FIELD, '')
            user.tfa_opts = json.dumps({
                "method": self.class_kind,
                "pre": token
            })
            user.save(update_fields=['tfa_opts'])
            tt = totp.TOTP(token)
            url = tt.provisioning_uri('____')
            url = url.replace("____", "%s:%s" % (urllib.quote(settings.BASE_CHAT_NAME), urllib.quote(un)))
            url += "&issuer=%s" % urllib.quote(settings.BASE_CHAT_NAME)
            import pyqrcode
            url = pyqrcode.create(url)
            buffer = io.BytesIO()
            url.svg(buffer, scale=8)
            url = "data:image/svg+xml;base64," + base64.b64encode(buffer.getvalue())
            return self.class_kind, url
        return None

    def get_lock(self):
        return []

    def get_method(self):
        return self.class_kind


class TfaEmail(TfaBase):
    class_kind = 'EMAIL'
    class_ts = 360

    def enable_user_factor(self, user, request):
        user.tfa_opts = json.dumps({
            "on": True,
            "method": self.class_kind
        })
        user.save(update_fields=['tfa_opts'])

    def check_user_factor(self, code, user, request, first=False):
        o = Otp.read(user, self.class_kind, code)
        if o is not None:
            return True

    def prepare_user_factor(self, user, request, init=False):
        val = get_random_string()
        Otp.write(user, self.class_kind, val, ts=self.class_ts)

        context = {
            'otp': val,
        }

        template = get_template('users/email/otp_password.html')
        text = get_template('users/email/otp_password.txt')
        subject = get_template('users/email/otp_password.subject')

        text_content = text.render(context)
        html_content = template.render(context)
        subject_content = subject.render(context)
        msg = EmailMultiAlternatives(
            subject_content, text_content,
            settings.DEFAULT_FROM_EMAIL,
            [user.email])

        msg.attach_alternative(html_content, "text/html")
        msg.send(fail_silently=True)

        return self.class_kind, self.class_ts

    def get_lock(self):
        return []

    def get_method(self):
        return self.class_kind
