"""nicechat URL Configuration"""

from django.urls import include, path
from nicechat.admins import ncadmin
from django.conf import settings
from django.conf.urls.static import static
from django.utils.module_loading import import_string

from dj_rest_auth.views import PasswordChangeView
from rest_framework import routers
from drf_yasg.views import get_schema_view
from drf_yasg import openapi
from rest_framework_simplejwt.views import TokenVerifyView, TokenRefreshView
from dj_rest_auth.registration.views import VerifyEmailView

from apps.billing.api import TransactionViewSet, PlanViewSet, CompanyPlanView, PaymentMethodViewSet, \
    BillingClinetToken, ConfigView, ValidatePromoCodeView, InvoiceViewSet
from apps.company.api import CompanyViewSet
from apps.users.api import UserViewSet
from apps.website.api import WebsiteViewSet, AgentViewSet, SendInstallInstructions, WidgetThemesViewSet, WidgetSettingsViewSet, DomainListViewSet
from apps.visitors.api import OperationSystemsViewSet, BrowsersViewSet, BatamStatisticsViewSet
from apps.balancer.api import ArbiterS2S, UIPropsView
from apps.website.views import HealthCheck, PropsView
from apps.visitors.views import MakeWidget
from apps.users.views import PasswordResetView, NcRegisterView, GetJWTToken, ConfirmEmailView, BlankView, \
    PasswordResetConfirmView

api_router = routers.SimpleRouter()
api_router.register('users', UserViewSet, 'users')
api_router.register(r'companies', CompanyViewSet)
api_router.register(r'websites', WebsiteViewSet)
api_router.register(r'websites/(?P<site_id>[A-Z0-9]+)/domains', DomainListViewSet)

api_router.register(r'agents', AgentViewSet)
api_router.register(r'transactions', TransactionViewSet)
api_router.register(r'plans', PlanViewSet)
api_router.register(r'payment-methods', PaymentMethodViewSet)
api_router.register(r'invoice', InvoiceViewSet)
api_router.register(r'browsers', BrowsersViewSet, 'browsers')
api_router.register(r'operation-systems', OperationSystemsViewSet, 'operation-systems')
api_router.register(r'widget-themes', WidgetThemesViewSet)
api_router.register(r'widget-settings', WidgetSettingsViewSet)

ext = []
extg = []
for up in settings.UPDATE_URLS:
    cls = import_string(up)
    update_route = getattr(cls, 'update_route', None)
    update_route_v2 = getattr(cls, 'update_route_v2', None)
    if update_route_v2:
        update_route_v2(api_router, api_router)
    elif update_route:
        update_route(api_router)
    update_flat_route = getattr(cls, 'update_flat_route', None)
    if update_flat_route:
        ext__ = []
        if update_flat_route(ext__):
            extg += ext__
        else:
            ext += ext__

api_urlpatterns = [
    path('', include(api_router.urls)),
    path('company-plan/', CompanyPlanView.as_view()),
    path('billing-client-token/', BillingClinetToken.as_view()),
    path('validate-promo-code/', ValidatePromoCodeView.as_view()),
    path('config/', ConfigView.as_view()),
    path('send-instructions/', SendInstallInstructions.as_view()),
]

if ext:
    api_urlpatterns += ext

if settings.DEBUG:
    from rest_framework import permissions
    schema_view = get_schema_view(
        openapi.Info(
            title="Nice.chat API",
            default_version='v1',
            description="Nice chat inner API",
            terms_of_service="https://nice.chat/",
            contact=openapi.Contact(email="stas@nice.chat"),
        ),
        public=True,
        permission_classes=(permissions.AllowAny,),
    )
    api_urlpatterns += [
        path('swagger<format>/', schema_view.without_ui(cache_timeout=0), name='schema-json'),
        path('swagger/', schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
        path('redoc/', schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
    ]

if settings.BILLING_ENABLED:
    from nicechat.utils.billing import register_payments
    pcls = import_string(settings.BILLING_BACKEND)
    register_payments(pcls, api_urlpatterns)


urlpatterns = [
    # DRF API urls
    path('api/v1/', include(api_urlpatterns)),

    path('api/v1/rest-auth/registration/', NcRegisterView.as_view(), name='rest_register'),
    path('api/v1/rest-auth/registration/verify-email/', VerifyEmailView.as_view(), name='rest_verify_email'),
    path('api/v1/rest-auth/registration/account-confirm-email/<str:key>/', ConfirmEmailView.as_view()),

    path('api/v1/rest-auth/password-change/', PasswordChangeView.as_view()),
    path('api/v1/rest-auth/auth/', GetJWTToken.as_view()),
    path('api/v1/rest-auth/password-reset/', PasswordResetView.as_view()),
    path('api/v1/rest-auth/password-reset-confirm/', PasswordResetConfirmView.as_view()),
    
    path('api/v1/rest-auth/token-verify/', TokenVerifyView.as_view()),
    path('api/v1/rest-auth/token-refresh/', TokenRefreshView.as_view()),

    path('widget/', include('apps.visitors.urls')),
    path('api/v1/users/', include(('apps.users.urls', 'users'), namespace='users')),
    path('api/v1/hc/', HealthCheck.as_view(), name='healthcheck'),
    path('api/v1/props/', PropsView.as_view(), name='props_view'),
    path('api/v1/props.js', PropsView.as_view(), name='props_view_js'),
    path('widget.js', MakeWidget.as_view(), name='get_widget_base'),
]

if settings.NODE_MODE == settings.NODE_MODE_ARBITER:
    urlpatterns.append(path('arbiter/admin/', ncadmin.urls))
    urlpatterns.append(path('arbiter/', include(('apps.balancer.urls', 'arbiter'), namespace='arbiter')))
    urlpatterns.append(path('arb_s2s/<slug:cmd>/', ArbiterS2S.as_view()))

if settings.NODE_MODE == settings.NODE_MODE_STANDALONE:
    urlpatterns.append(path('admin/', ncadmin.urls))

urlpatterns.append(path('api/v1/uiprops.js', UIPropsView.as_view()))
urlpatterns.append(path('', BlankView.as_view()))

if extg:
    urlpatterns += extg

if settings.DEBUG:
    # This is only needed when using runserver.
    urlpatterns = \
        static(settings.STATIC_URL, document_root=settings.STATIC_ROOT) \
        + static('/media/', document_root=settings.MEDIA_ROOT) \
        + urlpatterns
