"""nicechat URL Configuration"""

from django.urls import include, path
from nicechat.admins import ncadmin
from django.conf import settings
from django.conf.urls.static import static
from django.utils.module_loading import import_string

from rest_framework import routers
from drf_yasg.views import get_schema_view
from drf_yasg import openapi

from apps.website.api import QuickReplyViewSet, ProductsCategoryViewSet, ProductViewSet, \
    ProductImport, FormsViewSet, FormRowsViewSet, StatusViewSet, \
    ImportCsvFileUpload, CallbacksViewSet, WebsiteTokensViewSet
from apps.visitors.api import VisitorViewSet, StatisticsViewSet, OrdersViewSet
from apps.marketplaces.api import MarketplaceImportViewSet, MarketplaceImportResultViewSet, \
    MarketplaceCategoryViewSet, MarketplaceProductViewSet, MarketplaceOrderViewSet
from apps.chat.api import MessageViewSet, ChatViewSet, ArchivedChatViewSet
from apps.website.views import HealthCheck
from apps.chat.views import AttachDownload
from apps.users.views import BlankView
from apps.balancer.api import ArbiterS2SWorker

api_router = routers.SimpleRouter()
api_router.register(r'visitors/(?P<site_id>[A-Z0-9]+)', VisitorViewSet, 'visitors')
api_router.register(r'websites/(?P<site_id>[A-Z0-9]+)/tokens', WebsiteTokensViewSet, 'ws_tokens')
api_router.register(r'visitors/statistics', StatisticsViewSet, 'visitors-statistics')
api_router.register(r'statuses', StatusViewSet)
api_router.register(r'hooks', CallbacksViewSet)
api_router.register(r'quick-reply', QuickReplyViewSet)
api_router.register(r'messages', MessageViewSet)
api_router.register(r'site_(?P<site_id>[A-Z0-9]+)/messages', MessageViewSet)
api_router.register(r'products-categories', ProductsCategoryViewSet)
api_router.register(r'products', ProductViewSet)
api_router.register(r'orders', OrdersViewSet)
api_router.register(r'archive', ArchivedChatViewSet)

api_router.register(r'chat', ChatViewSet)
api_router.register(r'chat/(?P<chat_id>[A-Z0-9]+)/messages', MessageViewSet, 'messages')
api_router.register(r'forms', FormsViewSet, 'forms')
api_router.register(r'forms/(?P<form_id>[A-Z0-9]+)/rows', FormRowsViewSet, 'rows')
# add URL like /api/v1/<site_id>/chat/[id/]
api_router.register(r'(?P<site_id>[A-Z0-9]+)/chat', ChatViewSet)
api_router.register(r'(?P<site_id>[A-Z0-9]+)/chat/(?P<chat_id>[A-Z0-9]+)messages', MessageViewSet, 'messages_with_id')

ext = []
extg = []
for up in settings.UPDATE_URLS:
    cls = import_string(up)
    update_route = getattr(cls, 'update_route', None)
    update_route_v2 = getattr(cls, 'update_route_v2', None)
    if update_route_v2:
        update_route_v2(api_router, api_router)
    elif update_route:
        update_route(api_router)
    update_flat_route = getattr(cls, 'update_flat_route', None)
    if update_flat_route:
        ext__ = []
        if update_flat_route(ext__):
            extg += ext__
        else:
            ext += ext__

api_urlpatterns = [
    path('', include(api_router.urls)),
    path('product_import/', ProductImport.as_view()),
    path('product_import_csv/', ImportCsvFileUpload.as_view()),
    path('attach/<int:chat_id>/<int:msg_id>/<str:file_name>', AttachDownload.as_view()),
    path('attach/<int:msg_id>/<str:file_name>', AttachDownload.as_view()),
]

if ext:
    api_urlpatterns += ext

api_marketplaces = routers.SimpleRouter()
api_marketplaces.register(r'import-products', MarketplaceImportViewSet, basename='import-products')
api_marketplaces.register(r'import-products-result', MarketplaceImportResultViewSet, basename='import-products-result')
api_marketplaces.register(r'categories', MarketplaceCategoryViewSet)
api_marketplaces.register(r'products', MarketplaceProductViewSet)
api_marketplaces.register(r'orders', MarketplaceOrderViewSet)

if settings.DEBUG:
    from rest_framework import permissions
    schema_view = get_schema_view(
        openapi.Info(
            title="Nice.chat API",
            default_version='v1',
            description="Nice chat inner API",
            terms_of_service="https://nice.chat/",
            contact=openapi.Contact(email="stas@nice.chat"),
        ),
        public=True,
        permission_classes=(permissions.AllowAny,),
    )
    api_urlpatterns += [
        path('swagger<format>/', schema_view.without_ui(cache_timeout=0), name='schema-json'),
        path('swagger/', schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
        path('redoc/', schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
    ]


if settings.NC_VIDEO_ENABLE:
    from nicechat.utils.video import register_video
    vcls = import_string(settings.NC_VIDEO_BACKEND)
    register_video(vcls, api_urlpatterns)


urlpatterns = [
    # DRF API urls
    path('<str:worker_node>/widget/', include('apps.visitors.urls')),
    path('<str:worker_node>/api/v1/', include(api_urlpatterns)),
    path('<str:worker_node>/marketplaces/v1/', include(api_marketplaces.urls)),

    path('<str:worker_node>/api/v1/hc/', HealthCheck.as_view(), name='healthcheck')
]

if settings.NODE_MODE == settings.NODE_MODE_WORKER:
    urlpatterns.append(path(f'{settings.WORKER_NODE_NAME}/admin/', ncadmin.urls))
    urlpatterns.append(path('arb_s2s/<slug:cmd>/', ArbiterS2SWorker.as_view()))

urlpatterns.append(path('', BlankView.as_view()))

if extg:
    urlpatterns += extg

if settings.DEBUG:
    # This is only needed when using runserver.
    urlpatterns = \
        static(settings.STATIC_URL, document_root=settings.STATIC_ROOT) \
        + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT) \
        + urlpatterns
