import base64
import json
import os
import random
import time
import platform
import traceback
import hashlib
import socket
import requests
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7


DB_CLEAN_DELTA = 86400 * 14
HOST_AUTO_INACTIVE=120


def load_private_key(keystr):
    if not isinstance(keystr, str):
        return None
    key = base64.b32decode(keystr)
    if len(key) == 60:
        key = key[4:]
    if len(key) != 56:
        return None
    return x448.X448PrivateKey.from_private_bytes(key)


def load_public_key(keystr):
    if not isinstance(keystr, str):
        return None
    key = base64.b32decode(keystr)
    if len(key) == 60:
        key = key[4:]
    if len(key) != 56:
        return None
    return x448.X448PublicKey.from_public_bytes(key)


def dump_public_key(bytekey, is_arbiter):
    pubkey = bytekey.public_key().public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    )

    if is_arbiter:
        return base64.b32encode(b'\x04B\xfa\x04' + pubkey).decode()
    else:
        return base64.b32encode(b'}\x02\xb4\x08' + pubkey).decode()


def make_new_node_key():
    from django.conf import settings
    if settings.NODE_PRIVATE_KEY:
        _nodekey = x448.X448PrivateKey.generate()
        nodekey = _nodekey.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
        )
        nodekey = base64.b32encode(b'\xb3\xa2\xa2D' + nodekey).decode()

        pubkey = _nodekey.public_key().public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )
        pubkey = base64.b32encode(b'}\x02\xb4\x08' + pubkey).decode()

        return nodekey, pubkey

def get_fullnode_key(nodekey):
    from django.conf import settings

    pubkey = settings.NODE_PRIVATE_KEY.public_key().public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    )
    pubkey = base64.b32encode(b'\x04B\xfa\x04' + pubkey).decode()
    #or '}\x02\xb4\x08' - PUBLIC

    return f'{nodekey}.{pubkey}'

def arbiter_request(hosts, cmd, data={}, keys=None):
    if not hosts or not cmd:
        return False

    for _ in range(1, 5):
        try:
            ret = _arbiter_request_exec(hosts, cmd, data, keys)
            if ret:
                return ret
        except Exception as e:
            traceback.print_exception(e)
            from django.conf import settings
            if settings.SENTRY_TRACE:
                settings.SENTRY_TRACE(
                    e
                )
    return False

def _arbiter_request_exec(hosts, cmd, data={}, keys=None):
    if isinstance(hosts, str):
        host = hosts
    elif isinstance(hosts, list):
        host = random.choice(hosts)
    else:
        return False
    url_to = f'http://{host}:8000/arb_s2s/{cmd}/'
    data = json.dumps(data)
    if not keys:
        from django.conf import settings
        keys = [settings.NODE_PRIVATE_KEY, settings.NODE_ARBITER_PUB, settings.NODE_PUBLIC_KEY]

    _data = arb_encrypt(data, keys[0], keys[1])
    try:
        req = requests.post(url_to, data=_data, headers={'X-Sign': keys[2]}, timeout=10)
    except requests.exceptions.ConnectionError:
        return False
    if req.status_code == 200:
        ret_data, is_correct = arb_decrypt(req.content, keys[0], keys[1])
        return json.loads(ret_data) if is_correct else {}
    else:
        return False

def update_data_workernode(nodes, method, data):
    for _ in range(1, 5):
        try:
            ret = _update_data_workernode_exec(nodes, method, data)
            return ret
        except Exception as e:
            traceback.print_exception(e)
            from django.conf import settings
            if settings.SENTRY_TRACE:
                settings.SENTRY_TRACE(
                    e
                )
    return False

def _update_data_workernode_exec(nodes, method, data):
    from django.conf import settings
    if settings.NODE_MODE != settings.NODE_MODE_ARBITER:
        return

    from apps.balancer.models import ArbiterNodeHost
    hosts = list(ArbiterNodeHost.objects.filter(node__in=nodes))
    if hosts:
        h = random.choice(hosts)
        return host_exec(h, method, data)


def host_exec(h, method, data):
    from django.conf import settings
    data = json.dumps(data)
    _data = arb_encrypt(data, settings.NODE_PRIVATE_KEY, load_public_key(h.host_pubkey) )
    req_content = None
    for x in h.hostinfo.order_by('timestamp').filter(timestamp__gte=int(time.time()-60)):
        url_to = f'http://{x.address}:8000/arb_s2s/{method}/'
        try:
            req = requests.post(url_to, data=_data, headers={'X-Sign': settings.NODE_PUBLIC_KEY}, timeout=10)
            req_content = req.content
            if req_content:
                break
        except requests.RequestException:
            continue
    if not req_content:
        for j in h.workers.filter(data_timestamp__gte=time.time() - 60):
            print(j,j.hostname)
            _url_to = f'http://{j.hostname}:8000/arb_s2s/{method}/'
            try:
                req = requests.post(_url_to, data=_data, headers={'X-Sign': settings.NODE_PUBLIC_KEY}, timeout=10)
                req_content = req.content
                if req_content:
                    break
            except requests.RequestException:
                continue
            if not req_content:
                return None
    print('Sync', method, h.host_address, req.status_code)
    if not req_content:
        return False

    _data, _is_correct = arb_decrypt(req_content, settings.NODE_PRIVATE_KEY, load_public_key(h.host_pubkey))
    return False if not _is_correct else _data
            

def arb_encrypt(body, source_private_key, target_pub_key):
    if isinstance(body, str):
        body = body.encode()

    if isinstance(target_pub_key, str):
        target_pub_key = load_public_key(target_pub_key)
    ex_key = source_private_key.exchange(target_pub_key)
    hs = os.urandom(16)
    iv = os.urandom(16)
    padding = PKCS7(128).padder()
    padded_data = padding.update(body)
    padded_data += padding.finalize()
    enc_key = HKDF(algorithm=hashes.SHA256(),length=32,salt=None, info=hs,).derive(ex_key)

    cipher = Cipher(algorithms.AES(enc_key), modes.CBC(iv))
    encryptor = cipher.encryptor()
    ct = encryptor.update(padded_data) + encryptor.finalize()
    return hs + iv + ct


def arb_decrypt(ct, target_private_key, source_pub_key):
    try:
        hs =  ct[0:16]
        iv =  ct[16:32]
        ct = ct[32:]
    
        if isinstance(source_pub_key, str):
            source_pub_key = load_public_key(source_pub_key)
        ex_key = target_private_key.exchange(source_pub_key)
        enc_key = HKDF(algorithm=hashes.SHA256(),length=32,salt=None, info=hs,).derive(ex_key)
        unpadding = PKCS7(128).unpadder()

        cipher = Cipher(algorithms.AES(enc_key), modes.CBC(iv))
        decryptor = cipher.decryptor()
        data = decryptor.update(ct) + decryptor.finalize()
        udata = unpadding.update(data)
        udata += unpadding.finalize()
        return udata.decode(), True
    except ValueError as e:
        return False, False
    except Exception as e:
        import traceback
        traceback.print_exception(e)
        return False, False


def arbiter_get_hosts(settings):
    if settings.NODE_MODE == settings.NODE_MODE_WORKER:
        return settings.NODE_ARBITER_HOST
    else:
        try:
            return socket.gethostbyname(platform.node())
        except socket.error:
            return '127.0.0.1'


def balancer_make_sign(balancer_sign):
    _t = int(time.time()).to_bytes(4, 'big')
    _s = os.urandom(26)
    _h = hashlib.sha384(_t + _s + balancer_sign.encode()).digest()
    res = base64.b64encode(_s + _t + _h, altchars=b'-_')
    return res


def balancer_check_sign(sign, balancer_sign):
    try:
        _dt = base64.b64decode(sign, altchars=b'-_')
    except Exception:
        return False
    _s = _dt[0:26]
    _t = _dt[26:30]
    _h = _dt[30:]
    _hv = hashlib.sha384(_t + _s + balancer_sign.encode()).digest()
    if _h != _hv:
        return False
    
    _t = int.from_bytes(_t, 'big')
    if abs(time.time() - _t) > 150:
        return False
    
    return True


def mark_node_status(content, host=None, request=None):
    from apps.balancer import models
    from django.conf import settings
    event = content.get('event', None)
    cl_host = content.get('node', None)
    mode = content.get('mode', 'worker')
    pid = content.get('pid', 0)

    http_req = content.get('http_req', 0)
    http_s2s = content.get('http_s2s', 0)
    ws_req = content.get('ws_req', 0)
    active_ws = content.get('active_ws', 0)
    stats = content.get('stats', {})

    host_active = event in ['stats', 'newhost']
    
    client = request.scope.get('client', None) if request else None
    client = client[0] if client and len(client) > 0 else None
    
    if client:
        try:
            cl_host = cl_host if client == socket.gethostbyname(cl_host) else client
        except socket.error:
            cl_host = client

    if not host and mode == settings.NODE_MODE_ARBITER:
        host = models.ArbiterNodeHost.objects.filter(node__mode=settings.NODE_MODE_ARBITER).first()
        if not host:
            node, _ = models.ArbiterNode.objects.get_or_create(
                mode=settings.NODE_MODE_ARBITER, defaults={ 'name': 'arbiter' })
            host, _ = models.ArbiterNodeHost.objects.get_or_create(
                node=node, defaults={ 'host_pubkey': settings.NODE_PUBLIC_KEY, 'host_address': 'arbiter' })
    elif not host:
        return
            
    node = host.node

    record_key=f"{node.name}.{cl_host}.{mode}.{pid}"

    stat_head, _ = models.ArbiterNodeWorkersStatHead.objects.get_or_create(
        record_key=record_key,
        defaults={
            'worker_hostname': cl_host,
            'worker_process': pid,
            'worker_node': node.name,
            'worker_host': host.host_pubkey,
            'worker_mode': mode,
        }
    )

    models.ArbiterNodeWorkersStatData.objects.create(
        head = stat_head,
        data_timestamp=int(time.time()),
        stat_http_req=http_req,
        stat_http_s2s=http_s2s,
        stat_ws_req=ws_req,
        stat_active_ws=active_ws,
    )

    if host_active:
        try:
            stats = json.dumps(stats) if stats else '{}'
        except Exception:
            stats = '{}'
        stat_head, need_update_map = models.ArbiterNodeWorkersStat.objects.update_or_create(
            record_key=record_key,
            node=node,
            defaults={
                'host': host,
                'hostname': cl_host,
                'data_timestamp': int(time.time()),
                'stat_http_req': http_req,
                'stat_http_s2s': http_s2s,
                'stat_ws_req': ws_req,
                'stat_active_ws': active_ws,
                'stats': stats,
            }
        )
    else:
        host_a = models.ArbiterNodeWorkersStat.objects.filter(
            record_key=record_key,
            node=node,
        )
        if host_a.exists():
            need_update_map = True
            host_a.delete()


    if mode != settings.NODE_MODE_ARBITER and host.host_address != cl_host:
        host.host_address = cl_host
        host.save()
        need_update_map = True

    qs = models.ArbiterHostMap.objects.filter(
        timestamp__lte=int(time.time()) - DB_CLEAN_DELTA
    )
    if qs.exists():
        need_update_map = True
        qs.delete()

    set_hosts_inactive = models.ArbiterNodeWorkersStat.objects.filter(
        data_timestamp__lte=int(time.time()) - HOST_AUTO_INACTIVE,
    )
    if set_hosts_inactive.exists():
        need_update_map = True
        set_hosts_inactive.delete()

    if event == 'stopworker':
        need_update_map = True

    if need_update_map:
        data = [(stat['node__mode'], stat['node__name'], stat['hostname'],) 
            for stat in models.ArbiterNodeWorkersStat.objects.all().distinct().values('node__mode', 'node__name', 'hostname')]        
        data = sorted(data)
        ex_nodes = list(set(d[1] for d in data))
        for node in models.ArbiterNode.objects.all():
            if not node.name in ex_nodes:
                data.append((node.mode, node.name, False))

        _a = models.ArbiterHostMap.objects.last()
        if not _a or json.dumps(data) != json.dumps(_a.data) or time.time() - _a.timestamp > 7200:            
            models.ArbiterHostMap.objects.create(
                timestamp=int(time.time()),
                data=data
            )
