import time
import json
import string
import random
import re
import hashlib
from datetime import datetime
from django.utils import timezone
from django.core.files.base import ContentFile
from django.core.files.storage import get_storage_class



def get_user_relations(user):
    from apps.users.models import User, UserSiteRole
    from apps.website.models import Website
    ret = {}
    ret['user'] = {
        'pk': user.pk,
        'email': user.email,
        'first_name': user.first_name,
        'last_name': user.last_name,

        'company_id': user.company_id,
        'label': user.label,
        'contacts': user.contacts,

        'avatar': user.avatar.path if user.avatar else None,
        'position': user.position,
        'phone_number': user.phone_number,

        'timezone': user.timezone,
        'interface_language': user.interface_language,
        'tfa_opts': user.tfa_opts,
    }
    if user.is_superuser:
        ret['user']['is_superuser']=True
        ret['user']['is_staff']=True
    ret['websites'] = list(user.website_set.all().values_list('pk', flat=True))
    ret['w2s'] = []
    for usr in UserSiteRole.objects.filter(user=user, site_id__in=ret['websites']):
        ret['w2s'].append({
            'site_id': usr.site_id,
            'role': usr.role
        })
    return ret


def set_user_with_relations(data):
    from apps.users.models import User, UserSiteRole
    from apps.website.models import Website

    user = data['user']

    user, created = User.objects.update_or_create(
        pk=user['pk'],
        defaults=user
    )
    wss = {ws.pk:ws for ws in Website.objects.filter(pk__in=data['websites'])}
    for ws in wss.values():
        ws.users.add(user)
    
    for delws in user.website_set.exclude(pk__in=data['websites']):
        delws.users.remove(user)

    crec = list(UserSiteRole.objects.filter(user=user).values_list('site_id', flat=True))
    for rec in data['w2s']:
        site = wss.get(rec['site_id'])
        if site:
            usr = UserSiteRole.objects.filter(site=site, user=user).first()
            if usr:
                usr.role = rec['role']
                usr.save()
            else:
                usr = UserSiteRole.objects.create(
                    user=user, 
                    site=site,
                    role=rec['role']
                )
            if site.pk in crec:
                crec.remove(site.pk)
    UserSiteRole.objects.filter(site_id__in=crec, user=user).delete()

    return {'user_id': user.pk, 'created': created}


def get_website_relations(website):
    from apps.company.models import Company
    from apps.billing.models import CompanyPlan, Plan
    from apps.website.models import Website
    company = website.company
    try:
        company_plan = company.billing_plan
    except Company.billing_plan.RelatedObjectDoesNotExist:
        company_plan = None
    plan = company_plan.plan if company_plan else None

    ret = {}
    ret['company'] = {
        'pk': company.pk,
        'label': company.label,
        'owner_id': company.owner_id,
        'name': company.name,

        'contact_person': company.contact_person,
        'contact_email_main': company.contact_email_main,
        'contact_phone_main': company.contact_phone_main,
        'contact_info': company.contact_info,

        'billing_id': company.billing_id,
        'billing_provider': company.billing_provider,
        'allow_video': company.allow_video,
    }
    ret['company_plan'] = {
        'company_id': company_plan.company_id,
        'plan_id': company_plan.plan_id,

        'agents_count': company_plan.agents_count,
        'begin': int(company_plan.begin.timestamp()) if company_plan.begin else 0,
        'end': int(company_plan.end.timestamp()) if company_plan.end else 0,
        'payment_period': company_plan.payment_period,

        'billing_subscription_id': company_plan.billing_subscription_id,
        'billing_provider': company_plan.billing_provider,
        'notes': company_plan.notes,
    } if company_plan else {}

    ret['plan'] = {
        'pk': plan.pk,
        'name': plan.name,
        'month_price': float(plan.month_price) if plan.month_price else 0.0,
        'annual_price': float(plan.annual_price) if plan.annual_price else 0.0,

        'features': {p.name:p.title for p in plan.features.all()},

        'sites_num': plan.sites_num,
        'agents_num': plan.agents_num,
        'products_num': plan.products_num,
        'chat_stats_period': plan.chat_stats_period,
        'sales_stats_period': plan.sales_stats_period,
        'visitors_stats_period': plan.visitors_stats_period,
        'chat_history_period': plan.chat_history_period,
        'visitor_data_period': plan.visitor_data_period,
        'billing_prefix': plan.billing_prefix,
    } if plan else {}

    ret['website'] = {
        'pk': website.pk,
        'company_id': website.company_id,
        'title': website.title,
        'domain': website.domain,
        'icon': website.icon.path if website.icon else None,

        'cart_callback': website.cart_callback,
        'error_count': website.error_count,
        'online': website.online,
        'props': website.props,
    }
    return ret


def set_website_with_relations(data):
    from apps.company.models import Company
    from apps.billing.models import CompanyPlan, Plan, PlanFeature
    from apps.website.models import Website
    from apps.users.models import User

    _company = data['company']
    _plan = data['plan']
    _company_plan = data['company_plan']
    _website = data['website']

    if _company.get('owner_id'):
        owner_id = _company.get('owner_id')
        if not User.objects.filter(pk=owner_id).first():
            User.objects.create(
                pk=owner_id,
                email=f'tmp{owner_id}@tmp.com',
            )


    company, _ = Company.objects.update_or_create(
        pk=_company['pk'],
        defaults=_company
    )

    features = _plan.get('features', {})
    features_objs = []
    for nm, tl in features.items():
        pf, _ = PlanFeature.objects.update_or_create(
            name=nm,
            defaults={
                'title': tl
            }
        )
        features_objs.append(pf)

    del _plan['features']

    plan, _ = Plan.objects.update_or_create(
        pk=_plan['pk'],
        defaults=_plan
    )
    for pf in features_objs:
        plan.features.add(pf)
    plan.features.exclude(name__in=features.keys()).delete()
    
    _company_plan['begin'] = timezone.make_aware(datetime.fromtimestamp(_company_plan['begin']),
                             timezone=timezone.get_current_timezone()) if _company_plan.get('begin') else timezone.now()
    _company_plan['end'] = timezone.make_aware(datetime.fromtimestamp(_company_plan['end']),
                             timezone=timezone.get_current_timezone()) if _company_plan.get('end') else timezone.now()

    company_plan, _ = CompanyPlan.objects.update_or_create(
        company_id=_company_plan['company_id'],
        defaults=_company_plan
    )

    website, _ = Website.objects.update_or_create(
        pk=_website['pk'],
        defaults=_website
    )


def node_syncing():
    from django.conf import settings

    from nicechat.utils.arbiter import arbiter_request
    if settings.NODE_MODE != settings.NODE_MODE_WORKER:
        return
    if settings.NODE_NOT_SYNC:
        print("Node syncing skiped by settings")
        return
    process = True
    page = 0
    c_ws = 0
    c_usr = 0
    while process:
        arb_data = None
        while not arb_data:
            arb_data = arbiter_request(
                settings.NODE_ARBITER_HOST, 
                'upsync',
                data={'page': page},
            )
            if not arb_data:
                print('Wait...')
                time.sleep(3)
        for ws in arb_data.get('websites', []):
            set_website_with_relations(ws)
        for usr in arb_data.get('users', []):
            set_user_with_relations(usr)
        c_ws += len(arb_data.get('websites', []))
        c_usr += len(arb_data.get('users', []))
        if arb_data.get('next') == False:
            process = False
        else:
            page += 1
            print('syncing...')
    print('Sync completed.', c_ws, "website(s) and", c_usr, "user(s) fetched.")


class WatcherMiddleware:
    http_req = 0
    http_s2s = 0
    ws_req = 0
    active_ws = 0
    stats = {}

    # WIDGET_RE = re.compile("^/[^/]+/widget/(\d+)/")
    # AGENT_RE = re.compile("^/[^/]+/api/v1/websites/(\d+)/")

    def __init__(self, inner):
        self.inner = inner

    @staticmethod
    def _track_start(scope):
        try:
            _type = scope.get('type')
            _path = scope.get('path')
            if _type == 'http':
                if _path and 'arb_s2s' in _path:
                    WatcherMiddleware.http_s2s += 1
                else:
                    WatcherMiddleware.http_req += 1
                    for _h, _v in scope.get('headers', []):
                        if _h == b'x-site-id' and _v and _v.isdigit():
                            _v = int(_v)
                            WatcherMiddleware.stats[f"http:{_v}"] = 1 + WatcherMiddleware.stats.get(f"http:{_v}", 0)
                            return
                    WatcherMiddleware.stats[f"http:gen"] = 1 + WatcherMiddleware.stats.get(f"http:gen", 0)

            if _type == 'websocket':
                WatcherMiddleware.ws_req += 1
                WatcherMiddleware.active_ws += 1

        except Exception:
            pass

    @staticmethod
    def _track_stop(scope):

        try:
            _type = scope.get('type')
            if _type == 'websocket':
                WatcherMiddleware.active_ws += -1
        except Exception:
            pass

    async def __call__(self, scope, receive, send):

        WatcherMiddleware._track_start(scope)
        ret = await self.inner(scope, receive, send)
        WatcherMiddleware._track_stop(scope)

        return ret


def scenario_store(scenario, data, add_to_list, website_id, action):
    from django.conf import settings

    content = json.dumps({
        'scenario': scenario,
        'data': data
    }).encode()
    hash = hashlib.sha3_256(content).hexdigest()

    try:
        storage = get_storage_class(settings.CHAT_UPLOAD_STORAGE)()
        token = f'{website_id}-{action}'
        path = 'sync-%s-%s/%s/%s.dat' % (datetime.now().year, datetime.now().month, token, hash)
        if not storage.exists(path):
            result = storage.save(path, ContentFile(content))
        else:
            result = path
        add_to_list.append(str(result))
        return True
    except Exception as ex:
        print(ex)
        return False
    

def scenario_read(data_path):
    from django.conf import settings

    try:
        storage = get_storage_class(settings.CHAT_UPLOAD_STORAGE)()
        print(data_path, storage.exists(data_path))
        if storage.exists(data_path):
            with storage.open(data_path) as f:
                data = f.read()
                data = json.loads(data)
                return data['scenario'], data['data']
    except Exception as ex:
        print(ex)
    return None, None


def scenario_clear(data_path):
    from django.conf import settings

    try:
        storage = get_storage_class(settings.CHAT_UPLOAD_STORAGE)()
        print(data_path, storage.exists(data_path))
        if storage.exists(data_path):
            storage.delete(data_path)
            return True
    except Exception as ex:
        print(ex)
    return None
