from datetime import datetime
import time
import hashlib
import random
import base64
from rest_framework import status
from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import get_storage_class


def save_attach_file(request, file_field_name, chat_id, is_agent=False, is_json=False):

    try:
        if request.META.get('CONTENT_LENGTH'):
            size = int(request.META['CONTENT_LENGTH'])
        elif request.META.get('HTTP_X_FILE_SIZE'):
            size = int(request.META['HTTP_X_FILE_SIZE'])
        else:
            size = None
    except ValueError:
        size = None

    if is_json:
        name = request.data.get('filename')
        _file = None
    else:
        _file = request.FILES.get(file_field_name)
        name = _file.name


    if not size:
        return {'name': name, 'error': status.HTTP_400_BAD_REQUEST}, False

    if size > (settings.AGENT_ATTACH_LIMIT if is_agent else settings.VISITOR_ATTACH_LIMIT) * 1048576:
        return {'name': name, 'error': status.HTTP_413_REQUEST_ENTITY_TOO_LARGE}, False

    if not is_json:
        if not request.FILES or not file_field_name in request.FILES:
            return {'name': name, 'error': status.HTTP_400_BAD_REQUEST}, False

    if is_json:
        try:
            bc64 = request.data.get('file')
            if 'base64,' in bc64:
                bc64 = bc64.split(';base64,')
                ct = bc64[0].replace('data:', '')
                bc64 = bc64[1]
            else:
                ct = 'application/octetstream'
            content = base64.b64decode(bc64)
            size = len(content)
        except (ValueError, TypeError) as e:
            print(e)

            return {'name': name, 'error': status.HTTP_400_BAD_REQUEST}, False
    else:
        size = _file.size
        content = _file.file.getvalue()
        ct = None

    if (is_agent and not allowed_filetype(content)) or (not is_agent and not allowed_filetype(content)):
        return {'name': name, 'error': status.HTTP_415_UNSUPPORTED_MEDIA_TYPE}, False

    # create file at first
    # manual create now, CDN in future
    try:
        storage = get_storage_class(settings.CHAT_UPLOAD_STORAGE)()
    except ImportError as ie:
        print(ie)
        return {'name': name, 'error': status.HTTP_507_INSUFFICIENT_STORAGE}, False

    path = '%s-%s/%s/%s.dat' % (datetime.now().year, datetime.now().month, str(chat_id), time.time())

    try:
        result = storage.save(path, ContentFile(content))
    except Exception as ex:
        print(ex)
        return {'name': name, 'error': status.HTTP_507_INSUFFICIENT_STORAGE}, False

    try:
        from PIL import Image
        import io
        from uuid import uuid4
        image = Image.open(io.BytesIO(content))

        MAX_SIZE = (200, 200)
        
        image.thumbnail(MAX_SIZE)
        
        path2u = "%s-%s/%s.jpg" % (datetime.now().year, datetime.now().month, str(uuid4()))
        path2 = "%s/%s" % (settings.MEDIA_ROOT, path2u)
        if image.mode == "RGBA":
            new_image = Image.new("RGBA", image.size, "WHITE")
            new_image.paste(image, mask=image)
            image = new_image.convert("RGB")
        if image.mode == "P":
            image = image.convert("RGB")
        image.save(path2, format="JPEG")
        
        preview = '%s%s' % (settings.MEDIA_URL, path2u)
    except Exception as x:
        import traceback
        print(traceback.print_exc())
        preview = None

    return {
        'content_type': _file.content_type if not is_json else ct,
        'name': name,
        'size': size,
        'preview': preview,
        'storage': settings.CHAT_UPLOAD_STORAGE,
        'part_path': str(result),
        'path': "%s:%s" % (settings.CHAT_UPLOAD_STORAGE, str(result))}, True


def simple_save(content, ct, name, chat_id):

    size = len(content)

    if size > settings.AGENT_ATTACH_LIMIT * 1048576:
        return {'name': name, 'error': status.HTTP_413_REQUEST_ENTITY_TOO_LARGE}, False

    if not allowed_filetype(content):
        return {'name': name, 'error': status.HTTP_415_UNSUPPORTED_MEDIA_TYPE}, False

    try:
        storage = get_storage_class(settings.CHAT_UPLOAD_STORAGE)()
    except ImportError as ie:
        print(ie)
        return {'name': name, 'error': status.HTTP_507_INSUFFICIENT_STORAGE}, False

    path = '%s-%s/%s/%s.dat' % (datetime.now().year, datetime.now().month, str(chat_id), time.time())

    try:
        result = storage.save(path, ContentFile(content))
    except Exception as ex:
        print(ex)
        return {'name': name, 'error': status.HTTP_507_INSUFFICIENT_STORAGE}, False

    try:
        from PIL import Image
        import io
        from uuid import uuid4
        image = Image.open(io.BytesIO(content))

        MAX_SIZE = (200, 200)
        
        image.thumbnail(MAX_SIZE)
        
        path2u = "%s-%s/%s.jpg" % (datetime.now().year, datetime.now().month, str(uuid4()))
        path2 = "%s/%s" % (settings.MEDIA_ROOT, path2u)
        image.save(path2, format="JPEG")
        
        preview = '%s%s' % (settings.MEDIA_URL, path2u)
    except PIL.UnidentifiedImageError:
        preview = None
    except Exception as x:
        import traceback
        print(traceback.print_exc())
        preview = None

    return {
        'content_type': ct,
        'name': name,
        'size': size,
        'preview': preview,
        'storage': settings.CHAT_UPLOAD_STORAGE,
        'part_path': str(result),
        'path': "%s:%s" % (settings.CHAT_UPLOAD_STORAGE, str(result))}, True


def allowed_filetype(content):
    if not isinstance(content, bytes):
        return False
    if content.startswith(b'MZ'):  # win-executable
        return False
    if content.startswith(b'#!/'):  # shell script
        return False
    if content.startswith(b'Rar!'):  # rar-archive
        return False
    if content.startswith(b'\x1f\x8b\x08'):  # tar.gz-archive
        return False
    if content.startswith(b'\x7fELF'):  # *nix-executable
        return False
    if content.startswith(b'\xcf\xfa\xed\xfe') or content.startswith(b'\xce\xfa\xed\xfe') or \
            content.startswith(b'\xfe\xed\xfa\xce') or content.startswith(b'\xfe\xed\xfa\xcf') or \
            content.startswith(b'\xca\xfe\xba\xbe'):  # mach-o binary
        return False
    return True


def is_image(content):
    if not isinstance(content, bytes):
        return False
    if content.startswith(b'\x89PNG'):
        return True
    if content.startswith(b'GIF89a') or content.startswith(b'GIF87a'):
        return True
    if content.startswith(b'\xff\xd8') and content.endswith(b'\xff\xd9'):  # jpeg
        return True
    return False


def make_seed(msg_id):
    t = int(time.time() // 14400)
    f = f"{t}{settings.ATTACH_SIGN_SECRET}{msg_id}".encode()
    seed = bytes([random.randint(1,254), random.randint(1,254), random.randint(1,254)])
    f += seed
    f = seed + hashlib.sha3_384(f).digest()
    return base64.b64encode(f, altchars=b"_-").decode()


def verify_seed(msg_id, seed):
    dt = base64.b64decode(seed, altchars=b"_-")
    s = dt[0:3]
    dt = dt[3:]
    for q in range(1, -3, -1):
        t = int(time.time() // 14400) + q
        f = f"{t}{settings.ATTACH_SIGN_SECRET}{msg_id}".encode() + s
        if hashlib.sha3_384(f).digest() == dt:
            return True
    return False
