import json

from nicechat.celery import app
from django.utils.translation import gettext_lazy as _
from django.core.cache import caches, InvalidCacheBackendError
from django.forms.models import model_to_dict

from django.conf import settings
from django.core.mail import send_mail
import requests

EVENT_NEW_ORDER = 'new_order'
EVENT_NEW_VISITOR = 'new_visitor'
EVENT_NEW_CHAT = 'new_chat'
EVENT_FIRST_MESSAGE = 'first_message'
EVENT_UNANSWERED = 'unanswered'
EVENT_PUSH_FORM = 'push_form'

EVENT_CHOICES = (
    (EVENT_NEW_ORDER, _('New Order'), _('This event is triggered when a client makes a purchase')),
    (EVENT_NEW_VISITOR, _('New Visitor'), _('This event is triggered when a new Visitor was discovered ')),
    (EVENT_NEW_CHAT, _('New Chat'), _('This event is triggered when any Chat began')),
    (EVENT_FIRST_MESSAGE, _('Client first message'),
        _('This event is triggered when Client write first message')),
    (EVENT_UNANSWERED, _('You have unanswered chat'),
        _('This event is triggered when Client begun chat, but no Agents present')),
    (EVENT_PUSH_FORM, _('Form filled'), _('Client fill form')),
)


@app.task
def make_callback(event, website_id, obj_type, obj_id):
    from apps.website.models import Callbacks

    token = "callbacks-%s-%s" % (event, website_id)
    try:
        cache = caches['default']
        callbacks = cache.get(token)
    except InvalidCacheBackendError:
        callbacks = None

    if not callbacks:
        callbacks = [item for item in Callbacks.objects.filter(website_id=website_id, event=event,
                                                               is_active=True).values('target', 'params')]
        try:
            cache = caches['default']
            cache.set(token, callbacks, 7200)
        except InvalidCacheBackendError:
            pass

    for item in callbacks:
        mode = item.get('target')
        params = item.get('params')

        if not params:
            continue

        if mode == Callbacks.TARGET_SLACK:
            #do_slack(event, params, get_short_message(obj_type, obj_id))
            pass
        elif mode == Callbacks.TARGET_HTTP:
            dt = get_tech_message(obj_type, obj_id)
            dt['event'] = event
            do_http(event, params, dt)
        elif mode == Callbacks.TARGET_EMAIL_TECH:
            dt = get_tech_message(obj_type, obj_id)
            dt['event'] = event
            do_email(event, params, dt, True)
        elif mode == Callbacks.TARGET_EMAIL_HUMAN:
            do_email(event, params, get_short_message(obj_type, obj_id), False)


def get_tech_message(obj_type, obj_id):
    if obj_type == 'Chat':
        from apps.chat.models import Chat
        chat = Chat.objects.filter(pk=obj_id).first()
        if chat:
            obj = model_to_dict(chat, fields=['id', 'sys_lang_full', 'sys_browser', 'sys_os', 'sys_ip',
                                              'sys_lang', 'sys_time', 'geo_country' 'geo_city',
                                              'geo_lat', 'geo_lon', 'page_utm', 'page_ref'])

            obj['ts_open'] = chat.ts_open.replace(microsecond=0).isoformat() if chat.ts_open else None
            obj['ts_close'] = chat.ts_close.replace(microsecond=0).isoformat() if chat.ts_close else None
            obj['client_name'] = chat.get_client_name()
            obj['client_id'] = chat.get_client_pk()
            return obj
    elif obj_type == 'Client':
        from apps.visitors.models import Client
        instance = Client.objects.filter(pk=obj_id).first()
        if instance:
            return {
                'id': instance.pk,
                'client_name': instance.client_name,
                'first_ts': instance.first_ts.replace(microsecond=0).isoformat(),
                'last_activity_ts': instance.last_activity_ts.replace(microsecond=0).isoformat(),
                'visit_count': instance.visit_count,
                'chat_count': instance.chat_count,
                'void_chat_count': instance.void_chat_count,
                'unaccepted_chat_count': instance.unaccepted_chat_count,
                'purchase_count': instance.purchase_count,
                'product_view_count': instance.product_view_count,
            }
    elif obj_type == 'Order':
        from apps.visitors.models import Order
        instance = Order.objects.filter(pk=obj_id).first()
        if instance:
            obj = model_to_dict(instance)
            obj['order_date'] = instance.order_date.replace(microsecond=0).isoformat()
            obj['products'] = [model_to_dict(prod, fields=('original_id', 'title', 'price', 'currency', 'url', 'image_url',))
                               for prod in instance.products.all()]
            obj['agents'] = [model_to_dict(agent, fields=('first_name', 'last_name', 'email',))
                             for agent in instance.agents.all()]
            return obj
    elif obj_type == 'FormResult':
        from apps.visitors.models import FormResult
        instance = FormResult.objects.filter(pk=obj_id).first()
        if instance:
            obj = model_to_dict(instance)
            obj['created'] = instance.created.replace(microsecond=0).isoformat()
            return obj

    return {}


def get_short_message(obj_type, obj_id):
    if obj_type == 'Chat':
        from apps.chat.models import Chat
        instance = Chat.objects.filter(pk=obj_id).first()
        if instance:
            return u'"%(name)s" on chat #%(id)s on %(website)s. ' \
                   u'\n https://%(acc_domain)s/chats/%(id)s ' % {
                                                                        'website': instance.website.title,
                                                                        'name': instance.get_client_name(),
                                                                        'acc_domain': settings.DOMAIN_ACC,
                                                                        'id': instance.pk
                                                                    }
    elif obj_type == 'Client':
        from apps.visitors.models import Client
        instance = Client.objects.filter(pk=obj_id).first()
        if instance:
            return u'New user "%(name)s" was appeared on %(website)s. ' \
                   u'\n ID:%(id)s ' % {
                                          'website': instance.website.title,
                                          'name': instance.client_name,
                                          'id': instance.pk
                                      }
    elif obj_type == 'Order':
        from apps.visitors.models import Order
        instance = Order.objects.filter(pk=obj_id).first()
        if instance:
            return u'New Purchase #%(id)s was made by %(name)s on %(website)s.' \
                   u'\n https://%(acc_domain)s]/orders/%(id)s ' % {
                                                                        'website': instance.website.title,
                                                                        'name': instance.client.client_name,
                                                                        'acc_domain': settings.DOMAIN_ACC,
                                                                        'id': instance.pk
                                                                    }
    elif obj_type == 'FormResult':
        from apps.visitors.models import FormResult
        instance = FormResult.objects.filter(pk=obj_id).first()
        if instance:
            return u'New form "%(name)s" was filled on %(website)s. ' \
                   u'\n ID:%(id)s ' % {
                                          'website': instance.website.title,
                                          'name': instance.label,
                                          'id': instance.pk
                                      }
    return "unknown object %s with ID %s" % (obj_type, obj_id)


# def do_slack(event, params, data):
#     slack_to = params.get('slack_to')
#     if slack_to:

#         head = 'n/a'

#         for item in EVENT_CHOICES:
#             if item[0] == event:
#                 head = item[1]

#         message = "*%s*\n%s\n\n" % (head, data)

#         try:
#             slack_client = Slacker(token=None, incoming_webhook_url=slack_to)
#             slack_client.incomingwebhook.post({
#                 "username": "%s notification system" % settings.BASE_CHAT_NAME,
#                 "text": message,
#                 "icon_emoji": ":point_up:"
#             })
#         except Error as e:
#             print(e)


def do_http(event, params, data):
    url_to = params.get('url_to')
    if url_to:
        message = {
            'event': event,
            'data': data
        }
        try:
            requests.post(url_to, json=message, headers={'User-agent': '%s callbacks' % settings.BASE_CHAT_NAME})
        except Exception as e:
            print(e)


def do_email(event, params, data, is_tech):
    send_to = params.get('send_to')
    if send_to:
        if not is_tech:
            head_txt = 'n/a'
            description = params.get('description', '')

            for item in EVENT_CHOICES:
                if item[0] == event:
                    head_txt = item[1]

            message = "*%s*\n\n%s\n\n%s\n" % (head_txt, description, data)
        else:
            message = json.dumps({
                'event': event,
                'data': data
            })

        send_mail('[' + event + ']', message, 'callback@%s' % settings.DOMAIN_BASE,
                  [send_to], fail_silently=True)

