import re
import collections

from rest_framework_simplejwt.settings import api_settings
from rest_framework_simplejwt.exceptions import AuthenticationFailed, InvalidToken, TokenError

from django.db.models import Q
from django.contrib.auth import get_user_model

from apps.website.models import Website
from django.utils.translation import gettext_lazy as _


ADDRESS_DICT = (
    ('first', ('First name',),),
    ('last', ('Last name',),),
    ('address_1', ('Address', 'Address-1', 'Address 1',),),
    ('address_2', ('Address-2', 'Address 2',),),
    ('address_3', ('Address-3', 'Address 3',),),
    ('city', ('City',),),
    ('state', ('State',),),
    ('country', ('Country',),),
    ('zip', ('POST/ZIP','ZIP','POST','P.O.S.T.'),),
)

PRICE_NUM_PATTERNS = [
    r"^\s?([0-9\s\,]*)\.([0-9]{1,2})\s?$",
    r"^\s?([0-9\s\.]*)\,([0-9]{1,2})\s?$",
    r"^\s?([0-9\s]*)\s?$"]

PRICE_STRING_PATTERNS = [
    ("^\s?([A-Za-z\$]{1,3})\s([0-9\,\.\s]+)\s?$", 1, 0,),
    ("^\s?([0-9\,\.\s]+)\s([A-Za-z\$]{1,3})\s?$", 0, 1,),
    ("^\s?([0-9\,\.\s]+)\s?$", 0, None,)
]


def parse_float(price):
    for regexp in PRICE_NUM_PATTERNS:
        if re.match(regexp, price):
            grp = re.match(regexp, price).groups()
            if len(grp) == 2:
                s = "%s.%s" % (re.sub("[^0-9]", "", grp[0]), grp[1])
            elif len(grp) == 1:
                s = "%s.00" % re.sub("[^0-9]", "", grp[0])
            else:
                s = "0.0"
            try:
                return float(s)
            except ValueError:
                return False
    return False


def parse_price(price_str):
    price = False
    currency = False

    if isinstance(price_str, str):

        price_str = price_str.strip()

        for regexp, price_id, curr_id in PRICE_STRING_PATTERNS:

            mth = re.match(regexp, price_str)
            if mth:
                rez = mth.groups()
                price = parse_float(rez[price_id])
                currency = rez[curr_id].lower() if curr_id else 'usd'

        if currency == '$':
            currency = 'usd'
    elif isinstance(price_str, float):
        return price_str, 'usd'
    elif isinstance(price_str, int):
        return 0.0 + price_str, 'usd'

    return price, currency


def nice_address(in_addr):
    out_addr = collections.OrderedDict()

    for item, variants in ADDRESS_DICT:
        val = None
        for v in variants:
            if in_addr.get(v, None):
                val = in_addr.get(v)
                break
        out_addr[item] = val

    return out_addr


def get_current_site_from_request(request, raise_exception=True):
    """
    Fetches the current site based on 'site_id' GET parameter and considering other restrictions/checks:
    * the site has to belong to current user's company
    * user must be a member of this site OR the owner of site's company

    :param request: HttpRequest instance
    :param raise_exception: either raise an exceptions if any of the checks fails or to fail silently and return None
    :return: either corresponding apps.website.models.Website instance or None

    """
    site_id = request.GET.get('site_id', None)
    if not site_id:
        site_id = request.META.get('HTTP_X_SITE_ID', None)
    # print(request.kwargs, request.kwargs.get('worker_node/site_id'), )
    try:
        assert site_id
        assert request.user.is_authenticated

        current_site = Website.objects.get(
            Q(pk=site_id, company=request.user.company),
            Q(pk__in=request.user.website_set.values_list('pk', flat=True))
            | Q(company__owner=request.user)
        )
    except (Website.DoesNotExist, AssertionError, ValueError):
        if raise_exception:
            raise
        else:
            current_site = None

    return current_site


def auth_jwt(jwt_value):

    messages=[]
    for AuthToken in api_settings.AUTH_TOKEN_CLASSES:
        try:
            vl = AuthToken(jwt_value)
            
            try:
                user_id = vl[api_settings.USER_ID_CLAIM]
            except KeyError:
                return False, "Token contained no recognizable user identification"

            User = get_user_model()
            try:
                user = User.objects.get(**{api_settings.USER_ID_FIELD: user_id})
            except User.DoesNotExist:
                return False, "User not found"

            if not user.is_active:
                return False, "User is inactive"

            return True, user

        except TokenError as e:
            messages.append(
                {
                    "token_class": AuthToken.__name__,
                    "token_type": AuthToken.token_type,
                    "message": e,
                }
            )

    return False, "Token expired or incorrent"
