import os
import uuid
import hashlib
from datetime import datetime

from django.core.files.base import ContentFile
from PIL import ExifTags, Image
from django.conf import settings
from django.core.files.storage import get_storage_class
from .attach import is_image


def get_image_file_path(instance, filename):
    ext = filename.split('.')[-1]
    filename = "%s.%s" % (uuid.uuid4(), ext)

    path = 'misc'
    return os.path.join('images/{}/'.format(path), filename)


def image_orientation_fix(image):
    # try check image rotation by EXIF data
    try:
        # check file orientation
        for orientation in ExifTags.TAGS.keys():
            if ExifTags.TAGS[orientation] == 'Orientation':
                break

        exif = dict(image._getexif().items())
        if exif[orientation] == 3:
            image = image.rotate(180, expand=True)
        elif exif[orientation] == 6:
            image = image.rotate(270, expand=True)
        elif exif[orientation] == 8:
            image = image.rotate(90, expand=True)

            # clear EXIF data for current image

    except (AttributeError, KeyError, IndexError):
        # cases: image don't have EXIF key (Exif.Image.Orientation)
        pass

    return image


def image_post_processing(image):
    try:
        storage = get_storage_class(settings.DEFAULT_FILE_STORAGE)()
    except ImportError as ie:
        return False

    try:
        file_path = image.path
        image_obj = Image.open(storage.open(file_path))
        # try check image rotation by EXIF data
        rotated_image = image_orientation_fix(image_obj)
        # image resize to fit the limits
        rotated_image.thumbnail(settings.MAX_IMAGE_SIZE, Image.ANTIALIAS)
        rotated_image.save(file_path)
    except Exception as ex:
        return False


def image_post_create(site_id, content, name):

    if not is_image(content):
        return False, False

    ext = 'jpg'
    if name and isinstance(name, str):
        if name[-3:] in ['jpg', 'png', 'gif']:
            ext = name[-3:]

    h16 = hashlib.sha1(content).hexdigest()

    try:
        storage = get_storage_class(settings.DEFAULT_FILE_STORAGE)()
    except ImportError as ie:
        print(ie)
        return False, False

    path = '%s-%s/%s/%s.%s' % (datetime.now().year, datetime.now().month, str(site_id), h16, ext)

    try:
        result = storage.save(path, ContentFile(content))
        return result, storage.url(result)
    except Exception as ex:
        print(ex)

    return False, False
