import os
import re
import logging

Openable = (str, os.PathLike)
logger = logging.getLogger(__name__)


def read_env(env_file=None, encoding='utf8', **overrides):

    try:
        if isinstance(env_file, Openable):
            # Python 3.5 support (wrap path with str).
            with open(str(env_file), encoding=encoding) as f:
                content = f.read()
        else:
            with env_file as f:
                content = f.read()
    except OSError:
        logger.info(
            "%s not found - if you're not configuring your "
            "environment separately, check this.", env_file)
        return

    logger.debug('Read environment variables from: %s', env_file)

    def _keep_escaped_format_characters(match):
        """Keep escaped newline/tabs in quoted strings"""
        escaped_char = match.group(1)
        if escaped_char in 'rnt':
            return '\\' + escaped_char
        return escaped_char

    for line in content.splitlines():
        m1 = re.match(r'\A(?:export )?([A-Za-z_0-9]+)=(.*)\Z', line)
        if m1:
            key, val = m1.group(1), m1.group(2)
            # Look for value in quotes, ignore post-# comments
            # (outside quotes)
            m2 = re.match(r"\A\s*'(?<!\\)(.*)'\s*(#.*\s*)?\Z", val)
            if m2:
                val = m2.group(1)
            else:
                # For no quotes, find value, ignore comments
                # after the first #
                m2a = re.match(r"\A(.*?)(#.*\s*)?\Z", val)
                if m2a:
                    val = m2a.group(1)
            m3 = re.match(r'\A"(.*)"\Z', val)
            if m3:
                val = re.sub(r'\\(.)', _keep_escaped_format_characters,
                                m3.group(1))
            overrides[key] = str(val)
        elif not line or line.startswith('#'):
            # ignore warnings for empty line-breaks or comments
            pass
        else:
            logger.warning('Invalid line: %s', line)

    for key, value in overrides.items():
        os.environ.setdefault(key, value)

    return overrides
