import pytz

from abc import ABCMeta
from collections import namedtuple
from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils import timezone
from django.test import override_settings
from faker import Faker
from channels.tests import ChannelTestCase
from rest_framework.test import APITestCase, APIClient

from nicechat.utils.constants import UserRoles
from apps.users.models import UserSiteRole
from apps.billing.models import CompanyPlan, Plan
from apps.billing.utils import get_trial_plan
from apps.company.models import Company
from apps.website.models import Website, WidgetTheme
from apps.visitors.models import Statistics

fake = Faker()


@override_settings(CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
                   CELERY_ALWAYS_EAGER=True,
                   BROKER_BACKEND='memory')
class TestCaseWithUsers(APITestCase):
    __metaclass__ = ABCMeta

    def setUp(self):
        """
        Create a set of user objects and store them as current class/object attributes
        named 'user'+index (self.user1, self.user2 etc.)

        """
        super(TestCaseWithUsers, self).setUp()
        self.api_url = '/api/v1/'
        self.api_client = APIClient()

        # Creating users
        self.password = 'password1'

        UserData = namedtuple('UserData', 'email first_name last_name company_id')

        users_data = [
            UserData('u1@example.com', 'Some', 'User', 11),
            UserData('u2@example.com', 'Some', 'User', 22),
            UserData('u3@example.com', 'Another', 'User', 23),
            UserData('u4@example.com', 'Another', 'Admin', 33),
            UserData('u5@example.com', 'Some', 'Admin', 11),
        ]

        for idx, user_data in enumerate(users_data, start=1):
            attr_name = 'user{}'.format(idx)

            self.__setattr__(attr_name, get_user_model().objects.create_user(
                first_name=user_data.first_name,
                last_name=user_data.last_name,
                email=user_data.email,
                password=self.password,
                company_id=user_data.company_id,

            ))

    def login_user(self, user=None):
        if user:
            email = user.email
        else:
            email = self.user1.email

        login_result = self.client.login(username=email, password=self.password)

        if not login_result:
            raise Exception('User login failed')


class TestCaseWithSites(TestCaseWithUsers, ChannelTestCase):
    num_of_companies = 4
    users_per_company = 2
    site_per_company = 3
    domains = []

    fixtures = ['nicechat/data/db_dumps/plans.json']

    def setUp(self):
        """
        Create Company, several related Site objects (and one User for requirements)

        """
        super(TestCaseWithSites, self).setUp()

        WidgetTheme.objects.get_or_create(name=settings.DEFAULT_WIDGET_THEME)

        for company_index in range(0, self.num_of_companies):
            company_owner = get_user_model().objects.create_user(
                email="owner@company%i.com" % company_index, password=self.password,
                first_name="User", last_name="Owner %i" % company_index,
            )
            company = Company.objects.create(label="Company %i" % company_index, owner=company_owner)
            company_owner.company = company
            company_owner.save()

            setattr(self, "user_owner_%i" % company_index, company_owner)
            setattr(self, "company%i" % company_index, company)

            # company billing plan
            if (company_index + 1) % 4 == 0:
                # every third company gets 'enterprise' plan
                plan = Plan.objects.get(name='Enterprise')
                agents_count = 12
            elif (company_index + 1) % 3 == 0:
                # every third company gets 'business' plan
                plan = Plan.objects.get(name='Business')
                agents_count = 7
            elif (company_index + 1) % 2 == 0:
                # 'free' plan for every even company index
                plan = Plan.objects.get(name='Free')
                agents_count = 1
            else:
                # 'trial' for all the others
                plan = get_trial_plan()
                agents_count = 3

            CompanyPlan.objects.create(
                company=company,
                plan=plan,
                agents_count=agents_count,
                begin=timezone.now(),
                end=timezone.now() + timedelta(days=365),
            )

            # users
            for user_index in range(0, self.users_per_company):
                setattr(self, "user_%i_%i" % (company_index, user_index), get_user_model().objects.create_user(
                    email="user%i@company%i.com" % (user_index, company_index),
                    password=self.password,
                    first_name="User %i%i" % (company_index, user_index),
                    last_name="Operator %i%i" % (company_index, user_index),
                    company=company,
                ))

            # sites
            for site_index in range(0, self.site_per_company):
                site_attr_name = "website_%i_%i" % (company_index, site_index)
                for t in range(1, 50):
                    domain = fake.domain_name()
                    if domain in TestCaseWithSites.domains:
                        continue
                    else:
                        setattr(self, site_attr_name, Website.objects.create(
                            title=site_attr_name,
                            domain=domain,
                            company=company,
                        ))
                        TestCaseWithSites.domains.append(domain)
                        break

                # add odd index users to the odd index sites and vise versa
                for user_index in range(0, self.users_per_company):
                    if user_index % 2 == site_index % 2:
                        user = getattr(self, "user_%i_%i" % (company_index, user_index))
                        site = getattr(self, site_attr_name)

                        site.users.add(user)

                        # make odd index users admins and even index user agents of the site
                        UserSiteRole.objects.create(
                            user=user, site=site,
                            role=UserRoles.SITE_AGENT if user_index % 2 == 0 else UserRoles.SITE_ADMIN
                        )

    def tearDown(self):
        pass

        super(TestCaseWithSites, self).tearDown()

    def refresh_companies_from_db(self):
        for company_index in range(0, self.num_of_companies):
            getattr(self, "company%i" % company_index).refresh_from_db()


class TestCaseWithStatistics(TestCaseWithSites):
    def setUp(self):
        """
        Create Company, several related Site objects (and one User for requirements)

        """
        super(TestCaseWithStatistics, self).setUp()

        website_list = []
        for t in dir(self):
            if t.startswith('website_'):
                website_list.append(getattr(self, t))

        self.website_list = website_list
        self.BROWSER_1 = 'browser_1'
        self.BROWSER_2 = 'browser_2'
        self.OS_1 = 'os_1'
        self.OS_2 = 'os_2'
        self.OSMOB_1 = 'osmob_1'
        self.CITY_1 = 'city_1'
        self.CITY_2 = 'city_2'
        self.PAGE_1 = 'page_1'
        self.PAGE_2 = 'page_2'
        self.REF_1 = 'ref_1'
        self.REF_2 = 'ref_2'

        for site in website_list:
            Statistics.objects.create(
                website=site,
                period=Statistics.QUARTER,
                ts=datetime(2017, 05, 01, 10, 15, 0, tzinfo=pytz.UTC),
                visitors_count=2,
                browsers={self.BROWSER_1: 1, self.BROWSER_2: 1},
                desktop_os={self.OS_1: 1, self.OS_2: 1},
                mobile_os={},
                cities={self.CITY_1: 2},
                pages={self.PAGE_1: 2},
                referrers={self.REF_1: 2},
            )
            Statistics.objects.create(
                website=site,
                period=Statistics.QUARTER,
                ts=datetime(2017, 05, 01, 10, 30, 0, tzinfo=pytz.UTC),
                visitors_count=2,
                browsers={self.BROWSER_1: 2},
                desktop_os={self.OS_1: 1},
                mobile_os={self.OSMOB_1: 1},
                cities={self.CITY_1: 1, self.CITY_2: 1},
                pages={self.PAGE_1: 1, self.PAGE_2: 1},
                referrers={self.REF_1: 1, self.REF_2: 1},
            )
            Statistics.objects.create(
                website=site,
                period=Statistics.QUARTER,
                ts=datetime(2017, 05, 01, 11, 15, 0, tzinfo=pytz.UTC),
                visitors_count=2,
                browsers={self.BROWSER_1: 1, self.BROWSER_2: 1},
                desktop_os={self.OS_1: 1, self.OS_2: 1},
                mobile_os={},
                cities={self.CITY_1: 2},
                pages={self.PAGE_1: 2},
                referrers={self.REF_1: 2},
            )
            Statistics.objects.create(
                website=site,
                period=Statistics.QUARTER,
                ts=datetime(2017, 05, 05, 10, 30, 0, tzinfo=pytz.UTC),
                visitors_count=2,
                browsers={self.BROWSER_1: 2},
                desktop_os={self.OS_1: 1},
                mobile_os={self.OSMOB_1: 1},
                cities={self.CITY_1: 1, self.CITY_2: 1},
                pages={self.PAGE_1: 1, self.PAGE_2: 1},
                referrers={self.REF_1: 1, self.REF_2: 1},
            )
            Statistics.objects.create(
                website=site,
                period=Statistics.QUARTER,
                ts=datetime(2017, 04, 10, 10, 30, 0, tzinfo=pytz.UTC),
                visitors_count=2,
                browsers={self.BROWSER_1: 2},
                desktop_os={self.OS_1: 1},
                mobile_os={self.OSMOB_1: 1},
                cities={self.CITY_1: 1, self.CITY_2: 1},
                pages={self.PAGE_1: 1, self.PAGE_2: 1},
                referrers={self.REF_1: 1, self.REF_2: 1},
            )
